#!/bin/ksh
eval 'exec perl -x -S $0 ${1+"$@"}'
              if $running_under_some_shell;
#!/usr/bin/perl
# -*- Mode: cperl -*-
############################################################################
# A program to build fortran 90 module dependencies for all the files of current
# directory.
# Guesses the best order for making build, always putting
# dependencies before so that the recompilation of a module and one
# of it's dependencies always works if no side effect.
# Should also work with C++.
# Excludes main from processing, creating a specific rule that suits the local
# makefile file in current directory.
# Usage:
#   perl $0 -d => generates all dependencies
#   perl $0    => does not generate dependencies for faster recompilation.
#   perl $0 -f => generates dependencies for the main program only, use with caution.
#   perl $0 -d -f => generates all dependencies + program's global dependencies.
############################################################################
#    Copyright (C) 2003 Olivier Cessenat.
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
############################################################################
use File::Basename ;
use Getopt::Std;
getopts('bdfI:A:s:') ;
$debug = 0 ;
if (@ARGV) {
  (@dirs) = (@ARGV) ;
} else {
  (@dirs) = ('.') ;
}
(@aincludes) = split(':', $opt_A) ;
(@rincludes) = split(':', $opt_I) ;
$lib1 = '$(LIB)(' ;
$lib2 = ')' ;
$lib1 = ' ./objets$(OUT)$(TYPEREEL)/$(HOSTARCH)/' ;
$lib2 = '' ;
############################################################################
# Gets the absolute path to $2 from the $1 location: path becomes valid
# from the current working directory
sub getAbs {
  my $from = shift ;
  my $to = shift ;
  my $out ;
  my (@out) = split('/', $from) ;
  if ($to =~ /^\// ) {
    $out = $to ;
  } else {
    foreach my $l (@out) {
      if ($to =~ /^\.\.\/(.*)/ ) {
	$to = $1 ;
      }
    }
    $out = $to ;
  }
}
############################################################################
# Gets the relative location to $2 from $1
sub getRel {
  my $from = shift ;
  my $to = shift ;
  my $out ;
  if ($to =~ /^\// ) {
    $out = $to ;
  } else {
    # Tries to get ../$2 or something like this from $base...
    if ($from ne '.') {
      $to = "../$to" ;
    }
    while (dirname($from) ne '.') {
      $to = "../$to" ;
    }
    $out = $to ;
  }
}
############################################################################
# Tries to get the include file actual location according to the -I variables.
sub getInc {
  my $base = shift ;
  my $in  = shift ;
  my $out = shift ;
  if ((@aincludes)||(@rincludes)) {
    foreach $i (@aincludes) {
      if (-e "$i/$in") {
	$out = &getRel($base,"$i/$in") ;
      }
    }
    foreach $i (@rincludes) {
      my $j = $getrel{$i} ;
      if (-e "$j/$in") {
	$out = "$i/$in" ;
      }
    }
  } else {
    if (-e "$base/$in") {
      $out = $in ;
    }
  }
  return($out) ;
}
############################################################################
# Looks for include calls in CPP files
sub getDependcpp {
  my $f = shift ;
  my $s = shift ;
  my $d = shift ;
  $isamain{$s} = 0 ;
  if (open(FILE,"<$f")) {
    while(<FILE>) {
      # eliminates some comments, not all...
      next if (/^\s*\/\//) ;
      if (/^\s*\#include\s*([\"\'])(.*)\1/) {
	$inc = &getInc($d,$2);
	push(@{$dependencies{$s}},$inc) ;
      } elsif (/^\s*(int|void)\s*main\s*\(/) {
	# A not very smart manner to guess for main...
	$isamain{$s} = 1 ;
      } elsif (/^\s*CALL_FORTRAN\s*\(\s*(\w+)\s*\)/) {
	# Says we call a fortran stuff:
	$r = lc($1) ;
	$calls{$s}->{$r} = 1 ;
      }
    }
    close(FILE) ;
  }
}
############################################################################
# Looks for include calls in C files
sub getDependc {
  my $f = shift ;
  my $s = shift ;
  my $d = shift ;
  $isamain{$s} = 0 ;
  if (open(FILE,"<$f")) {
    while(<FILE>) {
      if (/^\s*\#include\s*([\"\'])(.*)\1/) {
	$inc = &getInc($d,$2);
	push(@{$dependencies{$s}},$inc) ;
      } elsif (/^\s*(int|void)\s*main\s*\(/) {
	$isamain{$s} = 1 ;
      } elsif (/^\s*CALL_FORTRAN\s*\(\s*(\w+)\s*\)/) {
	# Says we call a fortran stuff:
	$r = lc($1) ;
	$calls{$s}->{$r} = 1 ;
      }
    }
    close(FILE) ;
  }
}
############################################################################
# Looks for include calls in fortran 77 files
sub getDependf {
  my $f = shift ;
  my $s = shift ;
  my $d = shift ;
  $isamain{$s} = 0 ;
  if (open(FILE,"<$f")) {
    while(<FILE>) {
      if (/^\s*program\s*(\w+)/i) {
	$isamain{$s} = 1 ;
      } elsif (/^\s{6}\s*include\s*([\"\'])(.*)\1/i) {
	$inc = &getInc($d,$2);
	push(@{$dependencies{$s}},$inc) ;
      } elsif (/^\s{6}\s*use\s*(\w+)\b/i) {
	# Hum, not really safe, case management...
	my $r = lc($1) ;
	$uses{$s}->{$r} = 1 ;
      } elsif (/^\s{6}\s*module\s*(\w+)/i) {
	# Provides
	print STDOUT "File $s provides module $1\n" if $debug ;
	my $r = lc($1) ;
	$modules{$r} = "$s" ;
	$revmods{$s} = "$r" ;
      }
    }
    close(FILE) ;
  }
}
############################################################################
# Looks for include calls in fortran 90 files
sub getDependf90 {
  my $f = shift ;
  my $s = shift ;
  my $d = shift ;
  $isamain{$s} = 0 ;
  if (open(FILE,"<$f")) {
    while(<FILE>) {
      next if (/^\s*\#/) ;
      if (/^\s*program\s*(\w+)/i) {
	$isamain{$s} = 1 ;
      } elsif (/^\s*include\s*([\"\'])(.*)\1/i) {
	$inc = &getInc($d,$2);
	push(@{$dependencies{$s}},$inc) ;
      } elsif (/^\s*use\s*(\w+)\b/i) {
	# Hum, not really safe, case management...
	my $r = lc($1) ;
	$uses{$s}->{$r} = 1 ;
      } elsif (/^\s*module procedure\s*(\w+)/i) {
      } elsif (/^\s*module\s*(\w+)/i) {
	# Provides
	print STDOUT "File $s provides module $1\n" if $debug ;
	my $r = lc($1) ;
	$modules{$r} = "$s" ;
	$revmods{$s} = "$r" ;
      }
    }
    close(FILE) ;
  }
}
############################################################################
# Tries to guess the best sorting for the compilation order in mkobj.lst
sub trySort {
  my(@oldfiles) = (@_) ;
  my(@newfiles) = () ;
  my(@tmpfiles) = () ;
  # Get the files with dependencies:
  my $remains = 0 ;
  # my $debug = 1 ;

  foreach (@oldfiles) {
    if (exists $uses{$_}) {
      push(@tmpfiles, $_) ;
      $remains++ ;
    } else {
      push(@newfiles, $_) ;
      $dones{$_} = 1 ;
    }
  }
  my $ier = 0 ;
  my(%errs) = () ;
  my $level = 1 ;
  $noinfty = 1 ;
  while($remains && $noinfty) {
    $noinfty = 0 ;
    (@oldfiles) = (@tmpfiles) ;
    (@tmpfiles) = () ;
    $remains = 0 ;
    $level++ ;
    print STDOUT "level = $level @oldfiles\n" if $debug ;
    if ($ier) {
      print STDERR "At least one of the following modules are not set:\n" ;
      foreach my $module (keys %errs) {
	print STDOUT "\t$module, required by at least $errs{$module}\n" ;
      }
      print STDERR "initmak.pl: KO" ;
    }
    foreach (@oldfiles) {
      my $ko = 0 ;
      print STDOUT "Analysis for $_:\n" if $debug ;
      foreach my $module (keys %{$uses{$_}}) {
	if (exists $modules{$module}) {
	  my $file = $modules{$module} ;
	  print STDOUT "\tUsed module for $_ is $module defined in $file.$types{$file}\n"  if $debug;
	  my $dep = $modules{$module} ;
	  if (exists $dones{$dep}) {
	    print STDOUT "\tDepend for $module is $dep: already done\n" if $debug ;
	  } elsif ($dep eq $modules{$_}) {
	    print STDOUT "\tDepend for $module is $dep: equals self\n" if $debug ;
	  } else {
	    print STDOUT "\tDepend for $module is $dep: not already done (from $module)\n" if $debug ;
	    $ko = 1 ;
	    last ;
	  }
	} elsif (!exists $errs{$module}) {
	  print STDOUT "\tModule $module required by $_ not known\n" if $debug ;
	  $errs{$module} = "$_" ;
	  $ier++ ;
	}
      }
      if ($ko) {
	push(@tmpfiles, $_) ;
	if ($debug) {
	  my(@totos) = (keys %{$uses{$_}}) ;
	  print STDOUT "\t$_: uses list (@totos) not found in dones\n" if $debug;
	}
	$remains++ ;
     } else {
	push(@newfiles, $_) ;
	$dones{$_} = $level ;
	print STDOUT "\tDone for $_ at level $level is performed\n" if $debug ;
	$noinfty = 1 ;
      }
    }
  }
  if (!$noinfty) {
    print STDERR "trySort(*): Error on level = $level, $#oldfiles: @oldfiles\n";
    print STDERR "tmpfiles $#tmpfiles: @tmpfiles\n";
    print STDERR "remains is $remains\n";
    die "initmak.pl: KO" ;
  }
  # Now returns files sorted by level then by priority if any then by name:
  return(@newfiles) ;
}
############################################################################
sub bylevel {
  # First lower level functions then by alphabetical order
  $dones{$a} <=> $dones{$b} || $a cmp $b ;
}
############################################################################
# A more complicated routine that sorts the modules by level then by
# one level dependency, this is not very safe but could be sufficient, it
# depends on your code hierarchy:
sub glolevel {
  my $xa = 0 ;
  my $xb = 0 ;
  if ($dones{$a} == $dones{$b}) {
    # The following unsafe stuff is costly, run it only if same level:
    my(%depsa)=(%{$uses{$a}}) ;
    my(%depsb)=(%{$uses{$b}}) ;
    foreach my $moda (keys %depsa) {
      if ($modules{$moda} eq $b) {
	$xa = 1 ; 
      }
    }
    foreach my $modb (keys %depsb) {
      if ($modules{$modb} eq $a) {
	$xb = 1 ; 
      }
    }
  }
  $xa <=> $xb || $dones{$a} <=> $dones{$b} || $a cmp $b ;
}
############################################################################
sub basiclevel {
  my $xa = 0 ;
  my $xb = 0 ;
  # The following unsafe stuff is costly, run it only if same level:
  my(%depsa)=(%{$uses{$a}}) ;
  my(%depsb)=(%{$uses{$b}}) ;
  foreach my $moda (keys %depsa) {
    if ($modules{$moda} eq $b) {
      $xa = 1 ; 
    }
  }
  foreach my $modb (keys %depsb) {
    if ($modules{$modb} eq $a) {
      $xb = 1 ; 
    }
  }
  $xa <=> $xb || $a ;
}
############################################################################
# Prints the dependency rule for an f90 main program
# Returns the no dependency stuffs, then the ordered modules then the mains:
#  (@lnouse, @orders, @lmains90) ;
sub sortallf90 {
  my(@olds) = (@_) ;
  foreach my $sub (@olds) {
    if ($isamain{$sub}) {
      push(@lmains90, $sub) ;
    }
  }
  local(%writens) = () ;
  local($outline) = " " ;
  my $space = "  " ;
  foreach my $sub (@lmains90) {
    # last if ($sub eq "fdtd") ;
    print STDOUT "Analysis for $sub\n" if $debug ;
    &setUses($sub, 0, "$space") ;
  }
  foreach my $sub (@olds) {
    unless ($isamain{$sub}) {
      unless (exists $writens{$sub}) {
	unless (exists $revmods{$sub}) {
	  # Not a module, a global stuff
	  push(@lnouse90, $sub) ;
	}
      }
    }
  }
}
############################################################################
# Prints the dependency rule for an f90 main program
sub printMain {
  my $sub = shift ;
  my $typ = shift ;
  print LSTDEP "\n./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o:" ;
  # NAJ: do not use the dependencies list:
  if (1 || %{$uses{$sub}}) {
    print LSTDEP " \$(OBJF) \$(OBJs90) \$(OBJC) \$(OBJCPP)" ;
  }
  if ($opt_f || $type eq 'c') {
    # Individual dependency is written: force this for the C file...
    $lib1 = '$(LIB)(' ;
    $lib2 = ')' ;
    $lib1 = ' ./objets$(OUT)$(TYPEREEL)/$(HOSTARCH)/' ;
    $lib2 = '' ;
    &printrUses($sub) ;
  } elsif ($opt_d) {
    # Recursive analysis shall be performed, independent file are stored in OBJs90 
    &printUses($sub, 1) ;
  } else {
    # Simple compilation must be performed.
    # print LSTDEP " \$(OBJf90) \$(OBJF90) " ;
    &printUses($sub, 0) ;
  }
  print LSTDEP " $sub.$typ @{$dependencies{$sub}} \n" ;
  # Compilation rule:
  print LSTDEP "\t\@echo \"Compiled $sub library\"\n" ;
  # Release update:
  print LSTDEP "\tperl mkversion.pl -i '\$(HELP)' -h '\$(SECTION)\$(OUT)\$(TYPEREEL)_\$(HOSTARCH)' -v '\$(VERSION)' -d '\$(DATE)' -t '\$(TAG)'\n" ;
  if (0) {
    print LSTDEP "\t\$(MAKE) theversion TYPEREEL=\$(TYPEREEL) OUT=\$(OUT)\n" ;
  } else {
    if (exists $ldcmpl{"F90"}) {
      print LSTDEP "\t\$($ldcmpl{'F90'}) $ldopts{'F90'} -c ./Version.F90 -o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/Version.o\n" ;
    } else {
      print LSTDEP "\t\$(F90) -DTYPEREEL=\$(TYPEREEL) \$(FOPTS) \$(F90FLAGS) \$(OPT) \$(INCLUDES) -c ./Version.F90 -o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/Version.o\n" ;
    }
  }
  # Main program compilation:
  if (0) {
    print LSTOBJ "$sub.$typ:\n" ;
    if (exists $ldcmpl{$typ}) {
      print LSTOBJ "\t\$($ldcmpl{$typ}) $ldopts{$typ} -c ./$sub.$typ -o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o \n" ;
    } else {
      print LSTOBJ "\t\$(F90) -DTYPEREEL=\$(TYPEREEL) \$(FOPTS) \$(F90FLAGS) \$(OPT) \$(INCLUDES) -c ./$sub.$typ -o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o\n" ;
    }
  } else {
    if ($typ eq 'c') {
      # print STDOUT "$typ has ldcmpl=$ldcmpl{$typ}, opts=$ldopts{$typ}\n" ;
      if (exists $ldcmpl{$typ}) {
	print LSTDEP "\t\$($ldcmpl{$typ}) $ldopts{$typ} -c ./$sub.$typ -o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o \n" ;
      } else {
	print LSTDEP "\t\$(CC) \$(CFLAGS) \$(INCLUDES) -c ./$sub.$typ -o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o\n" ;
      }
    } elsif (exists $ldcmpl{$typ}) {
      print LSTDEP "\t\$($ldcmpl{$typ}) $ldopts{$typ} -c ./$sub.$typ -o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o \n" ;
    } else {
      print LSTDEP "\t\$(F90) -DTYPEREEL=\$(TYPEREEL) \$(FOPTS) \$(F90FLAGS) \$(OPT) \$(INCLUDES) -c ./$sub.$typ -o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o\n" ;
    }
  }
  # Linking rule:
  if ($typ eq 'c') {
    # When C calls fortran, set the compilation rule as followed
    if (%{$calls{$sub}}) {
      print LSTDEP "\t\$(LDC) \$(OPL) -o \$(BIN)a.$sub\$(OUT)\$(TYPEREEL)_\$(HOSTARCH) \$(BIBLIO) ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/Version.o \$(OBJC) \$(OBJF) \$(OBJf90) \$(OBJF90) \$(CEAUSER)\n" ;
    } else {
      print LSTDEP "\t\$(LDC) \$(OPL) -o \$(BIN)a.$sub\$(OUT)\$(TYPEREEL)_\$(HOSTARCH) \$(BIBLIO) ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o \$(CEAUSER) -nofor_main\n" ;
    }
  } else {
    if (%{$uses{$sub}}) {
      if ($opt_f || $type eq 'c') {
	print LSTDEP "\t\$(LD) \$(OPL) -o \$(BIN)a.$sub\$(OUT)\$(TYPEREEL)_\$(HOSTARCH) \$(BIBLIO) ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/Version.o" ;
	if (1 || %{$uses{$sub}}) {
	  print LSTDEP " \$(myOBJF) \$(myOBJs90) \$(myOBJC) \$(myOBJCPP)" ;
	}
	$lib1 = ' ./objets$(OUT)$(TYPEREEL)/$(HOSTARCH)/' ;
	$lib2 = '' ;
	&printrUses($sub) ;
	print LSTDEP " \$(CEAUSER)\n" ;
	$lib1 = '$(LIB)(' ;
	$lib2 = ')' ;
	$lib1 = ' ./objets$(OUT)$(TYPEREEL)/$(HOSTARCH)/' ;
	$lib2 = '' ;
      } else {
	print LSTDEP "\t\$(LD) \$(OPL) -o \$(BIN)a.$sub\$(OUT)\$(TYPEREEL)_\$(HOSTARCH) \$(BIBLIO) ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/Version.o \$(LIB) \$(CEAUSER)\n" ;
	print LSTDEP "\t#\$(LD) \$(OPL) -o \$(BIN)a.$sub\$(OUT)\$(TYPEREEL)_\$(HOSTARCH) \$(BIBLIO) ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/Version.o \$(myOBJF) \$(myOBJs90) \$(myOBJC) \$(myOBJf90) \$(myOBJF90) \$(CEAUSER)\n" ;
      }
    } else {
      print LSTDEP "\t\$(LD) \$(OPL) -o \$(BIN)a.$sub\$(OUT)\$(TYPEREEL)_\$(HOSTARCH) \$(BIBLIO) ./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$sub.o \$(CEAUSER)\n" ;
    }
  }
  print LSTDEP "\t\@echo \"OK, Generated \$(BIN)a.$sub\$(OUT)\$(TYPEREEL)_\$(HOSTARCH)\"\n" ;
}
############################################################################
# Tries to guess how to compile main program:
sub getRules {
  my $make = shift ;
  my $section = "" ;
  my (%compilers) = ( 'F77' => 'f', 'F90' => 'F90', 'CC' => 'c',) ;
  if (open(MK, "<$make")) {
    while(<MK>) {
      if ($section) {
	if (/\s+\$\((\w+)\)\s+(.*)/) {
	  if (exists $compilers{$1}) {
	    # Another rule may be built:
	    print STDOUT "\tNew rule($.) $section compiler $1 is $2\n" if $debug ;
	    $ldcmpl{$section} = $1 ;
	    $ldopts{$section} = $2 ;
	    if ($ldopts{$section} =~ /(.*)\s+\-c\s+(.*?)\s+\-o\s+.*\/(.*?)\s+(.*)$/) {
	      $ldopts{$section} = $1 . $4 ;
	    } elsif ($ldopts{$section} =~ /(.*)\s+\-c\s+(.*?)\s+\-o\s+.*\/(.*)$/) {
	      $ldopts{$section} = $1 ;
	    }
	  }
	}
      }
      if (/^\.(\w+)\.o\s*:\s*$/) {
	print STDOUT "New section $1\n" if $debug ;
	$section = $1 ;
      } elsif (/^\.\w+\.\w+\s*:\s*$/) {
	$section = "" ;
      } elsif (/^\w+\s*:\s*$/) {
	$section = "" ;
      }
    }
    close(MK) ;
  }
}
############################################################################
# Prints the dependency rule for an f90 subroutine
sub printRoutine {
  my $sub = shift ;
  my $typ = shift ;
  if ($sub =~ /mpi/i) {
    print LSTDEP "$lib1$sub.o$lib2:" ;
  } else {
    print LSTDEP "$lib1$sub.o$lib2:" ;
  }
  if ($opt_d) {
    &printUses($sub, 1) ;
  } else {
    &printUses($sub, 0) ;
  }
  print LSTDEP " $sub.$typ @{$dependencies{$sub}}\n" ;
}
############################################################################
# Prints dependencies for an f90 routine or program
sub printUses {
  my $sub = shift ;
  my $depend = shift ;
  if ($depend) {
    (%writens) = () ;
    foreach $module (sort bylevel keys %{$uses{$sub}}) {
      $dep = $modules{$module} ;
      print "File for module='$module' is $dep.\n" if ($depend == 2);
      unless (exists $writens{$dep}) {
	$writens{$dep} = 1 ;
	next if ($dep eq "Version") ;
	if ($dep =~ /mpi/i) {
	  print LSTDEP ' ' . $lib1 . $dep . '.o' . $lib2;
	} else {
	  print LSTDEP ' ' . $lib1 . $dep . '.o' . $lib2;
	}
      }
    }
  }
}
############################################################################
# Recursively builds the stuff:
sub setUses {
  my $sub = shift ;
  my $level = shift ;
  my $space = shift ;
  print "$space Analyzing $sub dependencies:\n" if $debug;
  foreach my $module (keys %{$uses{$sub}}) {
    my $dep = $modules{$module} ;
    unless (exists $writens{$dep}) {
      print "$space File for module='$module'($sub) is $dep.\n" if $debug;
      # $outline .= ' $(LIB)(' . $dep . '.o) ' . "@{$dependencies{$dep}}" ;
      $outline .= ' ' . $lib1 . $dep . '.o' . $lib2 . ' ' ;
      $writens{$dep} = ' ' . $lib1 . $dep . '.o' . $lib2 . ' ' ;
      if (%{$uses{$dep}}) {
	print STDOUT "$space further analysis for $dep\n" if $debug ;
	&setUses($dep, ($level+1), "$space  ")  ;
	push(@orders, $dep) ;
      } else {
	# Already performed stuff:
	print STDOUT "$space Storing $dep\n" if $debug ;
	push(@orders, $dep) ;
      }
    }
  }
  print "$space Performed $sub dependencies:\n" if $debug;
}
############################################################################
# Prints recursive dependencies for an f90 routine or program:
sub printrUses {
  my $sub = shift ;
  my $space = "  " ;
  local(%writens) = () ;
  local($outline) = " " ;
  local(@orders) = () ;
  # Case of C call... n,ot very clean manner:
  if (exists $calls{$sub}) {
    foreach my $lsub (keys %{$calls{$sub}}) {
      &setUses($lsub, 1, "$space") ;
    }
    foreach my $dep (@orders) {
      next if ($dep eq "Version") ;
      print LSTDEP " $writens{$dep}" ;
      print STDOUT " $writens{$dep}" if $debug ;
    }
    # Not very safe stuff, based on the fact the called routine MUST
    # be defined in a file with same name !!!!
    foreach my $lsub (keys %{$calls{$sub}}) {
      print LSTDEP "  $lib1$lsub.o$lib2" ;
    }
  } else {
    &setUses($sub, 0, "$space") ;
    print STDOUT "Final dep $sub: @orders\n" if $debug ;
#  local(%dones) = () ;
#  foreach my $dep (sort glolevel trySort keys %writens) {
    foreach my $dep (@orders) {
      next if ($dep eq "Version") ;
      print LSTDEP " $writens{$dep}" ;
      print STDOUT " $writens{$dep}" if $debug ;
    }
  }
  print STDOUT "\nRecursive dep is $outline\n" if $debug ;
}
############################################################################
# A spcial max format
sub max {
  my $max = shift(@_);
  foreach my $foo (@_) {
    my $b1 = $max ;
    my $a1 = $foo ;
    my $b2 = 0 ;
    my $a2 = 0 ;
    if ($max =~ /V(\d+)\_(\d+)/) {
      $b1 = $1 ; $b2 = $2 ;
    }
    if ($foo =~ /V(\d+)\_(\d+)/) {
      $a1 = $1 ; $a2 = $2 ;
    }
    if ($a1 > $b1) {
      $max = $foo ;
    } elsif ($a1 == $b1) {
      if ($a2 > $b2) {
	$max = $foo ;
      }
    }
  }
  return $max;
}
############################################################################
# Takes the most recent date of change
sub mergeDates {
  my $a = shift ;
  my $b = shift ;
  my $r = "" ;
  my (@months) = ("Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec") ;
  # Date format is Tue Jun 20 12:14:06 2006
  #                0   1   2   3        4
  my (@as) = split(/\s+/, $a) ;
  my (@bs) = split(/\s+/, $b) ;
  # Year compare:
  if ($bs[4] > $as[4]) {
    $r = $b ;
  } elsif ($bs[4] < $as[4]) {
    $r = $a ;
  } else {
    for (my $i=11; $i>=0; $i--) {
      if ($bs[1] eq $months[$i] || $as[1] eq $months[$i]) {
	if ($bs[1] eq $months[$i] && $as[1] eq $months[$i]) {
	  # Both months equal, look for day:
	  if ($bs[2] > $as[2]) {
	    $r = $b ;
	  } elsif ($bs[2] < $as[2]) {
	    $r = $a ;
	  } else {
	    # Look for time, hours:
	    my (@ad) = split(/:/, $as[3]) ;
	    my (@bd) = split(/:/, $bs[3]) ;
	    if ($bd[0] > $ad[0]) {
	      $r = $b ;
	    } elsif ($bd[0] < $ad[0]) {
	      $r = $a ;
	    } else {
	      # Look for minutes:
	      if ($bd[1] > $ad[1]) {
		$r = $b ;
	      } elsif ($bd[1] < $ad[1]) {
		$r = $a ;
	      } else {
		# Look for seconds:
		if ($bd[2] > $ad[2]) {
		  $r = $b ;
		} elsif ($bd[2] < $ad[2]) {
		  $r = $a ;
		} else {
		  # Times are absolutely equal
		  $r = $b ;
		}
	      }
	    }
	  }
	} elsif ($bs[1] eq $months[$i]) {
	  $r = $b ;
	} elsif ($as[1] eq $months[$i]) {
	  $r = $a ;
	}
	last ;
      }
    }
  }
#  print STDOUT "cmp $a with $b is $r\n";
  return($r) ;
}
############################################################################
# Guesses the current CVS version:
sub getVersion {
  my $version = "" ;
  my $tag = "" ;
  my $date = "" ;
  my(@versions) = () ;
  if (open(CVS,"CVS/Entries")) {
    while(<CVS>) {
      next unless (/\//) ;
      my (@a) = split(/\//, $_ ) ;
      my $fic = $a[1] ;
      my $ver = $a[2] ;
      my $d = $a[3] ;
      if ($date) {
	$date = mergeDates($d, $date) ;
      } else {
	$date = $d ;
      }
      my $t = $a[5] ;
      if ($t) {
	if ($t =~ /T(.*)/) {
	  $t = $1 ;
	  if ($tag) {
	    $tag = max($tag, $t) ;
	  } else {
	    $tag = $t ;
	  }
	}
      }
      if ($ver) {
	my (@b) = split(/\./, $ver) ;
	my $l = 0 ;
	foreach my $s (@b) {
	  $versions[$l] += $s ;
	  $l++ ;
	}
      }
    }
    close(CVS) ;
    if (@versions) {
      $version = join("_", @versions) ;
      print STDOUT "Versions = $version\n" if $debug ;
    }
  } else {
    print STDERR "No CVS version\n" ;
  }
  return($version,$date,$tag)
}
############################################################################
sub getList {
  if ($opt_s) {
    (@liste) = ("$opt_s") ;
  } else {
    (@liste) = grep { !/#\./ && -f $_ } readdir(DIR) ;
  }
}
############################################################################
sub printAll {
  open(LSTOBJ,">$dir/mkobj.lst") or die "Write permission $dir/mkobj.lst denied" ;
  open(LSTDEP,">$dir/mkdep.lst") or die "Write permission $dir/mkdep.lst denied" ;
  print LSTOBJ "VERSION = $version\n" ;
  print LSTOBJ "DATE = $date\n" ;
  print LSTOBJ "TAG = $tag\n" ;
  if (@listecpp) {
    print LSTOBJ "OBJCPP = " ;
    foreach (@listecpp) {
      next if ($isamain{$_}) ;
      print LSTOBJ "$lib1$_.o$lib2" ;
    }
    print LSTOBJ "\n" ;
    print LSTOBJ "myOBJCPP = " ;
    foreach (@listecpp) {
      next if ($isamain{$_}) ;
      print LSTOBJ "./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$_.o " ;
    }
    print LSTOBJ "\n" ;
    foreach (@listecpp) {
      next if ($isamain{$_}) ;
      print LSTDEP "$lib1$_.o$lib2: $_.cpp @{$dependencies{$_}} \n" ;
    }
  }
  if (@listec) {
    print LSTOBJ "OBJC = " ;
    foreach (@listec) {
      next if ($isamain{$_}) ;
      print LSTOBJ "$lib1$_.o$lib2 " ;
    }
    print LSTOBJ "\n" ;
    print LSTOBJ "myOBJC = " ;
    foreach (@listec) {
      next if ($isamain{$_}) ;
      print LSTOBJ "./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$_.o " ;
    }
    print LSTOBJ "\n" ;
    foreach (@listec) {
      next if ($isamain{$_}) ;
      print LSTDEP "$lib1$_.o$lib2: $_.c @{$dependencies{$_}} \n" ;
    }
  }
  if (@listef) {
    print LSTOBJ "OBJF = " ;
    foreach (@listef) {
      next if ($isamain{$_}) ;
      print LSTOBJ "$lib1$_.o$lib2" ;
    }
    print LSTOBJ "\n" ;
    print LSTOBJ "myOBJF = " ;
    foreach (@listef) {
      next if ($isamain{$_}) ;
      print LSTOBJ "./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$_.o " ;
    }
    print LSTOBJ "\n" ;
    foreach (@listef) {
      next if ($isamain{$_}) ;
      print LSTDEP "$lib1$_.o$lib2: $_.f @{$dependencies{$_}} \n" ;
    }
  }
  if (@listef90) {
    print LSTOBJ "\nOBJf90 = " ;
    foreach (@lnouse90,@orders) {
      next if ($types{$_} ne 'f90') ;
      if ($isamain{$_}) {
	print STDOUT "Main $_\n" if ($debug) ;
      } else {
	print LSTOBJ "$lib1$_.o$lib2 " ;
      }
    }
    print LSTOBJ "\n" ;
    print LSTOBJ "myOBJf90 = " ;
    foreach (@lnouse90,@orders) {
      next if ($types{$_} ne 'f90') ;
      unless ($isamain{$_}) {
	print LSTOBJ "./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$_.o " ;
      }
    }
    print LSTOBJ "\n" ;
  }
  if (@listeF90) {
    print LSTOBJ "\nOBJF90 = " ;
    foreach (@lnouse90,@orders) {
      next if ($types{$_} ne 'F90') ;
      next if ($_ eq "Version") ;
      if ($isamain{$_}) {
	print STDOUT "Main $_\n" if ($debug) ;
      } else {
	print LSTOBJ "$lib1$_.o$lib2 " ;
      }
    }
    print LSTOBJ "\n" ;
    print LSTOBJ "myOBJF90 = " ;
    foreach (@lnouse90,@orders) {
      next if ($types{$_} ne 'F90') ;
      next if ($_ eq "Version") ;
      unless ($isamain{$_}) {
	print LSTOBJ "./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$_.o " ;
      }
    }
    print LSTOBJ "\n" ;
  }
  if (@listeall90) {
    print LSTOBJ "\nOBJs90 = " ;
    foreach (@lnouse90) {
      next if ($isamain{$_}) ;
      next if (%{$uses{$_}}) ;
      print LSTOBJ "$lib1$_.o$lib2 " ;
    }
    print LSTOBJ "\n" ;
    print LSTOBJ "myOBJs90 = " ;
    foreach (@lnouse90) {
      next if ($isamain{$_}) ;
      next if (%{$uses{$_}}) ;
      print LSTOBJ "./objets\$(OUT)\$(TYPEREEL)/\$(HOSTARCH)/$_.o " ;
    }
    print LSTOBJ "\n" ;
  }
  if (@listeall90) {
    print LSTDEP "\n" ;
    # First prints routine dependencies,
    foreach (@lnouse90, @orders) {
      next if ($_ eq "Version") ;
      unless ($isamain{$_}) {
	&printRoutine($_, $types{$_}) ;
      }
    }
    # then main dependencies
    foreach (@lmains90) {
      if ($isamain{$_}) {
	&printMain($_, $types{$_}) ;
      }
    }
  }
  if (@listec || @listecpp) {
    # main dependencies for C and C++ programs:
    foreach (@listec,@listecpp) {
      next unless $_ ;
      if ($isamain{$_}) {
	&printMain($_, $types{$_}) ;
      }
    }
  }
  # Closes the target dependencies files
  close(LSTOBJ) ;
  close(LSTDEP) ;
}
############################################################################
# Main loop on all directories
sub mainRout {
  foreach $dir (@dirs) {
    unless (opendir(DIR,".")) {
      print STDERR "Read permission denied on directory $dir\n" ;
      next ;
    }
    # Gets the actual disk location for includes from current local directory
    (%getrel) = () ;
    if (@rincludes) {
      foreach my $i (@rincludes) {
	# Gets a location that exists on disk from current directory.
	$getrel{$i} = &getAbs($dir,$i) ;
      }
    }
    # Gets the directory's file list
    &getList() ;
    closedir(DIR) ;
    # Gets the list of possible makefiles:
    (@listemkf) = sort grep /makefile/i, @liste ;
    # Looking for makefile rules:
    foreach my $file (@listemkf) {
      print STDOUT "Read processing rules for $file\n" if $debug ;
      &getRules("$dir/$file") ;
    }
    ($version,$date,$tag) = &getVersion() ;
    (%isamain) = () ;
    (%dones) = () ;
    (%types) = () ;

    (@listecpp) = sort grep /\.cpp$/, @liste ;
    foreach my $file (@listecpp) {
      $file =~ s/\.cpp$// ;
      $types{$file} = 'cpp' ;
      &getDependc("$dir/$file.cpp", $file, $dir) ;
    }
    (@listec) = sort grep /\.c$/, @liste ;
    foreach my $file (@listec) {
      $file =~ s/\.c$// ;
      $types{$file} = 'c' ;
      &getDependc("$dir/$file.c", $file, $dir) ;
    }
    (@listef) = grep /\.f$/i, @liste ;
    foreach my $file (@listef) {
      if ($file =~ /^(.*)\.(f)$/i) {
	$types{$1} = $2 ;
      }
      $file =~ s/\.f$// ;
      &getDependf("$dir/$file.f", $file, $dir) ;
    }
    (@listef90) = sort grep /\.f90$/, @liste ;
    (@listeF90) = sort grep /\.F90$/, @liste ;
    foreach my $file (@listef90) {
      $file =~ s/\.f90$// ;
      $types{$file} = 'f90' ;
      &getDependf90("$dir/$file.f90", $file, $dir) ;
    }
    foreach my $file (@listeF90) {
      $file =~ s/\.F90$// ;
      $types{$file} = 'F90' ;
      &getDependf90("$dir/$file.F90", $file, $dir) ;
    }
    (@listeall90) = sort glolevel &trySort(@listef90, @listeF90); 
    # List of main programs in fortran 90:
    (@lmains90) = () ;
    # List of not used programs in fortran 90
    (@lnouse90) = () ;
    # List of dependencies:
    (@orders) = () ;
    &sortallf90(@listef90, @listeF90) ;
    print STDOUT "lnouse90=@lnouse90\n" if $debug ;
    print STDOUT "lmains90=@lmains90\n" if $debug ;
    print STDOUT "orders=@orders\n" if $debug ;
    # Open the target files for writing
    unless ($opt_b) {
      &printAll() ;
    }
  }
}
############################################################################
&mainRout() ;
############################################################################
1;
