#!/usr/bin/perl
#%########################################################
# Purpose: present a javascript index to a table of contents.
#%########################################################
use Getopt::Std;
use File::Copy ;
getopts('v:d');
$| = 1; # flush stdout with every print -- gives better feedback during
        # long computations
if (0) {
    $formulaire{'file'} = 'new_guide' ;
} else {
    $formulaire{'file'} = @ARGV[0] ;
}
$debug = 0 ;
$debug = 1 if $opt_d ;
#%########################################################
sub GetLevel {
    local($num) = @_ ;
    @lines = split(/\./,$num) ;
    $line = 1+$#lines ;
    return($line) ;
}
#%########################################################
@depth = ("part", "chapter", "section", "subsection", "subsubsection") ;
sub AddLevel {
    my($type,$previoussectionid) = @_ ;
    @lines = split(/\./,$previoussectionid) ;
    for ($i=0;$<=$#depth;$i++) {
	if ($type eq $depth[$i]) {
	    last ;
	}
    }
    $lines[$i]++ ;
    $sectionid = join('.', @lines[0..$i]);
    print "previoussectionid=$previoussectionid becomes $sectionid\n";
    return($sectionid,$i)
}
#%########################################################
sub ReadTocOld {
    local($file) = @_ ;
    $repertoire = "$file.html" ;
    open(FILE,"<$file.toc") or die "Toc file $file.toc must exist" ;
# Définit le int niveau de la section
    $uplevel = -1 ;
    $previouslevel = -1 ;
# Définit le string complet identifiant de la section
    $upsectionid = -1 ;
    $previoussectionid = -1 ;
    $father{$previoussectionid} = -1 ;
    while(<FILE>) {
	print STDOUT "Read $_" if $debug ;
	if (/contentsline\s*\{(part|chapter|section|subsection|subsubsection)\}\{(.*)\}\{\d+\}/) {
	    $type = $1 ;
	    $text = $2 ;
	    $ok = 0 ; 
	    print STDOUT "type=$type, text=$text\n" if $debug ;
	    if ($text =~ /\\numberline\s*\{([A-Z\d\.]+)\}/) {
		$sectionid = $1 ;
		$level = &GetLevel($sectionid) ;
		print STDOUT "level=$level\n" if $debug ;
		$ok = 1 ; 
	    } else {
		next ;
		($sectionid,$level) = &AddLevel($type,$previoussectionid) ;
		print STDOUT "(sectionid=$sectionid,level=$level)\n" if $debug;		
	    }
	    if ($level == $previouslevel) { # Même niveau qu'avant
		$father{$sectionid} = "$father{$previoussectionid}" ;
		push(@{$fic{"$father{$sectionid}"}},$sectionid) ;
		print STDOUT "sl father($sectionid)=$father{$sectionid}\n" if ($opt_v >=5) ;
	    } elsif ($level > $previouslevel) { # On descend d'un niveau
		$father{$sectionid} = "$previoussectionid" ;
		push(@{$fic{"$father{$sectionid}"}},$sectionid) ;
		print STDOUT "dl father($sectionid)=$father{$sectionid}\n" if ($opt_v >=5) ;
		$uplevel = $level ;
		print "upsectionid=$upsectionid " if ($opt_v >=5) ;
		$upsectionid = "$sectionid" ;
		print "becomes $upsectionid\n" if ($opt_v >=5) ;
	    } elsif ($level < $previouslevel) { # On monte d'un niveau
		$loop = 10 ;
		print "Up one or more levels:($sectionid-$level)[$father{$upsectionid}]\n" if ($opt_v >=5) ;
		while (&GetLevel($father{$upsectionid}) >= $level) {
		    $upsectionid = "$father{$upsectionid}" ;
		    $upsectionid = -1 unless defined $upsectionid ;
		    last unless ($loop--) ;
		}
		if ($loop) {
		    $father{$sectionid}  = "$father{$upsectionid}" ;
		    $father{$sectionid} = -1 unless defined $father{$sectionid} ;
		    $upsectionid = "$father{$sectionid}" ;
		    $upsectionid = -1 unless defined $upsectionid ;
		    $uplevel = $level ;
		    print "ul father($sectionid) = $father{$sectionid}\n" if ($opt_v >=5) ;
		    push(@{$fic{"$father{$sectionid}"}},$sectionid) ;
		}
	    }
	    $previoussectionid = $sectionid ;
	    $previouslevel = $level ;
	}
    }
}
#%########################################################
sub GetHtmlTitle {
    my($ftitle) = "" ;
    if (open(FILE,"<$_[0]")) {
	$ok = 0 ;
	$ftitle = "" ;
	while(<FILE>) {
	    if (/<title>(.*?)<\/title>/i) {
		$ftitle = $1 ;
		last ;
	    }
	    if (/^(.*?)<\/title>/i) {
		$ftitle .= $1 ;
		last ;
	    }
	    if ($ok) {
		chop($_);
		$ftitle .= $_ ;
	    }
	    if (/<title>(.*)/i) {
		$ftitle = $1 ;
		$ok = 1 ;
	    }
	}
	close(FILE) ;
	if ($ftitle =~ /^\s*(.*?)\s*$/) {
	    $ftitle = $1 ;
	}
	$ftitle =~ s/\"/\'/g; # "
	$ftitle =~ tr/
/ /;
	return($ftitle) ;
    }
}

#%########################################################
sub WriteTarget {
    $i = 0 ;
    $root = "lah1000:82";
    $root = "..";
    unless (open(TARGET,">$formulaire{'file'}/tree_items.js")) {
	print STDERR "Impossible d'écrire photosturism data base.";
	exit(1) ;
    }
    if ($opt_v >= 4) {
	foreach $file (keys %fic) {
	    print STDOUT "fic($file) = @{$fic{$file}}\n";
	}
    }
    $gentitle = &GetHtmlTitle("$formulaire{'file'}/$formulaire{'file'}.html") ;
    #print TARGET "foldersTree = gFld(\"<i>$gentitle</i>\",\"$formulaire{'file'}.html\")\n";
    print TARGET "var TREE_ITEMS = [\n\t[\"<i>$gentitle</i>\", \"$formulaire{'file'}.html\", \n";
    foreach $s0rep (@rootliste) {
	next unless $s0rep ;
	print TARGET &InsertFolder("aux1","foldersTree","$s0rep") ;
#	print TARGET "aux1 = insFld(foldersTree,gFld(\"$s0rep\",\"$s0rep\"))\n";
	$i = 0 ;
	print STDOUT "\tLevel root for $s0rep\n" if $debug;
	foreach $s1rep (@{$fic{"$s0rep"}}) {
	    next unless $s1rep ;
	    print STDOUT "\tLevel 1 for $s1rep\n" if $debug;
	    print TARGET &InsertFolder("aux2","aux1","$s1rep") ;
	    $i = 0 ;
	    foreach $s2rep (@{$fic{"$s1rep"}}) {
		next unless $s2rep ;
		print STDOUT "\tLevel 2 for $s2rep\n" if $debug;
		print TARGET &InsertFolder("aux3","aux2","$s2rep") ;
		$i = 0 ;
		foreach $s3rep (@{$fic{"$s2rep"}}) {
		    next unless $s3rep ;
		    print STDOUT "\tLevel 3 for $s3rep\n" if $debug;
		    print TARGET &InsertFolder("aux4","aux3","$s2rep") ;
		    print TARGET "],\n" ;
		}
		print TARGET "],\n" ;
	    }
	    print TARGET "],\n" ;
	}
	print TARGET "],\n" ;
    }
    print TARGET "]\n" ;
    print TARGET "];\n" ;
    close(TARGET);
}
#%########################################################
sub InsertFolder {
    my($f,$w,$sectionid) = @_ ;
    my($file) = "" ;
    if (exists $fic_file{$sectionid}) {
	if (exists $fic_section{$sectionid}) {
	    return("\n\t[\"$sectionid) ".&cleanHtml("$fic_titre{$sectionid}")."\", \"$fic_file{$sectionid}#$fic_section{$sectionid}\", ");
	} else {
	    return("\n\t[\"$sectionid) ".&cleanHtml("$fic_titre{$sectionid}")."\", \"$fic_file{$sectionid}\", ");
	}
    } else {
	return("\n\t[\"$sectionid\", null, ");
    }
}
#%########################################################
sub cleanHtml {
    my $var = shift ;
    $var =~ s/\</\&lt\;/g ;
    $var =~ s/\>/\&gt\;/g ;
    return($var) ;
}
	
#%########################################################
sub numerically {
    local(@x) = split(' ',$a);
    local(@y) = split(' ',$b);
    local($i, $result);
    for($i=0;$i<$#x;$i++) {
       last if ($result = ($x[$i] <=> $y[$i]));
    }
    $result
}
#%########################################################
# See latex2html...
sub ReadTocNew {
    local($dir) = "" ;
    local($delim) = '%:%' ;
    local(%section_info) = () ;
    local(@tmp_keys) = () ;
    require "$formulaire{'file'}/sections.pl";
    @tmp_keys = sort numerically keys %section_info;
    local($titreid) = "" ;
    local($sectionid) = "" ;
    my $key ;
    local $uplevel = -1 ;
    local $previouslevel = -1 ;
# Définit le string complet identifiant de la section
    local $upsectionid = -1 ;
    local $previoussectionid = -1 ;
    $father{$previoussectionid} = -1 ;
    while (@tmp_keys) {
	$key = shift @tmp_keys;
	my($depth, $file, $title, $body) = split($delim,$section_info{$key});
	if ($title =~ /^([A-Z\d\.]*)\s*(.*)/) {
	    $sectionid = $1 ;
	    $titreid = $2 ;
	    if ($sectionid =~ /(.*)\.$/) {
		$sectionid = $1 ;
	    }
	    $level = &GetLevel($sectionid) ;
	    &incrLevel() ;
	    $previoussectionid = $sectionid ;
	    $previouslevel = $level ;
	}
    }
}
#%########################################################
# See latex2html...
sub AnalyseTarget {
    local($dir) = "" ;
    local($delim) = '%:%' ;
    local(%section_info) = () ;
    local(@tmp_keys) = () ;
    require "$formulaire{'file'}/sections.pl";
    @tmp_keys = sort numerically keys %section_info;
    local($titreid) = "" ;
    local($sectionid) = "" ;
    my $key ;
    while (@tmp_keys) {
	$key = shift @tmp_keys;
	my($depth, $file, $title, $body) = split($delim,$section_info{$key});
	print STDOUT " $depth, $file, $title, $body\n" if $debug ;
	if ($title =~ /^([A-Z\d\.]*)\s*(.*)/) {
	    $sectionid = $1 ;
	    $titreid = $2 ;
	    if ($sectionid =~ /(.*)\.$/) {
		$sectionid = $1 ;
	    }
	    $level = &GetLevel($sectionid) ;
	    if ($sectionid) {
		$fic_titre{"$sectionid"} = "$titreid" ;
		$fic_file{"$sectionid"} = "$file" ;
		local($fic) = "$file" ;
		print "GetSubSection: $sectionid t='$titreid'\n" if $debug ;
		&GetSubSection() ;
	    }
	}
    }
}
#%########################################################
sub incrLevel {
    if ($level == $previouslevel) { # Même niveau qu'avant
	$father{$sectionid} = "$father{$previoussectionid}" ;
	push(@{$fic{"$father{$sectionid}"}},$sectionid) ;
	print STDOUT "sl father($sectionid)=$father{$sectionid}\n" if ($opt_v >=5) ;
    } elsif ($level > $previouslevel) { # On descend d'un niveau
	$father{$sectionid} = "$previoussectionid" ;
	push(@{$fic{"$father{$sectionid}"}},$sectionid) ;
	print STDOUT "dl father($sectionid)=$father{$sectionid}\n" if ($opt_v >=5) ;
	$uplevel = $level ;
	print "upsectionid=$upsectionid " if ($opt_v >=5) ;
	$upsectionid = "$sectionid" ;
	print "becomes $upsectionid\n" if ($opt_v >=5) ;
    } elsif ($level < $previouslevel) { # On monte d'un niveau
	$loop = 10 ;
	print "Up one or more levels:($sectionid-$level)[$father{$upsectionid}]\n" if ($opt_v >=5) ;
	while (&GetLevel($father{$upsectionid}) >= $level) {
	    $upsectionid = "$father{$upsectionid}" ;
	    $upsectionid = -1 unless defined $upsectionid ;
	    last unless ($loop--) ;
	}
	if ($loop) {
	    $father{$sectionid}  = "$father{$upsectionid}" ;
	    $father{$sectionid} = -1 unless defined $father{$sectionid} ;
	    $upsectionid = "$father{$sectionid}" ;
	    $upsectionid = -1 unless defined $upsectionid ;
	    $uplevel = $level ;
	    print "ul father($sectionid) = $father{$sectionid}\n" if ($opt_v >=5) ;
	    push(@{$fic{"$father{$sectionid}"}},$sectionid) ;
	}
    }
}
#%########################################################
sub GetSubSection {
    if (open(HTML,"<$formulaire{'file'}/$fic")) {
	$ul = 0 ;
	while(<HTML>) {
	    if ((/<A NAME="CHILD_LINKS">/) || (/<!--Table of Contents-->/)) {
		$ul = 1 ;
		print "\tOn for $sectionid file $fic: \n\t$titreid\n" if $debug ;
	    }
	    if (($ul) && ((/<!--End of Table of Child-Links-->/) || (/<!--End of Table of Contents-->/))) {
		$ul = 0 ;
		last ;
	    }
	    if (($ul) && (/HREF="([\w\.]*?)#(\w*)">([A-Z\d\.]*\d+)\.*? (.*?)<\/A>/)) {
		if (($3 eq $sectionid) && ($1 eq $fic) && ($4 eq $titreid)) {
		    $fic_section{"$sectionid"} = $2 ;
		    print "\tsub_sectionid($sectionid)=$2\n" if $debug ;
#		} else {
#		    print "\tnosub($_)\n" if $debug;
		}
	    }
	}
	close(HTML) ;
    }
}
#%########################################################
sub WriteEnvir {
    open(INDEX,">$formulaire{'file'}/anindex.html") or die "index MAIN non inscriptible";
    print INDEX "<html> <head>
<title>
$gentitle
</title>
</head>
<FRAMESET cols=\"300,*\"> 
  <frame SRC=\"tree.html\" NAME=\"index\">
  <frame SRC=\"$formulaire{'file'}.html\" name=\"source\">
</FRAMESET> 
</html>\n";
    close(INDEX) ;
    close(INDEX) ;
    open(INDEX,">$formulaire{'file'}/tree.html") or die "index LF non inscriptible";
    print INDEX "<html>
<head>
<style>
	a, A:link, a:visited, a:active, A:hover
		{color: #000000; text-decoration: none; font-family: Tahoma, Verdana; font-size: 12px}
</style>
</head>
<body bottommargin=\"0\" topmargin=\"0\" leftmargin=\"0\" rightmargin=\"0\" marginheight=\"0\" marginwidth=\"0\" bgcolor=\"white\">
<script language=\"JavaScript\" src=\"tree.js\"></script>
<script language=\"JavaScript\" src=\"tree_items.js\"></script>
<script language=\"JavaScript\" src=\"tree_tpl.js\"></script>
<table cellpadding=\"5\" cellspacing=\"0\" cellpadding=\"10\" border=\"0\" width=\"100%\">
<tr>
	<td>
	<script language=\"JavaScript\">
		new tree (TREE_ITEMS, TREE_TPL);
	</script>
	</td>
</tr>
</table>
</body>
</html>" ;
    close(INDEX) ;
}
#%########################################################    
sub WriteSystem {
    @liste = ("tree_tpl.js" ,"tree.js", "empty.gif", "line.gif", "base.gif", "folder.gif", "folderopen.gif", "page.gif", "joinbottom.gif", "join.gif", "plusbottom.gif", "plus.gif", "minusbottom.gif", "minusbottom.gif");
    if (exists $ENV{JAVATEX_DIR}) {
	$src = "$ENV{JAVATEX_DIR}/lib/perl/javatex" ;
    } elsif (exists $ENV{FTAGSHTML_DIR}) {
	$src = "$ENV{FTAGSHTML_DIR}/lib/perl/javatex" ;
    } elsif (-d "/usr/share/ftagshtml") {
	$src = "/usr/share/ftagshtml/lib/perl/javatex" ;
    } elsif (exists $ENV{CADMIC_dev}) {
	$src = "$ENV{CADMIC_dev}/lib/javatex" ;
    } elsif (exists $ENV{CADMIC_exploit}) {
	$src = "$ENV{CADMIC_exploit}/lib/javatex" ;
    }
    unless (-d "$src") {
	$src = "/net/lah220/usr/local/cessenat/lib/javatex" ;
    }
    print STDOUT "src=$src\n";
    foreach $file (@liste) {
	# print STDOUT "copy $src/$file to $formulaire{'file'}/$file\n" ;
	copy("$src/$file", "$formulaire{'file'}/$file");
    }
}
#%########################################################  
&ReadTocOld("$formulaire{'file'}") ;
(@rootliste) = () ;
foreach $fic (@{$fic{"-1"}}) {
    print STDOUT "rootliste extended with $fic" if $debug ;
    push(@rootliste,$fic) if ($fic ne "-1") ;
}
#print "Base root = @rootliste\n";
print STDOUT "Reading $formulaire{'file'}.toc file\n" if ($opt_v >= 1) ; 
&AnalyseTarget ;
print STDOUT "Writing $formulaire{'file'}/tree_items.js JavaScript DB File\n" if ($opt_v >= 1) ; 
&WriteTarget ;
print STDOUT "Writing Other Environment Files\n" if ($opt_v >= 1) ; 
&WriteEnvir ;
&WriteSystem ;
