package chtml ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A stuff container
###########################################################################
use strict;
use Carp ;
use CGI ;
use File::Basename ;
###########################################################################
use oref ;
use ihtml ;
use fthlabel ;
###########################################################################
use vars qw($debug $other $previous %MYOBJHASH) ;
$debug = 0 ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
sub getCurrent {
  if ($other && ref($other) eq 'chtml') {
    print STDOUT "file=".(caller(1))[1] . " rout=".(caller(1))[3] . " line=".(caller(1))[2] . "up=" . (caller(2))[3] ."\n" if $debug ;
    print STDOUT "prev(getCurrent)=$previous ". $other->get(-file)."\n" if $debug ;
    return($other) ;
  } else {
    return(undef);
  }
}
###########################################################################
sub setCurrent {
  if ($other) {
    if ($previous eq '</COMMENT>') {
      push(@{$other->{"HTML"}->[$.]}, '</COMMENT>') ;
    }
  }
  $other = shift ;
  $previous = $other->{"previous"} ;
  print STDOUT "prev(setCurrent)=$previous\n" if $debug ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = shift ;
  if (exists $MYOBJHASH{$name}) {
    $other = $MYOBJHASH{$name} ;
    $previous = $other->{"previous"} ;
    return($other) ;
  } else {
    # We refuse to create an object from no where:
    my(@list) = (keys %MYOBJHASH);
    print STDOUT "list=@list\n" if $debug ;
    croak("Violent build for $name is illegal, API error") ;
  }
}
###########################################################################
sub targetname {
  my $file = shift ;
  $file = basename($file) ;
  if ($file =~ /(.*)\.(.*)/) {
    $file = $1 . '_' . $2 ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  if ($args{-name}) {
    $name = $args{-name} ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
  } else {
    $first = 1 ;
    bless $self ;
    $previous = "" ;
    $MYOBJHASH{$name} = $self if $name ;
  }
  if ($args{-file}) {
    $args{-file} = &targetname($args{-file}) ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  print STDOUT (caller(0))[3] .": new chtlml " . $self->get(-name) . " from " . (caller(1))[3] . "\n" if $debug ;
  if ($other) {
    $other->{"previous"} = $previous ;
  }
  $other = $self ;
  return $self ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
############################################################################
# Adds a stuff:
sub addLaCoM {
  my $self = shift ;
  my $value = shift ;
  if ($value) {
    my $number = $. ;
    push(@{$self->{"HTML"}->[$number]}, $value) ;
  }
}
##########################################################################
# Adds a stuff:
sub addLaTeX {
  my $self = shift ;
  my $value = shift ;
  if ($value) {
    my $number = $. ;
    print STDOUT (caller(0))[2] . "addLaTeX: Adding $value at $number\n" if $debug ;
    if ($value =~ /[\<\>]/) {
      $value =~ s/\</&lt;/g ;
      $value =~ s/\>/&gt;/g ;
    }
    if ($value =~ /\\[\`\'\^\"]/) {
      $value =~ s/\\\^([aeiouAEIOU])/&${1}circ;/g ;
      $value =~ s/\\\'([aeiouAEIOU])/&${1}acute;/g ;
      $value =~ s/\\\`([aeiouAEIOU])/&${1}grave;/g ;
      $value =~ s/\\\"([aeiouAEIOU])/&${1}uml;/g ;
      $value =~ s/\\\~([aeiouAEIOU])/&${1}tilde;/g ;
      $value =~ s/\{*\\\^\{\\i\}\}*/&icirc;/g ;
    }
    if ($value =~ /\\/) {
      $value =~ s/\{\}//g ;
      $value =~ s/\\hfill\\noindent//g ;
      $value =~ s/\\hfill\\\\\\noindent/\<br\>/g ;
      $value =~ s/\\item\[(.*?)\]/\<li\>\<U\>${1}\<\/U\>/g ;
      $value =~ s/\\item\s*/\<li\>/g ;
      $value =~ s/\\begin\{enumerate\}/\<ol\>/g ;
      $value =~ s/\\end\{enumerate\}/\<\/ol\>/g ;
      $value =~ s/\\begin\{description\}/\<ul\>/g ;
      $value =~ s/\\end\{description\}/\<\/ul\>/g ;
      $value =~ s/\\begin\{itemize\}/\<ul\>/g ;
      $value =~ s/\\end\{itemize\}/\<\/ul\>/g ;
      $value =~ s/\\chapter\{(.*?)\}/\<H1\>$1\<\/H1\>/g ;
      $value =~ s/\\section\{(.*?)\}/\<H2\>$1\<\/H2\>/g ;
      $value =~ s/\\subsection\{(.*?)\}/\<H3\>$1\<\/H3\>/g ;
      # Parse the reference to a given label:
      while($value =~ /(.*?)\\(ref|label)\{(.*?)\}(.*)/) {
	$value = $4 ;
	my $name = $3 ;
	push(@{$self->{"HTML"}->[$number]}, $1) if $1;
	if ($2 eq 'ref') {
	  my $refe = new oref(-name => $name, -type => "fthlabel") ;
	  push(@{$self->{"HTML"}->[$number]}, $refe );
	} elsif ($2 eq 'label') {
	  new fthlabel(-name => $name, -file => $self->{"file"}) ;
	  # Hard coded label look:
	  push(@{$self->{"HTML"}->[$number]}, "<LABEL><A Name=$name>$name</A></LABEL>" );
	}
      }
      $value =~ s/\\ref\{(.*?)\}/\<A Href=$1\>$1\<\/A\>/g ;
      $value =~ s/\\label\{(.*?)\}/\<A Name=$1\>$1\<\/A\>/g ;
    }
    if ($value =~ /^\s*(.)\1\1+\s*$/) {
      # Three times the same character at least and nothing else:
      push(@{$self->{"HTML"}->[$number]}, "<HR></HR>") ;
    } elsif ($value =~ /^\|(.)\1\1+$/) {
      # Three times the same character at least and nothing else:
      push(@{$self->{"HTML"}->[$number]}, "<HR></HR>") ;
    } elsif ($value =~ /^\|(.)\1{3,}\w$/) {
      # Three times the same character at least and then only a marker
      push(@{$self->{"HTML"}->[$number]}, "<HR></HR>") ;
    } else {
      push(@{$self->{"HTML"}->[$number]}, $value) ;
    }
  }
}
###########################################################################
# Adds a stuff protecting from HTML acccents:
sub addProtect {
  my $self = shift ;
  my $value = shift ;
  if ($value) {
    my $number = $. ;
    push(@{$self->{"HTML"}->[$number]}, $value) ;
  }
}
###########################################################################
# Finds a stuff
sub findDef {
  my $self = shift ;
  my $name = shift ;
  my $object = shift ;
  my $j = $. ;
  my $k ;
  my $ok = 0 ;
  for (my $i=$j; $i>=0; $i--) {
    $k = $#{$self->{"HTML"}->[$i]} ;
    for (my $l=$k; $l>=0; $l--) {
      if ($self->{"HTML"}->[$i]->[$l] =~ /(.*?)\b$name\b(.*)/) {
	for (my $x=$k; $x>$l; $x--) {
	  my $y = ( $x + 2 ) ;
	  $self->{"HTML"}->[$i]->[$y] = $self->{"HTML"}->[$i]->[$x] ;
	}
	$self->{"HTML"}->[$i]->[$l] = $1 ;
	$l++ ;
	$self->{"HTML"}->[$i]->[$l] = $object ;
	$k = $l ;
	$l++ ;
	$self->{"HTML"}->[$i]->[$l] = $2 ;
	$j = $i ;
	$ok = 1 ;
	last ;
      }
    }
    last if $ok ;
  }
  if ($ok) {
    # Now sets the arglist...
    for (my $l=$k+1; $l<=$#{$self->{"HTML"}->[$j]}; $l++) {
      if ($self->{"HTML"}->[$j]->[$l] =~ /(.*?)\((.*)/) {
	my $bef = $1 ; my $aft = $2 ;
	# Checks the ARGLIST stuff is not wrongly wrapped:
	if ($aft =~ /^\s*\<\/(\w+)\s*(.*?)\>/ && $bef =~ /\<($1)\s*(.*?)\>\s*$/i) {
	  # The separator is inside a marker, take care not to split it !
	  my $m = $1 ;
	  print STDOUT "Found OK1 split: bef='$bef', aft='$aft', m=$m\n" if $debug ;
	  if ($bef =~ /(.*)(\<$m.*?\>\s*)$/i) {
	    $self->{"HTML"}->[$j]->[$l] = "$1<ARGLIST>$2($aft" ;
	    print STDOUT "Yields:'" . $self->{"HTML"}->[$j]->[$l] . "'\n" if $debug ;
	  }
	  #if ($aft =~ /(\s*\<\/$m.*?\>)(.*)/) {
	  #  $self->{"HTML"}->[$j]->[$l] .= "$1<ARGLIST>($aft" ;
	  #}
	} else {
	  $self->{"HTML"}->[$j]->[$l] = "$bef<ARGLIST>($aft" ;
	}
	$k = $l ;
	$ok = 2 ;
	last ;
      }
    }
    if ($ok != 2) {
      for (my $i=$j+1; $i<=$.; $i++) {
	$k = $#{$self->{"HTML"}->[$i]} ;
	for (my $l=$k; $l>=0; $l--) {
	  if ($self->{"HTML"}->[$i]->[$l] =~ /(.*?)\((.*)/) {
	    my $bef = $1 ; my $aft = $2 ;
	    # Checks the ARGLIST stuff is not wrongly wrapped:
	    if ($aft =~ /^\s*\<\/(\w+)\s*(.*?)\>/ && $bef =~ /\<($1)\s*(.*?)\>\s*$/i) {
	      # The separator is inside a marker, take care not to split it !
	      my $m = $1 ;
	      print STDOUT "Found OK2 split: bef='$bef', aft='$aft', m=$m\n" if $debug ;
	      if ($bef =~ /(.*)(\<$m.*?\>\s*)$/i) {
		$self->{"HTML"}->[$i]->[$l] = "$1<ARGLIST>$2($aft" ;
		print STDOUT "Yields:'" . $self->{"HTML"}->[$j]->[$l] . "'\n" if $debug ;
	      }
	      #if ($aft =~ /(\s*\<\/$m.*?\>)(.*)/) {
	      #  $self->{"HTML"}->[$j]->[$l] .= "$1<ARGLIST>($aft" ;
	      #}
	    } else {
	      $self->{"HTML"}->[$i]->[$l] = "$bef<ARGLIST>($aft" ;
	    }
	    $j = $i;
	    $k = $l ;
	    $ok = 2 ;
	    last ;
	  }
	}
	last if ($ok == 2) ;
      }
    }
    if ($ok == 2) {
      for (my $l=$k+1; $l<=$#{$self->{"HTML"}->[$j]}; $l++) {
	if ($self->{"HTML"}->[$j]->[$l] =~ /(.*)\)(.*)/) {
	  my $bef = $1 ; my $aft = $2 ;
	  # Checks the ARGLIST stuff is not wrongly wrapped:
	  if ($aft =~ /^\s*\<\/(\w+)\s*(.*?)\>/ && $bef =~ /\<($1)\s*(.*?)\>\s*$/i) {
	    # The separator is inside a marker, take care not to split it !
	    my $m = $1 ;
	    print STDOUT "Found OK3 split: bef='$bef', aft='$aft', m=$m\n" if $debug ;
	    if ($aft =~ /^(\s*\<\/$m.*?\>)(.*)/i) {
	      $self->{"HTML"}->[$j]->[$l] = "$bef)$1</ARGLIST>$2" ;
	      print STDOUT "Yields:'" . $self->{"HTML"}->[$j]->[$l] . "'\n" if $debug ;
	    }
	  } else {
	    $self->{"HTML"}->[$j]->[$l] = "$bef)</ARGLIST>$aft" ;
	  }
	  $k = $l ;
	  $ok = 3 ;
	  last ;
	}
      }
      if ($ok != 3) {
	for (my $i=$j+1; $i<=$.; $i++) {
	  $k = $#{$self->{"HTML"}->[$i]} ;
	  for (my $l=$k; $l>=0; $l--) {
	    if ($self->{"HTML"}->[$i]->[$l] =~ /(.*)\)(.*)/) {
	      my $bef = $1 ; my $aft = $2 ;
	      # Checks the ARGLIST stuff is not wrongly wrapped:
	      if ($aft =~ /^\s*\<\/(\w+)\s*(.*?)\>/ && $bef =~ /\<($1)\s*(.*?)\>\s*$/i) {
		# The separator is inside a marker, take care not to split it !
		my $m = $1 ;
		print STDOUT "Found OK4 split: bef='$bef', aft='$aft', m=$m\n" if $debug ;
		if ($aft =~ /^(\s*\<\/$m.*?\>)(.*)/i) {
		  $self->{"HTML"}->[$i]->[$l] = "$bef)$1</ARGLIST>$2" ;
		  print STDOUT "Yields:'" . $self->{"HTML"}->[$j]->[$l] . "'\n" if $debug ;
		}
	      } else {
		$self->{"HTML"}->[$i]->[$l] = "$bef)</ARGLIST>$aft" ;
	      }
	      $ok = 3 ;
	      last ;
	    }
	  }
	  last if ($ok == 3) ;
	}
      }
    }
  }
}
###########################################################################
# Adds a stuff:
sub addDef {
  my $self = shift ;
  my $value = shift ;
  print STDOUT (caller(0))[3] . " from=" . (caller(1))[3] . ':' . (caller(0))[2] . " prev=$previous, cur=$value". $other->get(-file)."\n" if $debug ;
  if ($value) {
    my $number = $. ;
    #$value .= (caller(1))[3] . (caller(1))[2] ;
    print STDOUT (caller(0))[2] . "addDef: Adding $value at $number\n" if $debug ;
    #print STDOUT "Pushing @{$self->{'HTML'}->[$number]}" if (exists $self->{'HTML'} && $self->{'HTML'}->[$number] && @{$self->{'HTML'}->[$number]}) ;
    #print STDOUT " whith $value\n" ;
    if ($value =~ /^\s*$/) {
      push(@{$self->{"HTML"}->[$number]}, $value) ;
      return ;
    }
    if ($previous) {
      if ($previous eq '</COMMENT>' && $value eq '<COMMENT>') {
	$previous = '<COMMENT>' ;
	return ;
      } elsif ($previous eq '<COMMENT>' && $value eq '</COMMENT>') {
	$previous = '</COMMENT>' ;
	return ;
      } elsif ($previous eq '</COMMENT>') {
	my $p = ( $number - 1 ) ; 
	push(@{$self->{"HTML"}->[$p]}, '</COMMENT>') ;
	if ($main::opt_latex) {
	  push(@{$self->{"HTML"}->[$p]}, '<PRE>') ;
	}
      }
    }
    if ($main::opt_latex && $value eq '<COMMENT>') {
      push(@{$self->{"HTML"}->[$number]}, '</PRE>') ;
    }
    $previous = $value ;
    if ($value =~ /[\<\>]/) {
      $value =~ s/\</&lt;/g ;
      $value =~ s/\>/&gt;/g ;
    }
    # push(@{$self->{"HTML"}->[$number]}, $value . (caller(1))[3]. (caller(0))[2]) ;
    # print STDOUT "$number: $value\n" ;
    push(@{$self->{"HTML"}->[$number]}, $value) ;
  }
}
###########################################################################
# Adds a stuff:
sub addRaw {
  my $self = shift ;
  my $value = shift ;
  print STDOUT "addRaw: prev=$previous, cur=$value, file=". $other->get(-file)."\n" if $debug ;
  if ($value) {
    my $number = $. ;
    print STDOUT (caller(0))[2] . ":addDef: Adding $value at $number\n" if $debug ;
    if ($previous) {
      if ($previous eq '</COMMENT>' && $value eq '<COMMENT>') {
	$previous = '<COMMENT>' ;
	return ;
      } elsif ($previous eq '<COMMENT>' && $value eq '</COMMENT>') {
	$previous = '</COMMENT>' ;
	return ;
      } elsif ($previous eq '</COMMENT>') {
	my $p = ( $number - 1 ) ; 
	push(@{$self->{"HTML"}->[$p]}, '</COMMENT>') ;
	if ($main::opt_latex) {
	  push(@{$self->{"HTML"}->[$p]}, '<PRE>') ;
	}
      }
    }
    if ($main::opt_latex && $value eq '<COMMENT>') {
      push(@{$self->{"HTML"}->[$number]}, '</PRE>') ;
    }
    $previous = $value ;
    # push(@{$self->{"HTML"}->[$number]}, $value . (caller(1))[3]. (caller(0))[2] ) ;
    # print STDOUT "$number: $value\n" ;
    push(@{$self->{"HTML"}->[$number]}, $value) ;
  }
}
###########################################################################
# Adds a stuff:
sub addPot {
  my $self = shift ;
  my $value = shift ;
  print STDOUT "addPot: prev=$previous, cur=$value, file=". $other->get(-file)."\n" if $debug ;
  if ($value) {
    while($value =~ /(.*?)(\<.*?\>)(.*)/) {
      $value = $3;
      $self->addRaw($1) if $1 ;
      $self->addRaw($2) ;
    }
    $self->addRaw($value) ;
  }
}
###########################################################################
# Adds a stuff:
sub addDefx3 {
  my $self = shift ;
  my (@array) = (@_) ;
  my $number = $. ;
  my $p = $number - $#array ;
  print STDOUT "addDef3: startup\n" if $debug || 1 ;
  foreach my $value (@array) {
    print STDOUT "\t" . (caller(1))[2] . "addDef3: Adding $value at $p\n" if $debug || 1 ;
    if ($value =~ /[\<\>]/) {
      $value =~ s/\</&lt;/g ;
      $value =~ s/\>/&gt;/g ;
    }
    # push(@{$self->{"HTML"}->[$p]}, "addDef3:" . (caller(1))[3]. (caller(0))[2] . $value) ;
    push(@{$self->{"HTML"}->[$p]}, $value) ;
    $p++ ;
  }
  print STDOUT "addDef3: endup\n" if $debug || 1 ;
}
###########################################################################
# Adds a shifted stuff:
sub addDefx2 {
  my $self = shift ;
  my $value = shift ;
  my $number = shift ;
  if ($value) {
    $number += $. ;
    print STDOUT (caller(0))[2] . "addDef2: Adding $value at $number\n" if $debug ;
    # push(@{$self->{"HTML"}->[$number]}, "addDef2:" . (caller(1))[3]. (caller(0))[2]. $value) ;
    push(@{$self->{"HTML"}->[$number]}, $value) ;
  }
}
###########################################################################
sub htmlprint {
  my $self = shift ;
  my $ihtml = shift ;
  if (exists $self->{"HTML"}) {
    my $handle = $ihtml->get(-handle) ;
    my $number = 0 ;
    my $j = 0 ;
    print STDOUT (caller(0))[3] . ": Self $self:". $self->get(-name) . " at file = " . $self->get(-file) . "\n" if $debug ;
    my $name = $self->get(-name) ;
    # my $file = $self->get(-file) ;
    # my $target = &ihtml::targetname($file) ;
    my $c = 0 ;
    my $p = 0 ;
    foreach my $array (@{$self->{"HTML"}}) {
      $number++ ;
      $j++ ;

      if ($array && @{$array}) {
	foreach my $stuff (@{$array}) {
	  if ($stuff) {
	    if ($stuff eq '<COMMENT>') {
	      $c=1;
	    } elsif ($stuff eq '</COMMENT>') {
	      $c=0;
	    }
	    if ($stuff eq '<ARGLIST>' || $stuff eq '<BRANCHEMENT>') {
	      $c=1;
	    } elsif ($stuff eq '</ARGLIST>' || $stuff eq '</BRANCHEMENT>') {
	      $c=0;
	    }
	  }
	  $ihtml->printstuff($stuff) ;
	}
	if (!$main::opt_nolinenum && $j >= $main::num_affiche) {
	  unless ($c || $p) {
	    if ($main::opt_nofile) {
	      print $handle "<DIV ALIGN=right><INFO>$number</INFO></DIV>" ;
	    } else {
	      print $handle "<DIV ALIGN=right><INFO>$name-\&gt;$number</INFO></DIV>" ;
	    }
	  }
	  $j = 0 ;
	}
	print $handle "\n" ;
      } else {
	# Void stuff but no line number:
	print $handle "\n" ;
      }
    }
  }
}
###########################################################################
1;

