package fthc ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for C files:
#   It contains the modules and types setting, as well as the interfaces
#   and procedures definitions. Also contains global procedures.
###########################################################################
use strict;
use Carp ;
use File::Basename ;
use vars qw($debug %hashrecursion %antirecursions @arrayrecursion %MYOBJHASH $opt_oldmm) ;
$debug = 0 ;
###########################################################################
use fthmod ;
use fthtype ;
use fthinter ;
use fthrout;
use fthinp;
use fthinc;
use ihtml;
use chtml;
use oref;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = lc(shift) ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    # We refuse to create an object from no where
    debug("Violent build for $name is illegal, API error") if $debug ;
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  if ($args{-name}) {
    $name = $args{-name} ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    print STDOUT "Reusing $name\n" if $debug ;
    $self = $MYOBJHASH{$name} ;
  } else {
    $first = 1 ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
    print STDOUT "Defining $name\n" if $debug ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
sub old {
  my $self = {} ;
  shift ;
  my(@args) = (@_) ;
  bless $self;
  if (@args) {
    $self->init(@args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
sub printHash {
  my $self = shift ;
  my $t = shift ;
  my $hash = shift ;
  while(my($cle,$val) = each(%{$hash})) {
    print STDOUT "$t$cle => $val, " ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "\t$cle = (" ;
      foreach my $sval (@array) {
	print STDOUT "$sval, " ;
      }
      print STDOUT ")\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	if (exists $array{$sval} && ref($array{$sval}) eq 'fthrout') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthinp') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthtype') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthinter') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'HASH') {
	  print STDOUT "$t\t$sval => {"  ;
	  $self->printHash("\t$t", $array{$sval}) ;
	  print STDOUT"}, " ;
	} else {
	  print STDOUT "$sval => $array{$sval}, " ;
	}
      }
      print STDOUT "},\n" ;
    } else {
      print STDOUT "\t$cle = $val\n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "TRACE FTHC: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
  print STDOUT "VALUE FTHC: $message\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
# Adds a routine to a current stuff:
sub addRout {
  my $self = shift ;
  my $name = shift ;
  my $value = shift ;
  $self->{"FPROCEDURES"}->{"$name"} = $value ;
  print STDOUT "Adding a routine for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds a type to a current stuff:
sub addType {
  my $self = shift ;
  my $name = shift ;
  my $value = shift ;
  $self->{"FTYPES"}->{"$name"} = $value ;
  print STDOUT "Adding a type for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds a LaTeX input file
sub addInput {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"INPUTS"}->{"$name"} = $value ;
  $value->addReferer($self) ;
  print STDOUT "Adding an input for $self(" . $self->get(-name) . "), name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds an include:
sub addInclude {
  my $self = shift ;
  my $name = shift ;
  my $value = shift ;
  $self->{"INCLUDES"}->{"$name"} = $value ;
  $value->addReferer($self) ;
  print STDOUT "Adding an include for $self, name=$name, object=$value(" . $value->get(-name) . ")\n" if $debug ;
}
###########################################################################
sub getInputs {
  my $self = shift ;
  if (exists $self->{"INPUTS"}) {
    return(%{$self->{"INPUTS"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
sub getIncludes {
  my $self = shift ;
  if (exists $self->{"INCLUDES"}) {
    return(%{$self->{"INCLUDES"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
sub setDepends {
  my $self = shift ;
  # void function...
}
###########################################################################
sub revCalls {
  my $self = shift ;
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    # Current file defines $name
    my $orout = $self->{"FPROCEDURES"}->{"$name"} ;
    print STDOUT (caller(0))[3] . ": Procedure $name points to $orout, defined in file ".$self->get(-name) .":\n" if $debug ;
    $orout->revCalls() ;
  }
}
###########################################################################
sub setCompiles {
  my $self = shift ;
  my (%hash) = () ;
  # Gets the local dependency list:
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    my $orout = $self->{"FPROCEDURES"}->{"$name"} ;
    my (@array) = $orout->setCompiles() ;
    foreach my $a (@array) {
      $hash{$a} = 1 ;
    }
  }
  my $file = $self->get(-file) ;
  # Delete myself I need to be the last in array.
  delete $hash{$file} if (exists $hash{$file}) ;
  foreach my $file (keys %hash) {
    push(@{$self->{"COMPORDER"}}, $file) ;
  }
}
###########################################################################
# Fake setting for global:
sub setCompOrder {
  my $self = shift ;
  (@{$self->{"COMPORDER"}}) = (@_) ;
}
###########################################################################
# Gets all the direct files dependencies for current, make it once only
sub recCompiles {
  my $self = shift ;
  my (%objs) = (@_) ;
  my $file = $self->get(-file) ;
  print STDOUT "\t" . (caller(0))[3] . " Recursive call tests for $file for ".$self->get(-name)."\n" if $debug;
  if ($file && $antirecursions{$file}) {
    if ($antirecursions{$file} == 1) {
      unless ($main::opt_v) {
	print STDOUT "\n" if $main::opt_show ;
      }
      print STDOUT "\t" . (caller(0))[3] . " Recursivity not allowed file $file for ".$self->get(-name)."\n" if $main::opt_show ;
      $antirecursions{$file} = 2 ;
    }
    return ;
  }
  $antirecursions{$file} = 1 if $file ;
  if (exists $self->{"COMPORDER"}) {
    foreach my $fic (@{$self->{"COMPORDER"}}) {
      unless (exists $hashrecursion{$fic}) {
	my $stuff = $objs{$fic} ;
	if ($stuff) {
	  $stuff->recCompiles(%objs) ;
	}
	unless (exists $hashrecursion{$fic}) {
	  push(@arrayrecursion, $fic) ;
	  $hashrecursion{$fic} = 1 ;
	}
      }
    }
  }
  if ($file) {
    unless (exists $hashrecursion{$file}) {
      $hashrecursion{$file} = 1 if $file ;
      push(@arrayrecursion, $file) ;
    }
  }
}
###########################################################################
# Gets all the direct files dependencies for current, make it once only
sub getCompiles {
  my $self = shift ;
  my (%objs) = (@_) ;
  if (!$self->get(-depends)) {
    local(%hashrecursion) = () ;
    local(@arrayrecursion) = () ;
    local(%antirecursions) = () ;
    $self->recCompiles(%objs) ;
    print STDOUT "\t" . (caller(0))[3] . " Arrayrecursion on $debug (".$self->get(-name).") = @arrayrecursion\n" if $debug && $self->get(-name) ;
    (@{$self->{"COMPILELIST"}}) = (@arrayrecursion) ;
    $self->set(-depends => 1) ;
  }
  if (exists $self->{"COMPILELIST"}) {
    return(@{$self->{"COMPILELIST"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Gets all the direct files dependencies for current, make it once only
sub oldCompiles {
  my $self = shift ;
  my (%objs) = (@_) ;
  print STDOUT (caller(1))[3] . ':' . (caller(0))[2] . " In " . $self->get(-name) . " fetching calls($debug)\n" if $debug ;
  if (!$self->get(-depends)) {
    my (%hash) = () ;
    if (exists $self->{"COMPORDER"}) {
      foreach my $file (@{$self->{"COMPORDER"}}) {
	next unless $file ;
	unless (exists $hash{$file}) {
	  my $stuff = $objs{$file} ;
	  my $name = undef ;
	  $name = $stuff->get(-file) if $stuff ;
	  if ($name && $self->get(-name)) {
	    unless (!$stuff || exists $hash{$name} || $name eq $self->get(-name)) {
	      print STDOUT (caller(0))[3] . "Fetching sub call to ".$self->get(-name)." in file $file(@{$self->{COMPORDER}})\n" if $debug ;
	      my(@array) = $stuff->oldCompiles(%objs) ;
	      foreach my $a (@array) {
		unless (exists $hash{$a}) {
		  $hash{$a} = $a ;
		  push(@{$self->{"COMPILELIST"}}, $a) ;
		}
	      }
	    }
	  }
	}
	unless (exists $hash{$file}) {
	  $hash{$file} = $file ;
	  push(@{$self->{"COMPILELIST"}}, $file) ;
	}
      }
    }
    my $file = $self->get(-file) ;
    if ($file) {
      unless (exists $hash{$file}) {
	# Add myself:
	$hash{$file} = $file ;
	push(@{$self->{"COMPILELIST"}}, $file) ;
      }
    }
    $self->set(-depends => 1) ;
  }
  if (exists $self->{"COMPILELIST"}) {
    return(@{$self->{"COMPILELIST"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
sub setCalls {
  my $self = shift ;
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    # Current file defines $name
    my $orout = $self->{"FPROCEDURES"}->{"$name"} ;
    print STDOUT (caller(0))[3] . ": Procedure $name points to $orout, defined in file ".$self->get(-name) .":\n" if $debug ;
    $orout->setCalls() ;
  }
}
###########################################################################
sub sayCalls {
  my $self = shift ;
  my $ier = shift ;
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    my $orout = $self->{"FPROCEDURES"}->{"$name"} ;
    print STDOUT (caller(0))[3] . ": Routine $name points to $orout, defined in file ".$self->get(-name) .":\n" if $main::opt_v;
    $orout->sayCalls($ier) ;
  }
}
#|#######################################################
sub C_Tags {
  my $ligne = shift ;
  if ($ligne =~ /^([\w_]+)[\s\(]+(\d+),\d+$/) {
    my $name = $1 ;
    my $number = $2 ;
    unless ($name =~ /\bstructdef\b/i) {
      return("void", $name, $number) ;
    }
  } elsif ($ligne =~ /^([\w_]*)\s*.*?\s*([\w_]+)[\s\(\=]+(\d+),\d+$/) {
    my $rtype = $1 ;
    my $name = $2 ;
    my $number = $3 ;
    return($rtype, $name, $number) ;
  } else {
    return ((undef));
  }
}

#|#######################################################
sub Lecture_Tags {
  my ($DIRE, $DIRS, @etags) = (@_);
  my $self ;
  if (0) {
    print"\t Building $DIRE/ftagshtml.TAGS for files @etags:\n";
    system("etags --ignore-indentation --no-defines --no-globals -o $DIRE/ftagshtml.TAGS @etags") ;
    if (open(TAGS,"<$DIRE/ftagshtml.TAGS")) {
      my $file ;
      my $rtype ;
      my $name ;
      my $number ;
      while (<TAGS>) {
	if (//) {
	  undef $file ;
	  next ;
	}
	if ($file) {
	  ($rtype, $name, $number) = &C_Tags($_);
	  if ($number) {
	    my $orout = new fthrout(-name => "$name", -file => $file, -masfic => $file, -global => 1, -number => $number) ;
	    print STDOUT "\t\t" . (caller(0))[3] . ": Creating from $self new routine $name=$orout line $number file $file\n" if $debug ; 
	    $self->addRout($name => $orout) ;
	    $orout->set(-theproto => $rtype) ;
	    $orout->set(-number => $number) ;
	  }
	} else {
	  if (/^([\w\-\._\/]+),\d+$/) {
	    $file = $1 ;
	    if ($file =~ /^$DIRS(.*)/) {
	      $file = $1 ;
	    }
	    $self = new fthc(-file => $file, -masfic => $file, -name => $file) ;
	    next ;
	  }
	}
      }
      close(TAGS);
    }
  }
  print"\t Reading C files implementation:\n";
  foreach $self (values %MYOBJHASH) {
    print STDOUT "." unless $debug ;
    $self->readc() ;
  }
  print STDOUT "\n" unless $debug ;
}
###########################################################################
# sort by line number:
sub bynumber {
 $a->{"number"} <=> $b->{"number"} ;
}
###########################################################################
# A routine for reading a C file.
sub htmlprint {
  my $self = shift ;
  my $outdir = shift ;
  my $reldir = shift ;
  my $file = $self->get(-file) ;
  my $on ;
  my $out ;
  my $outm ;
  my $outi ;
  my $outl ;
  my $outj ;
  my $masf ;
  $out = "$outdir/" . basename($file) ;
  print STDOUT (caller(0))[3] . ": Outfile($file) is $out\n" if $debug ;
  if ($out =~ /(.*)\.(.*)/) {
    # Implementation HTML file:
    $outm = $1 . '_' . $2 . '_c' . '.html' ;
    # Summary HTML file:
    $outi = $1 . '_' . $2 . '_i' . '.html' ;
    $outl = $1 . '_' . $2 . '_LF' . '.html' ;
    $outj = $1 . '_' . $2 . '.js' ;
    $masf = $1 . '_' . $2 . '.html' ;
  } else {
    $outm = $out . '_c' . '.html' ;
    $outi = $out . '_i' . '.html' ;
    $outl = $out . '_LF' . '.html' ;
    $outj = $out . '.js' ;
    $masf = $out . '.html' ;
  }
  my $ihtml ;
  if (open(MASTER,">$masf")) {
    my $handle = \*MASTER ;
    $ihtml = new ihtml(-name => $file, -handle => $handle, -file => $file, -reldir => $reldir) ;
    $ihtml->print_master() ;
    close(MASTER) ;
  }
  if (!$main::opt_nojava && open(OUFL,">$outl")) {
    my $handle = \*OUFL ;
    $ihtml->set(-handle => $handle) ;
    $ihtml->print_lf() ;
    close(OUFL) ;
  }
  if (!$main::opt_nojava && open(OUFJ,">$outj")) {
    my $js = \*OUFJ ;
    my $ligne = "" ;
    # Java tree building is initialized down there:
    $ligne .= $ihtml->start_tree() ."\n" ;
    $ligne .= $ihtml->addFolder("Descending tree", $file) . "\n" ;
    foreach my $orout (sort bynumber values %{$self->{"FPROCEDURES"}}) {
      $ligne .= $ihtml->irouts_down_tree($orout) ;
    }
    unless ($opt_oldmm) {
      $ligne .= "  ],\n" ;
    }
    $ligne .= $ihtml->addFolder("Ascending tree", $file) . "\n" ;
    foreach my $orout (sort bynumber values %{$self->{"FPROCEDURES"}}) {
      $ligne .= $ihtml->irouts_up_tree($orout) ;
    }
    unless ($opt_oldmm) {
      $ligne .= "  ],\n" ;
    }
    unless ($opt_oldmm) {
      $ligne .= "\t]\n];\n" ;
    }
    print $js $ligne ;
    close($js) ;
  }
  if (open(OUFI,">$outi")) {
    my $handle = \*OUFI ;
    my $title = "Interface for file $file" ;
    $ihtml->set(-handle => $handle) ;
    $ihtml->set(-title => $title) ;
    $ihtml->start_interface() ;
    my (@deps) = $self->getCompiles() ;
    if (@deps) {
      $ihtml->print_deps(@deps) ;
    }
    # Other stuffs:
    foreach my $orout (sort bynumber values %{$self->{"FPROCEDURES"}}) {
      my $name = $orout->get(-name) ;
      $ihtml->printi_routs($orout) ;
    }
    my(%hash) = $self->getIncludes() ;
    if (%hash) {
      print $handle "<BR>$file includes\n" ;
      foreach my $incname (sort keys %hash) {
	my $inco = $self->{"INCLUDES"}->{"$incname"} ;
	$ihtml->printi_incl($inco) ;
      }
    }
    if (exists $self->{"INPUTS"}) {
      print $handle "<BR>$file inputs\n" ;
      foreach my $inpname (sort keys %{$self->{"INPUTS"}}) {
	my $inpo = $self->{"INPUTS"}->{"$inpname"} ;
	$ihtml->printi_incp($inpo) ;
      }
    }
    $ihtml->end_html() ;
    close($handle) ;
  }
  if (open(OUFM,">$outm")) {
    my $handle = \*OUFM ;
    my $title = "Implementation of file $file" ;
    $ihtml->set(-title => $title) ;
    $ihtml->set(-handle => $handle) ;
    $ihtml->set(-js => 1) unless $main::opt_nojava;
    $ihtml->start_implementation() ;
    print $handle "<PRE>\n" ;
    # Get the chtml stuff:
    my $chtml = &chtml::getObject($file) ;
    $chtml->htmlprint($ihtml) ;
    print $handle "</PRE>\n" ;
    $ihtml->end_html() ;
    close($handle) ;
    $ihtml->free() ;
  }
}
###########################################################################
# A routine for reading a C file.
sub readc {
  my $self = shift ;
  my $file = shift ;
  my $in = shift ;
  return if $self->get(-isread) ;
  my $chtml = &chtml::getCurrent() ;
  my $oi = 0 ;
  my $orout = undef ;
  unless ($file) {
    $file = $self->get(-file) ;
  }
  unless ((ref $in) && (fileno $in)) {
    croak("No API for $file") unless $file ;
    if (-f "$file") {
      if (open(FILE,"<$file")) {
	$chtml = new chtml(-file => $file, -name => $file) ;
	$in = \*FILE ;
	$oi = 1 ;
      }
    }
  }
  if ((ref $in) && (fileno $in)) {
    my $c = 0 ;
    my $bs = 0 ; # Start counter for declaration
    my $bc = 0 ; # {} counter
    my $pc = 0 ; # () counter
    my $name = undef ;
    my $ip = "" ; # Stored declaration string
    my $cc = "" ; # Comment string
    my $str = "" ;
    my $nbbran = 0 ;
    my $proto = "" ;
    my $cname = "" ;
    my $args = "" ;
    my $suite = "" ;
    my $num ;
    my $nump ;
    my $numb ;
    my(@aloops) = () ;
    my $preif = 0 ;
    my(@ploops) = () ;
    my $nbpar = 0 ;
    my(@rloops) = () ;
    my $nbbra = 0 ;
    my(@bloops) = () ;
    print STDOUT (caller(0))[3] . ": OK reading file $file\n" if $debug ;
    while(<$in>) {
      # Parsing a C file below:
      chomp ;
      # Parsing the inputs:
      unless ($str || $c) {
	if (/^(\#include\s+)\<([\w\.\/]+)\>(.*)/i) {
	  $chtml->addDef($1 . '&lt;' . $2 . '&gt;' . $3) ;
	  next ;
	} elsif (/^(\#include\s+)([\"\'])([\w\.\/]+)\2(.*)/i) {
	  # Includes here:
	  $chtml->addDef($1) ;
	  $chtml->addRaw("<A Name=C_$.></A>");
	  my $incf = $3 ;
	  unless (-f "$incf") {
	    my $base = dirname($self->get(-file)) ;
	    if (-f "$base/$incf") {
	      $incf = "$base/$incf" ;
	    } elsif (-f "$base/" . basename($incf)) {
	      $incf = "$base/" . basename($incf) ;
	    } elsif ($main::opt_I) {
	      # Possible...
	      my @TAGSHTML_PATH = (split ( /:/, $main::opt_I)) ;
	      foreach $base (@TAGSHTML_PATH) {
		if (-f "$base/$incf") {
		  $incf = "$base/$incf" ;
		  last ;
		}
	      }
	    }
	  }
	  if (-f "$incf") {
	    # Read the file on the fly right now:
	    my $fh ;
	    my $other ;
	    if (open($fh,"<$incf")) {
	      print STDOUT (caller(0))[3] . ": Open $incf from $file\n" if $debug ;
	      $other = &fthinc::getObject($incf) ;
	      if ($other) {
		new chtml() ;
	      } else {
		new chtml(-file => $incf, -name => $incf) ;
		$other = new fthinc(-file => "$incf", -name => "$incf", -masfic => $self->get(-masfic), -father => $self) ;
	      }
	      $self->addInclude("$incf" => $other) ;
	      $chtml->addDef("$2") ;
	      $chtml->addRaw($other) ;
	      $chtml->addDef("$2$4") ;
	      $self->readc($incf, $fh) ;
	      close($fh);
	      &chtml::setCurrent($chtml) ;
	      print STDOUT (caller(0))[3] . ": Closed $incf($.) from $file\n" if $debug ;
	    }
	  } else {
	    my $refe = new oref(-name => "$3", -type => "fthinc") ;
	    $chtml->addDef("$2") ;
	    $chtml->addRaw($refe) ;
	    $chtml->addDef("$2$4") ;
	  }
	  next ;
	}
      }
      if (/^\#(ifdef|ifndef|if|elif|else|endif)(.*)/i) {
	# Preprocessing tests, this seems good code below:
	if ($1 eq "endif") {
	  $num = pop(@ploops) ;
	  if ($num) {
	    $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	    $preif-- ;
	  } else {
	    $chtml->addDef('#'."$1$2") ;
	  }
	} elsif ($1 eq "if" || $1 eq "ifdef" || $1 eq "ifndef") {
	  push(@ploops, $.) ;
	  $preif++ ;
	  $chtml->addRaw("<TEST><A Name=C_$.>".'#'."$1</A></TEST>$2") ;
	} else {
	  $num = pop(@ploops) ;
	  if ($num) {
	    push(@ploops, $num) ;
	    $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	  } else {
	    $chtml->addDef('#'."$1$2") ;
	  }
	}
	next ;
	$_ = "" ;
      }
      if (/^\#(define\s*)\b(\w+)\s*\((.*)\)(.*)/i) {
	# Special define treatment, stuff()...bid(){ favours creation of routine stuff instead of bid.
	# For interface definition, the semi-column saves the trick.
	$chtml->addRaw('#'."$1") ;
	if ($orout) {
	  my $lrout = &fthrout::getObject($2) ;
	  if ($lrout) {
	  } else {
	    print STDOUT "New func='$cname' proto='$proto' args='$args', context='$_' line $. file '$file', masfic='".$self->get(-masfic)."'\n" if $debug ;
	    $lrout = new fthrout(-name => "$2", -file => $file, -masfic => $self->get(-masfic), -global => 1, -number => $.) ;
	    $lrout->set(-number => $.) ;
	  }
	  $self->addRout($2 => $lrout) ;
	  $chtml->addRaw($lrout) ;
	  $chtml->addRaw("<ARGLIST>(" . $3 . ")</ARGLIST>");
	} else {
	  $orout = &fthrout::getObject($2) ;
	  if ($orout) {
	  } else {
	    print STDOUT "New func='$cname' proto='$proto' args='$args', context='$_' line $. file '$file', masfic='".$self->get(-masfic)."'\n" if $debug ;
	    $orout = new fthrout(-name => "$2", -file => $file, -masfic => $self->get(-masfic), -global => 1, -number => $.) ;
	    $orout->set(-number => $.) ;
	  }
	  $self->addRout($2 => $orout) ;
	  $chtml->addRaw($orout) ;
	  $chtml->addRaw("<ARGLIST>(" . $3 . ")</ARGLIST>");
	  # Make sure we do not believe being inside a routine:
	  $orout = undef ;
	}
	$_ = $4 ;
      }
      # Now parsing all the stuff
      print STDOUT "\tBefore loop, file='$file'($.): b='$bc', p='$pc', s='$str', c='$c', ip='$ip', heading='$_'\n" if $debug ;
      while (/^(.*?)(\{|\}|\'|\"|\(|\)|\/\*|\*\/|\/\/|\;)(.*)/i) {
	my $bef = $1 ;
	my $sep = $2 ;
	my $aft = $3 ;
	print STDOUT "\t\t Looping bef=$bef, sep=$sep, aft=$aft, p=$pc, b=$bc, c=$c, s=$str\n" if $debug ;
	$_ = $3 ;
	if ($bef =~ /^(.*?)\\\\$/) {
	  print STDOUT "\t\t Antislash special character encountered\n" if $debug ;
	} elsif ($bef =~ /^(.*?)\\$/) {
	  # Fake stuff:
	  print STDOUT "\t\t Wrong special character encountered\n" if $debug ;
	  if ($str) {
	    $chtml->addDef($bef.$sep) ;
	  } elsif ($c) {
	    # Comment is on:
	    $cc .= $bef . $sep ;
	    print STDOUT "\t\t Yet another comment stuff, cc=$cc\n" if $debug ;
	  } else {
	    print STDOUT "Normally ending $_\n" ;
	    $chtml->addDef($bef.$sep) ;
	  }
	  next ;
	}
	if ($bef =~ /\s*extern\s*$/ && $aft =~ /^C/) {
	  # Special marker for the extern C stuff:
	  $bs = 1 ;
	}
	# FROM:
	if ($str && $sep eq $str) {
	  # Indicates a string end
	  $str = "" ;
	  $chtml->addDef($bef.$sep) ;
	  $chtml->addRaw('</STRING>') ;
	} elsif ($c && $sep eq '*/') {
	  # Indicates a comment end
	  print STDOUT "\t\t\t Terminating comment\n" if $debug ;
	  $c = 0 ;
	  $cc .= $bef ;
	} elsif ($str) {
	  # Nothing special, continuing the string build
	  $chtml->addDef($bef.$sep) ;
	} elsif ($c) {
	  # Nothing special, continuing the comment build
	  $cc .= $bef . $sep ;
	} elsif ($sep eq '"' || $sep eq "'") {
	  $str = $sep ;
	  $chtml->addDef($bef) ;
	  $chtml->addRaw('<STRING>') ;
	  $chtml->addDef($sep) ;
	} elsif ($sep eq '//') {
	  $chtml->addDef($bef) ;
	  $cc = $aft ;
	  $_ = "" ;
	} elsif ($sep eq '/*' || $sep eq '*/') {
	  print STDOUT "\t\t\t Separator is comment: sep=$sep, c=$c\n" if $debug ;
	  if ($sep eq '/*') {
	    if ($c == 0) {
	      $c = 1 ;
	      print STDOUT "\t\t\t Previous stuff is $bef, now c=$c\n" if $debug ;
	      $chtml->addDef($bef) ;
	      # This is added to the comments list
	    } else {
	      $cc .= $bef . $sep ;
	    }
	  } elsif ($sep eq '*/') {
	    if ($c == 1) {
	      print STDOUT "\t\t\t Terminating comment\n" if $debug ;
	      $c = 0 ;
	      $cc .= $bef ;
	    } else {
	      # Should never happen
	      print STDOUT (caller(0))[3] . ": Error file $file meets end of comment inside instruction line $. $bef$sep$aft\n" ;
	      $chtml->addDef($bef . $sep) ;
	    }
	  }
	} elsif ($sep eq '{' || $sep eq '}') {
	  # print STDOUT "sep=$sep, bc=$bc, bs=$bs\n" ;
	  if ($sep eq '{') {
	    $num = 'b'.$. ;
	    if ($numb) {
	      if ($numb eq $num) {
		$num = $num . '_1' ;
	      } else {
		if ($numb =~ /\_(\d+)/) {
		  $num = $num . '_' . ($1 + 1) ;
		}
	      }
	    }
	    $numb = $num ;
	    push(@bloops, $numb) ;
	    $nbbra++ ;
	    # Indicates a first '{', this is a good candidate for start up:
	    if ($bc == $bs && $ip =~ /\)\s*$/) {
	      # Definition point raised and made over below:
	      $ip .= $bef ;
	      $cname = "" ;
	      # cfortran style, as it seems to be:
	      if ($ip =~ /(\w*)\s*\b[A-Z0-9\_]+\s*\(\s*(\w+)\s*\,\s*\2\s*\)\s*\((.*)\)(\s*)$/i) {
		print STDOUT "ip at $. line 883\n" if $debug ;
		$proto = $1 ;
		$cname = $2 ;
		$args = $3 ;
		$suite = $4 ;
	      } elsif ($ip =~ /(\w*)\s*\b[A-Z0-9\_]+\s*\((\s*\w+\s*)\)\s*\((.*)\)(\s*)$/) {
		print STDOUT "ip at $. line 889\n" if $debug ;
		$proto = $1 ;
		$cname = $2 ;
		$args = $3 ;
		$suite = $4 ;
	      } elsif ($ip =~ /(\w*)\s*\b([\w\_]+)\s*\((.*)\)(\s*)$/) {
		print STDOUT "ip at $. line 895, name=$2\n" if $debug ;
		$proto = $1 ;
		$cname = $2 ;
		$args = $3 ;
		$suite = $4 ;
	      } else {
		print STDERR "ip=$ip, ligne=$_, line=$.\n" if $debug ;
	      }
	      if ($cname) {
		$orout = &fthrout::getObject($cname) ;
		if ($orout) {
		  print STDOUT "Rout for $cname exists\n" if $debug ;
		} else {
		  print STDOUT "Rout for $cname created\n" if $debug ;
		  # etags parsing had not been performed !
		  # This ought to be considered as an internal error:
		  print STDOUT "New func='$cname' proto='$proto' args='$args', context='$_' line $. file '$file', masfic='".$self->get(-masfic)."'\n" if $debug ;
		  $orout = new fthrout(-name => "$cname", -file => $file, -masfic => $self->get(-masfic), -global => 1, -number => $.) ;
		  $self->addRout($cname => $orout) ;
		  $orout->set(-number => $.) ;
		}
		# At this level, proto parsing is allowed,
		if ($_ =~ /(.*)\b($cname)\s*\((.*)\)(.*)/) {
		  print STDOUT "Adding def=$1\n" if $debug ;
		  $chtml->addDef($1) ;
		  if ($orout) {
		    $chtml->addRaw($orout) ;
		  } else {
		    $chtml->addDef($cname) ;
		  }
		  $chtml->addRaw("<ARGLIST>(" . $args . ")</ARGLIST>");
		  $chtml->addDef($suite) ;
		} elsif ($orout) {
		  # Backwards search for the routine name:
		  print STDOUT "Search backwards cname=$cname for orout=$orout\n" if $debug ;
		  $chtml->findDef($cname => $orout) ;
		}
		$chtml->addRaw("<A Name=C_$num>$sep</A>") ;
		if ($orout) {
		  if ($cname eq 'main') {
		    $orout->set(-theproto => 2) ;
		  } elsif ($ip =~ /.*\s(.*?)\b(\w+)\s*\((.*)\)(.*)/) {
		    print STDOUT "ip at 931, proto=$1, val=$2\n" if $debug ;
		    $orout->set(-theproto => $1) ;
		  }
		}
	      } else {
		$chtml->addDef($bef) ;
		$chtml->addRaw("<A Name=C_$num>$sep</A>") ;
		print STDERR "fthc KO 899($file,$.): $ip, bef=$bef, sep=$sep, aft=$aft\n" ;
	      }
	      $ip = "" ;
	    } else {
	      # Another internal bloc starts:
	      $chtml->addDef($bef) ;
	      $chtml->addRaw("<A Name=C_$num>$sep</A>") ;
	    }
	    $bc++ ;
	    if ($bc == $bs) {
	      # Restart the ip stuff ?
	      $ip = "" ;
	    }
	  } elsif ($sep eq '}') {
	    $numb = pop(@bloops) ;
	    $chtml->addDef($bef) ;
	    if ($numb) {
	      $chtml->addRaw("<A Href=#C_$numb>$sep</A>") ;
	      $nbbra-- ;
	    } else {
	      $chtml->addRaw("$sep") ;
	    }
	    if ($bc == $bs + 1) {
	      $ip = "" ;
	    }
	    $bc-- ;
	    if ($bc == $bs) {
	      # A possibly dangerous stuff, delete the current routine:
	      $orout = undef ;
	    }
	    if ($bc == 0) {
	      $bs = 0 ;
	    }
	  }
	} elsif ($sep eq '(' || $sep eq ')') {
	  if ($bc == $bs) {
	    print STDOUT "Bloc starting ? ip = '$ip', bef='$bef', sep='$sep'\n" if $debug ;
	    $ip .= $bef . $sep ;
	  }
	  if (0) {
	    $chtml->addDef($bef.$sep) ;
	    print STDOUT "\t\t In $file($.): b=$bc, p=$pc, sep=$sep, let ip=$ip\n" if $debug ;
	    if ($sep eq '(') {
	      $pc++ ;
	    } elsif ($sep eq ')') {
	      $pc-- ; 
	    }
	  } elsif ($sep eq '(') {
	    $num = 'p'.$. ;
	    if ($nump) {
	      if ($nump eq $num) {
		$num = $num . '_1' ;
	      } else {
		if ($nump =~ /\_(\d+)/) {
		  $num = $num . '_' . ($1 + 1) ;
		}
	      }
	    }
	    $nump = $num ;
	    push(@rloops, $nump) ;
	    $nbpar++ ;
	    undef $name ;
	    if ($orout) {
	      my $prev = $bef.$sep.$aft ;
	      #if ($prev =~ /\s*([A-Z_]+)\s*\(\s*(\w+)\s*\)\s*\((.*)/) {
	      if ($prev =~ /\s*(CALL_FORTRAN)\s*\(\s*(\w+)\s*\)\s*\((.*)/) {
		#die "OK bef='$bef', sep='$sep', aft='$aft', 1='$1', 2='$2', 3='$3'\n" ;
		$name = $2 ;
		if (&fthrout::isInternC($name)) {
		  undef $name ;
		} else {
		  $aft = $3 ;
		  $_ = $3 ;
		  print STDOUT (caller(0))[3] . ": Calling $name from " . $orout->get(-name)." line $. file $file\n" if $debug ;
		  # detected another call:
		  my $called = new fthcall(-name => $name, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $orout, -number => $.) ;
		  $orout->addCall($name => $called) ;
		  # No possible discussion for C stuff, not prototype analysis for the moment:
		  # NAJ: setting a NAJ here has what effect ?
		  #$called->setLink(0, &fthrout::getObject($name)) ;
		  # $called->set(-thelink => &fthrout::getObject($name)) ;
		  $chtml->addDef($bef.'(') ;
		  $chtml->addRaw($called) ;
		  $chtml->addDef(')') ;
		  if ($nbbran == 0) {
		    $nbbran = ( $pc + 1 ) ;
		    $chtml->addRaw("<BRANCHEMENT>") ;
		  }
		  $chtml->addRaw("<A Name=C_$num>$sep</A>") ;
		  #$chtml->addRaw("(") ;
		}
	      } elsif ($bef =~ /(.*?)(\w+)\s*$/) {
		# Another possible call:
		$name = $2 ;
		$prev = $1 ;
		print STDOUT (caller(0))[3] . ": Calling $name from file $file line $.\n" if $debug  ;
		# NAJ: make a first step to read it first !
		if (&fthrout::isInternC($name)) {
		  if ($bef =~ /(.*)\b(if\s*)$/) {
		    $chtml->addDef("$1") ;
		    $chtml->addRaw("<TEST><A Name=C_$.>$2</A></TEST>") ;
		    $chtml->addRaw("<A Name=C_$num>$sep</A>") ;
		    #$chtml->addRaw($sep) ;
		  } elsif ($bef =~ /(.*)\b(for\s*)$/) {
		    $chtml->addDef("$1") ;
		    $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
		    $chtml->addRaw("<A Name=C_$num>$sep</A>") ;
		    #$chtml->addRaw($sep) ;
		  } else {
		    undef $name ;
		  }
		} else {
		  print STDOUT (caller(0))[3] . ": Calling $name from " . $orout->get(-name)." line $. file $file\n" if $debug ;
		  # detected another call:
		  my $called = new fthcall(-name => $name, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $orout, -number => $.) ;
		  $orout->addCall($name => $called) ;
		  # No possible discussion for C stuff, not prototype analysis for the moment:
		  # NAJ: setting a NAJ here has what effect ?
		  #$called->setLink(0, &fthrout::getObject($name)) ;
		  # $called->set(-thelink => &fthrout::getObject($name)) ;
		  $chtml->addDef("$prev") ;
		  $chtml->addRaw($called) ;
		  if ($nbbran == 0) {
		    $nbbran = ( $pc + 1 ) ;
		    $chtml->addRaw("<BRANCHEMENT>") ;
		  }
		  $chtml->addRaw("<A Name=C_$num>$sep</A>") ;
		  #$chtml->addRaw("(") ;
		}
	      }
	    }
	    unless ($name) {
	      $chtml->addDef($bef) ;
	      $chtml->addRaw("<A Name=C_$num>$sep</A>") ;
	    }
	    $pc++ ;
	  } elsif ($sep eq ')') {
	    $nump = pop(@rloops) ;
	    $chtml->addDef($bef) ;
	    if ($nump) {
	      $chtml->addRaw("<A Href=#C_$nump>$sep</A>") ;
	      $nbpar-- ;
	    } else {
	      $chtml->addRaw("$sep") ;
	    }
	    if ($nbbran && $nbbran == $pc ) {
	      $chtml->addRaw("</BRANCHEMENT>") ;
	      $nbbran = 0 ;
	    } else {
	    }
	    $pc-- ;
	  }
	} elsif ($sep eq ';') {
	  # Instruction separator, restart from the beginning...
	  if ($bc == $bs) {
	    $ip = "" ;
	  }
	  $chtml->addDef($bef.$sep) ;
	}
      }
      if ($c) {
	$cc .= $_ ;
      } elsif ($bc >= $bs + 1) {
	$chtml->addDef($_) ;
      } else {
	$ip .= $_ . " " ;
	$chtml->addDef($_) ;
      }
      if ($cc) {
	$chtml->addRaw('<COMMENT>') ;
	if ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  $chtml->addLaCoM("<A Name=C_$.></A>") ;
	}
	while ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  my $incp = "$3.$4" ;
	  $cc = $5 ;
	  $chtml->addLaTeX($1.$2." ") ;
#	  if ($incp =~ /^\// && ! -f  "$incp") {
#	    $incp = basename($incp) ;
#	  }
	  unless (-f "$incp") {
	    my $base = dirname($self->get(-file)) ;
	    my $nase = basename(dirname($incp)) ;
	    if (-f "$base/$incp") {
	      $incp = "$base/$incp" ;
	    } elsif (-f "$base/" . basename($incp)) {
	      $incp = "$base/" . basename($incp) ;
	    } elsif (-f "$nase/" . basename($incp)) {
	      $incp = "$nase/" . basename($incp) ;
	    } elsif ($main::opt_L) {
	      # LaTeX inputs directories
	      # Possible...
	      my @TAGSHTML_PATH = (split ( /:/, $main::opt_L)) ;
	      foreach $base (@TAGSHTML_PATH) {
		if (-f "$base/$incp") {
		  $incp = "$base/$incp" ;
		  last ;
		}
	      }
	    }
	  }
	  if (-f "$incp") {
	    my $other = new fthinp(-file => "$incp", -name => "$incp") ;
	    $self->addInput("$incp" => $other) ;
	    $chtml->addLaCoM($other) ;
	  } else {
	    my $refe = new oref(-name => $incp, -type => "fthinp") ;
	    $chtml->addLaCoM($refe) ;
	  }
	}
	$chtml->addLaTeX($cc) ;
	$chtml->addRaw("</COMMENT>") ;
	$cc = "" ;
      }
    }
  }
  close($in) if $oi ;
  $self->set(-isread => 1) ;
  if (1 == 0) {
    # Stupid old stuff:
    foreach my $incf (keys %{$self->{"INCLUDES"}}) {
      my $inco = &fthinc::getObject("$incf") ;
      my $other = new fthc(-file => "$incf", -masfic => $incf, -name => "$incf") ;
      $inco->set(-father => $other) ;
      $other->readc() ;
    }
  }
}
###########################################################################
1;

