package fthcall ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for fthcall 90 files.
###########################################################################
use strict;
use Carp ;
use vars qw($debug $nbself %MYOBJHASH) ;
$debug = 0 ;
$nbself = 0 ;
(%MYOBJHASH) = () ;
use fthrout ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = lc(shift) ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    # We refuse to create an object from no where
    croak("fthcall: Violent build for $name is illegal, API error") ;
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  bless $self;
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  $nbself++ ;
  if (exists $self->{"name"}) {
    my $name = lc($self->{"name"}) ;
    $MYOBJHASH{$name} = $self ;
  }
  return $self ;
}
###########################################################################
sub old {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  if ($args{-name}) {
    $name = $args{-name} ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
  } else {
    $first = 1 ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
# Gets the number of built objects:
sub getNbself {
  return($nbself) ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
    # Debugging:
    if ($1 eq 'thelink') {
      $self->{"causelink"} = (caller(1))[3] . ':' . (caller(0))[2] ;
    }
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	if ($sval && ref($sval) eq 'fthrout') {
	  print STDOUT "$sval'" . $sval->get(-name) ."', ";
	} else {
	  print STDOUT "$sval, " ;
	}
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	if (exists $array{$sval} && ref($array{$sval}) eq 'fthrout') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->get(-name)  ;
	  print STDOUT", " ;
	} else {
	  print STDOUT "$sval => '$array{$sval}', " ;
	}
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'fthrout')) {
      print STDOUT "$t$cle = $val'" . $val->get(-name) ."',\n";
    } else {
      print STDOUT "$t$cle = $val,\n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
# sort by line number:
sub bynumber {
 $a->{"number"} <=> $b->{"number"} ;
}
###########################################################################
sub addPossible {
  my $self = shift ;
  my $orout = shift ;
  my $finter = shift ;
  if (exists $self->{"MYPOSSCALLS"}) {
    foreach my $erout (@{$self->{"MYPOSSCALLS"}}) {
      if ($erout eq $orout) {
	# Nothing to be done, multiple definition is set
	return ;
      }
    }
  }
  push(@{$self->{"MYPOSSCALLS"}}, $orout) ;
  if ($finter) {
    my $name = $orout->get(-name) ;
    $self->{"MYPOSSINTERS"}->{"$name"} = $finter ;
  }
  $orout->addPossible($self) ;
}
###########################################################################
sub getInterface {
  my $self = shift ;
  my $name = shift ;
  if (exists $self->{"MYPOSSINTERS"}->{"$name"}) {
    return($self->{"MYPOSSINTERS"}->{"$name"}) ;
  } else {
    return(undef) ;
  }
}
###########################################################################
sub getPossibles {
  my $self = shift ;
  if (exists $self->{"MYPOSSCALLS"}) {
    return(@{$self->{"MYPOSSCALLS"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
sub okLink {
  my $self = shift ;
  if ($self->{"thelink"}) {
    return(0) ;
  } elsif (exists $self->{"OKS"} && $#{$self->{"OKS"}} == 0) {
    # Only one answer, return it even though it may be a far one...
    return($self->{"ok"}) ;
  } else {
    return(undef) ;
  }
}
###########################################################################
sub getLink {
  my $self = shift ;
  if ($self->{"thelink"}) {
    # Safe stuff: prototype is equal
    return($self->{"thelink"}) ;
  } elsif (exists $self->{"ok"}) {
    # One or more answers
    if ($#{$self->{"OKS"}} == 0) {
      # Only one answer, return it even though it may be a far one...
      return($self->{"alink"}) ;
    } else {
      # Multiple answers at same level:
      if ($debug) {
	print STDOUT (caller(0))[3] .": Called routine " . $self->get(-name) . " file " . $self->get(-file) . " line " . $self->get(-number) . " may refer to any of" ;
	foreach (@{$self->{'OKS'}}) {
	  print STDOUT " " . $_->get(-name) ;
	}
	print STDOUT "\n" ;
      }
      return(undef) ;
    }
  } else {
    return(undef) ;
  }
}
###########################################################################
sub debugLink {
  my $self = shift ;
  my $ligne = "Cause:" ;
  if ($self->{"causelink"}) {
    $ligne .= "(" . $self->{"causelink"} . ")" ;
  }
  if ($self->{"thelink"}) {
    $ligne .= " Direct." ;
  } elsif (exists $self->{"ok"}) {
    if ($#{$self->{"OKS"}} == 0) {
      $ligne .= " Tested." ;
    }
  }
  return($ligne) ;
}  
###########################################################################
sub setLink {
  my $self = shift ;
  my $ko = shift ;
  my $fthrout = shift ;
  print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Setting link for $self with ko=$ko to $fthrout " . $fthrout->get(-name) . "\n" if $debug ;
  if (exists $self->{"ok"}) {
    my $ok = $self->{"ok"} ;
    # Mark exists
    if ($ko < $ok) {
      # Better mark new visitor has lower number, it means is more suitable.
      $self->{"ok"} = $ko ;
      $self->{"alink"} = $fthrout ;
      delete $self->{"OKS"} if $self->{"OKS"} ;
      (@{$self->{"OKS"}}) = ($fthrout) ;
      $self->{"causelink"} = (caller(1))[3] . ':' . (caller(0))[2] ;
    } elsif ($ok == $ko) {
      my $addit = 1 ;
      foreach my $bidule (@{$self->{"OKS"}}) {
	if ($fthrout == $bidule) {
	  $addit = 0 ;
	}
      }
      # Same mark, manage multiple:
      push(@{$self->{"OKS"}}, $fthrout) if $addit ;
      $self->{"causelink"} = (caller(1))[3] . ':' . (caller(0))[2] ;
    }
  } else {
    # No mark, initialize:
    $self->{"ok"} = $ko ;
    $self->{"alink"} = $fthrout ;
    $self->{"causelink"} = (caller(1))[3] . ':' . (caller(0))[2] ;
    (@{$self->{"OKS"}}) = ($fthrout) ;
  }
}
###########################################################################
# Adds an argument to a calline:
sub addArgument {
  my $self = shift ;
  my $name = lc(shift) ;
  $self->{"ARGHASH"}->{"$name"} = 1 ;
  push(@{$self->{"ARGORDER"}}, $name) ;
  print STDOUT (caller(0))[3] . ": Adding an arg for $self(".$self->get(-name)."), name=$name file " . $self->get(-file) . " line $. (". (caller(1))[3] . " called at line ". (caller(1))[2] . ":" . (caller())[2] . " by ". (caller(2))[3] .")\n" if $debug ;
}
###########################################################################
# Gets the arguments for a given stuff
sub getfArguments {
  my $self = shift ;
  if (exists $self->{"ARGFULLNAME"}) {
    return(@{$self->{"ARGFULLNAME"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
sub setArgument {
  my $self = shift ;
  my $name = shift ;
  push(@{$self->{"ARGFULLNAME"}}, $name) ;
}
###########################################################################
sub getArgument {
  my $self = shift ;
  my $number = shift ;
  if (exists $self->{"ARGFULLNAME"} && $number <= $#{$self->{"ARGFULLNAME"}}) {
    print STDOUT "Returning ARGFULLNAME: @{$self->{'ARGFULLNAME'}} for $number is $self->{'ARGFULLNAME'}[$number]\n" if $debug ;
    return($self->{"ARGFULLNAME"}[$number]) ;
  } else {
    return("") ;
  }
}
###########################################################################
# Gets the arguments for a given stuff
sub getArguments {
  my $self = shift ;
  if (exists $self->{"ARGORDER"}) {
    return(@{$self->{"ARGORDER"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Gets the arguments for a given stuff
sub replaceArg {
  my $self = shift ;
  my $number = shift ;
  my $oldarg = shift ;
  my $newarg = shift ;
  if ($oldarg && $newarg && exists $self->{"ARGORDER"} && exists $self->{"ARGHASH"} && $self->{"ARGHASH"}->{"$oldarg"} && $self->{"ARGORDER"}[$number] eq $oldarg) {
    # Old argument exists
    #for (my $i=0; $i <= $#{$self->{"ARGORDER"}}; $i++) {
    # Replacement available
    if ($debug) {
      print STDOUT (caller(0))[3] . ": Replacing argument $number=$oldarg by $newarg for ".$self->{"ARGORDER"}[$number]." in " . $self->get(-name) . " line " . $self->get(-number). " file " . $self->get(-file) . "\n" ;
      $self->datadump("\t") ;
    }
    $self->{"ARGORDER"}[$number] = $newarg ;
    delete $self->{"ARGHASH"}->{"$oldarg"} ;
    $self->{"ARGHASH"}->{"$newarg"} = 1 ;
    if (exists $self->{"SCALAR"}->{"$oldarg"}) {
      $self->{"SCALAR"}->{"$newarg"} = $self->{"SCALAR"}->{"$oldarg"} ;
      delete $self->{"SCALAR"}->{"$oldarg"} ;
    }
    if (exists $self->{"ARGTYPES"}->{"$oldarg"}) {
      $self->{"ARGTYPES"}->{"$newarg"} = $self->{"ARGTYPES"}->{"$oldarg"} ;
      delete $self->{"ARGTYPES"}->{"$oldarg"} ;
    } else {
      $self->{"ARGTYPES"}->{"$newarg"} = 'unknown' ;
    }
    $self->{"OPTIONAL"}->{"$newarg"} = 1 ;
    if ($debug) {
      $self->datadump("\t") ;
    }
    return(1) ;
 } else {
    if ($debug) {
      print STDOUT (caller(0))[3] . ": Unable to replace argument $number=$oldarg by $newarg in " . $self->get(-name) . " line " . $self->get(-number). " file " . $self->get(-file) . "\n" ;
      $self->datadump("\t") ;
    }
    return(undef) ;
  }
}
###########################################################################
# Gets the argument's type for a given stuff, try to guess...
sub getType {
  my $self = shift ;
  my $name = lc(shift) ;
  return('void') unless defined($name) ;
  unless ($name =~ /^([\"\'])(.*?)\1$/) {
    $name = $1 if ($name =~ /^[-+](.*)/) ;
    # A trick to simplify the arguments analysis
    $name = $1 if ($name =~ /^(.*?)[\-\+\*\/]/) ;
  }
  if ($debug) {
    if (exists $self->{"ARGTYPES"}->{"$name"}) {
      print STDOUT "\t" . (caller(0))[3] . ": Fetching type $name for " . $self->get(-name) . " is " . $self->{"ARGTYPES"}->{"$name"} . "\n" ;
    } else {
      print STDOUT "\t" . (caller(0))[3] . ": Fetching type $name for " . $self->get(-name) . "\n" ;
    }
  }
  if (exists $self->{"ARGTYPES"}->{"$name"}) {
    return($self->{"ARGTYPES"}->{"$name"}) ;
  } elsif ($name =~ /^([\"\'])(.*?)\1$/) {
    return('character*'.length($2)) ;
  } elsif ($name =~ /^(\.)(.*?)\1$/) {
    return('logical') ;
  } elsif ($name =~ /\./) {
    return('real') ;
  } elsif ($name =~ /^\d+$/) {
    return('integer') ;
  } else {
    return('void') ;
  }
}
###########################################################################
# Adds an argument's type to a routine: debugging performed with calling perl routine
# name irst instead of usual this routine name.
sub setType {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  if (exists $self->{"SCALAR"} && exists $self->{"SCALAR"}->{"$name"} && $value =~ /dimension\(.*?\)/) {
    # Scalar called for a vector field as an input:
    print STDOUT (caller(1))[3] . ": dimension arg name $name used in scalar mode for " . $self->get(-name) . " originates to $value\n" if $debug ;
    $value =~ s/,*dimension\(.*?\)//g ;
    # Eliminate the pointer attribute...
    $value =~ s/,*pointer//g ;
  }
  # NAJ
  $self->{"ARGTYPES"}->{"$name"} = $value ;
  $name = $1 if ($name =~ /^[-+](.*)/) ;
  $name = $1 if ($name =~ /^(.*?)[-+]/) ;
  $self->{"ARGTYPES"}->{"$name"} = $value ;
  print STDOUT "\t" . (caller(0))[3] . " From upper perl=" . (caller(1))[3] . ":" . (caller(1))[2] . ": Argument $name is of type $value for called stuff ".$self->get(-name)."\n" if $debug ;
}
###########################################################################
# Sets a dependency for current
sub setDepend {
  my $self = shift ;
  my $number = shift ;
  my $ocall = shift ;
  $self->{"DEPENDS"}->{"$number"} = $ocall ;
}
###########################################################################
# Sets a dependency for current
sub getDepends {
  my $self = shift ;
  if (exists $self->{"DEPENDS"}) {
    return(%{$self->{"DEPENDS"}} ) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Sets a dependency for current
sub getDepend {
  my $self = shift ;
  my $number = shift ;
  if (exists $self->{"DEPENDS"} && exists $self->{"DEPENDS"}->{"$number"}) {
    return($self->{"DEPENDS"}->{"$number"}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Gets the proto for a call:
sub getProto {
  my $self = shift ;
  my (@args) = $self->getArguments() ;
  my (@proto) = () ;
  foreach my $arg (@args) {
    # croak("KO '@args': " . $self->get(-name)) unless $arg ;
    print STDOUT  "\t\t\t" . (caller(0))[3] . ": In call $self=" . $self->get(-name) . ", args of $arg are " . $self->getType("$arg") . "\n" if $debug ;
    push(@proto, "$arg=" . $self->getType("$arg")) ;
  }
  return(@proto) ;
}
###########################################################################
# Counts the number of parenthesis
sub countParenthesis {
  my $nbparenthe = shift ;
  my $ligne = shift ;
  while ($ligne =~ /^(.*?)([\(\)])(.*)/) {
    $ligne = $3 ;
    if ($2 eq '(') {
      $nbparenthe++ ;
    } else {
      $nbparenthe-- ; 
    }
  }
  return($nbparenthe, $ligne) ;
}
###########################################################################
sub setScalar {
  my $self = shift ;
  my $name = lc(shift) ;
  $self->{"SCALAR"}->{"$name"} = 1 ;
}
###########################################################################
sub readf77 {
  my $self = shift ;
  my $file = shift ;
  $_ = shift ;
  my $cont = 1 ;
  my $defarg = "" ;
  my $chtml = &chtml::getCurrent() ;
  my $curpos ;
  if ((ref $file) && (fileno $file)) {
    $self->set(-number => $.) ;
    $curpos = tell($file) ;
  }
  my $nbparenthe = 0 ;
  my (@arglist) = () ;
  my (@argfull) = () ;
  my $str = "" ;
  my $outs = "" ;
  my $outi = "" ;
  my $advance = 0 ;
  # Shall be the number of arguments on return:
  my $nbargs = undef ;
  while ($cont > 0) {
    # print STDOUT "cont=$cont\n" ;
    my $cc = "" ;
    # Parsing an f77 call file below: parsing ends at end of parenthesis
    if ($cont && /^([\s\d]{5}[\&\S]\s*)(.*)/) {
      my $ligne = $2 ;
      my $outp = $ligne ;
      if ($cont == 1) {
	$cont = 2 ;
      } else {
	$outi .= $1 ;
      }
      # Resetting the arglist in case of multiple line call:
      if ($ligne =~ /[\'\"\(\)\,\!]/) {
	# Eliminate the commas and parenthesis inside the strings !
	while ($ligne =~ /^(.*?)([\'\"\(\)\,\!])(.*)/) {
	  my $bef = $1 ;
	  my $sep = $2 ;
	  my $aft = $3 ;
	  if (defined($nbargs)) {
	    if ($sep eq ',' && $nbparenthe == 1) {
	      $argfull[$nbargs] .= $bef ;
	      print STDOUT "\t\t" . " 1) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all=@argfull\n" if $debug ;
	    } elsif ($sep eq ')' && $nbparenthe == 1) {
	      $argfull[$nbargs] .= $bef ;
	      print STDOUT "\t\t" . " 2) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all=@argfull\n" if $debug ;
	    } else {
	      $argfull[$nbargs] .= $bef . $sep ;
	      print STDOUT "\t\t" . " 3) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all=@argfull\n" if $debug ;
	    }
	  }
	  if ($str) {
	    if ($str eq $sep) {
	      $str = "" ;
	      # End of string:
	      $outs .= $bef . $sep ;
	      print STDOUT "\t" . (caller(0))[3] . ": String manager, bef=$1, sep=$2, aft=$3, adding argument $outs\n" if $debug ;
	      if ($advance) {
		push(@arglist, $outs) ;
		$advance = 0 ;
	      }
	      $outs = "" ;
	      $outi .= $bef . $sep ;
	      $outi .= "</STRING>" ;
	    } else {
	      # Nothing to do, inside the string...
	      $outs .= $bef . $sep ;
	      $outi .= $bef . $sep ;
	    }
	  } elsif ($sep eq '"' || $sep eq "'") {
	    $str = $sep ;
	    $outs = $sep ;
	    $outi .= $bef . "<STRING>" . $sep ;
	    print STDOUT "\t" . (caller(0))[3] . ": String start, bef=$1, sep=$2, aft=$3, pushing $outi\n" if $debug ;
	  } elsif ($sep eq '(' || $sep eq ")") {
	    print STDOUT "\t" . (caller(0))[3] . ": Bracket manager, bef=$1, sep=$2, aft=$3, advance=$advance, p=$nbparenthe\n" if $debug ;
	    # Bracket manager
	    if ($sep eq '(') {
	      if ($nbparenthe == 0) {
		$outi .= $bef . '<BRANCHEMENT>' . $sep ;
	      } else {
		$outi .= $bef . $sep ;
	      }
	      if ($nbparenthe == 0) {
		# Start of the stuff, initialize the arg list ?
		$advance = 1 ;
		$nbargs = 0 ;
		print STDOUT "\t" . (caller(0))[3] . ": Start of parenthesis file " . $self->get(-file) . ", line " . $self->get(-number) . ", end of line <$ligne> for <$_>.\n" if $debug ;
	      } elsif ($nbparenthe == 1) {
		# LHS is an argument and we add another parenthesis
		if ($advance && $bef =~ /(.*?)([\w\%\.]+)/) {
		  # Argument is $2...
		  print STDOUT "\t" . (caller(0))[3] . ": Pushing $2 to @arglist due to a=$advance and bef=$bef\n" if $debug ;
		  push(@arglist, $2 . '{') ;
		  $advance = 0 ;
		} elsif (!$advance && $bef =~ /^\%([\w\%\.]+)/) {
		  $arglist[$#arglist] .= '%' . $1 ;
		}
	      } else {
		print STDOUT "\t" . (caller(0))[3] . ": Ready to advance ($advance)\n" if $debug ;
		if ($advance && $aft =~ /(.*?)([\w\%\.]+)/) {
		  $defarg = $2 . '{' ;
		}
	      }
	      $nbparenthe++ ;
	    } elsif ($sep eq ')') {
	      $outi .= $bef . $sep ;
	      if ($nbparenthe == 1) {
		# Close parenthesis of the call
		$nbargs++ ;
		# End of the call stuff:
		print STDOUT "\t" . (caller(0))[3] . ": End of parenthesis file " . $self->get(-file) . ", line " . $self->get(-number) . ", end of line <$ligne> is <$3> for <$_>, parsing($advance)=$1\n" if $debug ;
		$outi .= "</BRANCHEMENT>" ;
		$_ = $aft ;
		# Allow the call toto syntax
		$cont = 0 unless $nbparenthe ;
		if ($advance && $bef =~ /(.*?)([\w\%\.]+)/) {
		  push(@arglist, $2) ;
		  $advance = 0 ;
		} elsif ($advance) {
		  if ($defarg) {
		    push(@arglist, $defarg) ;
		    $defarg = "" ;
		  }
		  $advance = 0 ;
		}
		$cont = 0 ;
		if (/^(\s*)\!(.*)/) {
		  $outi .= "$1" ;
		  $cc = $2 ;
		  $_ = "" ;
		}
		$ligne = "" ;
		last ;
	      } else {
		# Closing parenthesis
		if ($advance && !$defarg && $bef =~ /(.*?)([\w\%\.]+)/) {
		  $defarg = $2 . '{' ;
		}
	      }
	      $nbparenthe-- ;
	    }
	  } elsif ($sep eq ',') {
	    # A comma separating stuffs...
	    $outi .= $bef . $sep ;
	    if ($nbparenthe == 1) {
	      # Close parenthesis
	      #$nbargs++ ;
	      if ($advance && $bef =~ /(.*?)([\w\%\.]+)/) {
		# Argument is $2...
		print STDOUT "\t" . (caller(0))[3] . ':' . (caller(1))[2] . ": Adding the argument $2\n" if $debug ;
		push(@arglist, $2) ;
		$advance = 1 ;
	      } else {
		if ($defarg) {
		  push(@arglist, $defarg) ;
		  $defarg = "" ;
		}
		$advance = 1 ;
	      }
	      if ($bef =~ /^\%([\w\%\.]+)/) {
		$arglist[$#arglist] .= '%' . $1 ;
	      }
	      print STDOUT "\t" . (caller(0))[3] . ": Met comma, advance=$advance\n" if $debug ;
	    }
	  } elsif ($sep eq '!') {
	    $outi .= $bef . $sep ;
	    $cc .= $aft ;
	    $ligne = "" ;
	    last ;
	  } else {
	    # A comma separating stuffs...
	    print STDOUT "Strange event $_\n" ;
	    $outi .= $bef . $sep ;
	    if ($nbparenthe == 1) {
	      if ($bef =~ /(.*?)([\w\%\.]+)/) {
		# Argument is $2...
		print STDOUT "\t" . (caller(0))[3] . ": Adding an argument $2 from $1\n" ;
		$advance = 0 ;
		push(@arglist, $2) ;
	      }
	    }
	  }
	  $ligne = $aft ;
	}
      }
      $outi .= $ligne ;
      if ($ligne && $advance && $ligne =~ /(.*?)([\w\%\.]+)/) {
	push(@arglist, $2) ;
	$advance = 0 ;
      }
      $chtml->addPot("$outi") ;
      $outi = "" ;
    } elsif (/^c(.*)/i) {
      $cc = $1 ;
      $_ = "" ;
    } elsif (/^(\s*)\!(.*)/) {
      $chtml->addRaw($1) ;
      $cc = $2 ;
      $_ = "" ;
    } elsif (/^\s*$/) {
      # Only coments 
    } else {
      # Go back:
      if (ref $file && fileno $file) {
	my $pos = tell($file) - $curpos ; ;
	seek($file,-$pos,1);
	$.-- ;
      }
      $_ = "" ;
      $cont = -1 ;
    }
    if (ref $file && fileno $file) {
      $curpos = tell($file) ;
    }
    # Manage the cc stuff and add it to the output...
    if ($cc) {
      $chtml->addRaw('<COMMENT>') ;
      if ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	$chtml->addLaCoM("<A Name=C_$.></A>") ;
      }
      while ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	my $incp = "$3.$4" ;
	$cc = $5 ;
	$chtml->addLaTeX($1.$2." ") ;
	unless (-f "$incp") {
	  my $nase = basename(dirname($incp)) ;
	  my $base = dirname($self->get(-file)) ;
	  if (-f "$base/$incp") {
	    $incp = "$base/$incp" ;
	  } elsif (-f "$base/" . basename($incp)) {
	    $incp = "$base/" . basename($incp) ;
	  } elsif (-f "$nase/" . basename($incp)) {
	    $incp = "$nase/" . basename($incp) ;
	  } elsif ($main::opt_L) {
	    # LaTeX inputs directories
	    # Possible...
	    my @TAGSHTML_PATH = (split ( /:/, $main::opt_L)) ;
	    foreach $base (@TAGSHTML_PATH) {
	      if (-f "$base/$incp") {
		$incp = "$base/$incp" ;
		last ;
	      }
	    }
	  }
	}
	if (-f "$incp") {
	  my $other = new fthinp(-file => "$incp", -name => "$incp") ;
	  $self->get(-father)->addInput("$incp" => $other) ;
	  $chtml->addLaCoM($other) ;
	} else {
	  my $refe = new oref(-name => $incp, -type => "fthinp") ;
	  $chtml->addLaCoM($refe) ;
	}
      }
      $chtml->addLaTeX($cc) ;
      $chtml->addRaw("</COMMENT>") ;
    }
    if ($cont <= 0) {
      print STDOUT "\t" . (caller(0))[3] . ": In called " . $self->get(-name) . " \n" if $debug ;
      foreach my $name (@arglist) {
	my $s = 0 ;
	if ($name =~ /(.*)\s*\{(.*)$/) {
	  $s = 1 ;
	  $name = $1.$2 ;
	}
	if ($name =~ /[\'\"](.*)\1/) {
	  # NAJ: add the type
	} else {
	  $name = $1 if ($name =~ /^[\-\+\*\/](.*)/) ;
	  $name = $1 if ($name =~ /^(.*?)[\-\+\*\/]/) ;
	}
	if ($s) {
	  $self->addArgument("$name") ;
	  $self->setScalar("$name") ;
	  print STDOUT (caller(0))[3] . ": Argument $name is scalar in call ".$self->get(-name) . "(file ".$self->get(-file) .", number=".$self->get(-number) .")\n" if $debug ;
	} else {
	  $self->addArgument("$name") ;
	  print STDOUT (caller(0))[3] . ": Argument $name in call ".$self->get(-name) . "(file ".$self->get(-file) .", number=".$self->get(-number) .")\n" if $debug ;
	}
      }
      # Now parsing the full argument list:
      foreach my $name (@argfull) {
	print STDOUT "Setting argument $name for fthcall=" . $self->get(-name) . " file " . $self->get(-file) . " line " . $self->get(-number) . "\n" if $debug ;
	$self->setArgument($name) ;
      }
    } elsif ($cont > 0) {
      if (ref $file && fileno $file && defined($_ = <$file>)) {
	print STDOUT "\tReading advanced line=$_\n" if $debug ;
	chomp ;
      } else {
	print STDERR (caller(0))[3] .  "fthcall error line=$. fthcall= " . $self->get(-name) .  " file=" . $self->get(-file) . "\n" ;
	$cont = 0 ;
	return ;
      }
    }
  }
}
###########################################################################
# Convert an argument string into a guessed name
sub guessSimple {
  my $self = shift ;
  my $fname = shift ;
  my $name ;
  if ($fname =~ /^\s*([\"\'])(.+)\1/) {
    # Sounds like a character string:
    $name = $1 . $2 . $1 ;
  } elsif ($fname =~ /([\"\'])(.+)\1\s*$/) {
    # Sounds like a character string:
    $name = $1 . $2 . $1 ;
  } elsif ($fname =~ /^\s*([\w\%]+)\(.*?\)\%([\w\%]+)$/) {
    # Direct access to an element in an array structure
    $name = $1 . '%' . $2 ;
  } elsif ($fname =~ /^\s*([\w\%]+)\s*$/) {
    $name = $1 ;
  } elsif ($fname =~ /^\s*([\w\%]+)\s*\((.*)\)$/) {
    $name = $1 ;
    $self->setScalar("$name") ;
  } elsif ($fname =~ /^\s*\(.*\)\s*(.+)$/) {
    $name = $self->guessSimple($1) ;
  } elsif ($fname =~ /^\s*\(\s*(.*)\s*\)\s*$/) {
    $name = $self->guessSimple($1) ;
  } elsif ($fname =~ /^\s*([\w\%\.]+)([\-\+\*\/]+)\s*$/) {
    $name = $self->guessSimple($1) ;
  } elsif ($fname =~ /([\w\%\.]+)([\+\-\d\_\s]*)/) {
    $name = $1 ;
  } else {
    $name = $fname ;
  }
  return($name) ;
}

###########################################################################
sub parsef90 {
  my $self = shift ;
  my $file = shift ;
  my $ligne = shift ;
  my $cont = 100 ; # Sets a limit to the maximum number of arugments...
  my $nbparenthe = 0 ;
  my (@arglist) = () ;
  my (@argfull) = () ;
  my $str = "" ;
  # Shall be the number of arguments on return:
  my $nbargs = undef ;
  while ($cont > 0) {
    $cont-- ;
    print STDOUT "parsef90: cont=$cont, $ligne\n" if $debug ;
    # Eliminate the commas and parenthesis inside the strings !
    while ($ligne =~ /^(.*?)([\'\"\(\)\,])(.*)/) {
      my $bef = $1 ;
      my $sep = $2 ;
      my $aft = $3 ;
      if (defined($nbargs) && $nbparenthe == 1) {
	if ($sep eq ',') {
	  $argfull[$nbargs] .= $bef if ($bef ne "");
	  print STDOUT "\t\t" . " 1) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all='@argfull'\n" if $debug ;
	} elsif ($sep eq ')') {
	  $argfull[$nbargs] .= $bef if ($bef ne "") ;
	  print STDOUT "\t\t" . " 2) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all='@argfull'\n" if $debug ;
	} else {
	  $argfull[$nbargs] .= $bef . $sep ;
	  print STDOUT "\t\t" . " 3) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all='@argfull'\n" if $debug ;
	}
      } elsif (defined($nbargs)) {
	print STDOUT "\t\t" . " 5) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all='@argfull'\n" if $debug ;
	$argfull[$nbargs] .= $bef . $sep ;
      }
      if ($str) {
	if ($str eq $sep) {
	  $str = "" ;
	  # End of string:
	  print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "A: String manager, bef=$1, sep=$2, aft=$3\n" if $debug ;
	} else {
	  # Nothing to do, inside the string...
	}
      } elsif ($sep eq '"' || $sep eq "'") {
	$str = $sep ;
	print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "B: String start, bef=$1, sep=$2, aft=$3, pushing outi\n" if $debug ;
      } elsif ($sep eq '(' || $sep eq ')') {
	print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "C: Bracket manager, bef=$1, sep=$2, aft=$3, advance=advance, p=$nbparenthe\n" if $debug ;
	# Bracket manager
	if ($sep eq '(') {
	  if ($nbparenthe == 0) {
	    #$outi .= $bef . '<BRANCHEMENT>' . $sep ;
	  } else {
	    #$outi .= $bef . $sep ;
	  }
	  if ($nbparenthe == 0) {
	    # Start of the stuff, initialize the arg list ?
	    $nbargs = 0 ;
	    print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "D: Start of parenthesis(outi) file " . $self->get(-file) . ", line " . $self->get(-number) . ", end of line <$aft> for <$_>.\n" if $debug ;
	  } elsif ($nbparenthe == 1) {
	    # LHS is an argument and we add another parenthesis
	  } else {
	    print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "F: Ready to advance (advance)\n" if $debug ;
	  }
	  $nbparenthe++ ;
	} elsif ($sep eq ')') {
	  #$outi .= $bef . $sep ;
	  if ($nbparenthe == 1) {
	    # Close parenthesis
	    #$nbargs++ ;
	    # End of the call stuff:
	    print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "H: End of parenthesis file " . $self->get(-file) . ", line $. called from " . $self->get(-number) . ", bef=$bef, end of line <$ligne> lets aft=<$aft> for <$_>, parsing(advance)=$1\n" if $debug ;
	    #$outi .= "</BRANCHEMENT>" ;
	    $_ = $aft ;
	    # Terminate the stuff...
	    $cont = 0 ;
	    $ligne = "" ;
	    last ;
	  } else {
	    # Closing parenthesis
	  }
	  $nbparenthe-- ;
	}
      } elsif ($sep eq ',') {
	# A comma separating stuffs...
	#$outi .= $bef . $sep ;
	if ($nbparenthe == 1) {
	  # Add comma
	  $nbargs++ ;
	  print STDOUT "\t" . "Adding comma $nbargs is next arg, sep=$sep, bef was $bef.\n" if $debug ;
	  print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "M: Met comma, advance=advance\n" if $debug ;
	}
      } else {
	# A comma separating stuffs...
	print STDOUT "Strange event $_\n" ;
	#$outi .= $bef . $sep ;
	if ($nbparenthe == 1) {
	  if ($bef =~ /(.*?)([\w\%\.]+)/) {
	    # Argument is $2...
	    print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "T: Adding an argument $2 from $1\n" ;
	  }
	}
      }
      $ligne = $aft ;
    }
  }
  print STDOUT "\t" . (caller(0))[3] . ": In called " . $self->get(-name) . " argfull='@argfull', size=$#argfull\n" if $debug ;
  foreach my $fname (@argfull) {
    # Try to guess a simple name from it:
    if (defined($fname) && $fname ne "") {
      my $name = $self->guessSimple($fname) ;
      print STDOUT "\t" . (caller(0))[3] . ": Setting argument $name for fthcall=" . $self->get(-name) . " file " . $self->get(-file) . " line " . $self->get(-number) . " arg='$fname', name='$name'\n" if $debug ;
      $self->addArgument("$name") ;
      $self->setArgument($fname) ;
    }
  }
}

###########################################################################
sub readf90 {
  my $self = shift ;
  my $file = shift ;
  $_ = shift ;
  my $cont = 1 ;
  my $defarg = "" ;
  my $chtml = &chtml::getCurrent() ;
  my $curpos ;
  if ((ref $file) && (fileno $file)) {
    $self->set(-number => $.) ;
    $curpos = tell($file) ;
  }
  my $nbparenthe = 0 ;
  my (@arglist) = () ;
  my (@argfull) = () ;
  my $str = "" ;
  my $outs = "" ;
  my $outi = "" ;
  my $advance = 0 ;
  # Shall be the number of arguments on return:
  my $nbargs = undef ;
  my $num ;
  my $preif = 0 ;
  my(@ploops) = () ;
  while ($cont > 0) {
    print STDOUT "readf90: cont=$cont\n" if $debug ;
    my $cc = "" ;
    my $fcont = 0 ;
    # Parsing an f90 call file below: cont=3 means continuation symbol without
    # reporting it next line...
    if ($cont && /^(\s*\&\s*)(.*)/ || $cont == 3 && /(\s*)(.*)/) {
      my $ligne = $2 ;
      my $outp = $ligne ;
      if ($cont == 1) {
	$cont = 2 ;
      } else {
	$outi .= $1 ;
      }
      #print STDOUT "Parsing $ligne\n" if $debug;
      if ($ligne =~ /^\#(ifdef|ifndef|if|elif|else|endif)(.*)/i) {
	# Preprocessing tests, this seems good code below:
	if ($1 eq "endif") {
	  $num = pop(@ploops) ;
	  if ($num) {
	    $outi .= "<A Href=#C_$num>".'#'."$1</A>" ;
	    $preif-- ;
	  } else {
	    $outi .= '#'."$1$2" ;
	  }
	} elsif ($1 eq "if" || $1 eq "ifdef" || $1 eq "ifndef") {
	  push(@ploops, $.) ;
	  $preif++ ;
	  $outi .= "<TEST><A Name=C_$.>".'#'."$1</A></TEST>$2" ;
	} else {
	  $num = pop(@ploops) ;
	  if ($num) {
	    push(@ploops, $num) ;
	    $outi .= "<A Href=#C_$num>".'#'."$1</A>" ;
	  } else {
	    $outi .= '#'."$1$2" ;
	  }
	}
	$ligne = "" ;
	$fcont = 1 ;
      } elsif ($ligne =~ /[\'\"\(\)\,\!\&]/) {
	# Eliminate the commas and parenthesis inside the strings !
	while ($ligne =~ /^(.*?)([\'\"\(\)\,\!\&])(.*)/) {
	  my $bef = $1 ;
	  my $sep = $2 ;
	  my $aft = $3 ;
	  if (defined($nbargs) && $nbparenthe == 1) {
	    if ($sep eq ',') {
	      $argfull[$nbargs] .= $bef if ($bef ne "");
	      print STDOUT "\t\t" . " 1) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all='@argfull'\n" if $debug ;
	    } elsif ($sep eq ')') {
	      $argfull[$nbargs] .= $bef if ($bef ne "") ;
	      print STDOUT "\t\t" . " 2) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all='@argfull'\n" if $debug ;
	    } elsif ($sep eq '&' || $sep eq '!') {
	      $argfull[$nbargs] .= $bef if ($bef ne "") ;
	      print STDOUT "\t\t" . " 4) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all='@argfull'\n" if $debug ;
	    } else {
	      $argfull[$nbargs] .= $bef . $sep ;
	      print STDOUT "\t\t" . " 3) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all='@argfull'\n" if $debug ;
	    }
	  } elsif (defined($nbargs)) {
	    print STDOUT "\t\t" . " 5) Adding to argfull number $nbargs value $bef$sep to $argfull[$nbargs], brackets is $nbparenthe all='@argfull'\n" if $debug ;
	    $argfull[$nbargs] .= $bef . $sep ;
	  }
	  if ($str) {
	    if ($str eq $sep) {
	      $str = "" ;
	      # End of string:
	      $outs .= $bef . $sep ;
	      print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "A: String manager, bef=$1, sep=$2, aft=$3, adding argument $outs\n" if $debug ;
	      $outs = "" ;
	      $outi .= $bef . $sep ;
	      $outi .= "</STRING>" ;
	    } else {
	      # Nothing to do, inside the string...
	      $outs .= $bef . $sep ;
	      $outi .= $bef . $sep ;
	    }
	  } elsif ($sep eq '"' || $sep eq "'") {
	    $str = $sep ;
	    $outs = $sep ;
	    $outi .= $bef . "<STRING>" . $sep ;
	    print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "B: String start, bef=$1, sep=$2, aft=$3, pushing $outi\n" if $debug ;
	  } elsif ($sep eq '(' || $sep eq ')') {
	    
	    print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "C: Bracket manager, bef=$1, sep=$2, aft=$3, advance=$advance, p=$nbparenthe\n" if $debug ;
	    # Bracket manager
	    if ($sep eq '(') {
	      if ($nbparenthe == 0) {
		$outi .= $bef . '<BRANCHEMENT>' . $sep ;
	      } else {
		$outi .= $bef . $sep ;
	      }
	      if ($nbparenthe == 0) {
		# Start of the stuff, initialize the arg list ?
		$nbargs = 0 ;
		print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "D: Start of parenthesis($outi) file " . $self->get(-file) . ", line " . $self->get(-number) . ", end of line <$aft> for <$_>.\n" if $debug ;
	      } elsif ($nbparenthe == 1) {
		# LHS is an argument and we add another parenthesis
	      } else {
		print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "F: Ready to advance ($advance)\n" if $debug ;
	      }
	      $nbparenthe++ ;
	    } elsif ($sep eq ')') {
	      $outi .= $bef . $sep ;
	      if ($nbparenthe == 1) {
		# Close parenthesis for the current call
		# $nbargs++ ;
		# End of the call stuff:
		print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "H: End of parenthesis file " . $self->get(-file) . ", line $. called from " . $self->get(-number) . ", bef=$bef, end of line <$ligne> lets aft=<$aft> for <$_>, parsing($advance)=$1\n" if $debug ;
		$outi .= "</BRANCHEMENT>" ;
		$_ = $aft ;
		# Terminate the stuff...
		$cont = 0 ;
		if (/^(\s*)\!(.*)/) {
		  $outi .= "$1" ;
		  $cc = $2 ;
		  $_ = "" ;
		}
		$ligne = "" ;
		last ;
	      } else {
		# Closing parenthesis for a sub-element of the call
	      }
	      $nbparenthe-- ;
	    }
	  } elsif ($sep eq ',') {
	    # A comma separating stuffs...
	    $outi .= $bef . $sep ;
	    if ($nbparenthe == 1) {
	      # Add comma
	      $nbargs++ ;
	      $argfull[$nbargs] = "" ;
	      print STDOUT "\t" . "Adding comma $nbargs is next arg, sep=$sep, bef was $bef.\n" if $debug ;
	      print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "M: Met comma, advance=$advance\n" if $debug ;
	    }
	  } elsif ($sep eq '&') {
	    # To be continued ?
	    $fcont = 1 ; # Forces the continuation mode
	    $outi .= $bef . $sep ;
	    if ($nbparenthe == 1) {
	      # Maybe we need to add another argument to the current list:
	      if ($bef =~ /(.*?)([\w\%\.]+)/) {
		# Argument is $2...
		# $nbargs++ ;
	      }
	      print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "P: Met continuation, advance=$advance\n" if $debug ;
	    }
	  } elsif ($sep eq '!') {
	    $outi .= $bef ;
	    $cc .= $aft ;
	    $ligne = "" ;
	    if ($nbparenthe == 1) {
	      # Maybe we need to add another argument to the current list:
	      if ($bef =~ /(.*?)([\w\%\.]+)/) {
		# Argument is $2...
		# $nbargs++ ;
	      }
	      print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "S: Met comment, advance=$advance\n" if $debug ;
	    }
	    last ;
	  } else {
	    # A comma separating stuffs...
	    print STDOUT "Strange event $_\n" ;
	    $outi .= $bef . $sep ;
	    if ($nbparenthe == 1) {
	      if ($bef =~ /(.*?)([\w\%\.]+)/) {
		# Argument is $2...
		print STDOUT "\t" . (caller(0))[3] .':' . (caller(0))[2] . "T: Adding an argument $2 from $1\n" ;
	      }
	    }
	  }
	  $ligne = $aft ;
	}
      }
      $outi .= $ligne ;
      # Continue line only on continuation symbol:
      if ($ligne && $ligne =~ /(.*)\&/) {
	$cont = 3 ;
	$ligne = $1 ;
      } elsif ($fcont) {
	$cont = 3 ;
      } else {
 	$cont = 0 ;
      }
      # print STDOUT "addPot:$outi.\n" ;
      $chtml->addPot("$outi") ;
      $outi = "" ;
    } elsif (/^c(.*)/i) {
      $cc = $1 ;
      $_ = "" ;
    } elsif (/^(\s*)\!(.*)/) {
      $chtml->addRaw($1) ;
      $cc = $2 ;
      $_ = "" ;
    } elsif (/^\s*$/) {
      # Only coments : nothing to do.
    } else {
      # Went too far in fortran 90
      print STDOUT "\t" . (caller(0))[3] . " internal error file ".$self->get(-file)." line $. started at ".$self->get(-number)." skipped one line=<$_> in 90\n" ;
      if (ref $file && fileno $file) {
	my $pos = tell($file) - $curpos ; ;
	seek($file,-$pos,1);
	$.-- ;
      }
      $_ = "" ;
      $cont = -1 ;
    }
    if (ref $file && fileno $file) {
      $curpos = tell($file) ;
    }
    # Manage the cc stuff and add it to the output...
    if ($cc) {
      $chtml->addRaw('<COMMENT>') ;
      #$chtml->addProtect("fthcall:$.") ;
      if ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	$chtml->addLaCoM("<A Name=C_$.></A>") ;
	while ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  my $incp = "$3.$4" ;
	  $cc = $5 ;
	  $chtml->addLaTeX($1.$2." ") ;
	  unless (-f "$incp") {
	    my $nase = basename(dirname($incp)) ;
	    my $base = dirname($self->get(-file)) ;
	    if (-f "$base/$incp") {
	      $incp = "$base/$incp" ;
	    } elsif (-f "$base/" . basename($incp)) {
	      $incp = "$base/" . basename($incp) ;
	    } elsif (-f "$nase/" . basename($incp)) {
	      $incp = "$nase/" . basename($incp) ;
	    } elsif ($main::opt_L) {
	      # LaTeX inputs directories
	      # Possible...
	      my @TAGSHTML_PATH = (split ( /:/, $main::opt_L)) ;
	      foreach $base (@TAGSHTML_PATH) {
		if (-f "$base/$incp") {
		  $incp = "$base/$incp" ;
		  last ;
		}
	      }
	    }
	  }
	  if (-f "$incp") {
	    my $other = new fthinp(-file => "$incp", -name => "$incp") ;
	    $self->get(-father)->addInput("$incp" => $other) ;
	    $chtml->addLaCoM($other) ;
	  } else {
	    my $refe = new oref(-name => $incp, -type => "fthinp") ;
	    $chtml->addLaCoM($refe) ;
	  }
	}
      }
      $chtml->addLaTeX("$cc") ;
      #$chtml->addProtect("fin fthcall($.:$cc)") ;
      $chtml->addRaw('</COMMENT>') ;
      $_ = "" ;
      $cc = "" ;
    }
    if ($cont <= 0) {
      print STDOUT "\t" . (caller(0))[3] . ": In called " . $self->get(-name) . " argfull='@argfull', size=$#argfull\n" if $debug ;
      foreach my $fname (@argfull) {
	# Try to guess a simple name from it:
	if (defined($fname) && $fname ne "") {
	  my $name = $self->guessSimple($fname) ;
	  print STDOUT "\t" . (caller(0))[3] . ": Setting argument $name for fthcall=" . $self->get(-name) . " file " . $self->get(-file) . " line " . $self->get(-number) . " arg='$fname', name='$name'\n" if $debug ;
	  $self->addArgument("$name") ;
	  $self->setArgument($fname) ;
	}
      }
    } elsif ($cont > 0) {
      if (ref $file && fileno $file && defined($_ = <$file>)) {
	print STDOUT "\tReading advanced line=$_\n" if $debug ;
	chomp ;
      } else {
	print STDERR (caller(0))[3] .  "fthcall error line=$. fthcall= " . $self->get(-name) .  " file=" . $self->get(-file) . "\n" ;
	$cont = 0 ;
	return ;
      }
    }
  }
}
###########################################################################
1;

