package fthext ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A bad call containers.
###########################################################################
use strict;
use Carp ;
###########################################################################
use fthrout ;
use fthcall ;
use fthtype ;
use fthmod ;
use fthinter ;
use ihtml ;
###########################################################################
use vars qw($debug $nbself %MYOBJHASH) ;
$debug = 0 ;
$nbself = 0 ;
(%MYOBJHASH) = () ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = lc(shift) ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  if ($args{-name}) {
    $name = lc($args{-name}) ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
  } else {
    $nbself++ ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug;
  }
  return $self ;
}
###########################################################################
# Gets the number of built objects:
sub getNbself {
  return($nbself) ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	if ($sval && ref($sval) eq 'fthcall') {
	  print STDOUT "$sval'" . $sval->get(-name) ."', ";
	  $sval->datadump("$t\t") ;
	} else {
	  print STDOUT "$sval, " ;
	}
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	if (exists $array{$sval} && ref($array{$sval}) eq 'fthcall') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->get(-name)  ;
	  print STDOUT", " ;
	} else {
	  print STDOUT "$sval => '$array{$sval}', " ;
	}
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'fthcall')) {
      print STDOUT "$t$cle = $val'" . $val->get(-name) ."',\n";
    } else {
      $val = "" unless $val ;
      print STDOUT "$t$cle = $val,\n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
# Adds a call
sub addStuff {
  my $self = shift ;
  my $bidule = shift ;
  push(@{$self->{"STUFFS"}}, $bidule) ;
}
###########################################################################
# Gets all th contained stuffs...
sub getStuffs {
  my $self = shift ;
  if (exists $self->{"STUFFS"}) {
    return(@{$self->{"STUFFS"}}) ;
  } else {
    return((undef)) ;
  }
}
###########################################################################
# Gets a procedure:
sub getProcedure {
  my $self = shift ;
  my $name = lc(shift) ;
  my $orout ;
  if (exists $self->{"STUFFS"}) {
    foreach $orout (@{$self->{"STUFFS"}}) {
      if ($orout && ref($orout) eq 'fthcall') {
	if (lc($orout->get(-name)) eq $name) {
	  return($orout) ;
	}
      }
    }
  }
  return(undef)
}
###########################################################################
sub any {
  if (%MYOBJHASH) {
    return(1) ;
  } else {
    return(0) ;
  }
}
###########################################################################
1;

