package fthinc ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for fthinc 90 files.
###########################################################################
use strict;
use Carp ;
use File::Basename ;
use ihtml ;
use chtml ;
use vars qw($debug %MYOBJHASH) ;
$debug = 0 ;
(%MYOBJHASH) = () ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = lc(shift) ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    # We refuse to create an object from no where
    debug("Violent build for $name is illegal, API error") if $debug ;
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  if ($args{-name}) {
    $name = lc($args{-name}) ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
  } else {
    $first = 1 ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
sub printHash {
  my $self = shift ;
  my $t = shift ;
  my $hash = shift ;
  while(my($cle,$val) = each(%{$hash})) {
    print STDOUT "$t$cle => $val, " ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	print STDOUT "$sval, " ;
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	if (exists $array{$sval} && ref($array{$sval}) eq 'HASH') {
	  print STDOUT "$t\t$sval => {"  ;
	  $self->printHash("\t$t", $array{$sval}) ;
	  print STDOUT"}, " ;
	} else {
	  print STDOUT "$sval => $array{$sval}, " ;
	}
      }
      print STDOUT "},\n" ;
    } else {
      print STDOUT "$t$cle = $val, \n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
sub any {
  if (%MYOBJHASH) {
    return(1) ;
  } else {
    return(0) ;
  }
}
###########################################################################
sub alphabetic {
  lc($a) cmp lc($b) ||  $a cmp $b ;
}
###########################################################################
sub addReferer {
  my $self = shift ;
  my $father = shift ;
  my $fname = $father->get(-name) ;
  $self->{"REFERERS"}->{"$fname"} = $father ;
  push(@{$self->{"DEFLINE"}->{"$fname"}}, $.) ;
  print STDOUT "Adding a referer for $self, name= ".$self->get(-name).", object=$father(" . $father->get(-name). ")\n" if $debug ;
}
###########################################################################
# A routine for printing an include file:
sub htmlprint {
  my $self = shift ;
  my $outdir = shift ;
  my $reldir = shift ;
  my $file = $self->get(-file) ;
  my $out ;
  my $outm ;
  my $outi ;
  if (-f "$file") {
    $out = "$outdir/" . basename($file) ;
    print STDOUT (caller(0))[3] . ": Outfile($file) is $out\n" if $debug ;
    if ($out =~ /(.*)\.(.*)/) {
      # Implementation HTML file:
      $outm = $1 . '_' . $2 . '_c' . '.html' ;
      # Summary HTML file:
      $outi = $1 . '_' . $2 . '_i' . '.html' ;
    } else {
      $outm = $out . '_c' . '.html' ;
      $outi = $out . '_i' . '.html' ;
    }
  }
  my $ihtml ;
  if (open(OUFI,">$outi")) {
    my $handle = \*OUFI ;
    my $title = "Interface for file $file" ;
    $ihtml = new ihtml(-name => $file, -handle => $handle, -file => $file, -title => $title, -reldir => $reldir) ;
    $ihtml->start_interface() ;
    if (exists $self->{"REFERERS"}) {
      print $handle "<BR>$file is included by\n" ;
      my $cgi = $ihtml->get(-cgi) ;
      print $handle $cgi->start_ul()."\n" ;
      foreach my $fname (sort alphabetic keys %{$self->{"REFERERS"}}) {
	my $father = $self->{"REFERERS"}->{"$fname"} ;
	if (ref($father) eq 'fthmod' || ref($father) eq 'fthrout') {
	  print $handle $cgi->li() . $cgi->a({
	    href => &ihtml::targetname($father->get(-file)) . "_c.html#R_" . $father->get(-name),
	    target => 'source',
	  }, $father->get(-file)) . ": " ;
	} else {
	  print $handle $cgi->li() . $cgi->a({
	    href => &ihtml::targetname($father->get(-file)) . "_c.html",
	    target => 'source',
	  }, $father->get(-file)) . ": " ;
	}
	my $counter = 0 ;
	foreach my $number (@{$self->{"DEFLINE"}->{"$fname"}}) {
	  print $handle ", " if $counter ;
	  print $handle $cgi->a({
	    href => &ihtml::targetname($father->get(-file)) . "_c.html#C_$number",
	    target => 'source',
	  }, $number) ;
	  $counter++ ;
	}
	print $handle ".\n" ;
      }
      print $handle $cgi->end_ul()."\n" ;
    }
    if (exists $self->{"INCLUDES"}) {
      print $handle "<BR>$file includes\n" ;
      foreach my $incname (sort keys %{$self->{"INCLUDES"}}) {
	my $inco = $self->{"INCLUDES"}->{"$incname"} ;
	$ihtml->printi_incl($inco) ;
      }
    }
    if (exists $self->{"INPUTS"}) {
      print $handle "<BR>$file inputs\n" ;
      foreach my $inpname (sort keys %{$self->{"INPUTS"}}) {
	my $inpo = $self->{"INCLUDES"}->{"$inpname"} ;
	$ihtml->printi_incp($inpo) ;
      }
    }
    $ihtml->end_html() ;
    close($handle) ;
  }
  if (open(OUFM,">$outm")) {
    my $handle = \*OUFM ;
    $ihtml->set(-handle => $handle) ;
    my $title = "Implementation of file $file" ;
    $ihtml->set(-js => 0) ;
    $ihtml->start_implementation() ;
    print $handle "<PRE>\n" ;
    # Get the chtml stuff:
    my $chtml = &chtml::getObject($file) ;
    $chtml->htmlprint($ihtml) ;
    print $handle "</PRE>\n" ;
    $ihtml->end_html() ;
    close($handle) ;
    $ihtml->free() ;
  }
}
###########################################################################
1;

