package fthinp ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for fthinp 90 files.
###########################################################################
use strict;
use Carp ;
use File::Basename ;
use File::Copy ;
use ihtml ;
use chtml ;
use vars qw($debug %MYOBJHASH) ;
$debug = 0 ;
(%MYOBJHASH) = () ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = lc(shift) ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    # We refuse to create an object from no where
    debug("Violent build for $name is illegal, API error") if $debug ;
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  if ($args{-name}) {
    $name = lc($args{-name}) ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
  } else {
    $first = 1 ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
sub printHash {
  my $self = shift ;
  my $t = shift ;
  my $hash = shift ;
  while(my($cle,$val) = each(%{$hash})) {
    print STDOUT "$t$cle => $val, " ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	print STDOUT "$sval, " ;
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	if (exists $array{$sval} && ref($array{$sval}) eq 'HASH') {
	  print STDOUT "$t\t$sval => {"  ;
	  $self->printHash("\t$t", $array{$sval}) ;
	  print STDOUT"}, " ;
	} else {
	  print STDOUT "$sval => $array{$sval}, " ;
	}
      }
      print STDOUT "},\n" ;
    } else {
      print STDOUT "$t$cle = $val, \n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
sub any {
  if (%MYOBJHASH) {
    return(1) ;
  } else {
    return(0) ;
  }
}
###########################################################################
# Adds a LaTeX input file
sub addInput {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"INPUTS"}->{"$name"} = $value ;
  print STDOUT "Adding a type for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Gets the list of inputs:
sub getInputs {
  my $self = shift ;
  if (exists $self->{"INPUTS"}) {
    return(keys %{$self->{"INPUTS"}}) ;
  } else {
    return(undef) ;
  }
}
###########################################################################
sub read {
  my $self = shift ;
  my $file = shift ;
  my $in = shift ;
  return if $self->get(-isread) ;
  my $oi = 0 ;
  my $chtml = &chtml::getCurrent() ;
  unless ($file) {
    $file = $self->get(-file) ;
  }
  unless ((ref $in) && (fileno $in)) {
    croak("No API for $file") unless $file ;
    if (-f "$file") {
      if (open(FILE,"<$file")) {
	$chtml = new chtml(-file => $file, -name => $file) ;
	$in = \*FILE ;
	$oi = 1 ;
      }
    }
  }
  if ((ref $in) && (fileno $in)) {
    while(<$in>) {
      # Parsing a TEX file below:
      chomp ;
      if (/(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	$chtml->addRaw("<A Name=C_$.></A>") ;
      }
      while (/(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	$chtml->addLaTeX($1.$2." ") ;
	my $incp = "$3.$4" ;
	$_ = $5 ;
#	if ($incp =~ /^\// && ! -f  "$incp") {
#	  $incp = basename($incp) ;
#	}
	unless (-f "$incp") {
	  my $base = dirname($self->get(-file)) ;
	  my $nase = basename(dirname($incp)) ;
	  if (-f "$base/$incp") {
	    $incp = "$base/$incp" ;
	  } elsif (-f "$base/" . basename($incp)) {
	    $incp = "$base/" . basename($incp) ;
	  } elsif (-f "$nase/" . basename($incp)) {
	    $incp = "$nase/" . basename($incp) ;
	  } elsif ($main::opt_L) {
	    # LaTeX inputs directories
	    # Possible...
	    my @TAGSHTML_PATH = (split ( /:/, $main::opt_L)) ;
	    foreach $base (@TAGSHTML_PATH) {
	      if (-f "$base/$incp") {
		$incp = "$base/$incp" ;
		last ;
	      }
	    }
	  }
	}
	my $other ;
	if (-f "$incp") {
	  if ($self->get(-masfic)) {
	    $other = new fthinp(-file => "$incp", -name => "$incp", -masfic => $self->get(-masfic)) ;
	  } else {
	    $other = new fthinp(-file => "$incp", -name => "$incp") ;
	  }
	  $self->addInput("$incp" => $other) ;
	  $chtml->addDef($other) ;
	} else {
	  my $refe = new oref(-name => $incp, -type => "fthinp") ;
	  $chtml->addRaw($refe) ;
	}
      }
      $chtml->addLaTeX($_) ;
    }
  }
  close($in) if $oi ;
  $self->set(-isread => 1) ;
  for my $name (keys %{$self->{"INPUTS"}}) {
    my $other = $self->{"INPUTS"}->{"$name"} ;
    $other->read() ;
  }
}
###########################################################################
sub alphabetic {
  lc($a) cmp lc($b) ||  $a cmp $b ;
}
###########################################################################
sub addReferer {
  my $self = shift ;
  my $father = shift ;
  my $fname = $father->get(-name) ;
  $self->{"REFERERS"}->{"$fname"} = $father ;
  push(@{$self->{"DEFLINE"}->{"$fname"}}, $.) ;
  print STDOUT "Adding a referer for $self, name= ".$self->get(-name).", object=$father(" . $father->get(-name). ")\n" if $debug ;
}
###########################################################################
# A routine for printing an inputed file:
sub htmlprint {
  my $self = shift ;
  my $outdir = shift ;
  my $reldir = shift ;
  my $file = $self->get(-file) ;
  my $out ;
  my $outm ;
  my $outi ;
  if (-f "$file") {
    $out = "$outdir/" . basename($file) ;
    print STDOUT (caller(0))[3] . ": Outfile($file) is $out\n" if $debug ;
    if ($out =~ /(.*)\.(.*)/) {
      # Implementation HTML file:
      $outm = $1 . '_' . $2 . '_c' . '.html' ;
      # Summary HTML file:
      $outi = $1 . '_' . $2 . '_i' . '.html' ;
    } else {
      $outm = $out . '_c' . '.html' ;
      $outi = $out . '_i' . '.html' ;
    }
  }
  my $ihtml ;
  if (open(OUFI,">$outi")) {
    my $handle = \*OUFI ;
    my $title = "Interface for file $file" ;
    $ihtml = new ihtml(-name => $file, -handle => $handle, -file => $file, -title => $title, -reldir => $reldir) ;
    # $ihtml->datadump("\t") if $debug ;
    $ihtml->start_interface() ;
    if (exists $self->{"REFERERS"}) {
      print $handle "<BR>$file is inputed by\n" ;
      my $cgi = $ihtml->get(-cgi) ;
      print $handle $cgi->start_ul()."\n" ;
      foreach my $fname (sort alphabetic keys %{$self->{"REFERERS"}}) {
	my $father = $self->{"REFERERS"}->{"$fname"} ;
	if (ref($father) eq 'fthmod' || ref($father) eq 'fthrout') {
	  print $handle $cgi->li() . $cgi->a({
	    href => &ihtml::targetname($father->get(-file)) . "_c.html#R_" . $father->get(-name),
	    target => 'source',
	  }, $father->get(-name)) . ": " ;
	} else {
	  print $handle $cgi->li() . $cgi->a({
	    href => &ihtml::targetname($father->get(-file)) . "_c.html",
	    target => 'source',
	  }, $father->get(-file)) . ": " ;
	}
	my $counter = 0 ;
	foreach my $number (@{$self->{"DEFLINE"}->{"$fname"}}) {
	  print $handle ", " if $counter ;
	  print $handle $cgi->a({
	    href => &ihtml::targetname($father->get(-file)) . "_c.html#C_$number",
	    target => 'source',
	  }, $number) ;
	  $counter++ ;
	}
	print $handle "\n" ;
      }
      print $handle $cgi->end_ul()."\n" ;
    }
    $ihtml->end_html() ;
    close($handle) ;
  }
  my $hevea = 'hevea_' . basename($file);
  $hevea = &ihtml::targetname($hevea) ;
  if (open(OUFM,">$outm")) {
    my $handle = \*OUFM ;
    $ihtml->set(-handle => $handle) ;
    my $title = "Implementation of file $file" ;
    $ihtml->set(-js => 0) ;
    $ihtml->start_implementation() ;
    if ($main::opt_labels) {
      my $l2hres = basename($file) ;
      if (exists $main::external_labels{$l2hres}) {
	print $handle "<BR><A Href=$main::external_labels{$l2hres}>LaTeX2HTML Translation</A>\n";
      }
    }
    if ($main::opt_hevea) {
      print $handle "<BR><A Href=$hevea.html>Hevea Translation</A>\n";
      print $handle "<BR><A Href=file://$ENV{FTAGSHTML_DIR}/doc/browser.html>Correctly Configure Netscape.</A>\n<BR>";
    }
    print $handle "<PRE>\n" ;
    # Get the chtml stuff:
    my $chtml = &chtml::getObject($file) ;
    $chtml->htmlprint($ihtml) ;
    print $handle "</PRE>\n" ;
    $ihtml->end_html() ;
    close($handle) ;
    $ihtml->free() ;
  }
  # Time to write stuffs for hevea:
  if ($main::opt_hevea) {
    if (open (INPUT, "<$file")){
      my $eqs = 0 ;
      my $outs = "" ;
      my $linei ;
      my $LIGNEINP ;
      my $ligne ;
      # Meta struct Eqlign not understood by hevea... altering the stuff...
      while (defined($LIGNEINP=<INPUT>)) {
	if ($LIGNEINP =~ /\$\$/) {
	  $eqs = !$eqs ;
	  if ($eqs) {
	    $outs = "" ;
	  } else {
	    # Eject buffer:
	    $linei = $LIGNEINP ;
	    chomp($linei) ;
	    $outs .= $linei ;
	    if ($outs =~ /\$\$\s*\\Eqalign\{(.*)\}\$\$/) {
	      $ligne .= '\begin{align*}' . "\n" . $1 . "\n" . '\end{align*}' . "\n" ;
	    } else {
	      $ligne .= $outs ;
	    }
	  }
	}
	if ($eqs) {
	  $linei = $LIGNEINP ;
	  chomp($linei) ;
	  $outs .= $linei ;
	} elsif ($outs) {
	  $outs = "" ;
	} else {
	  $ligne .= $LIGNEINP ;
	}
	if ($LIGNEINP =~ /[<>]/) {
	  $LIGNEINP =~ s/</&lt;/g ;
	  $LIGNEINP =~ s/>/&gt;/g ;
	}
      }
      close(INPUT) ;
      my $hevealigne = &LatexHeveaProcess($ligne);
      &HeveaConstructFile($hevea,$hevealigne);
    }
  }
}
#|#######################################################
sub HeveaConstructFile {
  my $hevea = shift ;
  my $hevealigne = shift ;
  if (open (HEVEA, ">$main::DIRE/$hevea.tex")){
    print STDOUT "\tProcessing hevea file $hevea\n" if ($main::opt_v >= 2);
    print HEVEA "$hevealigne\n";
    close(HEVEA);
    # Now writing the Hevea output:
    system "cd $main::DIRE ; latex '\\nonstopmode\\input{${hevea}.tex}' >/dev/null >&1 && latex '\\nonstopmode\\input{${hevea}.tex}' >/dev/null >&1";
    system "cd $main::DIRE ; hevea -e '/' -o ${hevea}.html ${hevea}.tex 2>/dev/null >&1";
  }
}
#|#######################################################
sub LatexHeveaProcess {
  my $ligne = shift ;
  my $tete ;
  my $queue ;
$tete = '\\documentclass[leqno]{report}
\\usepackage[latin1]{inputenc}
\\usepackage{a4,psfig,alltt}
\\usepackage{fancyhdr}
\\usepackage[dvips]{graphicx}
\\usepackage[dvips]{changebar}
\\usepackage{heqn}
\\usepackage{epsfig}
\\usepackage{makeidx}
\\usepackage{verbatim}
\\usepackage{amsmath}
\\usepackage{amsthm}
\\usepackage{amsfonts}
%%\\usepackage{french}
\\usepackage{html,htmllist}
\\usepackage{colordvi}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%\\renewcommand{\\varepsilon}{\\epsilon}
%%\\renewcommand{\\vec}[1]{{\\stackrel{\\rightarrow}{\\displaystyle #1}}}
%%\\renewcommand{\\sqrt}[1]{\\left(#1\\right)^{1/2}}
%%\\renewcommand{\\binom}[2]{C_{#1}^{#2}}
%%\\renewcommand{\\text}[1]{{\\rm{#1}}}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%\\newtheorem{theorem}{\\bf Thorme}
%%\\newtheorem{hypothesis}{Hypothse}
%%\\newtheorem{proposition}{Proposition}
%%\\newtheorem{lemma}{\\bf Lemme}
%%\\newtheorem{corollary}{\\bf Corollaire}
%%\\newtheorem{notation}{Notation}
%%\\newtheorem{definition}{Dfinition}
%%\\newtheorem{conjecture}{Conjecture}
%%\\newtheorem{remark}{Remarque}
%%\\newtheorem{modif}{Modification de source}
%%\\include{def}
%%\\include{ocfrench}
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
';
  if ($ENV{def1}) {
    $tete .= '\\input '."$ENV{def1}\n" ;
  }
  if ($ENV{def2}) {
    $tete .= '\\input '."$ENV{def2}\n" ;
  }
  if ($ENV{INFO}) {
    $tete .= '\\input '."$ENV{INFO}\n" ;
  }
  if ($ENV{HEVEA_OVER}) {
    my $href ;
    if ($ENV{HEVEA_OVER} =~ /^.*?\/*(\w+)\.(.*)$/) {
      $href = $1 ;
    }
    copy("$ENV{HEVEA_OVER}", "$main::DIRE/.") unless (-e "$main::DIRE/$href.$2");
    $tete .= '\\include{' . "$href" . '}' . "\n" ;
  }
  $tete .= '
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
\\begin{document}
';
  $queue ='
\\end{document}
' ;
  $ligne="$tete"."$ligne"."$queue" ;
  return ($ligne);
}
#%###################################################
sub LaTeXProtect {
  my $Nom = shift;
  $Nom =~ s/_/\\_/g ;
  return($Nom) ;
}
###########################################################################
sub Ouvre_SLatex {
  if (open(SLATEX,">$main::DIRE/Sindex.tex")) {
    my $MonCode = &LaTeXProtect($main::RepCodeSource) ;
print SLATEX <<'INFORMATION';
\documentclass[a4paper,10pt,twoside]{report}
\usepackage{fancyhdr}
\usepackage{makeidx}
\usepackage{hyperref}
\pagestyle{fancy}
%%\usepackage{french}
\textheight 23.0cm
\textwidth 16.0cm
\topmargin -0.55cm
\headsep 0.7cm
\marginparwidth 0.cm
\marginparsep 0.cm
\oddsidemargin 0.4cm
\evensidemargin -0.925cm 
\pagestyle{fancy}
\addtolength{\headwidth}{0cm}
\lfoot{\today}
INFORMATION
print SLATEX "\\rfoot{$main::USER}\n" ;
print SLATEX "\\lhead{Presentation of code $MonCode.}\n" ;
print SLATEX <<'INFORMATION';
\begin{document}
INFORMATION
print SLATEX "\\author{$main::USER}\n" ;
print SLATEX "\\title{Presentation of code $MonCode.}\n" ;
print SLATEX <<'INFORMATION';
\maketitle
%%\setcounter{page}{2}
\tableofcontents
\input texindex.tex
\end{document}
INFORMATION
    close (SLATEX) ;
  }
}
#|#######################################################
sub Ouvre_LLatex {
  if (open(LLATEX,">$main::DIRE/Lindex.tex")) {
    my $MonCode = &LaTeXProtect($main::RepCodeSource) ;
print LLATEX <<'INFORMATION';
\documentclass[a4paper,12pt]{report}
\usepackage{fancyhdr}
\usepackage{makeidx}
\usepackage{hyperref}
\pagestyle{fancy}
\lfoot{\today}
INFORMATION
print LLATEX "\\rfoot{$main::USER}\n" ;
print LLATEX "\\lhead{Presentation of code $MonCode.}\n" ;
print LLATEX <<'INFORMATION';
\begin{document}
INFORMATION
print LLATEX "\\author{$main::USER}\n" ;
print LLATEX "\\title{Presentation of code $MonCode.}\n" ;
print LLATEX <<'INFORMATION';
\maketitle
\tableofcontents
\input texindex.tex
\end{document}
INFORMATION
    close (LLATEX) ;
  }
}
###########################################################################
sub open_latex {
  my $outdir = shift ;
  my $reldir = shift ;
  my $handle = undef ;
  &Ouvre_SLatex() ;
  &Ouvre_LLatex() ;
  if (open(LATEX,">$outdir/texindex.tex")) {
    $handle = \*LATEX ;
  }
  return($handle) ;
}
###########################################################################
1;

