package fthinter ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for fortran 90 interfaces.
###########################################################################
use strict;
use Carp ;
use vars qw($debug $nbself %MYOBJHASH) ;
$debug = 0 ;
$nbself = 0 ;
(%MYOBJHASH) = () ;
use fthrout ;
use fthmod ;
use fthobj ;
use oref ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = lc(shift) ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    # We refuse to create an object from no where
    croak("Violent build for $name is illegal, API error") ;
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  bless $self;
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug;
  }
  $nbself++ ;
  if (exists $self->{"name"}) {
    my $name = $self->{"name"} ;
    my $stuff = new fthobj(-name => $name) ;
    $name = lc($name) ;
    $MYOBJHASH{$name} = $self ;
    $stuff->addStuff($self) ;
  }
  return $self ;
}
###########################################################################
sub old {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  if ($args{-name}) {
    $name = $args{-name} ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
  } else {
    $first = 1 ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) ;
  }
  return $self ;
}
###########################################################################
# Gets the number of built objects:
sub getNbself {
  return($nbself) ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub save_hash {
  my $self = shift ;
  my $handle = shift ;
  my $name = $self->get(-name) ;
  if (exists $self->{"FPROCEDURES"}) {
    print $handle '$fthinters{"' . $name . '"} = (' ;
    foreach my $orout (values %{$self->{"FPROCEDURES"}}) {
      my $number = $orout->get(-number) ;
      my $file = $orout->get(-file) ;
      print $handle '"' . "$file:$number" . '", '  ;
    }
    print $handle ');' ."\n" ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	print STDOUT "$sval, " ;
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	print STDOUT "$sval => $array{$sval}, " ;
      }
      print STDOUT "},\n" ;
    } else {
      print STDOUT "$t$cle = $val,\n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "TRACE FTHINTER: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
  print STDOUT "VALUE FTHINTER: $message\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
sub any {
  if (%MYOBJHASH) {
    return(1) ;
  } else {
    return(0) ;
  }
}
###########################################################################
sub getProcedures {
  my $self = shift ;
  if (exists $self->{"PROCEDURES"}) {
    return(@{$self->{"PROCEDURES"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Provides an object target
sub getProcedure {
  my $self = shift ;
  my $name = lc(shift) ;
  if (exists $self->{"FPROCEDURES"}->{"$name"}) {
    return($self->{"FPROCEDURES"}->{"$name"}) ;
  } else {
    if ($debug) {
      print STDOUT "Bad call for $name:" ;
      $self->datadump("\t") ;
    }
    return() ;
  }
}
###########################################################################
# Adds a pointer to a module procedure for self:
sub addModuleProcedure {
  my $self = shift ;
  my $name = lc(shift) ;
  push(@{$self->{"PROCEDURES"}}, $name) ;
  print STDOUT "Adding a module procedure for $self, name=$name\n" if $debug ;
}
###########################################################################
# Adds a routine to a current interface:
sub addRout {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"FPROCEDURES"}->{"$name"} = $value ;
  print STDOUT "Adding a routine for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds interfaces to target right after reading has been performed:
sub addInter {
  my $self = shift ;
  my $omod = shift ;
  if (exists $self->{"FPROCEDURES"}) {
    foreach my $orout (values %{$self->{"FPROCEDURES"}}) {
      my $name = $orout->get(-name) ;
      my $inter = new fthinter(-name => $name, -auto => 1, -file => $self->get(-file), -masfic => $self->get(-masfic), -module => $self->get(-module)) ;
      push(@{$inter->{"PROCEDURES"}}, lc($name)) ;
      $inter->addRout($name => $orout) ;
      # Addds a link to the implemented routine:
      print STDOUT (caller(0))[3] .": Creating fake interface $name for module " . $self->get(-module) . "\n" if $debug ;
      $omod->addInter($name => $inter) ;
      # Inform the module it defines the interface:
      # $omod->addRout($orout) ;
      # May be stupid:
      # $orout->addInter($fthinter);
    }
  }
}
###########################################################################
sub read {
  my $self = shift ;
  my $handle = shift ;
  my $file = $self->get(-file) ;
  my $omod = $self->get(-module) ;
  my $chtml = &chtml::getCurrent() ;
  if ((ref $handle) && (fileno $handle)) {
    my $name = $self->get(-name) ;
    $self->set(-number => $.) unless $self->get(-number) ;
    while(<$handle>) {
      # Parsing an f90 interface file below:
      if (/^\s*end\s*interface/i) {
	$chtml->addDef($_) ;
	# End of interface creation
	print STDOUT "Created interface $name\n" if $debug ;
	return ;
      } else {
	# Proper management of input remains to be performed in the future:
	my $cc = "" ;
	if (/^(.*?)\!(.*)/) {
	  $_ = $1 ;
	  $cc = $2 ;
	}
	if (/^(\s*module\s+procedure\s+)(.*)/i) {
	  $chtml->addDef($1) ;
	  my $rhs = $2 ;
	  my (@array) = split(/\,/, $rhs) ;
	  my $c = 0 ;
	  foreach my $a (@array) {
	    $a =~ s/\s+//g ;
	    if ($a =~ /\&(.*)/) {
	      $chtml->addDef("&") ;
	      $a = $1 ;
	    }
	    if ($a =~ /(.*)\&/) {
	      $a = $1 ;
	    }
	    next unless $a ;
	    my $refe = new oref(-name => $omod->get(-name) . '::' . $a, -type => "uses") ;
	    $chtml->addDef(", ") if $c ;
	    $chtml->addRaw($refe) ;
	    push(@{$self->{"PROCEDURES"}}, lc($a)) ;
	    $c++ ;
	  }
	  if ($rhs =~ /(.+)\&/) {
	    $chtml->addDef("&") ;
	  }
	} elsif (/(subroutine|function)\s+(\w+)(.*)/i) {
	  my $rame = $2 ;
	  my $lame = lc($name) ;
	  # List of objects management
	  push(@{$self->{"PROCEDURES"}}, $rame) ;
	  my $orout ;
	  if ($omod && ref($omod) eq 'fthmod') {
	    $orout = new fthrout(-name => $rame, -file => $file, -masfic => $self->get(-masfic), -module => $omod->get(-name), -number => $.) ;
	  } elsif ($omod && !ref($omod)) {
	    $orout = new fthrout(-name => $rame, -file => $file, -masfic => $self->get(-masfic), -module => $omod, -number => $.) ;
	  } else {
	    print STDOUT (caller(0))[3] .": Bad omod=$omod API" if $main::opt_v ;
	    $orout = new fthrout(-name => $rame, -file => $file, -masfic => $self->get(-masfic), -number => $.) ;
	  }
	  $self->addRout($rame => $orout) ;
	  my $proto ;
	  if (/^(\s*)(.*?)(\s*function\s+)(\w+)(.*)/i) {
	    $chtml->addDef($1) ;
	    $proto = $2 ;
	    if ($proto) {
	      if ($proto =~ /\bpure\b/i || $proto =~ /\brecursive\b/i) {
		$chtml->addDef($proto) ;
	      } else {
		$orout->set(-theproto => $proto) ;
		print STDOUT "Defining proto to $rame as $proto\n" if $debug ;
		$chtml->addRaw("<DECLARE>$proto</DECLARE>") ;
	      }
	    }
	    $chtml->addDef($3) ;
	    my $refe = new oref(-name => $rame, -type => "irout", -fthrout => $orout) ;
	    $chtml->addRaw($refe) ;
	    # $chtml->addDef($orout) ;
	    #print STDOUT "Provide analysis for $5\n" ;
	    $orout->readf90subr($handle, "& $5") ;
	    if ($proto) {
	      unless ($proto =~ /\bpure\b/i || $proto =~ /\brecursive\b/i) {
		$orout->addArgument("$rame") ;
		$orout->addType("$rame" => "$proto") ;
	      }
	    }
	    unless ($orout->get(-theproto)) {
	      $orout->addArgument("$rame") ;
	      $orout->addType("$rame" => 'void') ;
	      $orout->set(-theproto => 'void') ;
	    }
	    if ($orout->getType($rame) eq 'void' && $proto) {
	      $orout->addType($rame => $proto) ;
	      $orout->setType($rame => $proto) ;
	      print STDOUT "Error bad proto $rame\n" if $debug ;
	    }
	  } elsif (/^(\s*)(.*?)(\s*subroutine\s+)$rame(.*)/i) {
	    $chtml->addDef($1.$2) ;
	    $orout->set(-theproto => 1) ;
	    $chtml->addDef($3) ;
	    my $refe = new oref(-name => $rame, -type => "irout", -fthrout => $orout) ;
	    $chtml->addRaw($refe) ;
	    # $chtml->addDef($orout) ;
	    $orout->readf90subr($handle, "& $4") ;
	  } else {
	    $chtml->addDef($_) ;
	  }
	} else {
	  $chtml->addDef($_) ;
	}
	if ($cc) {
	  $chtml->addRaw('<COMMENT>') ;
	  $chtml->addLaTeX("!$cc") ;
	  $chtml->addRaw('</COMMENT>') ;
	}
      }
    }
  }
}
###########################################################################
1;

