package fthlabel ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for fthlabel 90 files.
###########################################################################
use strict;
use Carp ;
use oref ;
use ihtml ;
use vars qw($debug $nbself %MYOBJHASH) ;
$debug = 0 ;
$nbself = 0 ;
(%MYOBJHASH) = () ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = shift ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    # We refuse to create an object from no where
    # croak("Violent build for $name is illegal, API error") ;
    debug("Violent build for $name is illegal, API error") if $debug ;
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  if (exists $args{"-name"}) {
    $name = $args{-name} ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
  } else {
    $nbself++ ;
    $first = 1 ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
# Gets the number of built objects:
sub getNbself {
  return($nbself) ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	print STDOUT "$sval, " ;
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	print STDOUT "$sval => $array{$sval}, " ;
      }
      print STDOUT "},\n" ;
    } else {
      print STDOUT "$t$cle = $val,\n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
sub any {
  if (%MYOBJHASH) {
    return(1) ;
  } else {
    return(0) ;
  }
}
#|#######################################################
sub latex2html {
# We shall build a database of external references for LaTeX2HTML
  if (open(LATEX2HTML,">$main::DIRE/latex2html.pl")) {
    print LATEX2HTML "# ftagshtml $main::version external references for\n";
    print LATEX2HTML "# LaTeX2HTML 98.1p1 release (March 2nd, 1998)\n";
    print LATEX2HTML "# Associate labels original text with physical files.\n\n";
    foreach my $self (values %MYOBJHASH) {
      my $file = $self->{"file"} ;
      my $target = &ihtml::targetname(${file}) ;
      my $name = $self->{"name"} ;
      print LATEX2HTML '$external_labels{"'.$name.'"} = "$URL/'.$target."\";\n";
    }
    print LATEX2HTML "\n1;\n";
    close(LATEX2HTML);
  }
}
###########################################################################
1;


