package fthmod ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for fthmod 90 files.
###########################################################################
use strict;
use Carp ;
use File::Basename ;
use vars qw($debug $nbself %MYOBJHASH) ;
$debug = 0 ;
$nbself = 0 ;
(%MYOBJHASH) = () ;
use fthtype;
use fthinter;
use fthrout;
use fthcall;
use fthbc;
use fthext;
use oref ;
use chtml ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = lc(shift) ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    # We refuse to create an object from no where
    debug("Violent build for $name is illegal, API error") if $debug ;
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  if ($args{-name}) {
    $name = lc($args{-name}) ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
    print STDERR "\n" . (caller(0))[3] . ": Redefining module $name from file " . $self->get(-file) . " in file " . $args{-file}."\n" ;
  } else {
    $first = 1 ;
    $nbself++ ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
# Gets the number of built objects:
sub getNbself {
  return($nbself) ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
sub printHash {
  my $self = shift ;
  my $t = shift ;
  my $hash = shift ;
  while(my($cle,$val) = each(%{$hash})) {
    print STDOUT "$t$cle => $val, " ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub save_hash {
  my $self = shift ;
  my $handle = shift ;
  my $name = $self->get(-name) ;
  if (0) {
    print $handle '$L_Declar{"' . $name . '"} = "' . $self->get(-number) . '" ;' . "\n" ;
    print $handle '$File_Of_Rout{"' . $name . '"} = "' . $self->get(-file) . '" ;' . "\n" ;
  } else {
    my $number = $self->get(-number) ;
    my $file = $self->get(-file) ;
    print $handle '$fthmods{"' . $name . '"} = ' ;
    print $handle '"' . "$file:$number" . '"'  ;
    print $handle ';' ."\n" ;
    print $handle '$names{"' . lc($name) . '"} = "' . $name . '";' ."\n" ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	print STDOUT "$sval, " ;
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	if (exists $array{$sval} && ref($array{$sval}) eq 'fthtype') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthinter') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthrout') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthcall') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthinp') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'HASH') {
	  print STDOUT "$t\t$sval => {"  ;
	  $self->printHash("\t$t", $array{$sval}) ;
	  print STDOUT"}, " ;
	} else {
	  print STDOUT "$sval => $array{$sval}, " ;
	}
      }
      print STDOUT "},\n" ;
    } else {
      print STDOUT "$t$cle = $val, \n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
sub any {
  if (%MYOBJHASH) {
    return(1) ;
  } else {
    return(0) ;
  }
}
###########################################################################
# Returns the implicit current rule pattern
sub getImplicitRules {
  my $self = shift ;
  if (exists $self->{"IMPLICITDEFS"}) {
    return(%{$self->{"IMPLICITDEFS"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Returns the implicit default rule pattern
sub ImplicitReset {
  my $self = shift ;
  if (exists $self->{"IMPLICITDEFS"}) {
    foreach my $rhs (keys %{$self->{"IMPLICITDEFS"}}) {
      delete $self->{"IMPLICITDEFS"}->{"$rhs"} ;
    }
    delete $self->{"IMPLICITDEFS"} ;
  }
}
###########################################################################
# Process the implicit line
sub ImplicitProcess {
  my $self = shift ;
  my (%hash) = (@_) ;
  my $string = "abcdefghijklmnopqrstuvwxyz" ;
  foreach my $cle (keys %hash) {
    if ($cle =~ /\-/) {
      # Replacing...
      if ($string =~ /([$cle]+)/i) {
	my $charlist = $1 ;
	while($charlist =~ /^(\w)(.*)/) {
	  $charlist = $2 ;
	  $self->{"IMPLICITDEFS"}->{$1} = $hash{$cle} ;
	}
      }
      
    } else {
      my $lcle = lc($cle) ;
      $self->{"IMPLICITDEFS"}->{$lcle} = $hash{$cle} ;
    }
  }
}
###########################################################################
# Adds a type to a current module:
sub addType {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"FTYPES"}->{"$name"} = $value ;
  print STDOUT "Adding a type for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
sub getTypes {
  my $self = shift ;
  if (exists $self->{"FTYPES"}) {
    return(%{$self->{"FTYPES"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
# Adds an interface to a current module:
sub addInter {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"FINTERFACES"}->{"$name"} = $value ;
  print STDOUT "Adding an interface for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Gets an interface from the current module:
sub getInter {
  my $self = shift ;
  my $name = lc(shift) ;
  if (exists $self->{"FINTERFACES"} &&exists $self->{"FINTERFACES"}->{"$name"}) {
    return($self->{"FINTERFACES"}->{"$name"}) ;
  } else {
    return(undef) ;
  }
}
###########################################################################
sub getInters {
  my $self = shift ;
  if (exists $self->{"FINTERFACES"}) {
    return(%{$self->{"FINTERFACES"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
# Adds a routine to a current module:
sub addRout {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"FPROCEDURES"}->{"$name"} = $value ;
  print STDOUT "Adding a routine for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
sub getRouts {
  my $self = shift ;
  if (exists $self->{"FPROCEDURES"}) {
    return(%{$self->{"FPROCEDURES"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
# Adds a call
sub addCall {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"FCALLS"}->{"$name"} = $value ;
  print STDOUT "Adding a call for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds a LaTeX input file
sub addInput {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"INPUTS"}->{"$name"} = $value ;
  $value->addReferer($self) ;
  print STDOUT "Adding an input for $self(" . $self->get(-name) . "), name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds an include:
sub addInclude {
  my $self = shift ;
  my $name =shift ;
  my $value = shift ;
  $self->{"INCLUDES"}->{"$name"} = $value ;
  $value->addReferer($self) ;
  print STDOUT "Adding a type for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
sub getInputs {
  my $self = shift ;
  if (exists $self->{"INPUTS"}) {
    return(%{$self->{"INPUTS"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
sub getIncludes {
  my $self = shift ;
  if (exists $self->{"INCLUDES"}) {
    return(%{$self->{"INCLUDES"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
# On the fly renaming is to be performed here : check syntax
# is correct: lhs is in module, rhs is the local name:
sub readUses {
  my $self = shift ;
  my $file = shift ;
  my $lmused = shift ;
  my $fmused = lc($lmused) ;
  $_ = shift ;
  print STDOUT "\t" . (caller(0))[3] . ": Reading $_ for used module $fmused\n" if $debug ;
  my $cont = 1 ;
  my $chtml = &chtml::getCurrent() ;
  if ((ref $file) && (fileno $file)) {
    while($cont) {
      my $cc = "" ;
      #print STDOUT "line=$. => $_\n" ;
      if (/^(.*?)\!(.*)/) {
	$_ = $1 ;
	$cc = $2 ;
      }
      if ($cont && /^(\s*\&\s*)(.*)/ || $cont >= 2 && /(\s*)(.*)/) {
	print STDOUT "\t\tRhs($.) for fmused($cont)=$2\n" if $debug ;
	my $rhs = $2 ;
	$chtml->addDef($1) if ($cont >= 2) ;
	$rhs =~ s/\s+//g ;
	if ($rhs =~ /(.*)\&/) {
	  $rhs = $1 ;
	  $cont = 2 ;
	} elsif ($rhs) {
	  $cont = 0 ;
	} else {
	  # Only spacing
	  $cont = 3 ;
	}
	my (@array) = split(/\,/, $rhs) ;
	my $counter = 0 ;
	#print STDOUT "Array=@array\n" ;
	foreach my $a (@array) {
	  $chtml->addDef(", ") if $counter ;
	  if ($a =~ /\&\s*$/) {
	    $cont = 2 ;
	    #print STDOUT "X: cont=$cont\n" ;
	    $chtml->addDef($a) if $counter ;
	  } elsif ($a =~ /\b(\w+)\b/) {
	    my $l = $a ;
	    my $r = $a ;
	    if ($a =~ /(.*)=>(.*)/) {
	      $l = $1 ;
	      $r = $2 ;
	    }
	    # On the right the name used for the call, on the left the
	    # actual routine name in it's module.
	    print STDOUT "\t\t" . (caller(0))[3] . ":SELUSES $r points to $fmused :: $l\n" if $debug ;
	    # Manage multiple override...
	    my $sr = lc($r) ;
	    my $sl = lc($l) ;
	    if ($self->{"SELUSES"}->{"$sr"}) {
	      $self->{"SELUSED"}->{"$sr"}->{"$fmused"} =  "$sl" ;
	    } else {
	      # $r is the routine name, fmused the package
	      $self->{"SELUSES"}->{"$sr"} =  $fmused . '::' . $sl ;
	    }
	    # NAJ: check the renaming at this level:
	    if ($l ne $r) {
	      $chtml->addRaw("<A Name=R_$l>$l</A> => ") ;
	    }
	    my $refe = new oref(-name => "$lmused" . '::' . "$r", -type => "uses") ;
	    $chtml->addRaw($refe) ;
	    $counter++ ;
	  }
	}
	$chtml->addDef(" &") if ($cont == 2) ;
      } else {
	print STDOUT (caller(0))[3] . ": KO $_\n" if $debug ;
      }
      if ($cc) {
	$chtml->addRaw('<COMMENT>') ;
	$chtml->addLaTeX($cc) ;
	$chtml->addRaw("</COMMENT>") ;
      }
      return unless $cont ;
      if (defined($_ = <$file>)) {
	chomp ;
      } else {
	print STDOUT (caller(0))[3] . ": error4 line=$. module='" . $self->get(-name) . "' file='" . $self->get(-file) . "'\n" if $debug ;
	$cont = 0 ;
      }
    }
  }
}
###########################################################################
# Sets the dependencies for a module:
sub getUses {
  my $self = shift ;
  if (exists $self->{"ALLUSES"}) {
    return(keys %{$self->{"ALLUSES"}}) ;
  } else {
    return(undef) ;
  }
}
###########################################################################
# Sets the dependencies for a module refusing the cyclic dependencies
sub getDepends {
  my $self = shift ;
  if (exists $self->{"RECUSES"}) {
    return(%{$self->{"RECUSES"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Sets the dependencies for a module refusing the cyclic dependencies
sub setCompiles {
  my $self = shift ;
  my (%hash) = () ;
  my $file ;
  if (exists $self->{"ALLUSES"}) {
    foreach my $fmused (keys %{$self->{"ALLUSES"}}) {
      my $omod = &fthmod::getObject($fmused) ;
      if ($omod) {
	$file = $omod->get(-file) ;
	$hash{$file} = 1 ;
      }
    }
  }
  if (exists $self->{"SELUSES"}) {
    while(my($sr, $value)=each %{$self->{"SELUSES"}}) {
      my $fmused = $self->{"SELUSES"}->{"$sr"} ;
      if ($fmused =~ /^(.*?)\:/) {
	$fmused = $1 ;
	my $omod = &fthmod::getObject($fmused) ;
	if ($omod) {
	  $file = $omod->get(-file) ;
	  $hash{$file} = 1 ;
	}
      }
      if (exists $self->{"SELUSED"} && exists $self->{"SELUSED"}->{"$sr"}) {
	foreach $fmused (keys %{$self->{"SELUSED"}->{"$sr"}}) {
	  my $omod = &fthmod::getObject($fmused) ;
	  if ($omod) {
	    $file = $omod->get(-file) ;
	    $hash{$file} = 1 ;
	  }
	}
      }
    }
  }
  # Now also sets the calls tree that is not that obvious:
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    my $drout = $self->{"FPROCEDURES"}->{"$name"} ;
    # Gets the absolute direct stuff:
    my (@sash) = $drout->setCompiles() ;
    foreach my $file (@sash) {
      $hash{$file} = 1 ;
    }
  }
  return(keys %hash) ;
}
###########################################################################
# Sets the dependencies for a module refusing the cyclic dependencies
sub recDepends {
  my $self = shift ;
  my $omod = shift ;
  my $rmod ;
  my(@array) = $omod->getUses() ;
  foreach my $mname (@array) {
    next unless $mname ;
    next if ($self->{"RECUSES"} && exists $self->{"RECUSES"}->{"$mname"}) ;
    $rmod = &fthmod::getObject($mname) ;
    next unless $rmod ;
    $self->{"RECUSES"}->{"$mname"} = $rmod ;
    $self->recDepends($rmod) ;
  }
}
###########################################################################
# Sets the dependencies for a module:
sub setDepends {
  my $self = shift ;
  my $omod ;
  if (exists $self->{"ALLUSES"}) {
    foreach my $mname (keys %{$self->{"ALLUSES"}}) {
      $omod = &fthmod::getObject($mname) ;
      next unless $omod ;
      $self->{"RECUSES"}->{"$mname"} = $omod ;
      $self->recDepends($omod) ;
    }
  }
  # Sets the interfaces to the current routines:
  if (exists $self->{"FINTERFACES"}) {
    foreach my $iname (keys %{$self->{"FINTERFACES"}}) {
      my $ointer = $self->{"FINTERFACES"}->{"$iname"} ;
      # An interface may point to one or more procedures:
      my(@procs) = $ointer->getProcedures() ;
      foreach my $pname (@procs) {
	my $lname = lc($pname) ;
	# If the module is correct, there must be a corresponding procedure
	if (exists $self->{"FPROCEDURES"}->{"$lname"}) {
	  my $orout = $self->{"FPROCEDURES"}->{"$lname"} ;
	  $orout->addInter($ointer);
	  print STDOUT (caller(0))[3] .": Interface $iname of module " . $self->get(-name) . " refers to module procedure " . $orout->get(-name) . " defined line " . $orout->get(-number) . " of file " . $orout->get(-file) ."\n" if $debug ;
	} else {
	  # Look for a global procedure to it:
	  my $orout = $ointer->getProcedure($iname) ;
	  if ($orout) {
	    my $crout = &fthrout::isGlobal($orout) ;
	    if ($crout) {
	      $orout->set(-thelink => $crout) ;
	      print STDOUT (caller(0))[3] .": Interface $iname of module " . $self->get(-name) . " refers to global procedure " . $crout->get(-name) . " defined line " . $crout->get(-number) . " of file " . $crout->get(-file) ."\n" if $debug ;
	    } else {
	      print STDOUT (caller(0))[3] .": Interface $iname of module " . $self->get(-name) . " refers to global procedure " . $orout->get(-name) . " whose implementation is hidden\n" if $debug ;
	    }
	  } elsif (exists $self->{"FINTERFACES"}->{"$lname"}) {
	    # Current interface has no module procedure but a
	    my $sinter = $self->{"FINTERFACES"}->{"$lname"} ;
	    my $orout = $sinter->getProcedure($lname) ;
	    if ($orout) {
	      # Provide a routine target to current module procedure:
	      $ointer->addRout($lname => $orout) ;
	      # Adds current name
	      $orout->addInter($iname => $ointer);
	      print STDOUT (caller(0))[3] .": Interface $iname of module " . $self->get(-name) . " refers to outer procedure " . $orout->get(-name) . " defined line " . $orout->get(-number) . " of file " . $orout->get(-file) ."\n" if $debug ;
	      if ($debug) {
		$sinter->datadump("\t") ;
		$orout->datadump("\t") ;
	      }
	    } else {
	      print STDOUT (caller(0))[3] .": Interface $iname of module " . $self->get(-name) . " to procedure $lname strangely is an interface.\n" if $debug ;
	    }
	  } else {
	    print STDOUT (caller(0))[3] .": Interface $iname of module " . $self->get(-name) . " to procedure $lname strangely has no implementation\n" if $debug ;
	  }
	}
      }
    }
  }
  # Sets the interfaces to the routines of the module precedures:
  if (exists $self->{"FPROCEDURES"}) {
    foreach my $lname (keys %{$self->{"FPROCEDURES"}}) {
      my $orout = $self->{"FPROCEDURES"}->{"$lname"} ;
      $orout->setDepends() ;
    }
  }
}
###########################################################################
# Gets the argument's type for a given stuff, try to guess...
sub getType {
  my $self = shift ;
  my $name = shift ;
  if (exists $self->{"ALLDEFS"}->{"$name"}) {
    return($self->{"ALLDEFS"}->{"$name"}) ;
  } else {
    return(undef) ;
  }
}
###########################################################################
# Finds the actual arg type:
sub findType {
  my $self = shift ;
  my $arg = lc(shift) ;
  my $valeur = $self->getType($arg) ;
  print STDOUT "\t" . (caller(1))[3] . ": Trying to find $arg in " . $self->get(-name) . "\n"  if $debug ;
  unless ($valeur) {
    foreach my $module (keys %{$self->{"RECUSES"}}) {
      my $omod = $self->{"RECUSES"}->{"$module"} ;
      next unless $omod ;
      $valeur = $omod->getType($arg) ;
      last if $valeur ;
    }
    if (exists $self->{"SELUSES"} && $self->{"SELUSES"}->{"$arg"}) {
      my $bidule = $self->{"SELUSES"}->{"$arg"} ;
      if ($bidule =~ /^(.*?)\:\:(.*)/) {
	my $sarg = $2 ;
	my $omod = &fthmod::getObject($1) ;
	if ($omod) {
	  $valeur = $omod->getType($sarg) ;
	}
      }
    }
  }
  print STDOUT "\t" . (caller(0))[3] . ": Obtained value for $arg in " . $self->get(-name) . ":$valeur\n"  if $debug && $valeur ;
  return($valeur) ;
}
###########################################################################
# Gets all the call for a given module:
sub getCalls {
  my $self = shift ;
  my (%hash) = () ;
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    my $drout = $self->{"FPROCEDURES"}->{"$name"} ;
    foreach my $ocall ($drout->getCalls()) {                                        
      my $orout = $ocall->getLink() ;
      if ($orout) {
	my $file = $orout->get(-file) ;
	my $number = $ocall->get(-number) ;
	my $rname = $orout->get(-name) ;
	my $mname = $orout->get(-module) ;
	if ($mname) {
	  $mname = lc($mname) ;
	  my $other = &fthmod::getObject($mname) ;
	  if ($other) {
	    $mname = $other->get(-name) ;
	  }
	} else {
	  $mname = "AAGLOBAL" ;
	}
	# Ready to build the array module:file:routine:line
	$hash{$mname}->{$rname}->{$number} = $ocall ;
      }
    }
  }
  return(%hash) ;
}
###########################################################################
# For every module procedure's called routine, provides it's definition point:
# Looks in current module, if not looks in procedure's internal setting:
sub setCalls {
  my $self = shift ;
  my $orout ;
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    # Current file defines $name
    my $drout = $self->{"FPROCEDURES"}->{"$name"} ;
    $self->setCalls2($drout) ;
  }
}
###########################################################################
# For every module procedure's called routine, provides it's definition point:
# Looks in current module, if not looks in procedure's internal setting:
sub setCalls2 {
  my $self = shift ;
  my $drout = shift ;
  my $orout ;
  my (@proto) = &fthrout::getProto($drout) ;
  my $name = $drout->get(-name) ;
  print STDOUT "\t\t\t" . (caller(0))[3] . " Routine $name points to $drout, defined in module ".$self->get(-name). ":\n" if $debug ;
  # Sets the calls for the contained procedures...
  my (@oldcalls) = ($drout->getCalls()) ;
  foreach my $ocall (@oldcalls) {
    # NAJ: useless ?
    next if $ocall->get(-thelink) ;
    # Make the call get ready at this level:
    print STDOUT "+" if ($main::opt_v >= 1) ;
    $drout->preCalls($ocall) ;
  }
  my (@newcalls) = ($drout->getInlined(@oldcalls)) ;
  foreach my $ocall (@newcalls) {
    print STDOUT (caller(0))[3] .": Trying to find call for ".$ocall->get(-name)." line ".$ocall->get(-number)." from file ".$self->get(-file)."\n" if $debug ;
    print STDOUT "." if ($main::opt_v >= 1) ;
    $drout->postCalls($ocall) ;
    $self->setCallers($drout, $ocall) ;
  }
  # Recursively handle privates:
  my (%privates) = $drout->tellRouts() ;
  foreach my $rame (keys %privates) {
    $self->setCalls2($privates{$rame}) ;
  }
}
###########################################################################
# For every module procedure's called routine, provides it's definition point:
# Looks in current module, if not looks in procedure's internal setting:
sub sayCalls {
  my $self = shift ;
  my $ier = shift ;
  my $orout ;
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    # Current file defines $name
    my $drout = $self->{"FPROCEDURES"}->{"$name"} ;
    print STDOUT "fthmod::sayCalls Routine $name points to $drout, defined in module ".$self->get(-name). ":\n" if $debug ;
    $self->sayCalls2($ier, $drout) ;
  }
}
###########################################################################
# For every module procedure's called routine, provides it's definition point:
# Looks in current module, if not looks in procedure's internal setting:
sub sayCalls2 {
  my $self = shift ;
  my $ier = shift ;
  my $drout = shift ;
  my (@proto) = &fthrout::getProto($drout) ;
  my $orout ;
  foreach my $ocall ($drout->getCalls()) {
    $orout = $ocall->getLink() ;
    if ($orout && $ier) {
      print STDOUT "OK good call " . $ocall->get(-name) . " from " . $self->get(-name) ."\n" ;
      $ocall->datadump("\t\t")  ;
      $orout->datadump("\t\t") ;
    } elsif (!$orout && !$ier && !$ocall->get(-fake)) {
      my $name = $ocall->get(-name) ;
      if (&fthrout::getObject($name) || &fthobj::getObject($name)) {
	print STDOUT "KO bad call " . $ocall->get(-name) . " from " . $self->get(-name) ."\n" ;
	$ocall->datadump("\t")  ;
      } else {
	print STDOUT "No reference to " . $ocall->get(-name) . " from " . $self->get(-name) ." line " . $ocall->get(-number) . "\n" ;
      }
    }
  }
  # Recursively handle privates:
  my (%privates) = $drout->tellRouts() ;
  foreach my $rame (keys %privates) {
    $self->sayCalls2($ier, $privates{$rame}) ;
  }
}
###########################################################################
# For every module procedure's called routine, provides it's definition point:
# Looks in current module, if not looks in procedure's internal setting:
sub revCalls {
  my $self = shift ;
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    # Current file defines $name
    my $drout = $self->{"FPROCEDURES"}->{"$name"} ;
    $self->revCalls2($drout) ;
  }
}
###########################################################################
# For every module procedure's called routine, provides it's definition point:
# Looks in current module, if not looks in procedure's internal setting:
sub revCalls2 {
  my $self = shift ;
  my $drout = shift ;
  my $orout ;
  foreach my $ocall ($drout->getCalls()) {
    $orout = $ocall->getLink() ;
    if ($orout) {
      if ($debug) {
	print STDOUT "\t" . (caller(0))[3] . ": call name=" . $ocall->get(-name) . " in file " . $self->get(-file) . " line ". $ocall->get(-number) ." points to rout = '$orout'\n" ;
	$orout->datadump("\t") ;
      }
      # routine is called by call from self:
      $orout->addCalled($ocall) ;
      if ($orout->get(-module)) {
	my $omod = &fthmod::getObject($orout->get(-module)) ;
	print STDOUT "Module for $orout name=" . $orout->get(-name) . " is ".$orout->get(-module)."\n" if $debug ;
	$omod->addCalled($self, $orout, $ocall) if $omod ;
      }
    }
  }
  # Recursively handle privates:
  my (%privates) = $drout->tellRouts() ;
  foreach my $rame (keys %privates) {
    $self->revCalls2($privates{$rame}) ;
  }
}
###########################################################################
# Adds an "is called by": self is called routine module
sub addCalled2 {
  my $self = shift ;
  my $orout = shift ;
  my $ocall = shift ;
  # Module is the calling module:
  my $module = "AAGLOBAL" ;
  # Father is the procedure name of the call location
  my $father = $ocall->get(-father) ;
  my $name = $father->get(-name) ;
  # Number is the line number for the call line
  my $number = $ocall->get(-number) ;
  $self->{"FCALLED"}->{"$module"}->{"$name"}->{"$number"} = $ocall ;
}
###########################################################################
# Adds an "is called by"...
sub addCalled {
  my $self = shift ;
  my $other = shift ;
  my $orout = shift ;
  my $ocall = shift ;
  my $module = $other->get(-name) ;
  my $father = $ocall->get(-father) ;
  my $name = $father->get(-name) ;
  my $number = $ocall->get(-number) ;
  $self->{"FCALLED"}->{"$module"}->{"$name"}->{"$number"} = $ocall ;
  print STDOUT (caller(0))[3] .": Module " . $self->get(-name)." is called by name=$module, object=$other\n" if $debug ;
}
###########################################################################
# Adds an "is called by"...
sub listCalled {
  my $self = shift ;
  if (exists $self->{"FCALLED"}) {
    return(%{$self->{"FCALLED"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
sub setCallers {
  my $self = shift ;
  my $drout = shift ;
  my $ocall = shift ;
  my $cname = $ocall->get(-name) ;
  my $ctype = $self->findType($cname) ;
  print STDOUT "\t" . (caller(0))[3] . ": Obtained value $ctype in " . $self->get(-name) . " for arg $cname\n" if $debug && $ctype ;
  # Accelerated return in case it is already known below:
  return if $ocall->get(-thelink) ;
  # Check for all module procedures:
  my $orout = $self->isRoutine($ocall) ;
  print STDOUT "\t\t" . (caller(0))[3] . ": Question " . $drout->get(-name) . $drout->get(-number) . " for call " . $ocall->get(-name) . $ocall->get(-number) . "\n" if $debug ;
  if ($orout) {
    $ocall->set(-thelink => $orout) ;
    print STDOUT "\t\t" . (caller(0))[3] . ": Setting thelink to " . $orout->get(-name) . $orout->get(-number) . " for call " . $ocall->get(-name) . $ocall->get(-number) . "\n" if $debug ;
  } else {
    # Could it be a private routine ?
    $orout = $drout->isPrivate($ocall) ;
    if ($orout) {
      print STDOUT "\t\t" . (caller(0))[3] . ": Setting private to " . $orout->get(-name) . $orout->get(-number) . " for call " . $ocall->get(-name) . $ocall->get(-number) . "\n" if $debug ;
      $ocall->set(-thelink => $orout) ;
    } else {
      # There might be a chance that the stuff is inside the used stuff:
      $orout = $drout->isModulprocedur($ocall) ;
      if ($orout) {
	$ocall->set(-thelink => $orout) ;
      } else {
	# Look for a direct access forcing connection:
	$orout = $drout->isDirect($ocall) ;
	if ($orout) {
	  $ocall->set(-thelink => $orout) ;
	} else {
	  # NAJ: not very clear stuff...
	  if (exists $drout->{"private"}) {
	    $self->setCallers($drout->{"private"}, $ocall) ;
	    return ;
	  } else {
	    # Look for a global procedure:
	    $orout = &fthrout::isGlobal($ocall) ;
	    if ($orout) {
	      $ocall->set(-thelink => $orout) ;
	    } elsif ($ctype && $ctype ne 'void') {
	      $ocall->set(-fake => $ctype) ;
	      return ;
	    }
	  }
	}
      }
    }
  }
  # Fasten search: eliminate void stuffs
  unless (&fthobj::getObject(lc($cname)) || &fthrout::getObject(lc($cname)) || ($ctype && $ctype ne 'void')) {
    unless ($cname) {
      $ocall->datadump("\t") ;
    }
    my $oext = new fthext(-name => "$cname") ;
    $oext->addStuff($ocall) ;
    return ;
  }
  unless ($ocall->getLink()) {
    # Last chance that the current function is a known return type :
    my $ctype = $drout->findType("$cname", 1) ;
    print STDOUT "\t" . (caller(0))[3] . ": Obtained value $ctype in " . $drout->get(-name) . " for arg $cname\n" if $ctype && $debug ;
    if ($ctype && $ctype ne 'void') {
      $ocall->set(-fake => $ctype) ;
      return ;
    } else {
      my $obc = new fthbc(-name => $cname) ;
      $obc->addStuff($ocall) ;
    }
  }
}
###########################################################################
# Intends to provide the reference to a the routine contained in self for
# a given call protocole:
sub isRoutine {
  my $self = shift ;
  my $ocall = shift ;
  my $cname = lc(shift) ;
  my $orout ;
  my $omod ;
  my $ko = 0 ;
  my $mname = $self->get(-name) ;
  $cname = lc($ocall->get(-name)) unless $cname ;
  my $ldebug = $debug ;
  # Procedures and interfaces direct call:
  $orout = $self->isSpecific($ocall) ;
  return($orout) if $orout ;
  # Selected use
  unless ($self->{"SELUSES"}->{"$cname"}) {
    print STDOUT "\t" . (caller(0))[3] . ": Module $mname does not select use of $cname:\n" if $ldebug ;
  }
  if ($self->{"SELUSES"}->{"$cname"}) {
    print STDOUT "\t" . (caller(0))[3] . ": Module $mname selects use of $cname:\n" if $ldebug ;
    # A specific import is to be found. Testing first encountered:
    if ($self->{"SELUSES"}->{"$cname"} =~ /(.*)::(.*)/) {
      $mname = lc($1) ; # Module name
      my $alias = lc($2) ; # Routine or interface name
      $omod = &fthmod::getObject($mname) ;
      if ($omod) {
	print STDOUT "\t\t" . (caller(0))[3] . ": Check whether selected module $mname contains $cname:\n" if $ldebug ;
	$orout = $omod->isSpecific($ocall, $alias) ;
	return($orout) if $orout ;
      }
    }
    # A specific import is to be found. Managing name conflicts:
    if (exists $self->{"SELUSED"}->{"$cname"}) {
      my (%hash) = (%{$self->{"SELUSED"}->{"$cname"}}) ;
      if (%hash) {
	foreach $mname (keys %hash) {
	  $omod = &fthmod::getObject($mname) ;
	  next unless $omod ;
	  my $alias = $self->{"SELUSED"}->{"$cname"}->{"$mname"} ;
	  print STDOUT "\t\t" . (caller(0))[3] . ": Check whether aliased module $mname contains $cname:\n" if $ldebug;
	  $orout = $omod->isSpecific($ocall, $alias) ;
	  return($orout) if $orout ;
	}
      }
      print STDOUT "$cname imported from " . $self->{"SELUSES"}->{"$cname"} . "\n" if $ldebug;
    }
  }
  # Recursive modules inquired once:
  if (exists $self->{"ALLUSES"}) {
    foreach $mname (keys %{$self->{"ALLUSES"}}) {
      $omod = $self->{"RECUSES"}->{"$mname"} ;
      next unless $omod ;
      print STDOUT "\t" . (caller(0))[3] . ": Module $mname used by " . $self->get(-name) . ":\n" if $ldebug;
      print STDOUT "\t\t" . (caller(0))[3] . ": Check whether module $mname contains $cname:\n" if $ldebug;
      # Ask the module: do you have such an object ?
      $orout = $omod->isSpecific($ocall) ;
      print STDOUT "\t\t" . (caller(0))[3] . ": Checked whether module $mname did contain $cname.\n" if $ldebug;
      if ($orout) {
	return($orout) ;
      }
    }
  }
  if (exists $self->{"RECUSES"}) {
    foreach $mname (keys %{$self->{"RECUSES"}}) {
      next if (exists $self->{"ALLUSES"}->{"$mname"}) ;
      $omod = $self->{"RECUSES"}->{"$mname"} ;
      next unless $omod ;
      print STDOUT "\t" . (caller(0))[3] . ": Module $mname used by " . $self->get(-name) . ":\n" if $ldebug;
      print STDOUT "\t\t" . (caller(0))[3] . ": Check whether module $mname contains $cname:\n" if $ldebug;
      # Ask the module: do you have such an object ?
      $orout = $omod->isSpecific($ocall) ;
      print STDOUT "\t\t" . (caller(0))[3] . ": Checked whether module $mname did contain $cname.\n" if $ldebug;
      if ($orout) {
	return($orout) ;
      }
    }
  }
  return(undef) ;
}
###########################################################################
# Intends to provide the reference to a the routine contained in self for
# procedures and interfaces
sub isSpecific {
  my $self = shift ;
  my $ocall = shift ;
  my $cname = lc(shift) ;
  my $orout ;
  my $ko = 0 ;
  my $mname = $self->get(-name) ;
  $cname = lc($ocall->get(-name)) unless $cname ;
  # Gets the prototype for the call:
  my (@cproto) = $ocall->getProto() ;
  print STDOUT "\t\t" . (caller(0))[3] . ": Is called $cname(@cproto) contained in module $mname ?\n" if $debug ;
  # Look for a direct procedure call:
  if (exists $self->{"FPROCEDURES"}->{"$cname"}) {
    # Routine name contained in the local procedure list
    $orout = $self->{"FPROCEDURES"}->{"$cname"} ;
    my (@mproto) = $orout->getProto() ;
    print STDOUT "\t\t" . (caller(0))[3] . ": Module $mname contains a procedure to $cname proto @mproto versus @cproto: $#mproto cmp $#cproto ie ".&fthrout::nargProto(\@mproto).':'.&fthrout::nargProto(\@cproto)."\n" if $debug ;
    #if ($#mproto == $#cproto) {
    if ($#mproto >= $#cproto && &fthrout::nargProto(\@mproto) <= &fthrout::nargProto(\@cproto)) {
      $ko = &fthrout::cmpProto(\@mproto, \@cproto) ;
      if ($ko) {
	$ocall->setLink($ko, $orout) ;
      } else {
	print STDOUT "\t\t\t" . (caller(0))[3] . ": Module procedure $cname has correct proto @mproto\n"  if $debug;
	return($orout) ;
      }
    } elsif ($debug) {
      print STDOUT "\t\t\t" . (caller(0))[3] . ": Call to $cname proto @mproto encounters @cproto\n"  if $debug;
    }
  }
  # Look for an interface:
  if (exists $self->{"FINTERFACES"}->{"$cname"}) {
    # Routine name contained in the interface.
    print STDOUT "\t\t\t" . (caller(0))[3] . ": Module $mname contains an interface to $cname:\n" if $debug ;
    foreach my $pname ($self->{"FINTERFACES"}->{"$cname"}->getProcedures()) {
      $pname = lc($pname) ;
      print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Testing module procedure $pname for module $mname, interface $cname:\n" if $debug ;
      if (exists $self->{"FPROCEDURES"}->{"$pname"}) {
	# Routine name contained in the module procedure list
	$orout = $self->{"FPROCEDURES"}->{"$pname"} ;
	my (@mproto) = $orout->getProto() ;
	print STDOUT "\t\t\t" . (caller(0))[3] . ": Module $mname contains a procedure to $pname(@mproto):$#mproto =? $#cproto\n" if $debug ;
	#if ($#mproto == $#cproto) {
	if ($#mproto >= $#cproto && &fthrout::nargProto(\@mproto) <= &fthrout::nargProto(\@cproto)) {
	  $ko = &fthrout::cmpProto(\@mproto, \@cproto) ;
	  if ($ko) {
	    $ocall->setLink($ko, $orout) ;
	    print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Interface $cname implements $pname with wrong proto $ko: @mproto differs from required @cproto\n"  if $debug;
	     # Tests the more relaxed analysis:
	    $ko = &fthrout::relProto(\@mproto, \@cproto) ;
	    unless ($ko) {
	      print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Interface $cname implements $pname with possible proto @mproto for @cproto\n" if $debug;
	      $ocall->addPossible($orout, $self->{"FINTERFACES"}->{"$cname"}) ;
	    }
	  } else {
	    print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Interface $cname implements $pname with correct proto @mproto\n"  if $debug;
	    return($orout) ;
	  }
	}
      } elsif ($self->{"FINTERFACES"}->{"$cname"}->getProcedure("$pname")) {
	# Routine name contained in the interface procedure list: may be defined outside the module
	# in a global file...
	$orout = $self->{"FINTERFACES"}->{"$cname"}->getProcedure("$pname") ;
	my (@mproto) = $orout->getProto() ;
	print STDOUT "\t\t\t" . (caller(0))[3] . ": Module $mname contains an interface to a procedure in $pname\n" if $debug ;
	#if ($#mproto == $#cproto) {
	if ($#mproto >= $#cproto && &fthrout::nargProto(\@mproto) <= &fthrout::nargProto(\@cproto)) {
	  $ko = &fthrout::cmpProto(\@mproto, \@cproto) ;
	  if ($ko) {
	    print STDOUT "\t\t\t" . (caller(0))[3] . ": Interface $cname implements $pname with wrong proto ko=$ko\n"  if $debug;
	    $ocall->setLink($ko, $orout) ;
	     # Tests the more relaxed analysis:
	    $ko = &fthrout::relProto(\@mproto, \@cproto) ;
	    unless ($ko) {
	      $ocall->addPossible($orout, $self->{"FINTERFACES"}->{"$cname"}) ;
	    }
	  } else {
	    # Reference to a global stuff:
	    if (&fthrout::getObject("$pname")) {
	      $orout = &fthrout::getObject("$pname") ;
	      print STDOUT "\t\t\t" . (caller(0))[3] . ": Interface $cname implements $pname with correct proto @mproto\n"  if $debug;
	    } else {
	      print STDOUT "\t\t\t" . (caller(0))[3] . ": Interface $cname implements $pname with correct proto @mproto\n"  if $debug;
	    }
	    return($orout) ;
	  }
	}
      }
    }
  }
}
###########################################################################
sub parseDef90 {
  my $self = shift ;
  my $ilhs = shift ;
  my $irhs = shift ;
  my $lhs = lc($ilhs) ;
  my $rhs = lc($irhs) ;
  my $shs ;
  my $name = $self->get(-name) ;
  my $chtml = &chtml::getCurrent() ;
  if ($lhs eq 'public' || $lhs eq 'private') {
    # This is not a type setting indeed:
    $chtml->addRaw("<DECLARE>$ilhs</DECLARE> :: <VARIABLE>$irhs</VARIABLE>") ;
    return ;
  }
  $lhs =~ s/\s+//g ;
  $rhs =~ s/\s+//g ;
  # Fortran argument definition equivalent form:
  $lhs =~ s/,*intent\(.*?\)//g ;
  $lhs =~ s/,*allocatable//g ;
  $lhs =~ s/,*save//g ;
  if ($lhs =~ /(.*?),*dimension\((.*?)\)(.*)/i) {
    my (@array) = split(/\,/, $2) ;
    $lhs = $1 . $3 . ',dimension(' . ',' x ( $#array + 0 ) . ')' ;
    $lhs =~ s/,*pointer//g ;
  }
  while ($rhs =~ /(.*)\((.*?)\)(.*)/) {
    my $a = $1 ;  my $b = $2 ;  my $c = $3 ;  
    $b =~ s/\,/\;/g ;
    $rhs = $a . '{' . $b . '}' . $c ;
  }
  my (@array) = split(/\,/, $rhs) ;
  foreach my $rhs (@array) {
    print STDOUT  (caller(0))[3] . ": Parsing rhs=<$rhs> \n" if $debug ;
    if ($rhs =~ /(.*)\s*\=\>{0,1}\s*([\w\.\+\-\*\/]*)/) {
      $rhs = $1 ;
    }
    if ($rhs =~ /(\w+)\{(.*?)\}/) {
      $rhs = $1 ;
      my (@larray) = split(/\,/, $2) ;
      $shs = $lhs . ',dimension(' . ',' x ( $#larray + 0 ) . ')' ; ;
    } else {
      $shs = $lhs ;
    }
    print STDOUT (caller(0))[3] . ": In " . $self->get(-name) . " line $. defined var=$rhs is of type $shs\n" if $debug ;
    $self->{"ALLDEFS"}->{"$rhs"} = $shs ;
  }
  if ($ilhs =~ /(.*type\s*\(\s*)(\w+)(\s*\).*)/i) {
    $chtml->addRaw("<DECLARE>$1") ;
    my $refe = new oref(-name => $2, -type => "fthtype") ;
    print STDOUT "Refering to another type object $2: $refe\n" if $debug ;
    $chtml->addRaw($refe) ;
    $chtml->addRaw("$3</DECLARE> :: <VARIABLE>$rhs</VARIABLE>") ;
  } elsif ($ilhs =~ /(.*class\s*\(\s*)(\w+)(\s*\).*)/i) {
    $chtml->addRaw("<DECLARE>$1") ;
    my $refe = new oref(-name => $2, -type => "fthtype") ;
    print STDOUT "Refering to another type object $2: $refe\n" if $debug ;
    $chtml->addRaw($refe) ;
    $chtml->addRaw("$3</DECLARE> :: <VARIABLE>$rhs</VARIABLE>") ;
  } else {
    $chtml->addRaw("<DECLARE>$ilhs</DECLARE> :: <VARIABLE>$irhs</VARIABLE>") ;
  }
}
###########################################################################
sub read {
  my $self = shift ;
  my $in = shift ;
  my $file = shift ;
  my $chtml = &chtml::getCurrent() ;
  if ((ref $in) && (fileno $in)) {
    my $name = $self->get(-name) ;
    $file = $self->get(-file) unless $file ;
    my $num ;
    my $preif = 0 ;
    my(@ploops) = () ;
    while(<$in>) {
      chomp ;
      # Parsing a fortran 90 module below:
      my $cc = "" ;
      if (/^(\#.*?)/) {
	# Some preprocessing stuff: passed to later analysis
      } elsif (/^(.*?)\!(.*)/) {
	$_ = $1 ;
	$cc = $2 ;
      }
      if (/^(\s*end\s*module\s+)$name\b/i) {
	# End of module creation
	print STDERR "Created module $name\n" if $debug ;
	$chtml->addDef("$1") ;
	my $refe = new oref(-name => $name, -type => "fthmod") ;
	$chtml->addRaw($refe) ;
	return ;
      } elsif (/^(\s*end)\b/i) {
	# End of module creation with somewhat illegal ending...
	# We ought t advise the user to use emacs instead of bullshit...
	print STDERR "Created module $name\n" if $debug ;
	$chtml->addDef("$1 module ") ;
	my $refe = new oref(-name => $name, -type => "fthmod") ;
	$chtml->addRaw($refe) ;
	return ;
      } else {
	if (/^\s*contains\b/i) {
	  $self->set(-contains => 1) ;
	  $chtml->addDef($_) ;
	} elsif (/^(\s*type\s*\,*\s*.*\:\:\s*)(\w+)([\s\;]*)/i) {
	  # Create a new type:
	  print STDOUT "Creating new ftype($.) = $2\n" if $debug ;
	  my $otype = new fthtype(-name => $2, -file => $file, -masfic => $self->get(-masfic), -number => $., -module => $self) ;
	  $self->addType($2 => $otype) ;
	  my $l = $3 ;
	  # Parse the $1 for the extends attribute:
	  if ($1 =~ /^(\s*type\s*\,\s*extends\s*)\(\s*(\w+)\s*\)(\s*\:\:\s*)/) {
	    my $refe = new oref(-name => $2, -type => "fthtype") ;
	    print STDOUT "Refering to another fthtype object " . $otype->get(-name) . ": $refe\n" if $debug ;
	    $chtml->addDef("$1(") ;
	    $chtml->addRaw($refe) ;
	    $chtml->addDef(")$3") ;
	  } else {
	    $chtml->addDef($1) ;
	  }
	  $chtml->addRaw($otype) ;
	  $chtml->addDef($l) ;
	  $otype->read($in) ;
	} elsif (/^(\s*type\s+)(\w+)([\s\;]*)/i) {
	  # Create a new type:
	  print STDOUT "Creating new ftype($.) = $2\n" if $debug ;
	  my $otype = new fthtype(-name => $2, -file => $file, -masfic => $self->get(-masfic), -number => $., -module => $self) ;
	  $self->addType($2 => $otype) ;
	  $chtml->addDef($1) ;
	  $chtml->addRaw($otype) ;
	  $chtml->addDef($3) ;
	  $otype->read($in) ;
	} elsif  (/^\s*(abstract)*\s*interface/i) {
	  if  (/^(\s*interface\s+)(operator\s*)\((.*?)\)(.*)/i) {
	    my $opname = "operator$3" ;
	    my $inter = new fthinter(-name => "$opname", -file => $file, -handle => $file, -masfic => $self->get(-masfic), -module => $self) ;
	    $chtml->addDef($1) ;
	    $chtml->addDef("$2(") ;
	    $chtml->addRaw($inter) ;
	    $chtml->addDef(")$4") ;
	    $inter->read($in) ;
	    $self->addInter("$opname" => $inter) ;
	  } elsif  (/^(\s*interface\s+)(\w+)/i) {
	    my $inter = new fthinter(-name => $2, -file => $file, -handle => $file, -masfic => $self->get(-masfic), -module => $self) ;
	    $chtml->addDef($1) ;
	    $chtml->addRaw($inter) ;
	    $inter->read($in) ;
	    $self->addInter($2 => $inter) ;
	  } elsif  (/^(\s*abstract\s+interface\s*)/i) {
	    # Basic interface to global stuff...
	    my $inter = new fthinter(-file => $file, -handle => $file, -masfic => $self->get(-masfic), -module => $self) ;
	    $chtml->addDef($1.' ') ;
	    $chtml->addRaw($inter) ;
	    $inter->read($in) ;
	    my $mname = $inter->get(-name) ;
	    #$self->addInter($mname => $inter) ;
	    $inter->addInter($self) ;
	  } elsif  (/^(\s*interface\s*)/i) {
	    # Basic interface to global stuff...
	    my $inter = new fthinter(-file => $file, -handle => $file, -masfic => $self->get(-masfic), -module => $self) ;
	    $chtml->addDef($1.' ') ;
	    $chtml->addRaw($inter) ;
	    $inter->read($in) ;
	    my $mname = $inter->get(-name) ;
	    #$self->addInter($mname => $inter) ;
	    $inter->addInter($self) ;
	  } else {
	    $chtml->addDef($_) ;
	  }
	} elsif  (/^\s*use/i) {
	  if  (/^(\s*use\s+)(\w+)\s*,\s*only\s*:\s*(.*)/i) {
	    my $fmused = $2 ; $_ = $3 ;
	    my $refe = new oref(-name => $2, -type => "fthmod") ;
	    $chtml->addDef("$1") ;
	    $chtml->addRaw($refe) ;
	    $chtml->addDef(", only : ") ;
	    $self->readUses($in, $fmused, "&" . $_) ;
	  } elsif  (/^(\s*use\s+)(\w+)\s*$/i) {
	    my $fmused = lc($2) ;
	    my $refe = new oref(-name => $2, -type => "fthmod") ;
	    $chtml->addDef("$1") ;
	    $chtml->addRaw($refe) ;
	    $self->{"ALLUSES"}->{"$fmused"} = 1 ;
	  } else {
	    $chtml->addDef($_) ;
	  }
	} elsif (/(subroutine|function)\s+(\w+)(.*)/i) {
	  my $rame = $2 ;
	  my $lame = lc($rame) ;
	  # List of objects management
	  my $orout = new fthrout(-name => $rame, -file => $file, -masfic => $self->get(-masfic), -module => $name, -number => $.) ;
	  $self->addRout($rame => $orout) ;
	  my $proto ;
	  if (/^(\s*)(.*?)(\s*function\s+)(\w+)(.*)/i) {
	    $chtml->addDef($1) ;
	    $proto = $2 ;
	    if ($proto) {
	      if ($proto =~ /\bpure\b/i || $proto =~ /\brecursive\b/i) {
		$chtml->addDef($proto) ;
	      } else {
		$orout->set(-theproto => $proto) ;
		print STDOUT "Defining proto to $rame as $proto\n" if $debug ;
		$chtml->addRaw("<DECLARE>$proto</DECLARE>") ;
	      }
	    }
	    $chtml->addDef($3) ;
	    $chtml->addRaw($orout) ;
	    $orout->readf90subr($in, "& $5") ;
	    if ($proto) {
	      unless ($proto =~ /\bpure\b/i || $proto =~ /\brecursive\b/i) {
		$orout->addArgument("$rame") ;
		$orout->addType("$rame" => "$proto") ;
		$orout->setType("$rame" => "$proto") ;
		$orout->set(-theproto => "$proto") ;
	      }
	    }
	    unless ($orout->get(-theproto)) {
	      $orout->addArgument("$rame") ;
	      $orout->addType("$rame" => 'void') ;
	      $orout->set(-theproto => 'void') ;
	    }
	    if ($orout->getType($rame) eq 'void' && $proto) {
	      $orout->addType($rame => $proto) ;
	      $orout->setType($rame => $proto) ;
	      print STDOUT "Error bad proto $rame\n" if $debug ;
	    }
	  } elsif (/^(\s*)(.*?)(\s*subroutine\s+)$rame(.*)/i) {
	    $chtml->addDef($1.$2) ;
	    $orout->set(-theproto => 1) ;
	    $chtml->addDef($3) ;
	    $chtml->addRaw($orout) ;
	    $orout->readf90subr($in, "& $4") ;
	  } else {
	    $chtml->addDef($_) ;
	  }
	} elsif (/^\#(ifdef|ifndef|if|elif|else|endif)(.*)/i) {
	  # Preprocessing tests, this seems good code below:
	  if ($1 eq "endif") {
	    $num = pop(@ploops) ;
	    if ($num) {
	      $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	      $preif-- ;
	    } else {
	      $chtml->addDef('#'."$1$2") ;
	    }
	  } elsif ($1 eq "if" || $1 eq "ifdef" || $1 eq "ifndef") {
	    push(@ploops, $.) ;
	    $preif++ ;
	    $chtml->addRaw("<TEST><A Name=C_$.>".'#'."$1</A></TEST>$2") ;
	  } else {
	    $num = pop(@ploops) ;
	    if ($num) {
	      push(@ploops, $num) ;
	      $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	    } else {
	      $chtml->addDef('#'."$1$2") ;
	    }
	  }
	} elsif (/^(\s*include\s+)([\"\'])([\w\.\/]+)\2(.*)/i || /^(\#include\s+)([\"\'])([\w\.\/]+)\2(.*)/i) {
	  # Includes here:
	  my $incf = $3 ;
	  $chtml->addDef("$1$2") ;
	  $chtml->addRaw("<A Name=C_$.></A>");
	  unless (-f "$incf") {
	    my $base = dirname($self->get(-file)) ;
	    if (-f "$base/$incf") {
	      $incf = "$base/$incf" ;
	    } elsif (-f "$base/" . basename($incf)) {
	      $incf = "$base/" . basename($incf) ;
	    } elsif ($main::opt_I) {
	      # Possible...
	      my @TAGSHTML_PATH = (split ( /:/, $main::opt_I)) ;
	      foreach $base (@TAGSHTML_PATH) {
		if (-f "$base/$incf") {
		  $incf = "$base/$incf" ;
		  last ;
		}
	      }
	    }
	  }
	  if (-f "$incf") {
	    my $fh ;
	    my $other ;
	    if (open($fh,"<$incf")) {
	      print STDOUT (caller(0))[3] . ": Open $incf from $file\n" if $debug ;
	      $other = &fthinc::getObject($incf) ;
	      if ($other) {
		new chtml() ;
	      } else {
		new chtml(-file => $incf, -name => $incf) ;
		$other = new fthinc(-file => "$incf", -name => "$incf", -masfic => $self->get(-masfic), -father => $self) ;
	      }
	      $self->addInclude("$incf" => $other) ;
	      $chtml->addRaw($other) ;
	      $chtml->addDef("$2$4") ;
	      $self->read($fh, $incf) ;
	      close($fh) ;
	      &chtml::setCurrent($chtml) ;
	      print STDOUT (caller(0))[3] . ": Closed $incf($.) from $file\n" if $debug ;
	    }
	  } else {
	    my $refe = new oref(-name => "$3", -type => "fthinc") ;
	    $chtml->addRaw($refe) ;
	    $chtml->addDef("$2$4") ;
	  }
	} elsif (/^(\s*)(.*?)\s*::\s*(.*)/) {
	  $chtml->addDef("$1") ;
	  $self->parseDef90($2, $3) ;
	} elsif (/^(\s*)(integer|real|character|complex|logical|dimension)(.*?)[\s\:]+([\w\,]+.*)/i || /^(\s*)(double)(\s+precision)[\s\:]+([\w\,]+.*)/i) {
	  my $lhs = $2 . $3  ; my $rhs = $4 ;
	  $chtml->addDef("$1") if $1 ;
	  $self->parseDef90($lhs, $rhs) ;
	} elsif (/^\s*common\s*\/\s*(\w+)\s*\/(.*)/i) {
	  $chtml->addDef($_) ;
	} elsif (/^\s*implicit/i) {
	  if (/^\s*implicit\s+none/i) {
	    # Delete all the implicit keys:
	    $self->ImplicitReset() ;
	    $chtml->addRaw("<DECLARE>") ;
	    $chtml->addDef($_) ;
	    $chtml->addRaw("</DECLARE>") ;
	  } elsif (/^\s*implicit\s+([\w\*]+)(.*?)[\s\:]+\((.*)\)/i) {
	    my $lhs = lc ( $1 . $2 ) ;
	    my $rhs = lc($3) ;
	    my (@array) = split(/\,/, $rhs) ;
	    foreach $rhs (@array) {
	      # Builds a regular expression:
	      $self->ImplicitProcess($rhs => $lhs) ;
	    }
	    $chtml->addRaw("<DECLARE>") ;
	    $chtml->addDef($_) ;
	    $chtml->addRaw("</DECLARE>") ;
	  } else {
	    $chtml->addDef($_) ;
	  }
	} else {
	  $chtml->addDef($_) ;
	}
      }
      if ($cc) {
	$chtml->addRaw('<COMMENT>') ;
	if ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  $chtml->addLaCoM("<A Name=C_$.></A>") ;
	}
	while ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  my $incp = "$3.$4" ;
	  $cc = $5 ;
	  $chtml->addLaTeX($1.$2." ") ;
	  unless (-f "$incp") {
	    my $base = dirname($self->get(-file)) ;
	    my $nase = basename(dirname($incp)) ;
	    if (-f "$base/$incp") {
	      $incp = "$base/$incp" ;
	    } elsif (-f "$base/" . basename($incp)) {
	      $incp = "$base/" . basename($incp) ;
	    } elsif (-f "$nase/" . basename($incp)) {
	      $incp = "$nase/" . basename($incp) ;
	    } elsif ($main::opt_L) {
	      # LaTeX inputs directories
	      # Possible...
	      my @TAGSHTML_PATH = (split ( /:/, $main::opt_L)) ;
	      foreach $base (@TAGSHTML_PATH) {
		if (-f "$base/$incp") {
		  $incp = "$base/$incp" ;
		  last ;
		}
	      }
	    }
	  }
	  if (-f "$incp") {
	    my $other = new fthinp(-file => "$incp", -name => "$incp", -masfic => $self->get(-masfic)) ;
	    $self->addInput("$incp" => $other) ;
	    $chtml->addLaCoM($other) ;
	  } else {
	    my $refe = new oref(-name => $incp, -type => "fthinp") ;
	    $chtml->addLaCoM($refe) ;
	  }
	}
	$chtml->addLaTeX($cc) ;
	print STDOUT "In " . $self->get(-file) . " line $. call for end of coment\n" if $debug ;
	$chtml->addRaw("</COMMENT>") ;
      }
    }
  }
}
###########################################################################
1;

