package fthobj ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A stuff container
###########################################################################
use strict;
use Carp ;
###########################################################################
use fthrout ;
use fthcall ;
use fthtype ;
use fthmod ;
use fthinter ;
use ihtml ;
###########################################################################
use vars qw($debug $nbself %MYOBJHASH) ;
$debug = 0 ;
$nbself = 0 ;
(%MYOBJHASH) = () ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = lc(shift) ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  if ($args{-name}) {
    $name = lc($args{-name}) ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
  } else {
    $nbself++ ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug;
  }
  return $self ;
}
###########################################################################
# Gets the number of built objects:
sub getNbself {
  return($nbself) ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub save_hash {
  my $self = shift ;
  my $handle = shift ;
  my $name = $self->get(-name) ;
  if (exists $self->{"STUFFS"}) {
    print $handle '(@{$fthobjs{"' . $name . '"}}) = (' ;
    foreach my $orout (@{$self->{"STUFFS"}}) {
      my $number = $orout->get(-number) ;
      my $file = $orout->get(-file) ;
      if ($file && $number) {
	# Weird no file number in no named interfaces.
	# Brutal hard correction, not trapping errors.
	print $handle '"' . "$file:$number" . '", '  ;
      }
    }
    print $handle ');' ."\n" ;
    print $handle '$names{"' . lc($name) . '"} = "' . $name . '";' ."\n" ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	if ($sval && ref($sval) eq 'fthrout') {
	  print STDOUT "$sval'" . $sval->get(-name) ."', ";
	} elsif ($sval && ref($sval) eq 'fthinter') {
	  $sval->datadump("$t\t") ;
	} else {
	  print STDOUT "$sval, " ;
	}
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	if (exists $array{$sval} && ref($array{$sval}) eq 'fthrout') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->get(-name)  ;
	  print STDOUT", " ;
	} elsif (exists $array{$sval} && ref($array{$sval}) eq 'fthinter') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->get(-name)  ;
	  print STDOUT", " ;
	} else {
	  print STDOUT "$sval => '$array{$sval}', " ;
	}
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'fthrout')) {
      print STDOUT "$t$cle = $val'" . $val->get(-name) ."',\n";
    } elsif (($val) && (ref($val) eq 'fthinter')) {
      print STDOUT "$t$cle = $val'" . $val->get(-name) ."',\n";
    } else {
      $val = "" unless $val ;
      print STDOUT "$t$cle = $val,\n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
# Adds a call
sub addStuff {
  my $self = shift ;
  my $bidule = shift ;
  push(@{$self->{"STUFFS"}}, $bidule) ;
}
###########################################################################
# Gets all the contained stuffs...
sub getStuffs {
  my $self = shift ;
  if (exists $self->{"STUFFS"}) {
    return(@{$self->{"STUFFS"}}) ;
  } else {
    return((undef)) ;
  }
}
###########################################################################
# Gets a procedure:
sub getProcedure {
  my $self = shift ;
  my $name = lc(shift) ;
  my $orout ;
  if (exists $self->{"STUFFS"}) {
    foreach $orout (@{$self->{"STUFFS"}}) {
      if ($orout && ref($orout) eq 'fthrout') {
	if (lc($orout->get(-name)) eq $name) {
	  return($orout) ;
	}
      }
    }
  }
  return(undef)
}
##########################################################################
# Computes all the return types for selves if any (requires functions of same
# return value).
sub setTypes {
  my $counter = 0 ;
  foreach my $self (values %MYOBJHASH) {
    $counter++ ;
    if (!$debug && $counter == 10) {
      print STDOUT "." ;
      $counter = 0 ;
    }
    $self->getType() ;
  }
  print STDOUT "\n" unless $debug ;
}
###########################################################################
# Providing the routine type correspoding to self if connected to a routine...
sub getType {
  my $self = shift ;
  if ($self->{"guessed"}) {
    return($self->{"guess"}) ;
  } else {
    $self->{"guessed"} = 1 ;
    $self->{"guess"} = undef ;
    my $val ;
    my $valeur = undef ;
    my(@array) = $self->getStuffs() ;
    my $ko = 0 ;
    my $ok ;
    # List the objects that may contain interesting stuff:
    foreach my $bidule (@array) {
      if (ref($bidule) eq 'fthrout') {
	# Routine target:
	$ok = $bidule->get(-theproto) ;
	if (!$ok || ($ok && $ok ne '1' && $ok ne '2')) {
	  # This is a function:
	  $val = $bidule->getReturn() ;
	  if ($valeur) {
	    unless ("$val" eq "$valeur") {
	      $valeur = "" ;
	      $ko = 1 ;
	      last ;
	    }
	  } elsif ($val) {
	    $valeur = $val ;
	  }
	} else {
	  $self->{"guessed"} = 2 ;
	}
      } elsif (ref($bidule) eq 'fthinter') {
	# Interface link:
	my(@procs) = $bidule->getProcedures() ;
	foreach my $name (@procs) {
	  my $other = &fthobj::getObject("$name") ;
	  if ($other) {
	    my(@array2) = $other->getStuffs() ;
	    # List the objects that may contain interesting stuff:
	    foreach my $bidule2 (@array2) {
	      if (ref($bidule2) eq 'fthrout') {
		# Routine target:
		$ok = $bidule2->get(-theproto) ;
		if (!$ok || ($ok && $ok ne '1' && $ok ne '2')) {
		  # This is a function:
		  $val = $bidule2->getReturn() ;
		  if ($valeur) {
		    unless ("$val" eq "$valeur") {
		      $valeur = "" ;
		      $ko = 1 ;
		      last ;
		    }
		  } elsif ($val) {
		    $valeur = $val ;
		  }
		} else {
		  $self->{"guessed"} = 2 ;
		}
	      }
	      last if $ko ;
	    }
	  }
	  last if $ko ;
	}
      }
      last if $ko ;
    }
    if (defined($valeur)) {
      print STDOUT (caller(0))[3] . ": Polymorphic object " . $self->get(-name) . " has a unique return type " . $valeur . "\n" if $debug;
    }
    $self->{"guess"} = $valeur ;
    return($self->{"guess"}) ;
  }
}
###########################################################################
sub any {
  if (%MYOBJHASH) {
    return(1) ;
  } else {
    return(0) ;
  }
}
###########################################################################
sub htmlprint {
  my $DIRE = shift ;
  my $reldir = shift ;
  # Create a fully new stuff:
  my $ihtml = new ihtml(-name => "inter/", -reldir => $reldir) ;
  while ( my($name, $self) = each %MYOBJHASH) {
    print STDOUT "." unless $debug ;
    if (exists $self->{"name"}) {
      $name = $self->{"name"} ;
    }
    my $target = &ihtml::targetinter($name) . "_i.html" ;
    if (open(INTERF,">$DIRE/$target")) {
      my $handle = \*INTERF ;
      $ihtml->set(-handle => $handle) ;
      $ihtml->set(-file => $target) ;
      $ihtml->set(-name => $name) ;
      $ihtml->set(-title => "Interface setting") ;
      $ihtml->start_interface("../") ;
      if (exists $self->{"STUFFS"}) {
	if ($self->{"guess"}) {
	  my $handle =  $ihtml->get(-handle) ;
	  print $handle "<BR>Default type is '" . $self->{"guess"} . "'\n" ;
	} else {
	  my $handle =  $ihtml->get(-handle) ;
	  print $handle "<BR>No default type\n" ;
	}
	$ihtml->print_interface($self);
      }
      $ihtml->end_html("../") ;
      close(INTERF) ;
    }
  }
  print STDOUT "\n" unless $debug ;
}
###########################################################################
1;

