package fthrout ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for fortran 90 procedures.
# This contains the calls and some data structure.
# This is contained by a module or may be a global stuff.
###########################################################################
use strict;
use Carp ;
use File::Basename ;
###########################################################################
use fthmod ;
use fthcall ;
use fthtype ;
use fthinter ;
use fthobj ;
use fthinc ;
use fthinp ;
use fthbc;
use fthext;
use oref ;
###########################################################################
use vars qw($debug $nbcalls $nbself $nbglo $nbmod $nbpri $nbfglo $nbfmod $nbfpri $nbrglo $nbrmod $nbrpri %MYOBJHASH %INTRF %INTRS %INTRC) ;
$debug = 0 ;
$nbcalls = 0 ;
$nbself = 0 ;
$nbglo = 0 ;
$nbmod = 0 ;
$nbpri = 0 ;
$nbrpri = 0 ;
$nbrmod = 0 ;
$nbrglo = 0 ;
$nbfpri = 0 ;
$nbfmod = 0 ;
$nbfglo = 0 ;
(%MYOBJHASH) = () ;
(%INTRF) = ('int' => 'integer', 'nint' => 'integer', 'mod' => 'integer', 'size' => 'integer', 
	    'ubound' => 'integer', 'lbound' => 'integer', 'ishft' => 'integer', 'isign' => 'integer', 'sign' => 'real',
	    'trim' => 'character*(*)', 'len' => 'integer', 'len_trim' => 'integer', 'index' => 'integer', 'kind' => 'integer',
	    'adjustl' => 'character*(*)',
	    'max' => 'void', 'min' => 'void', 'sum' => 'void', 'abs' => 'void', 'exp' => 'void', 
	    'real' => 'real', 'imag' => 'real', 'aimag' => 'real', 'conjg' => 'real', 
	    'float' => 'real',
	    'sqrt' => 'real', 'log' => 'real', 'log10' => 'real',
	    'sin' => 'real', 'cos' => 'real', 'tan' => 'real', 'sinh' => 'real', 'cosh' => 'real', 'tanh' => 'real',
	    'asin' => 'real', 'acos' => 'real', 'atan' => 'real', 'char' => 'character',
	    'cmplx' => 'cmplx', 'maxval' => 'void', 'minval' => 'void',
	    'epsilon' => 'real',
	    'deallocate' => 'void', 'null' => 'void',
	    'flush' => 'real',
	    'time' => 'real', 'etime' => 'real',
	    'exit' => 'integer',
	    'iargc' => 'integer',
	    'tiny' => 'real',
	    'huge' => 'real',
	    'secnds' => 'real',
	    'flush' => 'void',
	    ) ;
(%INTRS) = ('system_clock' => 1, 'get_command_argument' => 11, 'get_environment_variable' => 1, 
	    'system' => 1, 'getenv' => 1, 'getarg' => 1, 'exit' => 1, 'flush' => 1, 'iargc' => 1,
	    'cpu_time' => 1, 'date_and_time' => 1, 'etime' => 1,'etime' => 1,
	    'inspers' => 1, 's8mcur' => 1, 's8rmark' => 1, 's8arret' => 1,
	    'rename' => 1, 'unlink' => 1, 'mkdir' => 1, 'chown' => 1, 
    ) ;
(%INTRC) = ('fprintf' => 'void', 'printf' => 'void', 'flush' => 'void', 'fflush' => 'void',
	    'sizeof' => 'void', 'calloc' => 'void', 'malloc' => 'void',
	    'for' => 'void', 'if' => 'void', 'switch' => 'void', 'return' => 'void',
	    'free' => 'void',
    ) ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = lc(shift) ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  $nbself++ ;
  bless $self;
  my $name ;
  if (exists $args{"-name"}) {
    if (exists $args{"-global"}) {
      # Global unique routine point:
      $name = lc($args{-name}) ;
      if (exists $MYOBJHASH{$name}) {
	print STDERR "\n" . (caller(0))[3] . ": Redefining routine $name from file " . $MYOBJHASH{$name}->get(-file) . " in file " . $args{-file}."\n" ;
      }
      $MYOBJHASH{$name} = $self ;
      $nbglo++ ;
      print STDOUT (caller(0))[3] ." new global ($name)\n" if $debug ;
      $self->ImplicitProcess(&fthrout::ImplicitDefaults()) ;
    } elsif ($args{"-module"}) {
      # From module called, remains unique in fortran:
      $name = $args{-name} ;
      my $stuff = new fthobj(-name => $name) ;
      $stuff->addStuff($self) ;
      # $MYOBJHASH{$name} = $self unless (exists $MYOBJHASH{$name}) ;
      $name = lc( $args{"-module"} . '::' . $args{"-name"} ) ;
      if (exists $MYOBJHASH{$name}) {
	print STDERR "\t" . (caller(0))[3] . ": Redefining module routine $name from file " . $MYOBJHASH{$name}->get(-file) . " in file " . $args{-file}."\n" ;
      }
      $MYOBJHASH{$name} = $self ;
      if (exists $args{"-module"}) {
	my $father = &fthmod::getObject($args{"-module"}) ;
	$self->ImplicitProcess($father->getImplicitRules()) ;
      } else {
	$self->ImplicitProcess(&fthrout::ImplicitDefaults()) ;
      }
      $nbmod++ ;
      print STDOUT (caller(0))[3] . "new procedure($name)\n" if $debug ;
    } elsif ($args{"-private"}) {
      # From module called, remains unique in fortran:
      $name = $args{-name} ;
      my $stuff = new fthobj(-name => $name) ;
      $stuff->addStuff($self) ;
      # $MYOBJHASH{$name} = $self unless (exists $MYOBJHASH{$name}) ;
      $name = lc( 'private' . '::' . $args{"-private"}->get(-name) . '::' . $args{"-name"} ) ;
      if (exists $MYOBJHASH{$name}) {
	print STDERR "\n" . (caller(0))[3] . ": Redefining private routine $name from file " . $MYOBJHASH{$name}->get(-file) . " in file " . $args{-file}."\n" ;
      }
      $MYOBJHASH{$name} = $self ;
      $nbpri++ ;
      if (exists $args{"-private"}) {
	my $father = $args{"-private"} ;
	$self->ImplicitProcess($father->getImplicitRules()) ;
      }
      print STDOUT (caller(0))[3] . "new private($name)\n" if $debug ;
    }
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
sub old {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  if ($args{-name}) {
    $name = $args{-name} ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
  } else {
    $first = 1 ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) if (exists $self->{"debug"}) ;
  }
  return $self ;
}
###########################################################################
# Gets the number of built objects:
sub getNbself {
  return($nbself) ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub save_hash {
  my $self = shift ;
  my $handle = shift ;
  my $name = $self->get(-name) ;
  my $number = $self->get(-number) ;
  my $file = $self->get(-file) ;
  if (exists $self->{"module"}) {
    my $module = $self->{"module"} ;
    print $handle '$defmethod{"' . $module . '"}->{"' . $name . '"} = ' ;
    print $handle '"' . "$file:$number" . '"'  ;
    print $handle ';' ."\n" ;
    print $handle '$names{"' . lc($module) . '::' . lc($name) . '"} = "' . $name . '";' ."\n" ;
  } elsif (exists $self->{"private"}) {
    my $father = $self->{"private"} ;
    my $rame = $father->get(-name) ;
    print $handle '$defmethod{"' . $rame . '"}->{"' . $name . '"} = ' ;
    print $handle '"' . "$file:$number" . '"'  ;
    print $handle ';' ."\n" ;
    print $handle '$names{"' . lc($rame) . '::' . lc($name) . '"} = "' . $name . '";' ."\n" ;
  } elsif ($self->{"global"}) {
    # Not contained inside an fthobj:
    print $handle '$fthrouts{"' . $name . '"} = ' ;
    print $handle '"' . "$file:$number" . '"'  ;
    print $handle ';' ."\n" ;
    print $handle '$names{"' . lc($name) . '"} = "' . $name . '";' ."\n" ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	print STDOUT "$sval, " ;
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	if (exists $array{$sval} && ref($array{$sval}) eq 'fthcall') {
	  print STDOUT "$t\t$sval => "  ;
	  $array{$sval}->datadump("\t$t")  ;
	  print STDOUT", " ;
	} else {
	  print STDOUT "$sval => $array{$sval}, " ;
	}
      }
      print STDOUT "},\n" ;
    } else {
      print STDOUT "$t$cle = $val,\n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
sub any {
  if (%MYOBJHASH) {
    return(1) ;
  } else {
    return(0) ;
  }
}
###########################################################################
# Returns the implicit default rule pattern
sub ImplicitDefaults {
  return('C' => 'complex', 'I-N,P' => 'integer', 'A-B,D-H,O,Q,R-Z' => 'real') ;
}
###########################################################################
# Returns the implicit current rule pattern
sub getImplicitRules {
  my $self = shift ;
  if (exists $self->{"IMPLICITDEFS"}) {
    return(%{$self->{"IMPLICITDEFS"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Returns the implicit default rule pattern
sub ImplicitReset {
  my $self = shift ;
  if (exists $self->{"IMPLICITDEFS"}) {
    foreach my $rhs (keys %{$self->{"IMPLICITDEFS"}}) {
      delete $self->{"IMPLICITDEFS"}->{"$rhs"} ;
    }
    delete $self->{"IMPLICITDEFS"} ;
  }
}
###########################################################################
# Process the implicit line
sub ImplicitProcess {
  my $self = shift ;
  my (%hash) = (@_) ;
  my $string = "abcdefghijklmnopqrstuvwxyz" ;
  unless (%hash) {
    delete $self->{"IMPLICITDEFS"} ;
    return ;
  }
  foreach my $cle (keys %hash) {
    if ($cle =~ /\-/) {
      # Replacing...
      if ($string =~ /([$cle]+)/i) {
	my $charlist = $1 ;
	while($charlist =~ /^(\w)(.*)/) {
	  $charlist = $2 ;
	  $self->{"IMPLICITDEFS"}->{$1} = $hash{$cle} ;
	}
      }
      
    } else {
      my $lcle = lc($cle) ;
      $self->{"IMPLICITDEFS"}->{$lcle} = $hash{$cle} ;
    }
  }
}
###########################################################################
sub getCounter {
  my(%args) = (@_) ;
  print STDOUT (caller(0))[3] .": Type is " . $args{"-type"} ."\n" if $debug ;
  print STDOUT (caller(0))[3] .": Mode is " . $args{"-mode"} ."\n" if $debug ;
  if (exists $args{"-type"} && exists $args{"-mode"}) {
    if ($args{"-type"} eq "function") {
      if ($args{"-mode"} eq "glo") {
	return($nbfglo) ;
      } elsif ($args{"-mode"} eq "mod") {
	return($nbfmod) ;
      } elsif ($args{"-mode"} eq "pri") {
	return($nbfpri) ;
      }
    } elsif ($args{"-type"} eq "routine") {
      if ($args{"-mode"} eq "glo") {
	return($nbrglo) ;
      } elsif ($args{"-mode"} eq "mod") {
	return($nbrmod) ;
      } elsif ($args{"-mode"} eq "pri") {
	return($nbrpri) ;
      }
    }
  }
  if (exists $args{"-mode"}) {
    if ($args{"-mode"} eq "glo") {
      return($nbglo) ;
    } elsif ($args{"-mode"} eq "mod") {
      return($nbmod) ;
    } elsif ($args{"-mode"} eq "pri") {
      return($nbpri) ;
    }
  }
  return(0) ;
}
###########################################################################
sub getRouts {
  my(@array) = () ;
  $nbrpri = 0 ;
  $nbrmod = 0 ;
  $nbrglo = 0 ;
  if (%MYOBJHASH) {
    my $prok ;
    foreach my $self (values %MYOBJHASH) {
      $prok = $self->{'theproto'} ;
      if ($prok && $prok eq '1') {
	push(@array, $self)  ;
	if ($self->get(-global)) {
	  $nbrglo++ ;
	} elsif ($self->get(-module)) {
	  $nbrmod++ ;
	} elsif ($self->get(-private)) {
	  $nbrpri++ ;
	}
      }
    }
  }
  return(@array) ;
}
###########################################################################
sub getFuncs {
  my(@array) = () ;
  $nbfpri = 0 ;
  $nbfmod = 0 ;
  $nbfglo = 0 ;
  if (%MYOBJHASH) {
    my $prok ;
    foreach my $self (values %MYOBJHASH) {
      $prok = $self->{'theproto'} ;
      if (!$prok || ($prok && $prok ne '1' && $prok ne '2')) {
	push(@array, $self)  ;
	if ($self->get(-global)) {
	  $nbfglo++ ;
	} elsif ($self->get(-module)) {
	  $nbfmod++ ;
	} elsif ($self->get(-private)) {
	  $nbfpri++ ;
	}
      }
    }
  }
  return(@array) ;
}
###########################################################################
sub getPros {
  my(@array) = () ;
  if (%MYOBJHASH) {
    my $prok ;
    foreach my $self (values %MYOBJHASH) {
      $prok = $self->{'theproto'} ;
      if ($prok && "$prok"eq '2') {
	push(@array, $self)  ;
      }
    }
  }
  return(@array) ;
}
###########################################################################
# sort by line number:
sub bynumber {
 $a->{"number"} <=> $b->{"number"} || lc($a) cmp lc($b) || $a cmp $b ;
}
###########################################################################
# Sets an interface to self:
sub addInter {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"FINTERFACES"}->{"$name"} = $value ;
  if ($debug) {
    if ($value) {
      print STDOUT (caller(0))[3] .": Adding an interface for $self, name=$name, object reference=$value from " . (caller(1))[3] . (caller(1))[2] . "\n" if $debug ;
    } else {
      print STDOUT (caller(0))[3] .": Adding an interface for $self, name=$name from " . (caller(1))[3] . (caller(1))[2] . "\n" if $debug ;
    }
  }
}
###########################################################################
# Adds a routine to a current module:
sub addRout {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"FPROCEDURES"}->{"$name"} = $value ;
  print STDOUT (caller(0))[3] .": Adding a routine for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
sub tellRouts {
  my $self = shift ;
  if (exists $self->{"FPROCEDURES"}) {
    return(%{$self->{"FPROCEDURES"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
# Adds a LaTeX input file
sub addInput {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"INPUTS"}->{"$name"} = $value ;
  $value->addReferer($self) ;
  print STDOUT (caller(0))[3] .": Adding an input for $self(" . $self->get(-name) . "), name=$name, object=$value\n" if $debug ;
}
###########################################################################
# Adds an include:
sub addInclude {
  my $self = shift ;
  my $name =shift ;
  my $value = shift ;
  $self->{"INCLUDES"}->{"$name"} = $value ;
  $value->addReferer($self) ;
  print STDOUT (caller(0))[3] .": Adding a type for $self, name=$name, object=$value\n" if $debug ;
}
###########################################################################
sub getInputs {
  my $self = shift ;
  if (exists $self->{"INPUTS"}) {
    return(%{$self->{"INPUTS"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
sub getIncludes {
  my $self = shift ;
  if (exists $self->{"INCLUDES"}) {
    return(%{$self->{"INCLUDES"}}) ;
  } else {
    return (()) ;
  }
}
###########################################################################
sub addPossible {
  my $self = shift ;
  my $ocall = shift ;
  push(@{$self->{"MYPOSSCALLS"}}, $ocall) ;
}
###########################################################################
sub getPossibles {
  my $self = shift ;
  if (exists $self->{"MYPOSSCALLS"}) {
    my(@possibles) = (@{$self->{"MYPOSSCALLS"}}) ;
    # Now return all of them but the ones already contained in listCalled:
    my(@new) = () ;
    foreach my $ocall (@possibles) {
      # Gets the name of the calling context:
      my $urout = $ocall->get(-father) ;
      if ($urout) {
	# Gets the calling context fulkly qualified name in FCALLED hash:
	my $name = $urout->get(-name) ;
	my $file = $urout->get(-file) ;
	my $number = $ocall->get(-number) ;
	if (exists $self->{"FCALLED"}->{"$name:$file:$number"}) {
	  # Already provided
	  next ;
	}
      }
      push(@new, $ocall) ;
    }
    if (@new) {
      return(@new) ;
    } else {
      return(()) ;
    }
  } else {
    return(()) ;
  }
}
###########################################################################
# Adds an "is called by"...
sub addCalled {
  my $self = shift ;
  my $ocall = shift ;
  my $father = $ocall->get(-father) ;
  my $name = $father->get(-name) ;
  my $number = $ocall->get(-number) ;
  my $file = $ocall->get(-file) ;
  $self->{"FCALLED"}->{"$name:$file:$number"} = $ocall ;
  if ($self->get(-module) && !$father->get(-module)) {
    # Current called routine is inside a module but the called location father is not in a module
    my $omod = &fthmod::getObject($self->get(-module)) ;
    print STDOUT (caller(0))[3] .": Module $omod\n" if $debug ;
    $omod->addCalled2($self, $ocall) ;
  }
  print STDOUT (caller(0))[3] .": Routine $self:" . $self->get(-name)." is called by routine=$name, file=$file, line=$number, object=$ocall\n" if $debug ;
}
###########################################################################
# Adds a call
sub addCall {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $nbcalls++ ;
  $self->{"FCALLS"}->{"$name:$nbcalls"} = $value ;
  if ($debug) {
    if ($value) {
      print STDOUT (caller(0))[3] .": Adding a call for " . $self->get(-name).", name=$name, line =$., object=$value\n" if $debug ;
    } else {
      print STDOUT (caller(0))[3] .": Adding a call for " . $self->get(-name).", name=$name, line =$.\n" if $debug ;
    }
  }
}
###########################################################################
# Adds an "is called by"...
sub listCalled {
  my $self = shift ;
  if (exists $self->{"FCALLED"}) {
    return(%{$self->{"FCALLED"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
sub getCalls {
  my $self = shift ;
  return(values %{$self->{"FCALLS"}}) ;
}
###########################################################################
# Gets the actual called routines: unique ID is built on the fly, hash reference as key was refused.
sub getCalled {
  my $self = shift ;
  my (%list) = () ;
  if (exists $self->{"FCALLS"}) {
    foreach my $ocall (sort bynumber values %{$self->{"FCALLS"}}) {
      my $orout = $ocall->getLink() ;
      print STDOUT "\t" . (caller(0))[3] . ": getCalled for " . $self->get(-name) . "\n" if $debug ;
      my $name ;
      my $file ;
      my $number ;
      my $cle ;
      if ($orout) {
	$name = $orout->get(-name) ;
	$file = $orout->get(-file) ;
	$number = $orout->get(-number) ;
	$cle = lc("$name:$file:$number") ;
	unless (exists $list{$cle}) {
	  (@{$list{$cle}}) = $orout ;
	}
	push(@{$list{$cle}},$ocall) ;
      } elsif (!$ocall->get(-fake)) {
	$name = $ocall->get(-name) ;
	$file = $ocall->get(-file) ;
	$number = $ocall->get(-number) ;
	$file = "aaaaa" ;
	$number = 0 ;
	$cle = lc("$name:$file:$number") ;
	unless (exists $list{$cle}) {
	  (@{$list{$cle}}) = "$name" ;
	}
	push(@{$list{$cle}},$ocall) ;
      } else {
	print STDOUT "\t" . (caller(0))[3] . ": getCalled(*) for " . $self->get(-name) . " is $ocall = ".$ocall->get(-name)."\n" if $debug ;
      }
    }
  }
  return(%list) ;
}
###########################################################################
# Adds an argument to a routine:
sub addArgument {
  my $self = shift ;
  my $name = lc(shift) ;
  $self->{"ARGHASH"}->{"$name"} = 1 ;
  push(@{$self->{"ARGORDER"}}, $name) ;
  print STDOUT (caller(1))[3] . ":" . (caller(1))[2] . ":" . (caller(0))[2] .": Adding an argument for " . $self->get(-name) . ", name=$name\n" if $debug ;
}
###########################################################################
# Adds an argument's type to a routine:
sub addType {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = shift ;
  $self->{"ARGTYPES"}->{"$name"} = $value ;
  print STDOUT "Adding an arg for $self, name=$name, value=$value\n" if $debug ;
}
###########################################################################
# Gets the arguments for a given stuff
sub getArguments {
  my $self = shift ;
  if (exists $self->{"ARGORDER"}) {
    return(@{$self->{"ARGORDER"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Gets the return type
sub getReturn {
  my $self = shift ;
  return(undef) unless (exists  $self->{"ARGORDER"}) ;
  my $i = $#{$self->{"ARGORDER"}} ;
  my $arg = $self->{"ARGORDER"}->[$i] ;
  print STDOUT "\t" . (caller(0))[3] . ": Routine " . $self->get(-name). " argument $i is $arg in list @{$self->{'ARGORDER'}}\n" if $debug;
  return($self->getType("$arg")) ;
}
###########################################################################
# Finds the actual arg type:
sub findType {
  my $self = shift ;
  my $arg = lc(shift) ;
  my $fast = shift ;
  my $valeur = undef ;
  if (exists $self->{"ALLDEFS"}->{"$arg"}) {
    $valeur = $self->{"ALLDEFS"}->{"$arg"} ;
    print STDOUT "\t\t" . (caller(0))[3] . ": argument $arg auto defined as $valeur in file " . $self->get(-file) . " routine " . $self->get(-name) . "\n" if $debug && $valeur ;
  } elsif (exists $self->{"module"}) {
    # any possible father argument ?
    my $module = $self->{"module"} ;
    my $omod = &fthmod::getObject($module) ;
    if ($omod) {
      $valeur = $omod->findType($arg) ;
    }
    print STDOUT "\t\t" . (caller(0))[3] . ": argument $arg auto defined as $valeur in file " . $omod->get(-file) . " father module " . $omod->get(-name) . "\n" if $debug && $valeur ;
  }
  unless ($valeur) {
    foreach my $module (keys %{$self->{"ALLUSES"}}) {
      my $omod = &fthmod::getObject($module) ;
      $valeur = $omod->findType($arg) if $omod ;
      print STDOUT "\t\t" . (caller(0))[3] . ": argument $arg auto defined as $valeur in file " . $omod->get(-file) . " son module " . $omod->get(-name) . "\n" if $debug && $valeur ;
      last if $valeur ;
    }
  }
  unless ($valeur) {
    if (exists $self->{"SELUSES"} && $self->{"SELUSES"}->{"$arg"}) {
      my $bidule = $self->{"SELUSES"}->{"$arg"} ;
      if ($bidule =~ /^(.*?)\:\:(.*)/) {
	my $sarg = $2 ;
	my $omod = &fthmod::getObject($1) ;
	if ($omod) {
	  $valeur = $omod->getType($sarg) ;
	}
      }
    }
  }
  unless ($valeur) {
    if (exists $self->{"private"}) {
      # Try to find an argument from father ?
      return($self->{"private"}->findType($arg)) ;
    }
  }
  return($valeur) ;
}
###########################################################################
# Gets the argument's type for a given stuff, try to guess...
sub sayType {
  my $self = shift ;
  my $name = lc(shift) ;
  if (exists $self->{"ALLDEFS"}->{"$name"}) {
    return(exists $self->{"ALLDEFS"}->{"$name"}) ;
  } elsif ($name =~ /^(\.)(.*?)\1$/) {
    return('logical') ;
  } elsif ($name =~ /\./) {
    return('real') ;
  } elsif ($name =~ /^\d+$/) {
    return('integer') ;
  } else {
    return("") ;
  }
}
###########################################################################
# Gets the argument's type for a given stuff, try to guess...
sub sayCalling {
  my $self = shift ;
  my $name = lc(shift) ;
  if (exists $main::names{$name}) {
    my $mon = $main::names{$name} ;
    if (exists $main::fthrouts{$mon} || exists $main::fthobjs{$mon}) {
      if (exists $self->{"ALLDEFS"}->{"$name"}) {
	my $t = $self->{"ALLDEFS"}->{"$name"} ;
	if ($t =~ /dimension/i) {
	  # Limited security provided at this level:
	  return(0) ;
	}
      }
      return(1) ;
    }
  }
  return(0) ;
}
###########################################################################
# Gets the argument's type for a given stuff, try to guess...
sub setType {
  my $self = shift ;
  my $name = lc(shift) ;
  my $value = lc(shift) ;
  $self->{"ALLDEFS"}->{"$name"} = $value ;
}
###########################################################################
# Gets the argument's type for a given stuff, try to guess...
sub getType {
  my $self = shift ;
  my $name = lc(shift) ;
  return('void') unless defined($name) ;
  unless ($name =~ /^([\"\'])(.*?)\1$/) {
    $name = $1 if ($name =~ /^[-+](.*)/) ;
    # A trick to simplify the arguments analysis
    $name = $1 if ($name =~ /^(.*?)[\-\+\*\/]/) ;
  }
  print STDOUT "\t" . (caller(0))[3] . ": Fetching type $name for " . $self->get(-name) . "\n" if $debug ;
  if (exists $self->{"ALLDEFS"}->{"$name"}) {
    print STDOUT "\t\t".$self->{"ALLDEFS"}->{"$name"}."\n" if $debug ;
    return($self->{"ALLDEFS"}->{"$name"}) ;
  } elsif (exists $self->{"private"}) {
    # Try to find an argument from father ?
    print STDOUT "\t\tPrivate\n" if $debug ;
    return($self->{"private"}->getType($name)) ;
  } elsif ($name =~ /^([\"\'])(.*?)\1$/) {
    print STDOUT "\t\tCharacter\n" if $debug ;
    return('character*'.length($2)) ;
  } elsif ($name =~ /^(\.)(.*?)\1$/) {
    print STDOUT "\t\tLogical\n" if $debug ;
    return('logical') ;
  } elsif ($name =~ /\./) {
    print STDOUT "\t\tReal\n" if $debug ;
    return('real') ;
  } elsif ($name =~ /^\d+$/) {
    print STDOUT "\t\tInteger\n" if $debug ;
    return('integer') ;
  } elsif (!($name =~ /\%/)) {
    my $type = &fthrout::isIntrinsec($name) ;
    return($type) if $type;
  }
  return('void') ;
}
###########################################################################
# Gets the implicit default type for current if any:
sub defType {
  my $self = shift ;
  my $name = lc(shift) ;
  unless ($name =~ /\%/) {
    # Attempt to manage the implicit feature below:
    # What a dangereous stuff !
    if (exists $self->{"IMPLICITDEFS"}) {
      # String start:
      if ($name =~ /^(\w)/i) {
	my $rhs = $1 ;
	if (exists $self->{"IMPLICITDEFS"}->{"$rhs"}) {
	  print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Matching $name with $rhs to type " . $self->{"IMPLICITDEFS"}->{"$rhs"} . "\n" if $debug ;
	  return($self->{"IMPLICITDEFS"}->{"$rhs"});
	}
      }
    }
  }
  return('void') ;
}
###########################################################################
# Gets the proto for a routine:
sub getProto {
  my $self = shift ;
  my (@args) = $self->getArguments() ;
  my (@proto) = () ;
  foreach my $arg (@args) {
    my $val = $self->getType("$arg") ;
    if ($val eq 'void' && exists $self->{"IMPLICITDEFS"}) {
      $val = $self->defType("$arg") ;
    }
    print STDOUT  "\t\t\t" . (caller(0))[3] . ": In routine " . $self->get(-name) . ", args of $arg are " . $val . "\n" if $debug ;
    push(@proto, "$arg=" . $val) ;
  }
  return(@proto) ;
}
###########################################################################
sub getLink {
  my $self = shift ;
  if ($self->{"thelink"}) {
    # Safe stuff: prototype is equal
    return($self->{"thelink"}) ;
  } elsif ($self->{"ok"}) {
    # One or more answers
    if ($#{$self->{"OKS"}} == 0) {
      # Only one answer, return it even though it may be a far one...
      return($self->{"alink"}) ;
    } else {
      # Multiple answers at same level:
      if ($debug) {
	print STDOUT (caller(0))[3] .": Called routine " . $self->get(-name) . " file " . $self->get(-file) . " line " . $self->get(-number) . " may refer to any of" ;
	foreach (@{$self->{'OKS'}}) {
	  print STDOUT " " . $_->get(-name) ;
	}
	print STDOUT "\n" ;
      }
      return(undef) ;
    }
  } else {
    return(undef) ;
  }
}
###########################################################################
sub setLink {
  my $self = shift ;
  my $ko = shift ;
  my $orout = shift ;
  print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Setting link for $self with ko=$ko to $orout " . $orout->get(-name) . "\n" if $debug ;
  if (exists $self->{"ok"}) {
    my $ok = $self->{"ok"} ;
    # Mark exists
    if ($ko < $ok) {
      # Better mark new visitor has lower number, it means is more suitable.
      $self->{"ok"} = $ko ;
      $self->{"alink"} = $orout ;
      delete $self->{"OKS"} if $self->{"OKS"} ;
      (@{$self->{"OKS"}}) = ($orout) ;
    } elsif ($ok == $ko) {
      my $addit = 1 ;
      foreach my $bidule (@{$self->{"OKS"}}) {
	if ($orout == $bidule) {
	  $addit = 0 ;
	}
      }
      # Same mark, manage multiple:
      push(@{$self->{"OKS"}}, $orout) if $addit ;
    }
  } else {
    # No mark, initialize:
    $self->{"ok"} = $ko ;
    $self->{"alink"} = $orout ;
    (@{$self->{"OKS"}}) = ($orout) ;
  }
}
###########################################################################
# On the fly renaming is to be performed here : check syntax
# is correct: lhs is in module, rhs is the local name:
sub readUses {
  my $self = shift ;
  my $file = shift ;
  my $lmused = shift ;
  my $fmused = lc($lmused) ;
  my $chtml = &chtml::getCurrent() ;
  $_ = shift ;
  print STDOUT "\t" . (caller(0))[3] . ": Reading $_ for used module $fmused\n" if $debug ;
  my $cont = 1 ;
  if ((ref $file) && (fileno $file)) {
    while($cont) {
      my $cc = "" ;
      #print STDOUT "line=$. => $_\n" ;
      if (/^(\#.*?)/) {
	# Some preprocessing stuff: passed to later analysis
      } elsif (/^(.*?)\!(.*)/) {
	$_ = $1 ;
	$cc = $2 ;
      }
      if ($cont && /^(\s*\&\s*)(.*)/ || $cont >= 2 && /(\s*)(.*)/) {
	print STDOUT "\t\tRhs($.) for fmused($cont)=$2\n" if $debug ;
	my $rhs = $2 ;
	$chtml->addDef($1) if ($cont >= 2) ;
	$rhs =~ s/\s+//g ;
	if ($rhs =~ /(.*)\&/) {
	  $rhs = $1 ;
	  $cont = 2 ;
	} elsif ($rhs) {
	  $cont = 0 ;
	} else {
	  # Only spacing
	  $cont = 3 ;
	}
	my (@array) = split(/\,/, $rhs) ;
	my $counter = 0 ;
	foreach my $a (@array) {
	  $chtml->addDef(", ") if $counter ;
	  if ($a =~ /\&\s*$/) {
	    $cont = 2 ;
	    $chtml->addDef($a) if $counter ;
	  } elsif ($a =~ /\b(\w+)\b/) {
	    my $l = $a ;
	    my $r = $a ;
	    if ($a =~ /(.*)=>(.*)/) {
	      $l = $1 ;
	      $r = $2 ;
	    }
	    # On the right the name used for the call, on the left the
	    # actual routine name in it's module.
	    print STDOUT "\t\t" . (caller(0))[3] . ": SELUSES $r points to $fmused :: $l\n" if $debug ;
	    # Manage multiple override...
	    my $sr = lc($r) ;
	    my $sl = lc($l) ;
	    if ($self->{"SELUSES"}->{"$sr"}) {
	      $self->{"SELUSED"}->{"$sr"}->{"$fmused"} =  "$sl" ;
	    } else {
	      # Syntax: module_name :: routine_in_module
	      $self->{"SELUSES"}->{"$sr"} =  $fmused. '::' . $sl ;
	    }
	    if ($l ne $r) {
	      $chtml->addRaw("<A Name=R_$l>$l</A> => ") ;
	    }
	    my $refe = new oref(-name => "$lmused" . '::' . "$r", -type => "uses") ;
	    $chtml->addRaw($refe) ;
	    $counter++ ;
	  }
	}
	$chtml->addDef(" &") if ($cont == 2) ;
      } else {
	print STDOUT (caller(0))[3] . ": KO $_\n" if $debug ;
      }
      if ($cc) {
	$chtml->addRaw('<COMMENT>') ;
	$chtml->addLaTeX($cc) ;
	$chtml->addRaw("</COMMENT>") ;
      }
      return unless $cont ;
      if (defined($_ = <$file>)) {
	chomp ;
     } else {
	print STDOUT (caller(0))[3] . ": error5 line=$. routine='" . $self->get(-name) . "' file='" . $self->get(-file) . "'\n" if $debug || $main::opt_v ;
	$cont = 0 ;
      }
    }
  }
}
###########################################################################
# Returns the number of compulsary argument for a proto:
sub nargProto {
  my($oprot) = (@_) ;
  my (@oproto) = (@$oprot) ;
  my $number = ( $#oproto + 1 ) ;
  for (my $i=0; $i<=$#oproto; $i++) {
    my $o = $oproto[$i] ;
    print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Testing o=$o\n" if $debug ;
    if ($o =~ /optional/i) {
      $number = $i ;
      last ;
    }
  }
  return($number) ;
}
###########################################################################
sub cmpProto {
  my($mprot, $cprot) = (@_) ;
  my (@mproto) = (@$mprot) ;
  my (@cproto) = (@$cprot) ;
  my $ko = 0 ;
  my $nbargs = ( $#cproto + 1 ) ;
  for (my $i=0; $i<=$#cproto; $i++) {
    my $c = $cproto[$i] ; my $m = $mproto[$i] ;
    # NAJ: make so that dimension() makes no difference...
    $c = $1 if ($c =~ /=(.*)/) ;
    $m = $1 if ($m =~ /=(.*)/) ;
    print STDOUT "\t\t" . (caller(1))[3] . ": Comparing arg $i : $c versus $m\n" if $debug ;
    $c =~ s/,*parameter//g ;
    $m =~ s/,*parameter//g ;
    $c =~ s/,*save//g ;
    $m =~ s/,*save//g ;
    $c =~ s/,*optional//g ;
    $m =~ s/,*optional//g ;
    if ($c ne $m) {
      if ($c =~ /(pointer|target)/) {
	$c =~ s/,*pointer//g ;
	$m =~ s/,*pointer//g ;
	$c =~ s/,*target//g ;
	$m =~ s/,*target//g ;
	if ($c ne $m) {
	  $ko += 2 * 3**( $nbargs - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] differs from sub $mproto[$i], mark1=$ko\n" if $debug ;
	} else {
	  # Provide a slightly better mark
	  $ko += 3**( $nbargs - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] pointers differs from sub $mproto[$i], mark2=$ko\n" if $debug ;
	}
      } elsif ($c =~ /character/i && $m =~ /character/i) {
	my $c2 = $c ;
	my $m2 = $m ;
	if ($c =~ /(.*)character\s*\*\s*[\d]+\s*(.*)/) {
	  $c2 = $1.$2 ;
	}
	if ($m =~ /(.*)character\s*\*\s*[\d]+\s*(.*)/) {
	  $m2 = $1.$2 ;
	}
	if ($m =~ /character\(len\=\*\)/) {
	  # Accept the stuff
	} elsif ($m =~ /character\*\(\*\)/) {
	  # Accept the stuff
	} elsif ($c2 ne $m2) {
	  $ko += 2 * 3**( $nbargs - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] differs from sub $mproto[$i], mark1=$ko\n" if $debug ;
	} else {
	  # Provide a slightly better mark
	  $ko += 3**( $nbargs - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] characters differs from sub $mproto[$i], mark2=$ko\n" if $debug ;
	}
      } elsif ($m =~ /optional/) {
	$m =~ s/,*optional//g ;
	if ($c ne $m) {
	  $ko += 2 * 3**( $nbargs - $i ) ;
	}
      } else {
	# Highest the error number, lowest the chance it fits
	$ko += 2 * 3**( $nbargs - $i ) ;
	print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] differs from sub $mproto[$i], mark3=$ko\n" if $debug ;
      }
    }
  }
  unless ($ko) {
    print STDOUT "\t\t" . (caller(1))[3] . ": Protos are equal: @mproto\n" if $debug ;
  }
  return ($ko) ;
}
###########################################################################
# A relaxed proto compare:
sub relProto {
  my($mprot, $cprot) = (@_) ;
  my (@mproto) = (@$mprot) ;
  my (@cproto) = (@$cprot) ;
  my $ko = 0 ;
  my $nbargs = ( $#cproto + 1 ) ;
  for (my $i=0; $i<=$#cproto; $i++) {
    my $c = $cproto[$i] ; my $m = $mproto[$i] ;
    unless ($m) {
      $ko += 3**( $nbargs - $i ) ;
      last ;
    }
    # NAJ: make so that dimension() makes no difference...
    $c = $1 if ($c =~ /=(.*)/) ;
    $m = $1 if ($m =~ /=(.*)/) ;
    print STDOUT "\t\t" . (caller(1))[3] . ": Comparing arg $i : '$c' versus '$m'\n" if $debug ;
    $c =~ s/,*parameter//g ;
    $m =~ s/,*parameter//g ;
    $c =~ s/,*save//g ;
    $m =~ s/,*save//g ;
    $c =~ s/,*optional//g ;
    $m =~ s/,*optional//g ;
    if ($c ne $m) {
      $c = $1 if ($c =~ /^type\((\w+)\)/) ;
      $m = $1 if ($m =~ /^type\((\w+)\)/) ;
      $c = $1 if ($c =~ /^class\((\w+)\)/) ;
      $m = $1 if ($m =~ /^class\((\w+)\)/) ;
      $c = $1 if ($c =~ /^(integer|real|character|complex|logical)\(/) ;
      $m = $1 if ($m =~ /^(integer|real|character|complex|logical)\(/) ;
    }
    print STDOUT "\t\t\t" . (caller(1))[3] . ": Re-Comparing arg $i : '$c' versus '$m'\n" if $debug ;
    if ($c ne $m) {
      if ($c =~ /(pointer\target)/) {
	$c =~ s/,*pointer//g ;
	$m =~ s/,*pointer//g ;
	$c =~ s/,*target//g ;
	$m =~ s/,*target//g ;
	if ($c ne $m) {
	  $ko += 2 * 3**( $nbargs - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] differs from sub $mproto[$i], mark1=$ko\n" if $debug ;
	} else {
	  # Provide a slightly better mark
	  $ko += 3**( $nbargs - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] pointers differs from sub $mproto[$i], mark2=$ko\n" if $debug ;
	}
      } elsif ($c =~ /character/i && $m =~ /character/i) {
	my $c2 = $c ;
	my $m2 = $m ;
	if ($c =~ /(.*)character\s*\*\s*[\d]+\s*(.*)/) {
	  $c2 = $1.$2 ;
	}
	if ($m =~ /(.*)character\s*\*\s*[\d]+\s*(.*)/) {
	  $m2 = $1.$2 ;
	}
	if ($m =~ /character\(len\=\*\)/) {
	  # Accept the stuff
	} elsif ($m =~ /character\*\(\*\)/) {
	  # Accept the stuff
	} elsif ($c2 ne $m2) {
	  $ko += 2 * 3**( $nbargs - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] differs from sub $mproto[$i], mark1=$ko\n" if $debug ;
	} else {
	  # Provide a slightly better mark
	  $ko += 3**( $nbargs - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] characters differs from sub $mproto[$i], mark2=$ko\n" if $debug ;
	}
      } elsif ($m =~ /optional/) {
	$m =~ s/,*optional//g ;
	if ($c ne $m) {
	  $ko += 2 * 3**( $nbargs - $i ) ;
	}
      } else {
	# Highest the error number, lowest the chance it fits
	$ko += 2 * 3**( $nbargs - $i ) ;
	print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] differs from sub $mproto[$i], mark3=$ko\n" if $debug ;
      }
    }
    # print STDOUT "ko=$ko\n" ;
  }
  # NAJ518:
  unless ($ko) {
    print STDOUT "\t\t" . (caller(1))[3] . ": Protos are equal: @mproto\n" if $debug ;
  }
  return ($ko) ;
}
###########################################################################
sub cmpOldProto {
  my($mprot, $cprot) = (@_) ;
  my (@mproto) = (@$mprot) ;
  my (@cproto) = (@$cprot) ;
  my $ko = 0 ;
  for (my $i=0; $i<=$#mproto; $i++) {
    my $c = $cproto[$i] ; my $m = $mproto[$i] ;
    # NAJ: make so that dimension() makes no difference...
    $c = $1 if ($c =~ /=(.*)/) ;
    $m = $1 if ($m =~ /=(.*)/) ;
    if ($c ne $m) {
      if ($c =~ /pointer/) {
	$c =~ s/,*pointer//g ;
	$m =~ s/,*pointer//g ;
	if ($c ne $m) {
	  $ko += ( $#mproto + 1 - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] differs from sub $mproto[$i], mark1=$ko\n" if $debug ;
	  last ;
	} else {
	  # Provide a slightly better mark
	  $ko = ( $#mproto + 0.5 - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] pointers differs from sub $mproto[$i], mark2=$ko\n" if $debug ;
	}
      } elsif ($c =~ /character/i && $m =~ /character/i) {
	my $c2 = $c ;
	my $m2 = $m ;
	if ($c =~ /(.*)character\s*\*\s*[\d]+\s*(.*)/) {
	  $c2 = $1.$2 ;
	}
	if ($m =~ /(.*)character\s*\*\s*[\d]+\s*(.*)/) {
	  $m2 = $1.$2 ;
	}
	if ($m =~ /character\(len\=\*\)/) {
	  # Accept the stuff
	} elsif ($c2 ne $m2) {
	  $ko = ( $#mproto + 1 - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] differs from sub $mproto[$i], mark1=$ko\n" if $debug ;
	  last ;
	} else {
	  # Provide a slightly better mark
	  $ko = ( $#mproto + 0.5 - $i ) ;
	  print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] characters differs from sub $mproto[$i], mark2=$ko\n" if $debug ;
	}
      } else {
	# Highest the error number, lowest the chance it fits
	$ko = ( $#mproto + 1 - $i ) ;
	print STDOUT "\t\t" . (caller(1))[3] . ": Argument $i / $#mproto call proto $cproto[$i] differs from sub $mproto[$i], mark3=$ko\n" if $debug ;
	last ;
      }
    }
  }
  unless ($ko) {
    print STDOUT "\t\t" . (caller(1))[3] . ": Protos are equal: @mproto\n" if $debug ;
  }
  return ($ko) ;
}
###########################################################################
# Sets the dependencies for a routine refusing the cyclic dependencies
sub recDepends {
  my $self = shift ;
  my $omod = shift ;
  my $rmod ;
  my(@array) = $omod->getUses() ;
  foreach my $mname (@array) {
    next unless $mname ;
    next if ($self->{"RECUSES"} && exists $self->{"RECUSES"}->{"$mname"}) ;
    $rmod = &fthmod::getObject($mname) ;
    next unless $rmod ;
    $self->{"RECUSES"}->{"$mname"} = $rmod ;
    $self->recDepends($rmod) ;
  }
}
###########################################################################
# Sets the dependencies for a module refusing the cyclic dependencies
sub getDepends {
  my $self = shift ;
  if (exists $self->{"RECUSES"}) {
    return(%{$self->{"RECUSES"}}) ;
  } else {
    return(()) ;
  }
}
###########################################################################
# Sets the dependencies for a module refusing the cyclic dependencies
sub setCompiles {
  my $self = shift ;
  my (%hash) = () ;
  my $file ;
  if (exists $self->{"ALLUSES"}) {
    foreach my $fmused (keys %{$self->{"ALLUSES"}}) {
      print STDOUT (caller(0))[3] . ": Addding module dependence $fmused to ".$self->get(-name)."\n" if $debug ;
      my $omod = &fthmod::getObject($fmused) ;
      if ($omod) {
	$file = $omod->get(-file) ;
	$hash{$file} = 1 ;
      }
    }
  }
  if (exists $self->{"SELUSES"}) {
    while(my($sr, $value)=each %{$self->{"SELUSES"}}) {
      my $fmused = $self->{"SELUSES"}->{"$sr"} ;
      print STDOUT (caller(0))[3] . ": Addding selected module dependence $fmused to ".$self->get(-name)."\n" if $debug ;
      if ($fmused =~ /^(.*?)\:/) {
	$fmused = $1 ;
	my $omod = &fthmod::getObject($fmused) ;
	print STDOUT "\t" . (caller(0))[3] . " fmused=$fmused of $omod\n" if $debug ;
	if ($omod) {
	  $file = $omod->get(-file) ;
	  $hash{$file} = 1 ;
	}
      }
      if (exists $self->{"SELUSED"} && exists $self->{"SELUSED"}->{"$sr"}) {
	foreach $fmused (keys %{$self->{"SELUSED"}->{"$sr"}}) {
	  my $omod = &fthmod::getObject($fmused) ;
	  if ($omod) {
	    $file = $omod->get(-file) ;
	    $hash{$file} = 1 ;
	  }
	}
      }
    }
  }
  # Now also sets the calls tree that is not that obvious:
  foreach my $ocall (values %{$self->{"FCALLS"}}) {
    my $orout = $ocall->getLink() ;
    if ($orout) {
      $file = $orout->get(-file) ;
      $hash{$file} = 1 ;
    }
  }
  return(keys %hash) ;
}
###########################################################################
# Sets the dependencies for a routine:
sub setDepends {
  my $self = shift ;
  my $omod ;
  if (exists $self->{"ALLUSES"}) {
    foreach my $mname (keys %{$self->{"ALLUSES"}}) {
      $omod = &fthmod::getObject($mname) ;
      next unless $omod ;
      $self->{"RECUSES"}->{"$mname"} = $omod ;
      $self->recDepends($omod) ;
    }
  }
  # Sets the interfaces to the current routines:
  if (exists $self->{"FINTERFACES"}) {
    foreach my $iname (keys %{$self->{"FINTERFACES"}}) {
      my $ointer = $self->{"FINTERFACES"}->{"$iname"} ;
      next unless $ointer ;
      # An interface may point to one or more procedures:
      my(@procs) = $ointer->getProcedures() ;
      foreach my $pname (@procs) {
	my $lname = lc($pname) ;
	# If the module is correct, there must be a corresponding procedure
	if (exists $self->{"FPROCEDURES"}->{"$lname"}) {
	  my $orout = $self->{"FPROCEDURES"}->{"$lname"} ;
	  $orout->addInter($ointer);
	} else {
	  # Look for a global procedure to it:
	  my $orout = $ointer->getProcedure($iname) ;
	  if ($orout) {
	    my $crout = &fthrout::isGlobal($orout) ;
	    if ($crout) {
	      $orout->set(-thelink => $crout) ;
	    }
	  } else {
	    # Time to look for a global procedure
	    print STDOUT (caller(0))[3] .": No implementation for interface $iname to procedure $pname in module " . $self->get(-name) . "\n" if $debug ;
	  }
	}
      }
    }
  }
}
###########################################################################
sub isRoutine {
  my $self = shift ;
  my $ocall = shift ;
  my $orout ;
  $orout = $self->isPrivate($ocall) ;
  print STDOUT  "\t\t\t" . (caller(0))[3] . ": Q1, Is a module procedure $ocall(".$ocall->get(-name).':'.$ocall->get(-number).") for $self(".$self->get(-name).")?\n" if $debug ;
  if ($debug && $orout) {
    print STDOUT "\t\t\t" . (caller(0))[3] . ": Yes OK1, connect to " . $orout->get(-name). ':' . $orout->get(-number) ."\n" ;
  }
  return($orout) if $orout ;
  $orout = $self->isModulprocedur($ocall) ;
  print STDOUT  "\t\t\t" . (caller(0))[3] . ": Q2, Is a module procedure $ocall(".$ocall->get(-name).':'.$ocall->get(-number).") for $self(".$self->get(-name).")?\n" if $debug ;
  if ($debug && $orout) {
    print STDOUT "\t\t\t" . (caller(0))[3] . ": Yes OK2, connect to " . $orout->get(-name). ':' . $orout->get(-number) ."\n" ;
  }
  return($orout) if $orout ;
  return(undef) ;
}
###########################################################################
sub isModulprocedur {
  my $self = shift ;
  my $ocall = shift ;
  my $orout = undef ;
  my $cname = lc($ocall->get(-name)) ;
  my $mname ;
  my $omod ;
  my $ldebug = $debug ;
  $ocall->datadump("\t") if $debug ;
  # Direct call ?
  #     $orout = &getObject($ocall->get(-name)) ;
  # Loops on modules that are directly called:
  foreach $mname (keys %{$self->{"ALLUSES"}}) {
    $omod = &fthmod::getObject($mname) ;
    next unless $omod ;
    # $omod->datadump("\t\t") if $debug ;
    print STDOUT "\t" . (caller(0))[3] . ": Module $mname used by " . $self->get(-name) . ":\n" if $ldebug;
    print STDOUT "\t\t" . (caller(0))[3] . ": Check whether module $mname contains $cname:\n" if $ldebug ;
    # Ask the module: do you have such an object ?
    $orout = $omod->isRoutine($ocall) ;
    print STDOUT "\t\t" . (caller(0))[3] . ": Checked whether module $mname did contain $cname.\n" if $ldebug;
    if ($orout) {
      return($orout) ;
    }
  }
  if ($debug) {
    my(@a) = (keys %{$self->{"SELUSES"}}) ;
    print STDOUT "\t\t" . (caller(0))[3] . ": Selected uses of $cname is @a.\n" ;
  }
  
  if ($self->{"SELUSES"}->{"$cname"}) {
    # A specific import is to be found. Testing first encountered:
    if ($self->{"SELUSES"}->{"$cname"} =~ /(.*)::(.*)/) {
      $mname = lc($1) ;
      my $alias = lc($2) ;
      $omod = &fthmod::getObject($mname) ;
      if ($omod) {
	print STDOUT "\t\t" . (caller(0))[3] . ": Check whether selected module $mname contains an interface to $cname:\n" if $ldebug;
	$orout = $omod->isRoutine($ocall, $alias) ;
	if ($orout && $debug) {
	  print STDOUT "\t\t" . (caller(0))[3] . ": Yes OK4, connect to " . $orout->get(-name). ':' . $orout->get(-number) ."for caller " . $ocall->get(-name). ':' . $ocall->get(-number) . "\n" ;
	}
      }
      return($orout) if $orout ;
    }
    # A specific import is to be found. Managing name conflicts:
    if (exists $self->{"SELUSED"}->{"$cname"}) {
      my (%hash) = (%{$self->{"SELUSED"}->{"$cname"}}) ;
      if (%hash) {
	foreach $mname (keys %hash) {
	  $omod = &fthmod::getObject($mname) ;
	  next unless $omod ;
	  my $alias = $self->{"SELUSED"}->{"$cname"}->{"$mname"} ;
	  print STDOUT "\t\t" . (caller(0))[3] . ": Check whether aliased module $mname contains $cname:\n" if $ldebug;
	  $orout = $omod->isRoutine($ocall, $alias) ;
	  return($orout) if $orout ;
	}
      }
      print STDOUT "$cname imported from " . $self->{"SELUSES"}->{"$cname"} . "\n" if $ldebug;
    }
  }
  ##############################################
  # Recursive modules inquired once:
  if (exists $self->{"ALLUSES"}) {
    foreach $mname (keys %{$self->{"ALLUSES"}}) {
      $omod = $self->{"RECUSES"}->{"$mname"} ;
      next unless $omod ;
      print STDOUT "\t" . (caller(0))[3] . ": Module $mname used by " . $self->get(-name) . ":\n" if $ldebug;
      print STDOUT "\t\t" . (caller(0))[3] . ": Check whether module $mname contains $cname:\n" if $ldebug;
      # Ask the module: do you have such an object ?
      $orout = $omod->isSpecific($ocall) ;
      print STDOUT "\t\t" . (caller(0))[3] . ": Checked whether module $mname did contain $cname.\n" if $ldebug;
      if ($orout) {
	return($orout) ;
      }
    }
  }
  if (exists $self->{"RECUSES"}) {
    foreach $mname (keys %{$self->{"RECUSES"}}) {
      next if (exists $self->{"ALLUSES"}->{"$mname"}) ;
      $omod = $self->{"RECUSES"}->{"$mname"} ;
      next unless $omod ;
      print STDOUT "\t" . (caller(0))[3] . ": Module $mname used by " . $self->get(-name) . ":\n" if $ldebug;
      print STDOUT "\t\t" . (caller(0))[3] . ": Check whether module $mname contains $cname:\n" if $ldebug;
      # Ask the module: do you have such an object ?
      $orout = $omod->isSpecific($ocall) ;
      print STDOUT "\t\t" . (caller(0))[3] . ": Checked whether module $mname did contain $cname.\n" if $ldebug;
      if ($orout) {
	return($orout) ;
      }
    }
  }
  return(undef) ;
}
###########################################################################
# Checks for private routines:
sub isPrivate {
  my $self = shift ;
  my $ocall = shift ;
  my $orout = undef ;
  if ($self->{"FPROCEDURES"}) {
    my $cname = lc($ocall->get(-name)) ;
    if (exists $self->{"FPROCEDURES"}->{"$cname"}) {
      my (@cproto) = $ocall->getProto() ;
      $orout = $self->{"FPROCEDURES"}->{"$cname"} ;
      my (@mproto) = $orout->getProto() ;
      print STDOUT "\t\t" . (caller(0))[3] . ": On call $cname, testing private routine of " . $self->get(-name) . ", #args=$#mproto,$#cproto\n" if $debug ;
      #if ($#mproto == $#cproto) {
      if ($#mproto >= $#cproto && &fthrout::nargProto(\@mproto) <= &fthrout::nargProto(\@cproto)) {
	my $ko = &cmpProto(\@mproto, \@cproto) ;
	if ($ko) {
	  $ocall->setLink($ko, $orout) ;
	} else {
	  return($orout) ;
	}
      }
    }
  } else {
    print STDOUT "\t\t" . (caller(0))[3] . ": No private for " . $self->get(-name) . "\n" if $debug ;
  }
  return(undef) ;
}
###########################################################################
sub isGlobal {
  my $ocall = shift ;
  my $cname = lc($ocall->get(-name)) ;
  my $orout = &fthrout::getObject("$cname") ;
  my $ldebug = $debug ;
  # Look for a global:
  if ($orout) {
    my $ko = 0 ;
    my (@cproto) = $ocall->getProto() ;
    my (@mproto) = $orout->getProto() ;
    #if ($#mproto == $#cproto) {
    if ($#mproto >= $#cproto && &fthrout::nargProto(\@mproto) <= &fthrout::nargProto(\@cproto)) {
      $ko = &cmpProto(\@mproto, \@cproto) ;
      if ($ko) {
	print STDOUT "\t\t" . (caller(0))[3] . ": Global procedure $cname implements call with improper proto (may not be known) '@mproto', called proto is '@cproto'\n" if $ldebug;
	$ocall->setLink($ko, $orout) ;
      } else {
	print STDOUT "\t\t" . (caller(0))[3] . ": Global procedure $cname implements call with correct proto @cproto\n" if $ldebug;
	return($orout) ;
      }
    } else {
      $ocall->setLink( ( $#cproto + 1 ), $orout) ;
      print STDOUT "\t\t" . (caller(0))[3] . ": Call to $cname proto '@mproto' encounters very different '@mproto'\n" if $ldebug;
    }
  }
  return(undef) ;
}
###########################################################################
# Look for a direct access in global list mode:
sub isDirect {
  my $self = shift ;
  my $ocall = shift ;
  my $orout = undef ;
  my $cname = lc($ocall->get(-name)) ;
  my $ldebug = $debug ;
  my $ko ;
  print STDOUT "\t\t" . (caller(0))[3] . ": Looking for direct access to $cname from " . $self->get(-name) . "\n" if $ldebug ;
  my $fobj = &fthobj::getObject("$cname");
  # Look for a direct access to the procedure name:
  if ($fobj) {
    my (@bidules) = $fobj->getStuffs() ;
    print STDOUT "\t\t\t" . (caller(0))[3] . ": Found bidules = @bidules\n" if $ldebug ;
    if (@bidules) {
      my (@cproto) = $ocall->getProto() ;
      # Direct call should not be favored that way !
      #if ($bidules[0] && ref($bidules[0]) eq 'fthrout') {
	# $orout = $bidules[0] ;
      #}
      print STDOUT "\t\t\t" . (caller(0))[3] . ": Testing interface $fobj on (@bidules)\n" if $ldebug ;
      foreach my $truc (@bidules) {
	next unless $truc ;
	if ($truc && ref($truc) eq 'fthrout') {
	  # This could be it:
	  my (@mproto) = $truc->getProto() ;
	  print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Global procedure $cname tests proto of " . $truc->get(-name) . " ie=@mproto\n" if $ldebug;
	  #if ($#mproto == $#cproto) {
	  if ($#mproto >= $#cproto && &fthrout::nargProto(\@mproto) <= &fthrout::nargProto(\@cproto)) {
	    $orout = $truc ;
	    $ko = &cmpProto(\@mproto, \@cproto) ;
	    if ($ko) {
	      print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Global procedure $cname implements call with incorrect proto @mproto\n" if $ldebug;
	      $ocall->setLink($ko, $orout) ;
	    } else {
	      print STDOUT "\t\t\t\t" . (caller(0))[3] . ": Global procedure $cname implements call with correct proto @mproto\n" if $ldebug;
	      return($truc) ;
	    }
	  }
	}
      }
    }
  }
  return($orout) ;
}
###########################################################################
sub postCalls {
  my $self = shift ;
  my $ocall = shift ;
  my $number = 0 ;
  my (%hash) = $ocall->getDepends() ;
  foreach my $arg ($ocall->getArguments()) {
    my $type = $ocall->getType("$arg") ;
    if (!$type || $type eq 'void') {
      my $scall = $hash{"$number"} ;
      if ($scall) {
	# Gets the return type:
	my $orout = $scall->getLink() ;
	if ($orout) {
	  # Sets the type for the argument using the type from the function's prototype:
	  print STDOUT "\t" . (caller(0))[3] .": Final set for called " . $ocall->get(-name) . " line " . $ocall->get(-number) . " file " .$ocall->get(-file) . " argument=$arg points to file " .$orout->get(-file). " line " . $orout->get(-number) . " proto='".$orout->get(-theproto)."' to return type=" . $orout->getType($orout->get(-theproto)) . "\n" if $debug ;
	  $ocall->setType($arg => $orout->getType($orout->get(-theproto))) ;
	}
      }
    }
    $number++ ;
  }
}
###########################################################################
sub preCalls2 {
  my $self = shift ;
  my $ocall = shift ;
  my $arg = shift ;
  my $number = shift ;
  if ($arg =~ /^(\w+)\%(\w+)$/) {
    # A direct access to an unknown class:
    print STDOUT (caller(0))[3] . ": Direct private access to $arg, trying to set actual type\n" if $debug ;
    my $point = $1 ; my $var = $2 ;
    my $ko = 1 ;
    my $answer = $self->findType("$point") ;
    if ($answer) {
      print STDOUT (caller(0))[3] . ": Provided by $point value for $arg is $answer\n" if $debug ;
      if ($answer =~ /type\s*\(\s*(.*?)\s*\)/i) {
	$point = $1 ;
	if (my $ft = &fthtype::getObject("$point")) {
	  # Gets the type for $var in object $ft ;
	  print STDOUT "\t" . (caller(0))[3] . ": OK Type for $arg is $point%$var :: " . $ft->getType("$var")."\n" if $debug ;
	  $ocall->setType($arg => $ft->getType("$var")) ;
	  $ko = 0 ;
	}
      }
    }
    if ($ko) {
      print STDOUT "\t" . (caller(0))[3] . ": Unknown type for $arg is $point($var)\n" if $debug ;
      $ocall->setType($arg => $self->getType($arg)) ;
    }
  } elsif ($arg =~ /^(\w+)\%([\w\%]+)$/) {
    # Multi level analysis:
    my $point = $1 ;
    my $var = $2 ;
    print STDOUT (caller(0))[3] . ": Did you ever hear of encapsulation for $arg?\n" if $debug ;
    my $ft = undef ;
    my $answer = $self->findType("$point") ;
    print STDOUT "\t\tFetched structure $point brings $answer\n" if $debug ;
    if ($answer && $answer =~ /(type|class)\s*\(\s*(.*?)\s*\)/i) {
      $ft = &fthtype::getObject("$2") ;
      while ($ft && $var =~ /^(\w+)\%([\w\%]+)/) {
	# Many levels direct access: mad programming !
	$point = $1 ;
	$var = $2 ;
	$answer = $ft->getType("$point") ;
	print STDOUT "\t\t\tFetched structure $point brings $answer\n" if $debug ;
	if ($answer && $answer =~ /(type|class)\s*\(\s*(.*?)\s*\)/i) {
	  $ft = &fthtype::getObject("$2") ;
	  unless ($ft) {
	    last ;
	  }
	}
      }
    }
    if ($ft) {
      $answer = $ft->getType("$var") ;
      print STDOUT "\t\tSubStructure $1 brings $ft, analyze $var says $answer\n" if $debug ;
      $ocall->setType($arg => $answer) ;
    } elsif ($ocall->getType("$arg") eq 'void') {
      # From upper call: last chance to set the value...
      my $answer = $self->findType("$arg") ;
      $ocall->setType($arg => $answer) if $answer;
    }
  } elsif (0 && $arg =~ /^(\w+)\%(\w+)\%(\w+)$/) {
    # DELETED SECTION #
    # Two levels direct access: mad programming !
    print STDOUT (caller(0))[3] . ": Did you ever hear of encapsulation for $arg?\n" if $debug ;
    my $point = $1 ; my $x1 = $2 ; my $var = $3 ;
    my $answer = $self->findType("$point") ;
    print STDOUT "\t\tFetched structure $point brings $answer\n" if $debug ;
    if ($answer && $answer =~ /(type|class)\s*\(\s*(.*?)\s*\)/i) {
      my $ft = &fthtype::getObject("$2") ;
      if ($ft) {
	$point = $x1 ;
	$answer = $ft->getType("$point") ;
	print STDOUT "\t\tStructure $2 brings $ft, analyze $x1 says $answer\n" if $debug ;
	if ($answer && $answer =~ /(type|class)\s*\(\s*(.*?)\s*\)/i) {
	  print STDOUT "\t\tLooking for $2\n" if $debug ;
	  $ft = &fthtype::getObject("$2") ;
	  if ($ft) {
	    $answer = $ft->getType("$var") ;
	    print STDOUT "\t\tSubStructure $2 brings $ft, analyze $var says $answer\n" if $debug ;
	    $ocall->setType($arg => $answer) ;
	  }
	}
      }
    } elsif ($ocall->getType("$arg") eq 'void') {
      # From upper call: last chance to set the value...
      my $answer = $self->findType("$arg") ;
      $ocall->setType($arg => $answer) if $answer;
    }
  } elsif ($ocall->getType("$arg") eq 'void') {
    # From upper call: last chance to set the value...
    my $answer = $self->findType("$arg") ;
    my $type = $self->getType("$arg") ;
    &fthcall::setDebug(1) if $debug ;
    my $fullarg = $ocall->getArgument($number) ;
    &fthcall::setDebug(0) if $debug ;
    print STDOUT "\t\tYet another final chance to set the correct type for arg=$arg, number=$number, fullarg=$fullarg\n" if $debug ;
    if ($answer) {
      print STDOUT "\t\tYeah, $arg bound to $answer\n" if $debug ;
      $ocall->setType($arg => $answer) ;
    } elsif ($type && &fthrout::isIntrinsec($arg) && 'void' eq &fthrout::isIntrinsec($arg)) {
      # Intrinsec function not pointed ?
      print STDOUT "\t\tisIntrinsec: argument <$arg> bound to fullarg=<$fullarg>\n" if $debug ;
      if ($fullarg && $fullarg =~ /\s*(\w+?)\s*\((.*)\)/) {
	$answer = $self->getType($2) ;
	if ($answer && $answer ne 'void') {
	  print STDOUT "\t\t\tGetting $answer for arg=$arg thanks to fullarg=$fullarg\n" if $debug ;
	  $ocall->setType($arg => $answer) ;
	} else {
	  # argument may be a function ?
	  print STDOUT "\t\t\tTesting fullarg=$fullarg\n" if $debug ;
	  if ($fullarg =~ /(.*?\s*\w+?)\s*\((\w+?)\s*\((.*)/) {
	    my $fobj = &fthobj::getObject("$2") ;
	    print STDOUT "\t\t\tTesting object=$2\n"if $debug ;
	    if ($fobj) {
	      print STDOUT "\t\t\t\tIs an object !!!!\n" if $debug ;
	      if ($ocall->replaceArg($number, $arg, lc($2))) {
		print STDOUT "\t\t\t\tReplacing $1 with $2\n"if $debug ;
		$arg = $2 ;
		#$fullarg = $2 . '(' . $3 . '))' . $4 ;
		$fullarg = $2 . '(' . $3 ;
	      }
	    }
	  }
	}
      }
    }
    if (!$answer || $answer eq 'void') {
      if ($type && $type ne 'void' && $type ne 'unknown') {
	print STDOUT "\t\tYeah, arg $arg is built-in like $type\n" if $debug ;
	$ocall->setType($arg => $type) ;
	$answer = $type ;
      }
    }
    print STDOUT "\t\tNot yet known arg($number)=$arg for fullarg=<$fullarg>\n" if $debug ;
    if (!$answer || $answer eq 'void') {
      $type = &fthrout::isIntrinsec($arg) ;
      if ($type && $type ne 'void') {
	print STDOUT "\t\tFound implicit function return value $type for $arg\n" if $debug;
	$ocall->setType($arg => $type) ;
	$answer = $type ;
      }
    }
    print STDOUT "\t\t\t$answer\n" if $debug  && $answer;
    if (!$answer || $answer eq 'void') {
      # Last chance to get the value from the fobj :
      my $fobj = &fthobj::getObject("$arg") ;
      $answer = undef ;
      if ($fobj) {
	print STDOUT "\t\t" . (caller(0))[3] . ": argument $arg in object " . $fobj->get(-file) . " name " . $fobj->get(-name) . " very last chance\n" if $debug ;
	$answer = $fobj->getType() ;
      }
      if ($answer) {
	print STDOUT "\t\t" . (caller(0))[3] . ": argument $arg in object " . $fobj->get(-file) . " name " . $fobj->get(-name) . " is defined as $answer\n" if $debug ;
	$ocall->setType($arg => $answer) ;
      }
      # Better try to analyze the stuff down there:
      if ($fobj && $fullarg && $fullarg =~ /\s*(\w+?)\s*\((.*)\)/) {
	my $scall = new fthcall(-name => $1, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self) ;
	$scall->readf90("", "& ($2)") ;
	# Do not increase the list of called routines:
	$self->addCall($1 => $scall) ;
	# Sets the void arg:
	if ($answer) {
	  $scall->addArgument($1) ;
	  $scall->setType($1 => $answer) ;
	} else {
	  $scall->addArgument($1) ;
	  $scall->setType($1 => 'void') ;
	}
	$scall->set(-number => $ocall->get(-number));
	# Sets the type to the calling stuff:
	# NAJ: what is that standing for:
	$ocall->setDepend($number, $scall) ;
	$self->preCalls($scall) ;
	print STDOUT (caller(0))[3] . ": Called " . $ocall->get(-name) . " line " . $ocall->get(-number) . " in file " . $self->get(-file) . " Fullarg($number)='$fullarg' void $1 ruft ($2)\n" if $debug;
      } elsif ($fullarg && $fullarg =~ /^\s*(\w+)\s*\=\s*(.*)/) {
	print STDOUT (caller(0))[3] . ": Called " . $ocall->get(-name) . " line " . $ocall->get(-number) . " in file " . $self->get(-file) . " arg=$arg meets possible equal $1: $2\n" if $debug ;
	if (lc($arg) ne lc($2)) {
	  # Avoid recursion:
	  if ($ocall->replaceArg($number, lc($1), lc($2))) {
	    # Recursive call:
	    print STDOUT "\t\t" . (caller(0))[3] . ": argument $arg in object " . $ocall->get(-name) . " line " . $ocall->get(-number) . " in file " . $self->get(-file) . " recursive analysis\n" if $debug ;
	    $self->preCalls2($ocall, lc($2), $number) ;
	  }
	}
      } else {
	# Try the implicit ?
	$type = $self->defType($arg) ;
	if ($type && $type ne 'void' && $type ne 'unknown') {
	  print STDOUT "\t\tUgly programmer, arg $arg is implicit like $type\n" if $debug ;
	  $ocall->setType($arg => $type) ;
	}
      }
    }
  } else {
    # Call to self built-in stuff:
    print STDOUT (caller(0))[3] . ": Called " . $ocall->get(-name) . " line " . $ocall->get(-number) . " in file " . $self->get(-file) . " unknown says arg=$arg is of type " . $ocall->getType("$arg") . "\n" if $debug;
    $ocall->setType("$arg" => $ocall->getType("$arg")) ;
  }
}
###########################################################################
sub preCalls {
  my $self = shift ;
  my $ocall = shift ;
  my (@args) = $ocall->getArguments() ;
  if ($debug) {
    print STDOUT (caller(0))[3] . ": preCalls before analysis to call " . $ocall->get(-name) . ", file " . $ocall->get(-file) . " line " . $ocall->get(-number) . ": '@args'.\n" ;
    $ocall->datadump("\t") ;
  }
  my $number = 0 ;
  foreach my $arg (@args) {
    if ($debug) {
      print STDOUT (caller(0))[3] . ": Direct private access to $arg: " . $ocall->getType("$arg") . "\n" ;
    }
    if ($ocall->getType("$arg") eq 'unknown' || $ocall->getType("$arg") eq 'void') {
      print STDOUT "\t" . (caller(0))[3] . ": New analysis for $number argument $arg\n" if $debug ;
      $self->preCalls2($ocall, $arg, $number) ;
    }
    $number++ ;
  }
  if ($debug) {
    my (@a) = $ocall->getArguments() ;
    print STDOUT (caller(0))[3] . ": preCalls after analysis to call " . $ocall->get(-name) . ", file " . $ocall->get(-file) . " line " . $ocall->get(-number) . ": '@a'.\n" ;
    $ocall->datadump("\t") ;
  }
}
###########################################################################
sub revCalls {
  my $self = shift ;
  my $orout ;
  # Procesing the private stuffs:
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    my $drout = $self->{"FPROCEDURES"}->{"$name"} ;
    print STDOUT "\t" . (caller(0))[3] . ": call $name = '$drout'\n" if $debug;
    $drout->revCalls() ;
  }
  # Loop on the local calls:
  foreach my $ocall (values %{$self->{"FCALLS"}}) {
    $orout = $ocall->getLink() ;
    if ($orout) {
      if ($debug) {
	print STDOUT "\t" . (caller(0))[3] . ": call name=" . $ocall->get(-name) . " in file " . $self->get(-file) . " line ". $ocall->get(-number) ." points to fthrout = '$orout'\n" ;
	$orout->datadump("\t") ;
      }
      $orout->addCalled($ocall) ;
    }
  }
}
###########################################################################
sub setCallers {
  my $self = shift ;
  my $ocall = shift ;
  my $orout ;
  my $cname = $ocall->get(-name) ;
  # Fast return: using interface is stupid at this level !
  $orout = &fthrout::getObject(lc($cname)) ;
  if ($orout) {
    if ($orout->get(-global)) {
      $ocall->set(-thelink => $orout) ;
    }
  }
  my $ctype = $self->findType("$cname", 1) ;
  print STDOUT "\t" . (caller(0))[3] . ": Obtained value $ctype in " . $self->get(-name) . " for arg $cname\n" if $ctype && $debug ;
  # We may get the prototype for the stuff:
  if ($debug) {
    my (@proto) = &getProto($ocall) ;
    print STDOUT "\t" . (caller(0))[3] . ": Routine L(1)" . $self->get(-name) . " calls ". $ocall->get(-name) . "(@proto)\n"  ;
  }
  $orout = $self->isRoutine($ocall) ;
  if ($orout) {
    print STDOUT "\t" . (caller(0))[3] . ": Routine L(2)" . $self->get(-name) . " calls ". $ocall->get(-name) . " met by " . $orout->get(-name) . "\n" if $debug ;
    $ocall->set(-thelink => $orout) ;
  } else {
    $orout = $self->isDirect($ocall) ;
    if ($orout) {
      print STDOUT "\t" . (caller(0))[3] . ": Routine L(3)" . $self->get(-name) . " calls ". $ocall->get(-name) . " direct by $orout:" . $orout->get(-name) . "\n" if $debug ;
      $ocall->set(-thelink => $orout) ;
    } else {
      # NAJ: not very clear stuff: ask the father what it can do...
      if (exists $self->{"private"}) {
	$self->{"private"}->setCallers($ocall) ;
	return ;
      } else {
	$orout = &fthrout::isGlobal($ocall) ;
	if ($orout) {
	  print STDOUT "\t" . (caller(0))[3] . ": Routine L(4)" . $self->get(-name) . " calls ". $ocall->get(-name) . " global by $orout:" . $orout->get(-name) . "\n" if $debug ;
	  $ocall->set(-thelink => $orout) ;
	} elsif ($ctype && $ctype ne 'void') {
	  $ocall->set(-fake => $ctype) ;
	  return ;
	}
      }
    }
  }
  unless (&fthobj::getObject(lc($cname)) || &fthrout::getObject(lc($cname)) || ($ctype && $ctype ne 'void')) {
    # Fasten the search: no routine by this name or no available interface...
    my $oext = new fthext(-name => $cname) ;
    $oext->addStuff($ocall) ;
    return ;
  }
  unless ($ocall->getLink()) {
    # Fasten the search: no routine by this name or no available interface...
    my $obc = new fthbc(-name => $cname) ;
    my $bidule = &fthrout::getObject(lc($cname)) ;
    if ($bidule) {
      $ocall->set(-badcall => $bidule) ;
    } else {
      $bidule = &fthobj::getObject(lc($cname)) ;
      if ($bidule) {
	$ocall->set(-badcall => $bidule) ;
      }
    }
    $obc->addStuff($ocall) ;
    # NAJ518: time to look for possible interfaces call ? Or too early ?
    my $fobj = &fthobj::getObject($ocall->get(-name)) ;
    if (0 && $fobj) {
      my (@cproto) = $ocall->getProto() ;
      my (@linters) = $fobj->getStuffs() ;
      if (@linters) {
	foreach my $finter (@linters) {
	  if (ref($finter) eq 'fthinter') {
	    my (@procs) = $finter->getProcedures() ;
	    foreach my $lpname (@procs) {
	      my $lorout = &fthrout::getObject($lpname) ;
	      unless ($lorout) {
		# Another attempt:
		my $lfobj = &fthobj::getObject($lpname) ;
		$lorout = $lfobj->getProcedure($lpname) if $lfobj ;
	      }
	      if ($lorout) {
		# Compare the protos ?
		my(@mproto) = $lorout->getProto() ;
		my $ko = &fthrout::relProto(\@mproto, \@cproto) ;
		unless ($ko) {
		  $ocall->addPossible($lorout) ;
		}
	      }
	    }
	  }
	}
      }
    }
  }
}
###########################################################################
# Returns an ordered array of inlined called stuffs:
sub getInlined {
  my $self = shift ;
  my (@oldcalls) = (@_) ;
  (@{$self->{"NCALLS"}}) = () ;
  foreach my $ocall (@oldcalls) {
    $self->addInlined($ocall) ;
  }
  return(@{$self->{"NCALLS"}});
}
###########################################################################
sub addInlined {
  my $self = shift ;
  my $ocall = shift ;
  my(%hash) = $ocall->getDepends() ;
  if (%hash) {
    foreach my $number (keys %hash) {
      my $scall = $hash{"$number"} ;
      $self->addInlined($scall) ;
    }
  }
  push(@{$self->{"NCALLS"}}, $ocall) ;
}
###########################################################################
sub setCalls {
  my $self = shift ;
  # Loop on the local calls:
  my (@oldcalls) = (values %{$self->{"FCALLS"}}) ;
  # Processing the private stuffs:
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    my $drout = $self->{"FPROCEDURES"}->{"$name"} ;
    print STDOUT (caller(0))[3] .": Dealing with calls for ".$self->get(-name)." is $drout from file ".$self->get(-file)."\n" if $debug ;
    $drout->setCalls() ;
  }
  foreach my $ocall (@oldcalls) {
    print STDOUT "+" if ($main::opt_v >= 1) ;
    print STDOUT (caller(0))[3] .": Find call for ".$ocall->get(-name)." line ".$ocall->get(-number)." from file ".$self->get(-file)."\n" if $debug ;
    $self->preCalls($ocall) ;
  }
  my (@newcalls) = ($self->getInlined(@oldcalls)) ;
  foreach my $ocall (@newcalls) {
    print STDOUT (caller(0))[3] .": Trying to find call for ".$ocall->get(-name)." line ".$ocall->get(-number)." from file ".$self->get(-file)."\n" if $debug ;
    print STDOUT "." if ($main::opt_v >= 1) ;
    $self->postCalls($ocall) ;
    $self->setCallers($ocall) ;
  }
}
###########################################################################
sub sayCalls {
  my $self = shift ;
  my $ier = shift ;
  my $orout ;
  # Procesing the private stuffs:
  foreach my $name (keys %{$self->{"FPROCEDURES"}}) {
    my $drout = $self->{"FPROCEDURES"}->{"$name"} ;
    $drout->sayCalls() ;
  }
  # Loop on the local calls:
  foreach my $ocall (values %{$self->{"FCALLS"}}) {
    my (@proto) = &getProto($ocall) ;
    print STDOUT "\tfthrout::sayCalls for " . $self->get(-name) . " called ". $ocall->get(-name) . "(@proto)\n" if $debug ;
    $orout = $ocall->getLink() ;
    if ($orout && $ier) {
      print STDOUT "OK good call " . $ocall->get(-name) . " from " . $self->get(-name) ."\n" ;
      $ocall->datadump("\t\t")  ;
      $orout->datadump("\t\t")  ;
    } elsif (!$orout && !$ier && !$ocall->get(-fake)) {
      my $name = $ocall->get(-name) ;
      if (&fthrout::getObject($name) || &fthobj::getObject($name)) {
	print STDOUT "KO bad call " . $ocall->get(-name) . " from " . $self->get(-name) ."\n" ;
	$ocall->datadump("\t")  ;
      } else {
	print STDOUT "No reference to " . $ocall->get(-name) . " from " . $self->get(-name) ." line " . $ocall->get(-number) . "\n" ;
      }
    }
  }
}
###########################################################################
sub isInternC {
  my $name = lc(shift) ;
  my $ko = undef ;
  if (exists $INTRC{$name}) {
    $ko = $INTRC{$name} ;
  } else {
    $ko = &fthrout::isIntrinsec($name) ;
  }
}
###########################################################################
sub isIntrinsec {
  my $name = lc(shift) ;
  my $ko = undef ;
  if (exists $INTRF{$name}) {
    $ko = $INTRF{$name} ;
  } elsif (($name =~ /^c(.*)/ && exists $INTRF{"$1"}) || ($name =~ /^d(.*)/ && exists $INTRF{"$1"}) || ($name =~ /^cd(.*)/ && exists $INTRF{"$1"})) {
    $ko = $INTRF{$1} ;
  }
  return($ko) ;
}
###########################################################################
sub isExternal {
  my $name = lc(shift) ;
  my $ko = 0 ;
  if (($name =~ /^(date)/i) || exists $INTRS{$name}) {
    $ko = 1 ;
  }
  return($ko) ;
}
###########################################################################
# Counts the number of parenthesis
sub countParenthesis {
  my $nbparenthe = shift ;
  my $ligne = shift ;
  while ($ligne =~ /^(.*?)([\(\)])(.*)/) {
    $ligne = $3 ;
    if ($2 eq '(') {
      $nbparenthe++ ;
    } else {
      $nbparenthe-- ; 
    }
  }
  return($nbparenthe, $ligne) ;
}
###########################################################################
sub readf77func {
  my $self = shift ;
  my $file = shift ;
  my $res = shift ; 
  my $cname = shift ; 
  my $args = shift ;
  my $lname = lc($cname) ;
  my $ocall = undef ; 
  my $ko = &fthrout::isIntrinsec($lname) ;
  my $chtml = &chtml::getCurrent() ;
  if (1 == 0 && exists $self->{"ALLDEFS"}->{"$lname"}) {
    # Not function, known as a variable instead, no more control at this level since contradictory
    # with function calls through ():
    $ko = 1 ;
  }
  if ($ko) {
    # Push the stuff anyway
    $chtml->addDef($cname.'(') ;
  } else {
    # Calls a routine:
    print STDOUT (caller(0))[3] .": Creating function call $res = $cname ($args)\n" if $debug ;
    my $ocall = new fthcall(-name => $cname, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self) ;
    $chtml->addRaw($ocall) ;
    $self->addCall($cname => $ocall) ;
    $ocall->readf77($file, "     & ( $args") ;
    # Post process the arguments:
    foreach my $arg ($ocall->getArguments()) {
      my $val = $self->getType($arg) ;
      if ((!$val || $val eq 'void') && exists $self->{"IMPLICITDEFS"}) {
	print STDOUT (caller(0))[3] . ": Void arg $arg to type '" . $val . "'\n" if $debug ;
	$val = $self->defType($arg) ;
      }
      print STDOUT "\t" . (caller(0))[3] . ": Setting main arg $arg to type " . $val . "\n" if $debug ;
      $ocall->setType($arg => $val) ;
    }
    if ($res eq 'void') {
      # Unable to guess
      $ocall->addArgument($cname) ;
      $ocall->setType($cname => 'void') ;
    } elsif ($self->getType($res) ne 'void') {
      $ocall->addArgument($res) ;
      $ocall->setType($res => $self->getType($res)) ;
    } else {
      print STDOUT "\t" . (caller(0))[3] .": Seems called ".$ocall->get(-name)." results $res of type void in " . $self->get(-name)."\n"  if $debug ;
      $ocall->addArgument($cname) ;
      my $val = $self->getType($cname) ;
      if ((!$val || $val eq 'void') && exists $self->{"IMPLICITDEFS"}) {
	$val = $self->defType($cname) ;
      }
      print STDOUT "\t" . (caller(0))[3] . ": Setting return arg $cname to type " . $val . "\n" if $debug ;
      $ocall->setType($cname => $val) ;
    }
  }
  return($ocall) ;
}
###########################################################################
sub readf77call {
  my $self = shift ;
  my $file = shift ;
  my $cname = shift ;
  my $ligne = shift ;
  my $chtml = &chtml::getCurrent() ;
  if (&isExternal($cname)) {
    # NAJ: not very nice design pattern
    if ($main::opt_nocolor) {
      $chtml->addRaw("$cname") ;
    } else {
      $chtml->addRaw("<FONT COLOR=PINK>$cname</FONT>") ;
    }
    if ($ligne=~ /^\((.*)/) {
      $_ = $1 ;
      $chtml->addRaw('(') ;
    } else {
      $_ = $ligne ;
    }
    return ;
  }
  my $ocall = new fthcall(-name => "$cname", -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self) ;
  print STDOUT (caller(0))[3] . ": Creating Fortran call $cname from ".(caller(1))[3].':'.(caller(0))[2]."\n" if $debug ;
  $chtml->addRaw($ocall) ;
  $self->addCall("$cname" => $ocall) ;
  $ocall->readf77($file, "     & $ligne") ;
  # Post process the arguments:
  foreach my $arg ($ocall->getArguments()) {
    my $val = $self->getType($arg) ;
    if ((!$val || $val eq 'void') && exists $self->{"IMPLICITDEFS"}) {
      print STDOUT (caller(0))[3] . ": Void arg $arg to type '" . $val . "'\n" if $debug ;
      $val = $self->defType($arg) ;
    }
    print STDOUT (caller(0))[3] . ": Setting arg $arg to type '" . $val . "'\n" if $debug ;
    $ocall->setType($arg => $val) ;
  }
}

###########################################################################
sub parsef90func {
  my $self = shift ;
  my $file = shift ;
  my $res = shift ; 
  my $cname = shift ; 
  my $args = shift ;
  my $number = shift ;
  my $lname = lc($cname) ;
  my $ocall = undef ; 
  # Calls a routine:
  print STDOUT (caller(0))[3] . ": Line $. file " . $self->get(-file) . ", creating function call $res = $cname args=<($args> from ".(caller(1))[3].':'.(caller(0))[2]."\n" if $debug ;
  $ocall = new fthcall(-name => $cname, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self, -number => $number) ;
  $self->addCall($cname => $ocall) ;
  $ocall->parsef90($file, "($args)") ;
  # Post process the arguments:
  my $argnum = 0 ;
  foreach my $arg ($ocall->getArguments()) {
    if ($arg =~ /^(\w+)\%(\w+)$/) {
      # Manage simple first level argument types.
      # Do note multiple structure level direct access is not
      # an encapsulated OO programming. It ought not to be encountered
      # only this%data ought to be accessed directly.
      my $point = $1 ; my $var = $2 ;
      my $ko = 1 ;
      if (exists $self->{"ALLDEFS"}->{"$point"}) {
	if ($self->{"ALLDEFS"}->{"$point"} =~ /type\s*\(\s*(.*)\s*\)/i) {
	  $point = $1 ;
	  if (my $ft = &fthtype::getObject("$point")) {
	    # Gets the type for $var in object $ft ;
	    print STDOUT (caller(0))[3] . ": OK Type for $arg is $point%$var :: " . $ft->getType("$var")."\n" if $debug ;
	    $ocall->setType($arg => $ft->getType("$var")) ;
	    $ko = 0 ;
	  }
	}
      }
      if ($ko) {
	print STDOUT (caller(0))[3] . ": Unknown type for $arg is $point($var)\n" if $debug ;
	$ocall->setType($arg => 'unknown') ;
      }
    } else {
      my $val = $self->getType($arg) ;
      if ((!$val || $val eq 'void') && exists $self->{"IMPLICITDEFS"}) {
	print STDOUT (caller(0))[3] . ": Void arg $arg to type '" . $val . "'\n" if $debug ;
	$val = $self->defType($arg) ;
      }
      print STDOUT (caller(0))[3] . ": Setting arg $arg to type '" . $val . "'\n" if $debug ;
      $ocall->setType($arg => $val) ;
    }
    # Now try to parse recursive inside calls ?
  }
  if ($res eq 'void') {
    # Unable to guess
    $ocall->addArgument($cname) ;
    $ocall->setType($cname => 'void') ;
  } elsif ($self->getType($res) ne 'void' || $res =~ /\%/) {
    $ocall->addArgument($res) ;
    $ocall->setType($res => $self->getType($res)) ;
  } else {
    print STDOUT "\t" . (caller(0))[3] .": Seems called ".$ocall->get(-name)." results $res of type void in " . $self->get(-name)."\n"  if $debug ;
    $ocall->addArgument($cname) ;
    my $val = $self->getType($cname) ;
    if ((!$val || $val eq 'void') && exists $self->{"IMPLICITDEFS"}) {
      $val = $self->defType($cname) ;
    }
    print STDOUT "\t" . (caller(0))[3] . ": Setting return arg $cname to type " . $val . "\n" if $debug ;
    $ocall->setType($cname => $val) ;
  }
  $ocall->datadump("\t") if $debug ;
 # die "OK" ;
}

###########################################################################
sub readf90func {
  my $self = shift ;
  my $file = shift ;
  my $res = shift ; 
  my $cname = shift ; 
  my $args = shift ;
  my $lname = lc($cname) ;
  my $ocall = undef ; 
  my $ko = 0 ;
  my $chtml = &chtml::getCurrent() ;
  if (1 == 0) {
    # Eliminating intrinsics calls is not that a good idea.
    $ko = &isIntrinsec("$lname") ;
    if (exists $self->{"ALLDEFS"}->{"$lname"}) {
      # Not function, known as a variable instead
      $ko = 1 ;
    }
  }
  if ($ko) {
    # Push the stuff anyway
    $chtml->addDef($cname.'(') ;
  } else {
    # Calls a routine:
    print STDOUT (caller(0))[3] . ": Line $. file " . $self->get(-file) . ", creating function call $res = $cname args=<($args> from ".(caller(1))[3].':'.(caller(0))[2]."\n" if $debug ;
    $ocall = new fthcall(-name => $cname, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self) ;
    $chtml->addRaw($ocall) ;
    $self->addCall($cname => $ocall) ;
    $ocall->readf90($file, "& ($args") ;
    # Post process the arguments:
    my $argnum = 0 ;
    foreach my $arg ($ocall->getArguments()) {
      if ($arg =~ /^(\w+)\%(\w+)$/) {
	# Manage simple first level argument types.
	# Do note multiple structure level direct access is not
	# an encapsulated OO programming. It ought not to be encountered
	# only this%data ought to be accessed directly.
	my $point = $1 ; my $var = $2 ;
	my $ko = 1 ;
	if (exists $self->{"ALLDEFS"}->{"$point"}) {
	  if ($self->{"ALLDEFS"}->{"$point"} =~ /(type|class)\s*\(\s*(.*)\s*\)/i) {
	    $point = $2 ;
	    if (my $ft = &fthtype::getObject("$point")) {
	      # Gets the type for $var in object $ft ;
	      print STDOUT (caller(0))[3] . ": OK Type for $arg is $point%$var :: " . $ft->getType("$var")."\n" if $debug ;
	      $ocall->setType($arg => $ft->getType("$var")) ;
	      $ko = 0 ;
	    }
	  }
	}
	if ($ko) {
	  print STDOUT (caller(0))[3] . ": Unknown type for $arg is $point($var)\n" if $debug ;
	  $ocall->setType($arg => 'unknown') ;
	}
      } else {
	my $val = $self->getType($arg) ;
	if ((!$val || $val eq 'void') && exists $self->{"IMPLICITDEFS"}) {
	  print STDOUT (caller(0))[3] . ": Void arg $arg to type '" . $val . "'\n" if $debug ;
	  $val = $self->defType($arg) ;
	}
	print STDOUT (caller(0))[3] . ": Setting arg $arg to type '" . $val . "'\n" if $debug ;
	$ocall->setType($arg => $val) ;
      }
      # Now try to parse recursive inside calls:
      if (1 == 1) {
	my $fname = $ocall->getArgument($argnum) ;
	if ($fname =~ /([\w\%]+?)\((.*)/) {
	  my $ligne = $fname ;
	  while($ligne =~ /([a-z_\%]\w*?)\((.*?)\)(.*)/i) {
	    print STDOUT "Calling parsef90func for possible call='$1' with arguments='$2'\n" if $debug ;
	    $ligne = $3 ;
	    unless ($1 =~ /\%/) { #NAJ518: $self->sayCalling($1) forces known, prevents external.
	      if ($self->sayCalling($1)) {
		$self->parsef90func($file, "void", "$1", "$2", $ocall->get(-number)) ;
	      }
	    }
	    last unless $ligne ;
	  }
	}
	$argnum++ ;
      }
    }
    if ($res eq 'void') {
      # Unable to guess
      $ocall->addArgument($cname) ;
      $ocall->setType($cname => 'void') ;
    } elsif ($self->getType($res) ne 'void' || $res =~ /\%/) {
      $ocall->addArgument($res) ;
      $ocall->setType($res => $self->getType($res)) ;
    } else {
      print STDOUT "\t" . (caller(0))[3] .": Seems called ".$ocall->get(-name)." results $res of type void in " . $self->get(-name)."\n"  if $debug ;
      $ocall->addArgument($cname) ;
      my $val = $self->getType($cname) ;
      if ((!$val || $val eq 'void') && exists $self->{"IMPLICITDEFS"}) {
	$val = $self->defType($cname) ;
      }
      print STDOUT "\t" . (caller(0))[3] . ": Setting return arg $cname to type " . $val . "\n" if $debug ;
      $ocall->setType($cname => $val) ;
    }
  }
  return($ocall) ;
}
###########################################################################
sub postFunProcess {
  my $self = shift ;
  my $ocall = shift ;
  my $file = shift ;
  my $res = shift ; 
  my $cname = shift ; 
  # Post process the arguments:
  my $argnum = 0 ;
  foreach my $arg ($ocall->getArguments()) {
    if ($arg =~ /^(\w+)\%(\w+)$/) {
      # Manage simple first level argument types.
      # Do note multiple structure level direct access is not
      # an encapsulated OO programming. It ought not to be encountered
      # only this%data ought to be accessed directly.
      my $point = $1 ; my $var = $2 ;
      my $ko = 1 ;
      if (exists $self->{"ALLDEFS"}->{"$point"}) {
	if ($self->{"ALLDEFS"}->{"$point"} =~ /(type|class)\s*\(\s*(.*)\s*\)/i) {
	  $point = $2 ;
	  if (my $ft = &fthtype::getObject("$point")) {
	    # Gets the type for $var in object $ft ;
	    print STDOUT (caller(0))[3] . ": OK Type for $arg is $point%$var :: " . $ft->getType("$var")."\n" if $debug ;
	    $ocall->setType($arg => $ft->getType("$var")) ;
	    $ko = 0 ;
	  }
	}
      }
      if ($ko) {
	print STDOUT (caller(0))[3] . ": Unknown type for $arg is $point($var)\n" if $debug ;
	$ocall->setType($arg => 'unknown') ;
      }
    } else {
      my $val = $self->getType($arg) ;
      if ((!$val || $val eq 'void') && exists $self->{"IMPLICITDEFS"}) {
	print STDOUT (caller(0))[3] . ": Void arg $arg to type '" . $val . "'\n" if $debug ;
	$val = $self->defType($arg) ;
      }
      print STDOUT (caller(0))[3] . ": Setting arg $arg to type '" . $val . "'\n" if $debug ;
      $ocall->setType($arg => $val) ;
    }
    # Now try to parse recursive inside calls:
    if (1 == 1) {
      my $fname = $ocall->getArgument($argnum) ;
      if ($fname =~ /([\w\%]+?)\((.*)/) {
	my $ligne = $fname ;
	while($ligne =~ /([a-z_\%]\w*?)\((.*?)\)(.*)/i) {
	  print STDOUT "Calling parsef90func for possible call='$1' with arguments='$2'\n" if $debug ;
	  $ligne = $3 ;
	  unless ($1 =~ /\%/) { #NAJ518: $self->sayCalling($1) forces known, prevents external.
	    if ($self->sayCalling($1)) {
	      $self->parsef90func($file, "void", "$1", "$2", $ocall->get(-number)) ;
	    }
	  }
	  last unless $ligne ;
	}
      }
      $argnum++ ;
    }
  }
  if ($res eq 'void') {
    # Unable to guess
    $ocall->addArgument($cname) ;
    $ocall->setType($cname => 'void') ;
  } elsif ($self->getType($res) ne 'void' || $res =~ /\%/) {
    $ocall->addArgument($res) ;
    $ocall->setType($res => $self->getType($res)) ;
  } else {
    print STDOUT "\t" . (caller(0))[3] .": Seems called ".$ocall->get(-name)." results $res of type void in " . $self->get(-name)."\n"  if $debug ;
    $ocall->addArgument($cname) ;
    my $val = $self->getType($cname) ;
    if ((!$val || $val eq 'void') && exists $self->{"IMPLICITDEFS"}) {
      $val = $self->defType($cname) ;
    }
    print STDOUT "\t" . (caller(0))[3] . ": Setting return arg $cname to type " . $val . "\n" if $debug ;
    $ocall->setType($cname => $val) ;
  }
}
###########################################################################
sub readf90call {
  my $self = shift ;
  my $file = shift ;
  my $cname = shift ;
  my $ligne = shift ;
  my $chtml = &chtml::getCurrent() ;
  if (&isExternal($cname)) {
    # NAJ: not very nice design pattern
    if ($main::opt_nocolor) {
      $chtml->addRaw("$cname") ;
    } else {
      $chtml->addRaw("<FONT COLOR=PINK>$cname</FONT>") ;
    }
    if ($ligne=~ /^\((.*)/) {
      $_ = $1 ;
      $chtml->addRaw('(') ;
    } else {
      $_ = $ligne ;
    }
    return ;
  }
  my $ocall = new fthcall(-name => $cname, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self) ;
  print STDOUT (caller(0))[3] .": Creating Fortran call $cname\n" if $debug ;
  $chtml->addRaw($ocall) ;
  $self->addCall($cname => $ocall) ;
  $ocall->readf90($file, "& $ligne") ;
  # Post process the arguments:
  $self->postCallProcess($ocall) ;
}
###########################################################################
sub postCallProcess {
  my $self = shift ;
  my $ocall = shift ;
  foreach my $arg ($ocall->getArguments()) {
    if ($arg =~ /^(\w+)\%(\w+)$/) {
      # Manage simple first level argument types.
      # Do note multiple structure level direct access is not
      # an encapsulated OO programming. It ought not to be encountered
      # only this%data ought to be accessed directly.
      my $point = $1 ; my $var = $2 ;
      my $ko = 1 ;
      if (exists $self->{"ALLDEFS"}->{"$point"}) {
	if ($self->{"ALLDEFS"}->{"$point"} =~ /(type|class)\s*\(\s*(.*)\s*\)/i) {
	  $point = $2 ;
	  if (my $ft = &fthtype::getObject("$point")) {
	    # Gets the type for $var in object $ft ;
	    print STDOUT (caller(0))[3] . ": OK Type for $arg is $point%$var :: " . $ft->getType("$var")."\n" if $debug ;
	    $ocall->setType($arg => $ft->getType("$var")) ;
	    $ko = 0 ;
	  }
	}
      }
      if ($ko) {
	print STDOUT (caller(0))[3] . ": Unknown type for $arg is $point($var)\n" if $debug ;
	$ocall->setType($arg => 'unknown') ;
      }
    } else {
      print STDOUT (caller(0))[3] . ": Inquire argument $arg for call " . $ocall->get(-name) . " says type=" . $self->getType($arg) ."\n" if $debug ;
      $ocall->setType($arg => $self->getType($arg)) ;
    }
  }
}
###########################################################################
# Reads an if stuff...
sub readf77if {
  my $self = shift ;
  my $file = shift ;
  $_ = shift ;
  my $cont = shift ;
  $cont = 1 unless $cont ;
  my $nbparenthe = 0 ;
  my $chtml = &chtml::getCurrent() ;
  if ((ref $file) && (fileno $file)) {
    my $name = $self->get(-name) ;
    # Sets the line number for definition:
    $self->set(-number => $.) ;
    while($cont) {
      # Parsing an f90 rout file below:
      if (/^\s*\d*\s*end\s*if\b/i) {
	# End of routine creation
	print STDOUT "Finished if then $name\n" if $debug ;
	return ;
      } elsif ($cont == 1) {
	if (/^.{5}[\&\S]/) {
	  # Count the number of parenthesis:
	  ($nbparenthe, $_) = &countParenthesis($nbparenthe, $_) ;
	  if (!$nbparenthe && !/then/i) {
	    # End of stuff if no then, later test the call...
	    print STDOUT "Finished if straight $name\n" if $debug ;
	    return ;
	  }
	} else {
	  $cont = 2 ;
	}
      }
      if ($cont == 2) {
	if (/^\s*\d*\s+if\s*\((.*)/i) {
	  # Recursive call:
	  $self->readf77if($file, "     & ($1") ;
	}
	if (/^\s*\d*\s+call\s+(\w+)(.*)/i) {
	  $self->readf77call($file, "$1", $2) ;
	} elsif (/^\s*\d*\s+(\w+)\s*=\s*(\w+)\((.*)/i  && !exists $self->{"ALLDEFS"}->{"$2"}) {
	  # Improved analysis: recursive function call for specific type !
	  my $rett = $1 ;
	  my $sec = $2 ;
	  my $line = $3 ;
	  if (/^(\s*\d*\s+$rett\s*=\s*)/) {
	    $chtml->addDef("$1") ;
	  }
	  $_ = $line ;
	  $self->readf77func($file, "$rett", "$sec", "$line") ;
	  while (/^(\s*[\*\+\-\/]\s*)(\w+)\((.*)/i  && !exists $self->{"ALLDEFS"}->{"$2"}) {
	    $_ = $3 ;
	    $chtml->addDef("$1") ;
	    $self->readf77func($file, "$rett", "$2", "$3") ;
	  }
	}
      }
      if (defined($_ = <$file>)) {
	next if (/^c/i) ;
	$_ = $1 if (/^(.*)\!/) ;
      } else {
	print STDOUT (caller(0))[3] . ": error1 line=$. rout='" . $self->get(-name) . "' file='" . $self->get(-file) . "'\n" if $debug || $main::opt_v ;
	$cont = 0 ;
	return ;
      }
    }
  }
}
###########################################################################
sub readf77subr {
  my $self = shift ;
  my $file = shift ;
  $_ = shift ;
  my $cont = shift ;
  $cont = 1 unless $cont ;
  my $doif = 0 ;
  my $parenthe = 0 ;
  my $num ;
  my(@aloops) = () ;
  my $preif = 0 ;
  my(@ploops) = () ;
  my $str ;
  my $c ;
  my $nbparenthe = 0 ;
  my $chtml = &chtml::getCurrent() ;
  if ((ref $file) && (fileno $file)) {
    my $name = $self->get(-name) ;
    # Sets the line number for definition:
    $self->set(-number => $.) unless $self->get(-number) ;
    while($cont) {
      # Parsing an f77 rout file below:
      my $cc = "" ;
      $c = 0 ;
      if (/^(\#.*?)/) {
	# Some preprocessing stuff: passed to later analysis
      } elsif (/^c(.*)/i) {
	$cc = $1 ;
	$_ = "" ;
      } elsif (/^(.*?)\!(.*)/) {
	$_ = $1 ;
	$cc = $2 ;
	$c = 1 ;
      }
      if (/^(\s*)(\d+)(\s*)(.*)/) {
	$chtml->addDef($1) ;
	my $t = $2 ;
	my $n = $3 ;
	$_ = $4 ;
	if ($t) {
	  $chtml->addRaw("<A Name=" . $self->get(-name) . "_L_$2>$2</A>") ;
	}
	if ($doif) {
	  my $ok = 0 ;
	  while($num = pop(@aloops)) {
	    $ok++ ;
	    if ($num =~ /^$t\:(.*)/) {
	      $chtml->addRaw("[<A Href=#C_$1>$ok</A>]") ;
	    } else {
	      push(@aloops, $num) ;
	      last ;
	    }
	  }
	}
	$chtml->addDef($n) ;
      }
      if (/^\s*end/i) {
	if (/^(\s*)(end\s*do)(.*)\b/i) {
	  $num = pop(@aloops) ;
	  $chtml->addDef($1) ;
	  if ($num && $2) {
	    $chtml->addRaw("<A Href=#C_$num>$2</A>") ;
	    $doif-- ;
	  } else {
	    $chtml->addRaw("$2") ;
	    print STDOUT (caller(0))[3] . " Bad pos $_ file ".$self->get(-file)." line $.\n" ;
	  }
	  $chtml->addDef($3) ;
	  print STDOUT (caller(0))[3] . "Retrieve do $_ line $., doif=$doif\n" if $debug ;
	} elsif (/^(\s*)(end\s*if)(.*)/i) {
	  $chtml->addDef($1.$2.$3) ;
	} elsif (/^(\s*)(end\s*select)/i) {
	  $chtml->addDef($1.$2) ;
	} elsif (/^(\s*)(end)\b(\s*)$/i) {
	  # End of routine creation
	  print STDOUT (caller(0))[3] . ": Created subroutine $name line $.\n" if $debug ;
	  $cont = 0 ;
	  $chtml->addDef($1) ;
	  $chtml->addDef($2." ") ;
	  my $ft = $self->get(-global) ;
	  unless ($ft) {
	    my $prok = $self->get(-theproto) ;
	    if ($prok && $prok eq '1') {
	      $ft = "subroutine" ;
	    } elsif (!$prok || ($prok && $prok ne '1' && $prok ne '2')) {
	      $ft = "function" ;
	    } elsif ($prok && "$prok"eq '2') {
	      $ft = "program" ;
	    }
	  }
	  if ($ft) {
	    $ft =~ s/\s+//g ;
	    $chtml->addDef("$ft ") ;
	  }
	  my $refe = new oref(-name => $name, -type => "fthrout") ;
	  $chtml->addRaw($refe) ;
	  $chtml->addDef($3) ;
	} else {
	  $chtml->addDef($_) ;
	}
      } elsif ($cont <= 2 && /^(.{5}[\&\S]\s*)(.*)/) {
	my $rhs = $2 ;
	if ($cont == 1) {
	  $cont = 2 ;
	} else {
	  $chtml->addDef($1) ;
	}
	if ($parenthe == 0) {
	  if ($rhs =~ /^\((.*)/) {
	    $rhs = $1 ;
	    $chtml->addRaw("<ARGLIST>") ;
	    $chtml->addRaw('(') ;
	    $parenthe++ ;
	  }
	}
	my $result = "" ;
	if ($rhs =~ /^(.*?)\)(.*)/) {
	  $chtml->addDef($1) ;
	  $rhs = $1 ;
	  $chtml->addDef(')') ;
	  if ($parenthe > 0) {
	    $chtml->addRaw("</ARGLIST>") ;
	  }
	  $parenthe-- ;
	  $result = $2 ;
	  $cont = 3 ;
	  $_ = $2 ;
	  $chtml->addDef($_) ;
	} else {
	  $chtml->addDef($rhs) if $rhs ;
	}
	$rhs =~ s/\s+//g ;
	my (@array) = split(/\,/, $rhs) ;
	my $s = 1 ;
	foreach my $a (@array) {
	  if ($a =~ /^(.*?)\&.+/) {
	    $s = 0 ;
	    $a = $1 ;
	    $cont = 2 ;
	    print STDOUT "New cont=$cont\n" if $debug ;
	  }
	  if ($a =~ /(\w+)\b/) {
	    # On the left the name used for the call, on the right the
	    # actual routine name in it's module.
	    $self->addArgument($1) ;
	  }
	}
      } elsif ($cont <= 2 && /^\s*$/) {
	# Only comments for fortran 77 file ?
	print STDOUT "Comment line($cont) on '$_'\n" if $debug ;
      } elsif ($cont) {
	if ($cont <= 3) {
	  $cont = 4 ;
	}
	if (/^\s*implicit/i) {
	  if (/^\s*implicit\s+none/i) {
	    # Delete all the implicit keys:
	    $self->ImplicitReset() ;
	    $chtml->addRaw("<DECLARE>") ;
	    $chtml->addDef($_) ;
	    $chtml->addRaw("</DECLARE>") ;
	  } elsif (/^\s*implicit\s+([\w\*]+)(.*?)[\s\:]+\((.*)\)/i) {
	    my $lhs = lc ( $1 . $2 ) ;
	    my $rhs = lc($3) ;
	    my (@array) = split(/\,/, $rhs) ;
	    # TODO: potential major bug:
	    # Managing type declarations conflicts not performed yet.
	    foreach $rhs (@array) {
	      # Builds a regular expression:
	      $self->ImplicitProcess($rhs => $lhs) ;
	    }
	    print STDOUT "Setting implicit $rhs to type $lhs\n" if $debug ;
	    $chtml->addRaw("<DECLARE>") ;
	    $chtml->addDef($_) ;
	    $chtml->addRaw("</DECLARE>") ;
	  } else {
	    $chtml->addDef($_) ;
	  }
	} elsif  (/^\s*interface/i) {
	  if (/^(\s*interface\s+)(\w+)/i) {
	    my $inter ;
	    if ($self->get(-module)) {
	      $inter = new fthinter(-name => $2, -file => $self->get(-file), -handle => $file, -masfic => $self->get(-masfic), -module => $self->get(-module)) ;
	    } else {
	      $inter = new fthinter(-name => $2, -file => $self->get(-file), -handle => $file, -masfic => $self->get(-masfic)) ;
	    }
	    $chtml->addDef($1) ;
	    $chtml->addRaw($inter) ;
	    $inter->read($file) ;
	    $self->addInter($2 => $inter) ;
	  } elsif  (/^(\s*interface\s*)/i) {
	    # Basic interface to global stuff..
	    my $inter ;
	    if ($self->get(-module)) {
	      $inter = new fthinter(-file => $self->get(-file), -handle => $file, -masfic => $self->get(-masfic), -module => $self->get(-module)) ;
	    } else {
	      $inter = new fthinter(-file => $self->get(-file), -handle => $file, -masfic => $self->get(-masfic)) ;
	    }
	    $chtml->addDef($1.' ') ;
	    $chtml->addRaw($inter) ;
	    $inter->read($file) ;
	    my $mname = $inter->get(-name) ;
	    #$self->addInter($mname => $inter) ;
	    $inter->addInter($self) ;
	  }  else {
	    $chtml->addDef($_) ;
	  }
	} elsif (/^(\s*)(.*?)\s*::\s*(.*)/) {
	  # Fortran 90 types declaration below:
	  # print STDOUT "** Adding ref($.) = '$1' on $chtml\n" ;
	  $chtml->addDef($1) ;
	  my $lhs = $2 ; my $rhs = $3 ;
	  $self->parseDef90($lhs, $rhs) ;
	  # print STDOUT "Parsing defs $1 $lhs :: $rhs\n" ;
	} elsif (/^(\s*)(integer|real|character|complex|logical|dimension)(.*?)[\s\:]+([\w\,]+.*)/i || /^(\s*)(double)(\s+precision)[\s\:]+([\w\,]+.*)/i) {
	  # Fortran 77 types declaration compatible with fortran 90
	  my $lhs = $2 . $3 ; my $rhs = $4 ;
	  $chtml->addDef("$1") ;
	  $self->parseDef77($lhs, $rhs) ;
	} elsif (/^(\s*common\s*)\/(\s*\w+\s*)\/(.*)/i) {
	  $chtml->addRaw("<DECLARE>$1</DECLARE>".'/'."<VARIABLE>$2</VARIABLE>".'/'.$3) ;
	} elsif (/^(\s*common)\b(.*)/i) {
	  $chtml->addRaw("<DECLARE>$1</DECLARE>$2") ;
	} elsif (/^(\s*parameter\s*)\(([\w\s\,\=]+)\)(.*)/i) {
	  # List $2 may be of the form "variable=value,"...
	  $chtml->addRaw("<DECLARE>$1</DECLARE>".'('."<VARIABLE>$2</VARIABLE>".')'.$3) ;
	} elsif (/^(\s*parameter\s*)\(([\w\s\,\=]+)(.*)/i) {
	  $chtml->addRaw("<DECLARE>$1</DECLARE>".'('."<VARIABLE>$2</VARIABLE>".$3) ;
	} elsif (/^\s*parameter\b/i) {
	  $chtml->addDef($_) ;
	} elsif (/^(\s*equivalence\s*)\((\w+.*?)\,(\w+.*?\)*)\)(.*)/i) {
	  $chtml->addRaw("<DECLARE>$1</DECLARE>".'('."<VARIABLE>$2,$3</VARIABLE>".')'.$4) ;
	} elsif (/^\s*equivalence\b/i) {
	  $chtml->addDef($_) ;
	} elsif (/^(\s*import\s+)(\w+)(.*)/i) {
	  # This is not a type setting of course
	  $chtml->addRaw("<DECLARE>$1</DECLARE><VARIABLE>$2</VARIABLE>".$3) ;
	} elsif (/^(\s*use\s+)(\w+)\s*,\s*only\s*:\s*(.*)/i) {
	  my $fmused = $2 ; $_ = $3 ;
	  $chtml->addDef("$1") ;
	  my $refe = new oref(-name => $2, -type => "fthmod") ;
	  $chtml->addRaw($refe) ;
	  $chtml->addDef(", only : ") ;
	  $self->readUses($file, $fmused, "&" . $_) ;
	} elsif (/^(\s*use\s+)(\w+)\s*$/i) {
	  my $fmused = lc($2) ;
	  $chtml->addDef("$1") ;
	  my $refe = new oref(-name => $2, -type => "fthmod") ;
	  $self->{"ALLUSES"}->{"$fmused"} = 1 ;
	  $chtml->addRaw($refe) ;
	} elsif (/^\s*$/) {
	  # Only comments 
	} elsif (/^\#(ifdef|ifndef|if|elif|else|endif)(.*)/i) {
	  # Preprocessing tests, this seems good code below:
	  if ($1 eq "endif") {
	    $num = pop(@ploops) ;
	    if ($num) {
	      $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	      $preif-- ;
	    } else {
	      $chtml->addDef('#'."$1$2") ;
	    }
	  } elsif ($1 eq "if" || $1 eq "ifdef" || $1 eq "ifndef") {
	    push(@ploops, $.) ;
	    $preif++ ;
	    $chtml->addRaw("<TEST><A Name=C_$.>".'#'."$1</A></TEST>$2") ;
	  } else {
	    $num = pop(@ploops) ;
	    if ($num) {
	      push(@ploops, $num) ;
	      $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	    } else {
	      $chtml->addDef('#'."$1$2") ;
	    }
	  }
	  $_ = "" ;
	} elsif (/^(\s*include\s+)([\"\'])([\w\.\/]+)\2(.*)/i || /^(\#include\s+)([\"\'])([\w\.\/]+)\2(.*)/i) {
	  # There may be includes inside routines definitions, especially for f77 common:
	  my $incf = $3 ;
	  $chtml->addDef("$1$2") ;
	  $chtml->addRaw("<A Name=C_$.></A>");
	  unless (-f "$incf") {
	    my $base = dirname($self->get(-file)) ;
	    if (-f "$base/$incf") {
	      $incf = "$base/$incf" ;
	    } elsif (-f "$base/" . basename($incf)) {
	      $incf = "$base/" . basename($incf) ;
	    } elsif ($main::opt_I) {
	      # Possible...
	      my @TAGSHTML_PATH = (split ( /:/, $main::opt_I)) ;
	      foreach $base (@TAGSHTML_PATH) {
		if (-f "$base/$incf") {
		  $incf = "$base/$incf" ;
		  last ;
		}
	      }
	    }
	  }
	  if (-f "$incf") {
	    my $fh ;
	    my $other ;
	    if (open($fh,"<$incf")) {
	      print STDOUT (caller(0))[3] . ": Open $incf from $file\n" if $debug ;
	      $other = &fthinc::getObject($incf) ;
	      if ($other) {
		new chtml() ;
	      } else {
		new chtml(-file => $incf, -name => $incf) ;
		$other = new fthinc(-file => "$incf", -name => "$incf", -masfic => $self->get(-masfic), -father => $self) ;
	      }
	      $self->addInclude("$incf" => $other) ;
	      $chtml->addRaw($other) ;
	      $chtml->addDef("$2$4") ;
	      $self->readf77subr($fh, "", $cont) ;
	      close($fh);
	      &chtml::setCurrent($chtml) ;
	      print STDOUT (caller(0))[3] . ": Closed $incf($.) from $file\n" if $debug ;
	    }
	  } else {
	    my $refe = new oref(-name => $3, -type => "fthinc") ;
	    $chtml->addRaw($refe) ;
	    $chtml->addDef("$2$4") ;
	  }
	} elsif (/^\#/) {
	  # Preprocessing infos:
	  $chtml->addDef($_) ;
	} else {
	  if (/do/i) {
	    if (/^(\s*)(do)(\s*)(\D\w+)(\s*\=.*)/i) {
	      # NAJ: not counted as a do statement yet
	      $chtml->addDef($1) ;
	      $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
	      $chtml->addDef($3) ;
	      $_ = $4.$5 ;
	      $doif++ ;
	      print "Add1 do $_ line $., file ".$self->get(-file)." doif=$doif\n" if $debug ;
	      push(@aloops, $.) ;
	    } elsif (/^(\s*)(do)(\s*while)(.*)/i) {
	      $chtml->addDef($1) ;
	      $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
	      $chtml->addDef($3) ;
	      $_ = $4 ;
	      $doif++ ;
	      print "Add1 do $_ line $., doif=$doif\n" if $debug ;
	      push(@aloops, $.) ;
	    } elsif (/^(\s*\w+\s*\:\s*)(do)\b(\s*)(.*)/i) {
	      $chtml->addDef($1) ;
	      $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
	      $chtml->addDef($3) ;
	      $_ = $4 ;
	      $doif++ ;
	      print "Add1 do $_ line $., doif=$doif\n" if $debug ;
	      push(@aloops, $.) ;
	    } elsif (/^(\s*)(do)(\s*)(\d+)(\s*\w+\s*\=.*)/i) {
	      $chtml->addDef($1) ;
	      $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
	      $chtml->addDef($3) ;
	      # Add a link to the label 4:
	      $chtml->addRaw("<A Href=#" . $self->get(-name) . "_L_$4>$4</A>") ;
	      $_ = $5 ;
	      $doif++ ;
	      print "Add1 do $_ line $., doif=$doif\n" if $debug ;
	      push(@aloops, "$4:" . $.) ;
	    } elsif (/^(\s*)(do)(\s*)$/i) {
	      # NAJ: not counted as a do statement yet
	      $chtml->addDef($1) ;
	      $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
	      $_ = $3 ;
	      $doif++ ;
	      print "Add1 do $_ line $., doif=$doif\n" if $debug ;
	      push(@aloops, $.) ;
	    }
	  }
	  if (/^(\s*)(if)(\s*\(.*)/i) {
	    #$self->readf77if($file, "     & ($1") ;
	    $chtml->addDef($1) ;
	    $chtml->addRaw("<TEST>") ;
	    $chtml->addDef("$2") ;
	    $chtml->addRaw("</TEST>") ;
	    $_ = $3 ;
	  } elsif (/^(\s*)(go\s*to\s*)(\d+)(.*)/i) {
	    $chtml->addDef($1.$2) ;
	    $chtml->addRaw("<A Href=#" . $self->get(-name) . "_L_$3>$3</A>") ;
	    $_ = $4 ;
	  }
	  if (!$main::reuse) {
	    if (/^(\s*)(write|print|stop)(.*)/i) {
	      # Writing line is not of much interest for us...
	      $chtml->addDef($_) ;
	    } elsif (/^(.*?call\s+)(\w+)(.*)/i) {
	      # Calls a routine:
	      $chtml->addDef($1) ;
	      $self->readf77call($file, "$2", "$3") ;
	      $chtml->addDef($_) ;
	    } elsif (/^\s*(\w+)\s*=\s*(\w+)\((.*)/i  && !$self->sayType($2) && !&fthrout::isIntrinsec($2)) {
	      # Improved analysis: recursive function call !
	      my $rett = $1 ;
	      my $sec = $2 ;
	      my $line = $3 ;
	      if (/^(\s*$rett\s*=\s*)/) {
		$chtml->addDef("$1") ;
	      }
	      $_ = $line ;
	      my $ocall = $self->readf77func($file, "$rett", "$sec", "$line") ;
	      while (/^(\s*[\*\+\-\/]\s*)([\w]+)\((.*)/i  && !$self->sayType($2) && !&fthrout::isIntrinsec($2)) {
		$_ = $3 ;
		$chtml->addDef("$1") ;
		$self->readf77func($file, "$rett", "$2", "$3") ;
	      }
	      $chtml->addDef($_) ;
	    } else {
	      $chtml->addDef($_) ;
	    }
	  } elsif ($main::reuse) {
	    # Restart mode
	    if ($c && $cc && ($str || /[\'\"]/)) {
	      $_ .= '!' . $cc ;
	      $cc = "" ;
	    }
	    if (/^(.*?)([\'\"\(\)\!])(.*)/) {
	      my $ligne = "" ;
	      while (/^(.*?)([\'\"\(\)\!])(.*)/) {
		my $bef = $1 ;
		my $sep = $2 ;
		my $aft = $3 ;
		$_ = $aft ;
		if ($bef =~ /^(.*?)\\\\$/) {
		  print STDOUT "\t\t Antislash special character encountered\n" if $debug ;
		} elsif ($bef =~ /^(.*?)\\$/) {
		  if ($str) {
		    $ligne .= "$bef$sep" ;
		    if ($bef =~ /[\<\>]/) {
		      $bef =~ s/\</&lt;/g ;
		      $bef =~ s/\>/&gt;/g ;
		    }
		  } else {
		    # Should never happen
		    print STDOUT (caller(0))[3] .": Strange event file ".$self->get(-file)." line $. text = <$bef$sep$aft>\n" ;
		    $ligne .= "$bef$sep" ;
		  }
		  next ;
		}
		# Continuing FROM
		if ($str && $sep eq $str) {
		  if ($bef =~ /[\<\>]/) {
		    $bef =~ s/\</&lt;/g ;
		    $bef =~ s/\>/&gt;/g ;
		  }
		  $str = "" ;
		  $ligne .= "$bef$sep</STRING>" ;
		} elsif ($str) {
		  # Nothing special, continuing the string build
		  if ($bef =~ /[\<\>]/) {
		    $bef =~ s/\</&lt;/g ;
		    $bef =~ s/\>/&gt;/g ;
		  }
		  $ligne .= "$bef$sep" ;
		} elsif ($sep eq '"' || $sep eq "'") {
		  $str = $sep ;
		  $ligne .= "$bef<STRING>$sep" ;
		} elsif ($sep eq '!') {
		  $cc = $aft ;
		  $_ = "" ;
		  last ;
		} elsif ($sep eq '(' || $sep eq ')') {
		  if ($sep eq '(') {
		    $nbparenthe++ ;
		    if ($bef =~ /(.*?call\s+)(\w+)$/i) {
		      $chtml->addRaw($ligne.$1) ;
		      $ligne = "" ;
		      print STDOUT (caller(0))[3] .": Creating Fortran call $2\n" if $debug ;
		      $self->readf77call($file, "$2", "$sep$aft") ;
		      $nbparenthe-- ;
		    } elsif ($bef =~ /\s*([\%\w]+)\s*=\>{0,1}\s*(\w+)$/i && $self->sayCalling($2)) { # !$self->sayType($2) && !&fthrout::isIntrinsec($2)) {
		      my $rett = $1 ;
		      my $sec = $2 ;
		      # This method may lead to using arrays instead of functions...
		      if ($bef =~ /(.*)\b$sec\b/) {
			$chtml->addRaw("$ligne$1") ;
		      }
		      #if ($bef =~ /^(.*?$rett\s*=\>{0,1}\s*)$sec$/) {
		      #	$chtml->addRaw("$ligne$1") ;
		      #	$ligne = "" ;
		      #}
		      print STDOUT (caller(0))[3] .": Creating legacy Fortran routine $sec from $bef line $.\n" if $debug ;
		      my $ocall = $self->readf77func($file, "$rett", "$sec", "$aft") ;
		      $nbparenthe-- ;
		      while (/^(\s*[\*\+\-\/]\s*)([\w]+)\((.*)/i  && $self->sayCalling($2)) { # !$self->sayType($2) && !&fthrout::isIntrinsec($2)) {
			$_ = $3 ;
			$chtml->addRaw("$1") ;
			print STDOUT (caller(0))[3] .": Creating Fortran routine $2\n" if $debug ;
			$self->readf77func($file, "$rett", "$2", "$3") ;
		      }
		    } elsif ($bef =~ /^(.*?)(\w+)$/i && $self->sayCalling($2)) {# !$self->sayType($2) && $self->sayCalling($2)) {
		      my $rett = $1 ;
		      my $sec = $2 ;
		      $chtml->addRaw("$ligne$1") ;
		      $ligne = "" ;
		      print STDOUT (caller(0))[3] .": Creating auto Fortran routine $sec from $bef line $.\n" if $debug ;
		      my $ocall = $self->readf77func($file, "void", "$sec", "$aft") ;
		      $nbparenthe-- ;
		    } else {
		      $ligne .= "$bef$sep" ;
		    }
		  } elsif ($sep eq ')') {
		    $ligne .= "$bef$sep" ;
		    $nbparenthe-- ;
		  }
		}
	      }
	      if ($_ && /(.*?call\s+)(\w+)(.*)/i) {
		print STDOUT (caller(0))[3] .": Creating Fortran call $2\n" if $debug ;
		$chtml->addRaw($ligne.$1) ;
		$ligne = "" ;
		$self->readf77call($file, "$2", "$3") ;
		# Calls a routine:
	      } else {
		$chtml->addRaw($ligne.$_) ;
		$ligne = "" ;
	      }
	    } elsif (/^(.*?call\s+)(\w+)(.*)/i) {
	      print STDOUT (caller(0))[3] .": Creating Fortran call $2 line $.\n" if $debug ;
	      $chtml->addRaw($1) ;
	      $self->readf77call($file, "$2", "$3") ;
	    } else {
	      $chtml->addDef($_) ;
	    }
	  }
	}
      }
      if ($cc) {
	if ($cc =~ /^\|\_*S*(.*)/) {
	  $cc = $1 ;
	}
      }
      if ($cc) {
	$chtml->addRaw('<COMMENT>') ;
	if ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  $chtml->addLaCoM("<A Name=C_$.></A>") ;
	}
	while ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  my $incp = "$3.$4" ;
	  $cc = $5 ;
	  $chtml->addLaTeX($1.$2." ") ;
	  unless (-f "$incp") {
	    my $base = dirname($self->get(-file)) ;
	    my $nase = basename(dirname($incp)) ;
	    if (-f "$base/$incp") {
	      $incp = "$base/$incp" ;
	    } elsif (-f "$base/" . basename($incp)) {
	      $incp = "$base/" . basename($incp) ;
	    } elsif (-f "$nase/" . basename($incp)) {
	      $incp = "$nase/" . basename($incp) ;
	    } elsif ($main::opt_L) {
	      # LaTeX inputs directories
	      # Possible...
	      my @TAGSHTML_PATH = (split ( /:/, $main::opt_L)) ;
	      foreach $base (@TAGSHTML_PATH) {
		if (-f "$base/$incp") {
		  $incp = "$base/$incp" ;
		  last ;
		}
	      }
	    }
	  }
	  if (-f "$incp") {
	    my $other = new fthinp(-file => "$incp", -name => "$incp") ;
	    $self->addInput("$incp" => $other) ;
	    $chtml->addLaCoM($other) ;
	  } else {
	    my $refe = new oref(-name => $incp, -type => "fthinp") ;
	    $chtml->addLaCoM($refe) ;
	  }
	}
	$chtml->addLaTeX($cc) ;
	$chtml->addRaw("</COMMENT>") ;
      }
      return unless $cont ;
      if (defined($_ = <$file>)) {
	chomp ;
      } else {
	print STDOUT (caller(0))[3] . ": error2 doif=$doif, aloops=@aloops, line=$. rout='" . $self->get(-name) . "' file='" . $self->get(-file) . "'\n" if $debug || $main::opt_v;
	return ;
      }
    }
  }
}
###########################################################################
sub parseDef77 {
  my $self = shift ;
  my $ilhs = shift ;
  my $irhs = shift ;
  my $lhs = lc($ilhs) ;
  my $rhs = lc($irhs) ;
  my $shs ;
  my $name = $self->get(-name) ;
  my $chtml = &chtml::getCurrent() ;
  if ($lhs eq 'public' || $lhs eq 'private') {
    # This is not a type setting indeed:
    $chtml->addRaw("<DECLARE>$ilhs</DECLARE> <VARIABLE>$irhs</VARIABLE>") ;
    return ;
  }
  $lhs =~ s/\s+//g ;
  $rhs =~ s/\s+//g ;
  # Fortran argument definition equivalent form:
  $lhs =~ s/,*intent\(.*?\)//g ;
  $lhs =~ s/,*allocatable//g ;
  $lhs =~ s/,*save//g ;
  # Special fortran 77 dimension command means real type.
  print STDOUT  (caller(0))[3] . ": Parsing lhs=<$lhs> \n" if $debug ;
  if ($lhs =~ /^dimension(.*)/i) {
    if ($1) {
      $lhs = "$1" ;
    } else {
      $lhs = "" ;
    }
  }
  if ($lhs =~ /^(.*?),*dimension\((.*?)\)(.*)/i) {
    my (@array) = split(/\,/, $2) ;
    $lhs = $1 . $3 . ',dimension(' . ',' x ( $#array + 0 ) . ')' ;
    $lhs =~ s/,*pointer//g ;
  }
  while ($rhs =~ /(.*)\((.*?)\)(.*)/) {
    my $a = $1 ;  my $b = $2 ;  my $c = $3 ;  
    $b =~ s/\,/\;/g ;
    $rhs = $a . '{' . $b . '}' . $c ;
  }
  my (@array) = split(/\,/, $rhs) ;
  foreach $rhs (@array) {
    if ($rhs =~ /(.*)\s*\=\s*([\w\.\+\-\*\/]*)/) {
      $rhs = $1 ;
    }
    if ($rhs =~ /(\w+)\{(.*?)\}/) {
      $rhs = $1 ;
      my (@larray) = split(/\,/, $2) ;
      $shs = $lhs . ',dimension(' . ',' x ( $#larray + 0 ) . ')' ; ;
    } else {
      $shs = $lhs ;
    }
    unless ($lhs) {
      # Type is implicit:
      $shs = $self->defType($rhs) . $shs ;
    }
    print STDOUT (caller(0))[3] . ": In " . $self->get(-name) . " line $. defined var=$rhs is of type $shs\n" if $debug ;
    if ($self->{"ARGHASH"}->{"$rhs"}) {
      $self->{"ARGTYPES"}->{"$rhs"} = $shs ;
    } elsif (lc($rhs) eq lc($name) && ! exists $self->{"theproto"}) {
      $self->addArgument($name) ;
      $self->{"ARGTYPES"}->{"$name"} = $shs ;
      $self->{"theproto"} = $name ;
    } 
    $self->{"ALLDEFS"}->{"$rhs"} = $shs ;
  }
  if ($ilhs =~ /^(.*type\s*\(\s*)(\w+)(\s*\).*)/i) {
    # print STDOUT "** Adding ref($.) = '$1' on $chtml\n" ;
    $chtml->addRaw("<DECLARE>") ;
    $chtml->addDef("$1") ;
    my $refe = new oref(-name => $2, -type => "fthtype") ;
    # print STDOUT "Refering to another type object $2: $refe\n" ;
    $chtml->addRaw($refe) ;
    $chtml->addDef("$3") ;
    $chtml->addRaw("</DECLARE>") ;
    $chtml->addRaw(" <VARIABLE>$irhs</VARIABLE>") ;
  } elsif ($ilhs =~ /^(.*class\s*\(\s*)(\w+)(\s*\).*)/i) {
    # print STDOUT "** Adding ref($.) = '$1' on $chtml\n" ;
    $chtml->addRaw("<DECLARE>") ;
    $chtml->addDef("$1") ;
    my $refe = new oref(-name => $2, -type => "fthtype") ;
    # print STDOUT "Refering to another type object $2: $refe\n" ;
    $chtml->addRaw($refe) ;
    $chtml->addDef("$3") ;
    $chtml->addRaw("</DECLARE>") ;
    $chtml->addRaw(" <VARIABLE>$irhs</VARIABLE>") ;
  } else {
    #print STDOUT "Parsing($.) $ilhs :: $irhs\n" ;
    $chtml->addRaw("<DECLARE>$ilhs</DECLARE> <VARIABLE>$irhs</VARIABLE>") ;
  }
}
###########################################################################
sub parseDef90 {
  my $self = shift ;
  my $ilhs = shift ;
  my $irhs = shift ;
  my $lhs = lc($ilhs) ;
  my $rhs = lc($irhs) ;
  my $shs ;
  my $name = $self->get(-name) ;
  my $chtml = &chtml::getCurrent() ;
  if ($lhs eq 'public' || $lhs eq 'private') {
    # This is not a type setting indeed:
    $chtml->addRaw("<DECLARE>$ilhs</DECLARE> :: <VARIABLE>$irhs</VARIABLE>") ;
    return ;
  }
  $lhs =~ s/\s+//g ;
  $rhs =~ s/\s+//g ;
  # Fortran argument definition equivalent form:
  $lhs =~ s/,*intent\(.*?\)//g ;
  $lhs =~ s/,*allocatable//g ;
  $lhs =~ s/,*save//g ;
  if ($lhs =~ /^(.*?),*dimension\((.*?)\)(.*)/i) {
    my (@array) = split(/\,/, $2) ;
    $lhs = $1 . $3 . ',dimension(' . ',' x ( $#array + 0 ) . ')' ;
    $lhs =~ s/,*pointer//g ;
  }
  while ($rhs =~ /(.*)\((.*?)\)(.*)/) {
    my $a = $1 ;  my $b = $2 ;  my $c = $3 ;  
    $b =~ s/\,/\;/g ;
    $rhs = $a . '{' . $b . '}' . $c ;
  }
  my (@array) = split(/\,/, $rhs) ;
  foreach my $rhs (@array) {
    if ($rhs =~ /(.*)\s*\=\>{0,1}\s*([\w\.\+\-\*\/]*)/) {
      $rhs = $1 ;
    }
    if ($rhs =~ /(\w+)\{(.*?)\}/) {
      $rhs = $1 ;
      my (@larray) = split(/\,/, $2) ;
      $shs = $lhs . ',dimension(' . ',' x ( $#larray + 0 ) . ')' ; ;
    } else {
      $shs = $lhs ;
    }
    print STDOUT (caller(0))[3] . ": In " . $self->get(-name) . " line $. defined var=$rhs is of type $shs\n" if $debug ;
    if ($self->{"ARGHASH"}->{"$rhs"}) {
      $self->{"ARGTYPES"}->{"$rhs"} = $shs ;
    } elsif (lc($rhs) eq lc($name) && ! exists $self->{"theproto"}) {
      $self->addArgument($name) ;
      $self->{"ARGTYPES"}->{"$name"} = $shs ;
      $self->{"theproto"} = $name ;
    } 
    $self->{"ALLDEFS"}->{"$rhs"} = $shs ;
    # print STDOUT (caller(0))[3] .": Args of ".$self->get(-name)." is $rhs :: $lhs\n" if $debug ;
  }
  if ($ilhs =~ /^(.*type\s*\(\s*)(\w+)(\s*\).*)/i) {
    # print STDOUT "** Adding ref($.) = '$1' on $chtml\n" ;
    $chtml->addRaw("<DECLARE>") ;
    $chtml->addDef("$1") ;
    my $refe = new oref(-name => $2, -type => "fthtype") ;
    # print STDOUT "Refering to another type object $2: $refe\n" ;
    $chtml->addRaw($refe) ;
    $chtml->addDef("$3") ;
    $chtml->addRaw("</DECLARE>") ;
    $chtml->addRaw(" :: <VARIABLE>$irhs</VARIABLE>") ;
  } elsif ($ilhs =~ /^(.*class\s*\(\s*)(\w+)(\s*\).*)/i) {
    # print STDOUT "** Adding ref($.) = '$1' on $chtml\n" ;
    $chtml->addRaw("<DECLARE>") ;
    $chtml->addDef("$1") ;
    my $refe = new oref(-name => $2, -type => "fthtype") ;
    # print STDOUT "Refering to another type object $2: $refe\n" ;
    $chtml->addRaw($refe) ;
    $chtml->addDef("$3") ;
    $chtml->addRaw("</DECLARE>") ;
    $chtml->addRaw(" :: <VARIABLE>$irhs</VARIABLE>") ;
  } else {
    #print STDOUT "Parsing($.) $ilhs :: $irhs\n" ;
    $chtml->addRaw("<DECLARE>$ilhs</DECLARE> :: <VARIABLE>$irhs</VARIABLE>") ;
  }
}
###########################################################################
sub readf90subr {
  my $self = shift ;
  my $file = shift ;
  $_ = shift ;
  my $cont = shift ;
  $cont = 1 unless $cont ;
  my $contains = 0;
  my $parenthe = 0 ;
  my $doif = 0 ;
  my $num ;
  my(@aloops) = () ;
  my $preif = 0 ;
  my(@ploops) = () ;
  my $str ;
  my $nbparenthe = 0 ;
  my $chtml = &chtml::getCurrent() ;
  if ((ref $file) && (fileno $file)) {
    my $name = $self->get(-name) ;
    # Sets the line number for definition:
    $self->set(-number => $.) unless $self->get(-number) ;
    while($cont) {
      # Parsing an f90 rout file below:
      my $cc = "" ;
      if (/^\#(ifdef|ifndef|if|elif|else|endif)(.*)/i) {
	# Preprocessing tests, this seems good code below:
	if ($1 eq "endif") {
	  $num = pop(@ploops) ;
	  if ($num) {
	    $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	    $preif-- ;
	  } else {
	    $chtml->addDef('#'."$1$2");
	  }
	} elsif ($1 eq "if" || $1 eq "ifdef" || $1 eq "ifndef") {
	  push(@ploops, $.) ;
	  $preif++ ;
	  $chtml->addRaw("<TEST><A Name=C_$.>".'#'."$1</A></TEST>$2") ;
	} else {
	  $num = pop(@ploops) ;
	  if ($num) {
	    push(@ploops, $num) ;
	    $chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>");
	  } else {
	    $chtml->addDef('#'."$1$2") ;
	  }
	}
	$_ = "" ;
      } elsif (/^(.*?)\!(.*)/) {
	$_ = $1 ;
	$cc = $2 ;
      }
      if (/^(\s*)(\d+)(\s*)(.*)/) {
	$chtml->addDef($1) ;
	my $t = $2 ;
	my $n = $3 ;
	$_ = $4 ;
	if ($2) {
	  $chtml->addRaw("<A Name=" . $self->get(-name) . "_L_$2>$2</A>") ;
	}
	if ($doif) {
	  my $ok = 0 ;
	  while($num = pop(@aloops)) {
	    $ok++ ;
	    if ($num =~ /^$t\:(.*)/) {
	      $chtml->addRaw("[<A Href=#C_$1>$ok</A>]") ;
	    } else {
	      push(@aloops, $num) ;
	      last ;
	    }
	  }
	}
	$chtml->addDef($n) ;
      }
      if (/^\s*end/i) {
	if (/^(\s*)(end\s*do)(.*)\b/i) {
	  $num = pop(@aloops) ;
	  $chtml->addDef($1) ;
	  if ($num && $2) {
	    $chtml->addRaw("<A Href=#C_$num>$2</A>") ;
	    $doif-- ;
	  } else {
	    $chtml->addRaw("$2") ;
	    print STDOUT (caller(0))[3] . "Bad pos $_ file ".$self->get(-file)." line $.\n" ;
	  }
	  $chtml->addDef($3) ;
	  print STDOUT (caller(0))[3] . "Retrieve do $_ line $., doif=$doif\n" if $debug ;
	} elsif (/^(\s*)(end\s*if)(.*)/i) {
	  $chtml->addDef($1.$2.$3) ;
	} elsif (/^(\s*)(end\s*select)/i) {
	  $chtml->addDef($1.$2) ;
	} elsif (/^(\s*end\s*)(subroutine|function|program)/i) {
	  # End of rout creation
	  print STDOUT (caller(0))[3] . ": Terminated subroutine $name at $. ".$self->get(-number)."\n" if $debug ;
	  if ($cont <= 3) {
	    # Means dummy routine indeed...
	    $cont = 4 ;
	    # $chtml->addDef2("</ARGLIST>",-1) ;
	  }
	  $cont = 0 ;
	  $chtml->addDef("$1$2 ") ;
	  my $refe ;
	  if ($self->get(-module)) {
	    $refe = new oref(-name => $self->get(-module) . '::' . $name, -type => "fthrout") ;
	  } else {
	    $refe = new oref(-name => $name, -type => "fthrout") ;
	  }
	  $chtml->addRaw($refe) ;
	} elsif (/^(\s*)(end)\b(\s*)$/i) {
	  # Illegal end style:
	  # End of rout creation
	  print STDOUT (caller(0))[3] . ": Terminated subroutine $name\n" if $debug ;
	  $cont = 0 ;
	  $chtml->addDef($1) ;
	  $chtml->addDef($2 . " ") ;
	  # Only global procedures may end up like this:
	  my $ft = $self->get(-global) ;
	  unless ($ft) {
	    # Module procedures know their type:
	    my $prok = $self->get(-theproto) ;
	    if ($prok && $prok eq '1') {
	      $ft = "subroutine" ;
	    } elsif (!$prok || ($prok && $prok ne '1' && $prok ne '2')) {
	      $ft = "function" ;
	    } elsif ($prok && "$prok"eq '2') {
	      $ft = "program" ;
	    }
	  }
	  if ($ft) {
	    # Remove the spaces:
	    $ft =~ s/\s+//g ;
	    # $ft .= (caller(1))[3] . ':' . (caller(0))[2] ;
	    $chtml->addDef("$ft ") ;
	  }
	  my $refe = new oref(-name => $name, -type => "fthrout") ;
	  $chtml->addRaw($refe) ;
	} else {
	  $chtml->addDef($_) ;
	}
      } elsif ($cont <= 2 && /^\s*$/) {
	# Only comments for fortran 90 file, process later on
      } elsif ($cont <= 2 && /^(\s*\&\s*)(.*)/ || $cont == 2 && /^(.*?)(.*)/) {
	my $rhs = $2 ;
	if ($cont == 1) {
	  $cont = 2 ;
	} else {
	  $chtml->addDef($1) if $1 ;
	}
	if ($parenthe == 0) {
	  if ($rhs =~ /^\((.*)/) {
	    $rhs = $1 ;
	    $chtml->addRaw("<ARGLIST>") ;
	    $chtml->addRaw('(') ;
	    $parenthe++ ;
	  }
	}
	$cont = 3 ;
	my $result = "" ;
	if ($rhs =~ /^(.*?)\)(.*)/) {
	  $chtml->addDef($1) ;
	  $rhs = $1 ;
	  $chtml->addDef(')') ;
	  if ($parenthe > 0) {
	    $chtml->addRaw("</ARGLIST>") ;
	  }
	  $parenthe-- ;
	  $result = $2 ;
	  $_ = $2 ;
	  if ($result =~ /\&\s*$/) {
	    $cont = 2 ;
	  }
	  $chtml->addDef($_) ;
	} elsif ($rhs =~ /\&\s*$/) {
	  $cont = 2 ;
	  $chtml->addDef($rhs) ;
	} else {
	  $chtml->addDef($rhs) if $rhs ;
	}
	$rhs =~ s/\s+//g ;
	my (@array) = split(/\,/, $rhs) ;
	my $s = 1 ;
	foreach my $a (@array) {
	  if ($a =~ /(\w+)\b/) {
	    # On the left the name used for the call, on the right the
	    # actual routine name in it's module.
	    $self->addArgument($1) ;
	  }
	}
	if ($result =~ /\s*result\s*\(\s*(.*)\s*\)/i) {
	  # Last argument equals the routine name...
	  $self->set(-theproto => "$1") ;
	  $self->addArgument("$1") ;
	  print STDOUT (caller(0))[3] .": Adding function result $1 to ".$self->get(-name)." to list @{$self->{'ARGORDER'}}\n" if $debug ;
	}
      } else {
	if ($cont <= 3) {
	  $cont = 4 ;
	}
	if ($cont <= 4) {
	  if (/^\s*implicit/i) {
	    if (/^\s*implicit\s+none/i) {
	      # Delete all the implicit keys:
	      $self->ImplicitReset() ;
	      $chtml->addRaw("<DECLARE>") ;
	      $chtml->addDef($_) ;
	      $chtml->addRaw("</DECLARE>") ;
	    } elsif (/^\s*implicit\s+([\w\*]+)(.*?)[\s\:]+\((.*)\)/i) {
	      my $lhs = lc ( $1 . $2 ) ;
	      my $rhs = lc($3) ;
	      my (@array) = split(/\,/, $rhs) ;
	      foreach $rhs (@array) {
		# Builds a regular expression:
		$self->ImplicitProcess($rhs => $lhs) ;
	      }
	      $chtml->addRaw("<DECLARE>") ;
	      $chtml->addDef($_) ;
	      $chtml->addRaw("</DECLARE>") ;
	    } else {
	      $chtml->addDef($_) ;
	    }
	  } elsif  (/^\s*(abstract)*\s*interface/i) {
	    if (/^(\s*interface\s+)(\w+)/i) {
	      my $inter ;
	      if ($self->get(-module)) {
		$inter = new fthinter(-name => $2, -file => $self->get(-file), -handle => $file, -masfic => $self->get(-masfic), -module => $self->get(-module)) ;
	      } else {
		$inter = new fthinter(-name => $2, -file => $self->get(-file), -handle => $file, -masfic => $self->get(-masfic)) ;
	      }
	      $chtml->addDef($1) ;
	      $chtml->addRaw($inter) ;
	      $inter->read($file) ;
	      $self->addInter($2 => $inter) ;
	    } elsif  (/^(\s*abstract\s+interface\s*)/i) {
	      # Basic interface to global stuff..
	      my $inter ;
	      if ($self->get(-module)) {
		$inter = new fthinter(-file => $self->get(-file), -handle => $file, -masfic => $self->get(-masfic), -module => $self->get(-module)) ;
	      } else {
		$inter = new fthinter(-file => $self->get(-file), -handle => $file, -masfic => $self->get(-masfic)) ;
	      }
	      $chtml->addDef($1.' ') ;
	      $chtml->addRaw($inter) ;
	      $inter->read($file) ;
	      my $mname = $inter->get(-name) ;
	      #$self->addInter($mname => $inter) ;
	      $inter->addInter($self) ;
	    } elsif  (/^(\s*interface\s*)/i) {
	      # Basic interface to global stuff..
	      my $inter ;
	      if ($self->get(-module)) {
		$inter = new fthinter(-file => $self->get(-file), -handle => $file, -masfic => $self->get(-masfic), -module => $self->get(-module)) ;
	      } else {
		$inter = new fthinter(-file => $self->get(-file), -handle => $file, -masfic => $self->get(-masfic)) ;
	      }
	      $chtml->addDef($1.' ') ;
	      $chtml->addRaw($inter) ;
	      $inter->read($file) ;
	      my $mname = $inter->get(-name) ;
	      #$self->addInter($mname => $inter) ;
	      $inter->addInter($self) ;
	    }  else {
	      $chtml->addDef($_) ;
	    }
	  } elsif (/^\#(ifdef|ifndef|if|elif|else|endif)(.*)/i) {
	    # Preprocessing tests, this seems good code below:
	    if ($1 eq "endif") {
	      $num = pop(@ploops) ;
	      if ($num) {
		$chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
		$preif-- ;
	      } else {
		$chtml->addDef('#'."$1$2") ;
	      }
	    } elsif ($1 eq "if" || $1 eq "ifdef" || $1 eq "ifndef") {
	      push(@ploops, $.) ;
	      $preif++ ;
	      $chtml->addRaw("<TEST><A Name=C_$.>".'#'."$1</A></TEST>$2") ;
	    } else {
	      $num = pop(@ploops) ;
	      if ($num) {
		push(@ploops, $num) ;
		$chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	      } else {
		$chtml->addDef('#'."$1$2") ;
	      }
	    }
	    $_ = "" ;
#	    $chtml->addRaw("<TEST>fthrout:3067") ;
#	    $debug = 1 ;
#	    $chtml->addDef('#'.$1) ;
#	    $chtml->addRaw("</TEST>") ;
#	    $chtml->addDef($2) ;
	    $_ = "" ;
	  } elsif (/^(\s*include\s+)([\"\'])([\w\.\/]+)\2(.*)/i || /^(\#include\s+)([\"\'])([\w\.\/]+)\2(.*)/i) {
	    my $incf = $3 ;
	    $chtml->addDef("$1$2") ;
	    $chtml->addRaw("<A Name=C_$.></A>");
	    unless (-f "$incf") {
	      my $base = dirname($self->get(-file)) ;
	      if (-f "$base/$incf") {
		$incf = "$base/$incf" ;
	      } elsif (-f "$base/" . basename($incf)) {
		$incf = "$base/" . basename($incf) ;
	      } elsif ($main::opt_I) {
		# Possible...
		my @TAGSHTML_PATH = (split ( /:/, $main::opt_I)) ;
		foreach $base (@TAGSHTML_PATH) {
		  if (-f "$base/$incf") {
		    $incf = "$base/$incf" ;
		    last ;
		  }
		}
	      }
	    }
	    if (-f "$incf") {
	      my $fh ;
	      my $other ;
	      if (open($fh,"<$incf")) {
		print STDOUT (caller(0))[3] . ": Open $incf from $file\n" if $debug ;
		$other = &fthinc::getObject($incf) ;
		if ($other) {
		  new chtml() ;
		} else {
		  new chtml(-file => $incf, -name => $incf) ;
		  $other = new fthinc(-file => "$incf", -name => "$incf", -masfic => $self->get(-masfic), -father => $self) ;
		}
		$self->addInclude("$incf" => $other) ;
		$chtml->addRaw($other) ;
		$chtml->addDef("$2$4") ;
		$self->readf90subr($fh, "", $cont) ;
		close($fh);
		&chtml::setCurrent($chtml) ;
		print STDOUT (caller(0))[3] . ": Closed $incf($.) from $file\n" if $debug ;
	      }
	    } else {
	      my $refe = new oref(-name => $incf, -type => "fthinc") ;
	      $chtml->addRaw($refe) ;
	      $chtml->addDef("$2$4") ;
	    }
	  } elsif (/^(\s*)(.*?)\s*::\s*(.*)/) {
	    # Fortran 90 types declaration below:
	    # print STDOUT "** Adding ref($.) = '$1' on $chtml\n" ;
	    $chtml->addDef($1) ;
	    my $lhs = $2 ; my $rhs = $3 ;
	    $self->parseDef90($lhs, $rhs) ;
	    # print STDOUT "Parsing defs $1 $lhs :: $rhs\n" ;
	  } elsif (/^(\s*)(integer|real|character|complex|logical|dimension)(.*?)[\s\:]+([\w\,]+.*)/i || /^(\s*)(double)(\s+precision)[\s\:]+([\w\,]+.*)/i) {
	    my $lhs = $2 . $3  ; my $rhs = $4 ;
	    $chtml->addDef("$1") ;
	    $self->parseDef90($lhs, $rhs) ;
	  } elsif (/^(\s*common\s*)\/(\s*\w+\s*)\/(.*)/i) {
	    $chtml->addRaw("<DECLARE>$1</DECLARE>".'/'."<VARIABLE>$2</VARIABLE>".'/'.$3) ;
	  } elsif (/^(\s*common)\b(.*)/i) {
	    $chtml->addRaw("<DECLARE>$1</DECLARE>$2") ;
	  } elsif (/^(\s*parameter\s*)\(([\w\s\,\=]+)\)(.*)/i) {
	    $chtml->addRaw("<DECLARE>$1</DECLARE>".'('."<VARIABLE>$2</VARIABLE>".')'.$3) ;
	  } elsif (/^(\s*parameter\s*)\(([\w\s\,\=]+)(.*)/i) {
	    $chtml->addRaw("<DECLARE>$1</DECLARE>".'('."<VARIABLE>$2</VARIABLE>".$3) ;
	  } elsif (/^\s*parameter\b/i) {
	    $chtml->addDef($_) ;
	  } elsif (/^(\s*equivalence\s*)\((\w+.*?)\,(\w+.*?\)*)\)(.*)/i) {
	    $chtml->addRaw("<DECLARE>$1</DECLARE>".'('."<VARIABLE>$2,$3</VARIABLE>".')'.$4) ;
	  } elsif (/^\s*equivalence\b/i) {
	    $chtml->addDef($_) ;
	  } elsif (/^(\s*import\s+)(\w+)(.*)/i) {
	    # This is not a type setting of course
	    $chtml->addRaw("<DECLARE>$1</DECLARE><VARIABLE>$2</VARIABLE>".$3) ;
	  } elsif (/^(\s*use\s+)(\w+)\s*,\s*only\s*:\s*(.*)/i) {
	    my $fmused = $2 ; $_ = $3 ;
	    $chtml->addDef("$1") ;
	    my $refe = new oref(-name => $2, -type => "fthmod") ;
	    $chtml->addRaw($refe) ;
	    $chtml->addDef(", only : ") ;
	    $self->readUses($file, $fmused, "&" . $_) ;
	  } elsif (/^(\s*use\s+)(\w+)\s*$/i) {
	    my $fmused = lc($2) ;
	    $chtml->addDef("$1") ;
	    my $refe = new oref(-name => $2, -type => "fthmod") ;
	    $self->{"ALLUSES"}->{"$fmused"} = 1 ;
	    $chtml->addRaw($refe) ;
	  } elsif (/^\s*$/) {
	    # Only comments 
	  } elsif (/^\#/) {
	    # Preprocessing infos:
	    $chtml->addDef($_) ;
	  } else {
	    print STDOUT "File=".$self->get(-file)."line $.='$_' end of declaration\n" if $debug ;
	    $cont = 5 ;
	  }
	}
	print STDOUT "cont=$cont, reuse=$main::reuse line=$.:$_\n" if $debug ;
	if ($cont >= 5) {
	  if (/do/i) {
	    if (/^(\s*)(do)(\s*)(\D\w+)(\s*\=.*)/i) {
	      $chtml->addDef($1) ;
	      $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
	      $chtml->addDef($3) ;
	      $_ = $4.$5 ;
	      $doif++ ;
	      print "Add1 do $_ line $., doif=$doif\n" if $debug ;
	      push(@aloops, $.) ;
	    } elsif (/^(\s*)(do)(\s*while)(.*)/i) {
	      $chtml->addDef($1) ;
	      $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
	      $chtml->addDef($3) ;
	      $_ = $4 ;
	      $doif++ ;
	      print "Add1 do $_ line $., doif=$doif\n" if $debug ;
	      push(@aloops, $.) ;
	    } elsif (/^(\s*\w+\s*\:\s*)(do)\b(\s*)(.*)/i) {
	      $chtml->addDef($1) ;
	      $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
	      $chtml->addDef($3) ;
	      $_ = $4 ;
	      $doif++ ;
	      print "Add1 do $_ line $., doif=$doif\n" if $debug ;
	      push(@aloops, $.) ;
	    } elsif (/^(\s*)(do)(\s*)(\d+)(\s*\w+\s*\=.*)/i) {
	      $chtml->addDef($1) ;
	      $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
	      $chtml->addDef($3) ;
	      # Add a link to the label 4:
	      $chtml->addRaw("<A Href=#" . $self->get(-name) . "_L_$4>$4</A>") ;
	      $_ = $5 ;
	      $doif++ ;
	      print "Add1 do $_ line $., doif=$doif\n" if $debug ;
	      push(@aloops, "$4:" . $.) ;
	    } elsif (/^(\s*)(do)(\s*)$/i) {
	      # NAJ: not counted as a do statement yet
	      $chtml->addDef($1) ;
	      $chtml->addRaw("<LOOP><A Name=C_$.>$2</A></LOOP>") ;
	      $_ = $3 ;
	      $doif++ ;
	      print "Add1 do $_ line $., doif=$doif\n" if $debug ;
	      push(@aloops, $.) ;
	    }
	  }
	  if (/^\#(ifdef|ifndef|if|elif|else|endif)(.*)/i) {
	    # Preprocessing tests, this seems good code below:
	    if ($1 eq "endif") {
	      $num = pop(@ploops) ;
	      if ($num) {
		$chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
		$preif-- ;
	      } else {
		$chtml->addDef('#'."$1$2") ;
	      }
	    } elsif ($1 eq "if" || $1 eq "ifdef" || $1 eq "ifndef") {
	      push(@ploops, $.) ;
	      $preif++ ;
	      $chtml->addRaw("<TEST><A Name=C_$.>".'#'."$1</A></TEST>$2") ;
	    } else {
	      $num = pop(@ploops) ;
	      if ($num) {
		push(@ploops, $num) ;
		$chtml->addRaw("<A Href=#C_$num>".'#'."$1</A>") ;
	      } else {
		$chtml->addDef('#'."$1$2") ;
	      }
	    }
	    $_ = "" ;
	  } elsif (/^(\s*)(if)(\s*\(.*)/i) {
	    #$self->readf77if($file, "     & ($1") ;
	    $chtml->addDef($1) ;
	    $chtml->addRaw("<TEST>") ;
	    $chtml->addDef("$2") ;
	    $chtml->addRaw("</TEST>") ;
	    $_ = $3 ;
	  } elsif (/^(\s*)(go\s*to\s*)(\d+)(.*)/i) {
	    $chtml->addDef($1.$2) ;
	    $chtml->addRaw("<A Href=#" . $self->get(-name) . "_L_$3>$3</A>") ;
	    $_ = $4 ;
	  }
	  if (/^\s*(include|contains)/i || $contains) {
	    if (/^(\s*include\s+)([\"\'])([\w\.\/]+)\2(.*)/i || /^(\#include\s+)([\"\'])([\w\.\/]+)\2(.*)/i) {
	      # There may be includes inside routines definitions, especially for f77 common:
	      my $incf = $3 ;
	      $chtml->addDef("$1$2") ;
	      $chtml->addRaw("<A Name=C_$.></A>");
	      unless (-f "$incf") {
		my $base = dirname($self->get(-file)) ;
		if (-f "$base/$incf") {
		  $incf = "$base/$incf" ;
		} elsif (-f "$base/" . basename($incf)) {
		  $incf = "$base/" . basename($incf) ;
		} elsif ($main::opt_I) {
		  # Possible...
		  my @TAGSHTML_PATH = (split ( /:/, $main::opt_I)) ;
		  foreach $base (@TAGSHTML_PATH) {
		    if (-f "$base/$incf") {
		      $incf = "$base/$incf" ;
		      last ;
		    }
		  }
		}
	      }
	      if (-f "$incf") {
		my $fh ;
		my $other ;
		if (open($fh,"<$incf")) {
		  print STDOUT (caller(0))[3] . ": Open $incf from $file\n" if $debug ;
		  $other = &fthinc::getObject($incf) ;
		  if ($other) {
		    new chtml() ;
		  } else {
		    new chtml(-file => $incf, -name => $incf) ;
		    $other = new fthinc(-file => "$incf", -name => "$incf", -masfic => $self->get(-masfic), -father => $self) ;
		  }
		  $self->addInclude("$incf" => $other) ;
		  $chtml->addRaw($other) ;
		  $chtml->addDef("$2$4") ;
		  $self->readf90subr($fh, "", $cont) ;
		  close($fh);
		  &chtml::setCurrent($chtml) ;
		  print STDOUT (caller(0))[3] . ": Closed $incf($.) from $file\n" if $debug ;
		}
	      } else {
		my $refe = new oref(-name => $incf, -type => "fthinc") ;
		$chtml->addRaw($refe) ;
		$chtml->addDef("$2$4") ;
	      }
	    } elsif (/^(\s*contains)\b/i) {
	      $contains = 1 ;
	      $chtml->addDef($_) ;
	    } elsif ($contains) {
	      if (/(subroutine|function)\s+(\w+)(.*)/i) {
		my $rame = $2 ;
		my $lame = lc($rame) ;
		# List of objects management
		my $orout = new fthrout(-name => $rame, -file => $self->get(-file), -masfic => $self->get(-masfic),  -private => $self, -number => $.) ;
		$self->addRout($rame => $orout) ;
		my $proto ;
		if (/^(\s*)(.*?)(\s*function\s+)(\w+)(.*)/i) {
		  $chtml->addDef($1) ;
		  $proto = $2 ;
		  if ($proto) {
		    if ($proto =~ /\bpure\b/i || $proto =~ /\brecursive\b/i) {
		      $chtml->addDef($proto) ;
		    } else {
		      $orout->set(-theproto => $proto) ;
		      print STDOUT "Defining proto to $rame as $proto\n" if $debug ;
		      $chtml->addRaw("<DECLARE>$proto</DECLARE>") ;
		    }
		  }
		  $chtml->addDef($3) ;
		  $chtml->addRaw($orout) ;
		  $orout->readf90subr($file, "& $5") ;
		  if ($proto) {
		    unless ($proto =~ /\bpure\b/i || $proto =~ /\brecursive\b/i) {
		      $orout->addArgument("$rame") ;
		      $orout->addType("$rame" => "$proto") ;
		    }
		  }
		  unless ($orout->get(-theproto)) {
		    $orout->addArgument("$rame") ;
		    $orout->addType("$rame" => 'void') ;
		    $orout->set(-theproto => 'void') ;
		  }
		  if ($orout->getType($rame) eq 'void' && $proto) {
		    $orout->addType($rame => $proto) ;
		    $orout->setType($rame => $proto) ;
		    print STDOUT "Error bad proto $rame\n" if $debug ;
		  }
		} elsif (/^(\s*)(.*?)(\s*subroutine\s+)$rame(.*)/i) {
		  $chtml->addDef($1.$2) ;
		  $orout->set(-theproto => 1) ;
		  $chtml->addDef($3) ;
		  $chtml->addRaw($orout) ;
		  $orout->readf90subr($file, "& $4") ;
		} else {
		  $chtml->addDef($_) ;
		}
	      } else {
		$chtml->addDef($_) ;
	      }
	    }
	  } elsif (!$main::reuse) {
	    if (/^(\s*)(write|print|stop)(.*)/i) {
	      # Writing line is not of much interest for us...
	      $chtml->addDef($_) ;
	    } elsif (/^(\s*call\s+)(\w+)\%(\w+)(.*)/i) {
	      # This corresponds to a call of a "type" or "class" procedure:
	      print STDOUT (caller(0))[3] .": Creating Fortran class call $2\n" if $debug ;
	      my $a = $1 ; my $o = $2 ; my $f = $3 ; my $arg = $4 ;
	      if (!$4 || ($4 =~ /^\s*$/) || ($4 =~ /^\s*\(\s*\)\s*$/)) {
		# Call with no argument, an easy stuff:
		my $ocall = new fthcall(-name => $f, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self) ;
		$self->addCall($f => $ocall) ;
		# Adds the only one call argument:
		$ocall->addArgument($o) ;
		$ocall->setArgument($o) ;
		$chtml->addDef($a) ;
		$chtml->addRaw("<BRANCHEMENT>") ;
		$chtml->addDef($o) ;
		$chtml->addRaw("</BRANCHEMENT>") ;
		$chtml->addDef('%') ;
		$chtml->addRaw($ocall) ;
		$chtml->addDef($arg) ;
	      } else {
		# Call with arguments, more difficult to analyze, NDY:
		$chtml->addDef($_) ;
	      }
	    } elsif (/^(\s*call\s+)(\w+)(.*)/i) {
	      print STDOUT (caller(0))[3] .": Creating Fortran call $2\n" if $debug ;
	      $chtml->addDef($1) ;
	      $self->readf90call($file, "$2", "$3") ;
	      $chtml->addDef($_) ;
	    } elsif (/^\s*([\%\w]+)\s*=\>{0,1}\s*(\w+)\((.*)/i  && !$self->sayType($2) && !&fthrout::isIntrinsec($2)) {
	      # Improved analysis: recursive function call !
	      my $rett = $1 ;
	      my $sec = $2 ;
	      my $line = $3 ;
	      if (/^(\s*$rett\s*=\>{0,1}\s*)/) {
		$chtml->addDef("$1") ;
	      }
	      $_ = $line ;
	      my $ocall = $self->readf90func($file, "$rett", "$sec", "$line") ;
	      while (/^(\s*[\*\+\-\/]\s*)([\w]+)\((.*)/i  && !$self->sayType($2) && !&fthrout::isIntrinsec($2)) {
		$_ = $3 ;
		$chtml->addDef("$1") ;
		$self->readf90func($file, "$rett", "$2", "$3") ;
	      }
	      $chtml->addDef($_) ;
	    } elsif (/^(.*?)\b(get\w+)\s*\((.*)/i && !$self->sayType($2) && !&fthrout::isIntrinsec($2)) {
	      # Improved analysis: getStuff is a function.
	      # NAJ: a bug remains in case of multiple sub calls...
	      $chtml->addDef("$1") ;
	      my $sec = $2 ;
	      my $args = $3 ;
	      if ($1 =~ /call\s*$/i) {
		print STDOUT (caller(0))[3] .": Creating Fortran call $2\n" if $debug ;
		$self->readf90call($file, "$sec", "($args") ;
	      } else {
		$_ = $args ;
		my $ocall = $self->readf90func($file, "void", "$sec", "$args") ;
		while (/^(.*?)\b(get\w+)\s*\((.*)/i && !$self->sayType($2) && !&fthrout::isIntrinsec($2)) {
		  $chtml->addDef("$1") ;
		  $sec = $2 ;
		  $args = $3 ;
		  $_ = $args ;
		  $self->readf90func($file, "void", "$sec", "$args") ;
		}
	      }
	      $chtml->addDef($_) ;
	    } else {
	      $chtml->addDef($_) ;
	    }
	  } elsif ($main::reuse) {
	    # Restart mode
	    if ($cc && ($str || /[\'\"]/)) {
	      $_ .= '!' . $cc ;
	      $cc = "" ;
	    }
	    if (/^(.*?)([\'\"\(\)\!])(.*)/) {
	      my $ligne = "" ;
	      while (/^(.*?)([\'\"\(\)\!])(.*)/) {
		my $bef = $1 ;
		my $sep = $2 ;
		my $aft = $3 ;
		$_ = $aft ;
		if ($bef =~ /^(.*?)\\\\$/) {
		  print STDOUT "\t\t Antislash special character encountered\n" if $debug ;
		} elsif ($bef =~ /^(.*?)\\$/) {
		  if ($str) {
		    $ligne .= "$bef$sep" ;
		    if ($bef =~ /[\<\>]/) {
		      $bef =~ s/\</&lt;/g ;
		      $bef =~ s/\>/&gt;/g ;
		    }
		  } else {
		    # Should never happen
		    print STDOUT (caller(0))[3] .": Strange event file ".$self->get(-file)." line $. text = <$bef$sep$aft>\n" ;
		    $ligne .= "$bef$sep" ;
		  }
		  next ;
		}
		# Continuing FROM
		if ($str && $sep eq $str) {
		  if ($bef =~ /[\<\>]/) {
		    $bef =~ s/\</&lt;/g ;
		    $bef =~ s/\>/&gt;/g ;
		  }
		  $str = "" ;
		  $ligne .= "$bef$sep</STRING>" ;
		} elsif ($str) {
		  # Nothing special, continuing the string build
		  if ($bef =~ /[\<\>]/) {
		    $bef =~ s/\</&lt;/g ;
		    $bef =~ s/\>/&gt;/g ;
		  }
		  $ligne .= "$bef$sep" ;
		} elsif ($sep eq '"' || $sep eq "'") {
		  $str = $sep ;
		  $ligne .= "$bef<STRING>$sep" ;
		} elsif ($sep eq '!') {
		  $cc = $aft ;
		  $_ = "" ;
		  last ;
		} elsif ($sep eq '(' || $sep eq ')') {
		  if ($sep eq '(') {
		    $nbparenthe++ ;
		    if ($bef =~ /(.*?call\s+)(\w+)\%(\w+)$/i) {
		      # This corresponds to a call of a "type" or "class" procedure:
		      $chtml->addRaw($ligne.$1) ;
		      $ligne = "" ;
		      print STDOUT (caller(0))[3] .": Creating Fortran class call $2\n" if $debug ;
		      my $a = $1 ; my $o = $2 ; my $f = $3 ;
		      my $ocall = new fthcall(-name => $f, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self) ;
		      $self->addCall($f => $ocall) ;
		      # Adds the only one call argument:
		      $ocall->addArgument($o) ;
		      $ocall->setArgument($o) ;
		      # Sets the writer stuff:
		      $chtml->addDef($a) ;
		      $chtml->addRaw("<BRANCHEMENT>") ;
		      $chtml->addDef($o) ;
		      $chtml->addRaw("</BRANCHEMENT>") ;
		      $chtml->addDef('%') ;
		      $chtml->addRaw($ocall) ;
		      # Go on reading the stuff:
		      $ocall->readf90($file, "& $ligne") ;
		      # Post-process the call as usual in readf90call:
		      $self->postCallProcess($ocall) ;
		      $nbparenthe-- ;
		    } elsif ($bef =~ /(.*?call\s+)(\w+)$/i) {
		      $chtml->addRaw($ligne.$1) ;
		      $ligne = "" ;
		      print STDOUT (caller(0))[3] .": Creating Fortran call $2\n" if $debug ;
		      $self->readf90call($file, "$2", "$sep$aft") ;
		      $nbparenthe-- ;
		    } elsif ($bef =~ /\s*([\%\w]+)\s*=\>{0,1}\s*(\w+)$/i && $self->sayCalling($2)) {
		      my $rett = $1 ;
		      my $sec = $2 ;
		      print STDOUT (caller(0))[3] .": Creating legacy Fortran routine $sec from $bef line $.\n" if $debug;
		      if ($bef =~ /(.*)\b$sec\b/) {
			$chtml->addRaw("$ligne$1") ;
		      }
		      print STDOUT "Parsing file=$file, return=$rett, function=$sec, arguments=$aft\n" if $debug;
		      my $ocall = $self->readf90func($file, $rett, "$sec", "$aft") ;
		      $nbparenthe-- ;
		      while (/^(\s*[\*\+\-\/]\s*)([\w]+)\((.*)/i  && $self->sayCalling($2)) {
			$_ = $3 ;
			$chtml->addRaw("$1") ;
			print STDOUT (caller(0))[3] .": Creating Fortran routine $2\n" if $debug ;
			$self->readf90func($file, "$rett", "$2", "$3") ;
		      }
		    } elsif ($bef =~ /^(.*?)([\%\w]+)$/i && $self->sayCalling($2)) {
		      my $rett = $1 ;
		      my $sec = $2 ;
		      $chtml->addRaw("$ligne$1") ;
		      $ligne = "" ;
		      print STDOUT (caller(0))[3] .": Creating auto Fortran routine $sec from $bef line $.\n" if $debug ;
		      #my $ocall = $self->readf90func($file, "$rett", "$sec", "$aft") ;
		      # Working in a void context so far...
		      my $ocall = $self->readf90func($file, "void", "$sec", "$aft") ;
		      $nbparenthe-- ;
		    } elsif ($bef =~ /^(.*?)([\w]+)\%([\w]+)$/i && $self->sayCalling($3)) {
		      my $rett = $1 ;
		      my $sec = $3 ;
		      $chtml->addRaw("$ligne$1") ;
		      $chtml->addRaw("<BRANCHEMENT>") ;
		      $chtml->addDef($2) ;
		      $chtml->addRaw("</BRANCHEMENT>") ;
		      $chtml->addDef('%') ;
		      $ligne = "" ;
		      print STDOUT (caller(0))[3] .": Creating auto Fortran routine $sec from $bef line $.\n" if $debug ;
		      #my $ocall = $self->readf90func($file, "$rett", "$sec", "$aft") ;
		      # Working in a void context so far...
		      if (1 == 0) {
			my $ocall = $self->readf90func($file, $2, "$sec", "$aft") ;
		      } else {
			my $o = $2 ; my $f = $3 ;
			my $ocall = new fthcall(-name => $f, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self) ;
			$chtml->addRaw($ocall) ;
			# Adds the only one call argument:
			$ocall->addArgument($o) ;
			$ocall->setArgument($o) ;
			# Go on reading the stuff:
			$ocall->readf90($file, "& ($aft") ;
			# Post-process the function as usual in readf90func:
			# Is there a way to guess the return type through current context ?
			# This ought to be analyzed later on...
			# NAJ: aire_figure = f%aire[?aire]]]  too many brackets !!!
			if ($rett =~ /^\s*([\w\%]+)\s*\=\s*/) {
			  $self->postFunProcess($ocall, $file, $1, $f) ;
			} else {
			  $self->postFunProcess($ocall, $file, "void", $f) ;
			}
		      }
		      $nbparenthe-- ;
		    } else {
		      $ligne .= "$bef$sep" ;
		    }
		  } elsif ($sep eq ')') {
		    $ligne .= "$bef$sep" ;
		    $nbparenthe-- ;
		  }
		}
	      }
	      if ($_ && /(.*?call\s+)(\w+)\%(\w+)(.*)/i) {
		# This corresponds to a call of a "type" or "class" procedure:
		# die "OK" ;
		my $a = $1 ; my $o = $2 ; my $f = $3 ; my $arg = $4 ;
		if (!$4 || ($4 =~ /^\s*$/) || ($4 =~ /^\s*\(\s*\)\s*$/)) {
		  # Call with no argument, an easy stuff:
		  my $ocall = new fthcall(-name => $f, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self) ;
		  $self->addCall($f => $ocall) ;
		  # Adds the only one call argument:
		  $ocall->addArgument($o) ;
		  $ocall->setArgument($o) ;
		  $chtml->addDef($a) ;
		  $chtml->addRaw("<BRANCHEMENT>") ;
		  $chtml->addDef($o) ;
		  $chtml->addRaw("</BRANCHEMENT>") ;
		  $chtml->addDef('%') ;
		  $chtml->addRaw($ocall) ;
		  $chtml->addDef($arg) ;
		} else {
		  # Call with arguments, more difficult to analyze, NDY:
		  $chtml->addDef($_) ;
		}
	      } elsif ($_ && /(.*?call\s+)(\w+)(.*)/i) {
		print STDOUT (caller(0))[3] .": Creating Fortran call $2\n" if $debug ;
		$chtml->addRaw($ligne.$1) ;
		$ligne = "" ;
		$self->readf90call($file, "$2", "$3") ;
		# Calls a routine:
	      } else {
		$chtml->addRaw($ligne.$_) ;
		$ligne = "" ;
	      }
	    } elsif (/^(.*?call\s+)(\w+)\%(\w+)(.*)/i) {
	      # This corresponds to a call of a "type" or "class" procedure:
	      print STDOUT (caller(0))[3] .": Creating Fortran class call $3\n" if $debug ;
	      my $a = $1 ; my $o = $2 ; my $f = $3 ; my $arg = $4 ;
	      if (!$4 || ($4 =~ /^\s*$/) || ($4 =~ /^\s*\(\s*\)\s*$/)) {
		# Call with no argument, an easy stuff:
		my $ocall = new fthcall(-name => $f, -file => $self->get(-file), -masfic => $self->get(-masfic), -father => $self) ;
		$self->addCall($f => $ocall) ;
		# Adds the only one call argument:
		$ocall->addArgument($o) ;
		$ocall->setArgument($o) ;
		$chtml->addDef($a) ;
		$chtml->addRaw("<BRANCHEMENT>") ;
		$chtml->addDef($o) ;
		$chtml->addRaw("</BRANCHEMENT>") ;
		$chtml->addDef('%') ;
		$chtml->addRaw($ocall) ;
		$chtml->addDef($arg) ;
	      } else {
		# Call with arguments, more difficult to analyze, NDY:
		$chtml->addDef($_) ;
	      }
	    } elsif (/^(.*?call\s+)(\w+)(.*)/i) {
	      print STDOUT (caller(0))[3] .": Creating Fortran call $2\n" if $debug ;
	      $chtml->addRaw($1) ;
	      $self->readf90call($file, "$2", "$3") ;
	      # Calls a routine:
	    } else {
	      $chtml->addDef($_) ;
	    }
	  }
	}
      }
      if ($cc) {
	$chtml->addRaw('<COMMENT>') ;
	if ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  $chtml->addLaCoM("<A Name=C_$.></A>") ;
	}
	while ($cc =~ /(.*?\\)(reinput|verbatiminput|input)\s+([\w_\.\/]+)\.(tex|plt)(.*)/) {
	  my $incp = "$3.$4" ;
	  $cc = $5 ;
	  $chtml->addLaTeX($1.$2." ") ;
#	  if ($incp =~ /^\// && ! -f  "$incp") {
#	    $incp = basename($incp) ;
#	  }
	  unless (-f "$incp") {
	    my $nase = basename(dirname($incp)) ;
	    my $base = dirname($self->get(-file)) ;
	    if (-f "$base/$incp") {
	      $incp = "$base/$incp" ;
	    } elsif (-f "$base/" . basename($incp)) {
	      $incp = "$base/" . basename($incp) ;
	    } elsif (-f "$nase/" . basename($incp)) {
	      $incp = "$nase/" . basename($incp) ;
	    } elsif ($main::opt_L) {
	      # LaTeX inputs directories
	      # Possible...
	      my @TAGSHTML_PATH = (split ( /:/, $main::opt_L)) ;
	      foreach $base (@TAGSHTML_PATH) {
		if (-f "$base/$incp") {
		  $incp = "$base/$incp" ;
		  last ;
		}
	      }
	    }
	  }
	  if (-f "$incp") {
	    my $other = new fthinp(-file => "$incp", -name => "$incp") ;
	    $self->addInput("$incp" => $other) ;
	    $chtml->addLaCoM($other) ;
	  } else {
	    my $refe = new oref(-name => $incp, -type => "fthinp") ;
	    $chtml->addLaCoM($refe) ;
	  }
	}
	$chtml->addLaTeX($cc) ;
	$chtml->addRaw("</COMMENT>") ;
      }
      return unless $cont ;
      if (defined($_ = <$file>)) {
	chomp ;
      } else {
	print STDOUT (caller(0))[3] . ": error3 line=$. rout='" . $self->get(-name) . "' file='" . $self->get(-file) . "'\n" if $debug || $main::opt_v ;
	$cont = 0 ;
	return ;
      }
    }
  }
}
###########################################################################
1;

