package fthtype ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for fthtype 90 files.
###########################################################################
use strict;
use Carp ;
use oref ;
use vars qw($debug $nbself %MYOBJHASH) ;
$debug = 0 ;
$nbself = 0 ;
(%MYOBJHASH) = () ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = lc(shift) ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    # We refuse to create an object from no where
    # croak("Violent build for $name is illegal, API error") ;
    debug("Violent build for $name is illegal, API error") if $debug ;
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  if ($args{-name}) {
    $name = lc($args{-name}) ;
  } else {
    croak("Wrong API usage for $self") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
  } else {
    $nbself++ ;
    $first = 1 ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
# Gets the number of built objects:
sub getNbself {
  return($nbself) ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub save_hash {
  my $self = shift ;
  my $handle = shift ;
  my $name = $self->get(-name) ;
  if (0) {
    print $handle '$L_Declar{"' . $name . '"} = "' . $self->get(-number) . '" ;' . "\n" ;
    print $handle '$File_Of_Rout{"' . $name . '"} = "' . $self->get(-file) . '" ;' . "\n" ;
  } else {
    my $number = $self->get(-number) ;
    my $file = $self->get(-file) ;
    print $handle '$fthtypes{"' . $name . '"} = ' ;
    print $handle '"' . "$file:$number" . '"'  ;
    print $handle ';' ."\n" ;
    print $handle '$names{"' . lc($name) . '"} = "' . $name . '";' ."\n" ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	print STDOUT "$sval, " ;
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	print STDOUT "$sval => $array{$sval}, " ;
      }
      print STDOUT "},\n" ;
    } else {
      print STDOUT "$t$cle = $val,\n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
sub any {
  if (%MYOBJHASH) {
    return(1) ;
  } else {
    return(0) ;
  }
}
###########################################################################
# Gets the field's type in current structure
sub getType {
  my $self = shift ;
  my $name = lc(shift) ;
  return('void') unless $name ;
  if (exists $self->{"ALLDEFS"}->{"$name"}) {
    return($self->{"ALLDEFS"}->{"$name"}) ;
  } else {
    return('void') ;
  }
}
###########################################################################
sub parseDef90 {
  my $self = shift ;
  my $ilhs = shift ;
  my $irhs = shift ;
  my $lhs = lc($ilhs) ;
  my $rhs = lc($irhs) ;
  my $shs ;
  my $name = $self->get(-name) ;
  my $chtml = &chtml::getCurrent() ;
  if ($lhs eq 'public' || $lhs eq 'private') {
    # This is not a type setting indeed:
    $chtml->addRaw("<DECLARE>$ilhs</DECLARE> :: <VARIABLE>$irhs</VARIABLE>") ;
    return ;
  }
  $lhs =~ s/\s+//g ;
  $rhs =~ s/\s+//g ;
  # Fortran argument definition equivalent form:
  $lhs =~ s/,*intent\(.*?\)//g ;
  $lhs =~ s/,*allocatable//g ;
  $lhs =~ s/,*save//g ;
  if ($lhs =~ /(.*?),*dimension\((.*?)\)(.*)/i) {
    my (@array) = split(/\,/, $2) ;
    $lhs = $1 . $3 . ',dimension(' . ',' x ( $#array + 0 ) . ')' ;
    $lhs =~ s/,*pointer//g ;
  }
  while ($rhs =~ /(.*)\((.*?)\)(.*)/) {
    my $a = $1 ;  my $b = $2 ;  my $c = $3 ;  
    $b =~ s/\,/\;/g ;
    $rhs = $a . '{' . $b . '}' . $c ;
  }
  my (@array) = split(/\,/, $rhs) ;
  foreach my $rhs (@array) {
    if ($rhs =~ /(.*)\s*\=\>{0,1}\s*([\w\.\+\-\*\/]*)/) {
      $rhs = $1 ;
    }
    if ($rhs =~ /(\w+)\{(.*?)\}/) {
      $rhs = $1 ;
      my (@larray) = split(/\,/, $2) ;
      $shs = $lhs . ',dimension(' . ',' x ( $#larray + 0 ) . ')' ; ;
    } else {
      $shs = $lhs ;
    }
    # NAJ2003: what happens in case of procedure attribute ?
    print STDOUT (caller(0))[3] . ": In " . $self->get(-name) . " line $. defined var=$rhs is of type $shs\n" if $debug ;
    $self->{"ALLDEFS"}->{"$rhs"} = $shs ;
    push(@{$self->{"FIELDS"}}, "$shs :: $rhs") ;
    # print STDOUT (caller(0))[3] .": Args of ".$self->get(-name)." is $rhs :: $lhs\n" if $debug ;
  }
  if ($ilhs =~ /(.*type\s*\(\s*)(\w+)(\s*\).*)/i) {
    $chtml->addRaw("<DECLARE>$1") ;
    my $refe = new oref(-name => $2, -type => "fthtype") ;
    print STDOUT "Refering to another fthtype object $2: $refe\n" if $debug ;
    $chtml->addRaw($refe) ;
    $chtml->addRaw("$3</DECLARE> :: <VARIABLE>$irhs</VARIABLE>") ;
  } elsif ($ilhs =~ /(.*class\s*\(\s*)(\w+)(\s*\).*)/i) {
    $chtml->addRaw("<DECLARE>$1") ;
    my $refe = new oref(-name => $2, -type => "fthtype") ;
    print STDOUT "Refering to another fthtype object $2: $refe\n" if $debug ;
    $chtml->addRaw($refe) ;
    $chtml->addRaw("$3</DECLARE> :: <VARIABLE>$irhs</VARIABLE>") ;
  } elsif ($ilhs =~ /^\s*procedure\b/i) {
    # An option is to add another module procedure to the current module.
    # Another, a cleaner option, would be to add a "class" procedure where "class" is
    # related to the "type" or "abstract type", procedure call being parsed as object%procedure
    my $module = $self->get(-module) ;
    if ($module) {
      # let it be a module procedure, ought to be a fthproc, a suboject of self (fthtype)
      if ($irhs =~ /^\s*(\w+)$/) {
	my $inter = new fthinter(-name => $1, -file => $self->get(-file), -masfic => $self->get(-masfic), -module => $module) ;
	$chtml->addRaw("<DECLARE>$ilhs</DECLARE> :: ");
	my $refe = new oref(-name => $module->get(-name) . '::' . $1, -type => "uses") ;
	$chtml->addRaw($refe) ;
	$inter->addModuleProcedure($1) ;
      } elsif ($irhs =~ /^\s*(\w+)\s*\=\>\s*(\w+)\s*$/) {
	my $a = $2 ;
	$chtml->addRaw("<DECLARE>$ilhs</DECLARE> :: <VARIABLE>$1</VARIABLE> => ");
	my $inter = new fthinter(-name => $1, -file => $self->get(-file), -masfic => $self->get(-masfic), -module => $module) ;
	my $refe = new oref(-name => $module->get(-name) . '::' . $a, -type => "uses") ;
	$chtml->addRaw($refe) ;
	$inter->addModuleProcedure($a) ;
      } else {
	$chtml->addRaw("<DECLARE>$ilhs</DECLARE> :: <VARIABLE>$irhs</VARIABLE>") ;
      }
    } else {
      $chtml->addRaw("<DECLARE>$ilhs</DECLARE> :: <VARIABLE>$irhs</VARIABLE>") ;
    }
  } else {
    $chtml->addRaw("<DECLARE>$ilhs</DECLARE> :: <VARIABLE>$irhs</VARIABLE>") ;
  }
}
###########################################################################
sub read {
  my $self = shift ;
  my $file = shift ;
  my $chtml = &chtml::getCurrent() ;
  if ((ref $file) && (fileno $file)) {
    my $name = $self->get(-name) ;
    while(<$file>) {
      chomp ;
      # Parsing an f90 type file below:
      if (/^(\s*end\s*type\s+)$name\b/i) {
	# End of type creation: build a reference to the type creation:
	$chtml->addDef("$1") ;
	my $refe = new oref(-name => $name, -type => "fthtype") ;
	$chtml->addRaw($refe) ;
	print STDOUT "Created type $name\n" if $debug ;
	return ;
      } elsif (/^(\s*end\s*type\s*)/i) {
	# Badend of type syntax...
 	$chtml->addDef("$1") ;
	my $refe = new oref(-name => $name, -type => "fthtype") ;
	$chtml->addRaw($refe) ;
	print STDOUT "Created type $name\n" if $debug ;
	return ;
     } else {
	# Proper management of input remains to be performed in the future:
	my $cc = "" ;
	if (/^(.*?)\!(.*)/) {
	  $_ = $1 ;
	  $cc = $2 ;
	}
	if (/private/i) {
	  $self->set(-private => 1) ;
	  $chtml->addDef($_) ;
	} elsif (/^(\s*)(.*?)::(.*)/) {
	  $chtml->addDef("$1") ;
	  $self->parseDef90($2, $3) ;
	} elsif (/^(\s*)(integer|real|character|complex|logical|dimension)(.*?)[\s\:]+([\w\,]+.*)/i || /^(\s*)(double)(\s+precision)[\s\:]+([\w\,]+.*)/i) {
	  $chtml->addDef("$1") ;
	  $self->parseDef90($2 . $3, $4) ;
	} else {
	  $chtml->addDef($_) ;
	}
	if ($cc) {
	  $chtml->addRaw('<COMMENT>') ;
	  $chtml->addLaTeX($cc) ;
	  $chtml->addRaw('</COMMENT>') ;
	}
      }
    }
  }
}
###########################################################################
1;


