package ihtml ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A data structure for ihtml 90 files.
###########################################################################
use strict;
use Carp ;
use CGI ;
use File::Basename ;
use fthfortran ;
use fthc ;
use fthrout ;
use fthcall ;
use fthinc ;
use fthinp ;
use fthlabel ;
use fthbc;
use fthext;
use vars qw($debug $delbr %MYOBJHASH %hash $opt_oldmm) ;
$debug = 0 ;
(%MYOBJHASH) = () ;
###########################################################################
sub objectlist {
  return(values %MYOBJHASH) ;
}
###########################################################################
sub ghashlist {
  return(%MYOBJHASH) ;
}
###########################################################################
sub shashlist {
  (%MYOBJHASH) = (@_) ;
}
###########################################################################
# Gives an object by its name.
sub getObject {
  my $name = shift ;
  if (exists $MYOBJHASH{$name}) {
    return($MYOBJHASH{$name}) ;
  } else {
    # We refuse to create an object from no where
    debug("Violent build for $name is illegal, API error") if $debug ;
    return(undef) ;
  }
}
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  my $name ;
  my $first = 0 ;
  $delbr = 0 ;
  if ($args{-name}) {
    $name = $args{-name} ;
  } else {
    croak("Wrong API usage for $self, no -name argument") ;
  }
  if (($name) && (exists $MYOBJHASH{$name})) {
    $self = $MYOBJHASH{$name} ;
    print STDOUT (caller(0))[3] . ": Reusing CGI for $self=$name is an errror\n" ;
    my $q = new CGI() ;
    $self->{"cgi"} = $q ;
  } else {
    $first = 1 ;
    bless $self ;
    $MYOBJHASH{$name} = $self ;
    my $q = new CGI() ;
    print STDOUT "Create CGI for $self=$name\n" if $debug;
    $self->{"cgi"} = $q ;
  }
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug ;
  }
  return $self ;
}
###########################################################################
sub setDebug {
  $debug = shift ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
sub printHash {
  my $self = shift ;
  my $t = shift ;
  my $hash = shift ;
  while(my($cle,$val) = each(%{$hash})) {
    print STDOUT "$t$cle => $val, " ;
  }
}
###########################################################################
# Dumps all the data to stdout
sub datadump {
  my $self = shift ;
  my $t = shift ;
  print STDOUT "$self name ".$self->{"name"}."={\n ";
  while(my($cle,$val) = each(%{$self})) {
    if (($val) && (ref($val) eq 'ARRAY')) {
      my(@array) = (@{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (@array) {
	print STDOUT "$sval, " ;
      }
      print STDOUT "},\n" ;
    } elsif (($val) && (ref($val) eq 'HASH')) {
      my(%array) = (%{$val}) ;
      print STDOUT "$t$cle = {" ;
      foreach my $sval (keys %array) {
	if (exists $array{$sval} && ref($array{$sval}) eq 'HASH') {
	  print STDOUT "$t\t$sval => {"  ;
	  $self->printHash("\t$t", $array{$sval}) ;
	  print STDOUT"}, " ;
	} else {
	  print STDOUT "$sval => $array{$sval}, " ;
	}
      }
      print STDOUT "},\n" ;
    } else {
      print STDOUT "$t$cle = $val, \n" ;
    }
  }
  print "}\n" ;
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
sub print_header {
  my $self = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  print $handle $cgi->header() ;
}
###########################################################################
sub end_html {
  my $self = shift ;
  my $rel = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  $rel = "" unless $rel ;
  print $handle "<BR><P><BR><A Href=\"http://sourceforge.net/projects/ftagshtml/\"><IMG SRC=\"${rel}ftagshtml.jpg\" ALT=\"ftagshtml's logo\"></A>";
  print $handle sprintf "<BR>Code $main::RepCodeSource, using FtagsHTML-%5.3f by $main::USER", $main::version ;
  my $file = $self->{"file"} ;
  if (-f "$file") {
    print $handle ", file $file of " . $self->last_modified(). "." ;
  } else {
    my(@lt) = localtime();
    my($d,$m,$y) = @lt[3,4,5];
    $y += 1900 ; $m++ ;
    print $handle " on $d/$m/$y." ;
  }
  print $handle "\n" ;
  print $handle $cgi->end_html() ;
}
###########################################################################
sub print_lf {
  my $self = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $file = $self->{"file"} ;
  my $title = "Browse $file." ;
  my $target=  &targetname(${file}) ;
  if ($opt_oldmm) {
    print $handle "<html> <head> <title>$title</title>\n" ;
    print $handle "<link rel=\"stylesheet\" href=\"ftie4style.css\">\n";
    print $handle "<!-- Code for browser detection. DO NOT REMOVE.              -->\n";
    print $handle "<script src=\"ua.js\"></script>\n";
    print $handle "<!-- Infrastructure code for the TreeView. DO NOT REMOVE.    -->\n";
    print $handle "<script src=\"ftiens4.js\"></script>\n";
    print $handle "<!-- Scripts that define the tree. DO NOT REMOVE.            -->\n";
    print $handle "<script src=\"${target}.js\"></script>\n" ;
    print $handle "<script>initializeDocument()</script></head>\n" ;
    print $handle "<NOSCRIPT>" ;
    print $handle "   A tree for site navigation will open here if you enable JavaScript in your browser." ;
    print $handle "</NOSCRIPT>" ;
    print $handle "<body bgcolor=\"white\"></html>\n" ;
  } else {
    print $handle "<html> <head> <title>$title</title>\n" ;
    print $handle "<style>\n";
    print $handle "a, A:link, a:visited, a:active, A:hover\n";
    print $handle "{color: #000000; text-decoration: none; font-family: Tahoma, Verdana; font-size: 12px}\n";
    print $handle "</style></head>\n";
    print $handle "<body bottommargin=\"0\" topmargin=\"0\" leftmargin=\"0\" rightmargin=\"0\" marginheight=\"0\" marginwidth=\"0\" bgcolor=\"white\">\n";
    print $handle "<script language=\"JavaScript\" src=\"tree.js\"></script>\n";
    print $handle "<script language=\"JavaScript\" src=\"${target}.js\"></script>\n";
    print $handle "<script language=\"JavaScript\" src=\"tree_tpl.js\"></script>\n" ;
    print $handle "<table cellpadding=\"5\" cellspacing=\"0\" cellpadding=\"10\" border=\"0\" width=\"100%\">\n" ;
    print $handle "<tr><td><script language=\"JavaScript\">\n" ;
    print $handle "   new tree (TREE_ITEMS, TREE_TPL);\n" ;
    print $handle "</script></td></tr></table>\n" ;
    print $handle "</body></html>\n" ;
  }
}
###########################################################################
sub print_master {
  my $self = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $file = $self->{"file"} ;
  my $title = "Fichier $file." ;
  my $target=  &targetname(${file}) ;
  print $handle "<html> <head> <title>$title</title></head>\n" ;
  print $handle "<frameset COLS=\"30%,*\" BORDER=\"1\" MARGINHEIGHT=\"0\" MARGINWIDTH=\"0\">\n" ;
  if ($main::opt_nojava) {
    print $handle "  <frame SRC=\"${target}_i.html\" NAME=\"index\">\n" ;
  } else {
    print $handle "<FRAMESET ROWS=\"50%,*\">\n" ;
    print $handle "  <frame SRC=\"${target}_LF.html\" NAME=\"java\">\n" ;
    print $handle "  <frame SRC=\"${target}_i.html\" NAME=\"index\">\n" ;
    print $handle "</FRAMESET> \n" ;
  }
  print $handle "  <frame SRC=\"${target}_c.html\" NAME=\"source\">\n" ;
  print $handle "</FRAMESET>\n" ;
  print $handle "</html>\n" ;
}
###########################################################################
sub start_tree {
  my $self = shift ;
  my $file = $self->{"file"} ;
  my $target = &targetname(${file}) ;
  if ($opt_oldmm) {
    return("foldersTree = gFld(\"<i>Trees of $file</i>\",\"${target}_c.html\")") ;
  } else {
    return("var TREE_ITEMS = [\n\t[\"<i>Trees of $file</i>\", \"${target}_c.html\", ") ;
  }
}
###########################################################################
sub print_interface {
  my $self = shift ;
  my $fobj = shift ;
  my(@stuffs) = $fobj->getStuffs() ;
  my $name = $fobj->get(-name) ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  print $handle "<A Name='R_" . $name . "'>" . $name . "</A>: \n" ;
  print $handle $cgi->start_ol() ."\n" ;
  foreach my $stuff (@stuffs) {
    if (ref($stuff)) {
      print $handle $cgi->li() ;
      if (ref($stuff) eq 'fthrout') {
	print $handle $cgi->a({
	  href => "../" . &ihtml::targetname($stuff->get(-file)) . "_c.html#R_" . $stuff->get(-name) ,
	  target => 'source',
	},$stuff->get(-name)) ;
      } elsif (ref($stuff) eq 'fthinter') {
	# NAJ518: Adding a link to the interface definition file:
	my $module = $stuff->get(-module) ;
	if ($module && ref($module) eq 'fthmod') {
	  #print "Call fthmod" ;
	  print $handle $cgi->a({
	    href => "../" . &ihtml::targetname($stuff->get(-file)) . "_c.html#R_" . $stuff->get(-name) ,
	    target => 'source',
				},$module->get(-name)) ;
	  #print "Called fthmod" ;
	} else {
	  print $handle $cgi->a({
	    href => "../" . &ihtml::targetname($stuff->get(-file)) . "_c.html#R_" . $stuff->get(-name) ,
	    target => 'source',
				},$stuff->get(-name)) ;
	}
	my (@procs) = $stuff->getProcedures() ;
	if (@procs) {
	  print $handle ": " ;
	  my $ok = 0 ;
	  foreach my $pname (@procs) {
	    my $orout = &fthrout::getObject($pname) ;
	    print $handle ", " if $ok ;
	    unless ($orout) {
	      # Another attempt:
	      my $fobj = &fthobj::getObject($pname) ;
	      $orout = $fobj->getProcedure($pname) if $fobj ;
	    }
	    if ($orout) {
	      print $handle $cgi->a({
		href => "../" . &ihtml::targetname($orout->get(-file)) . "_c.html#R_" . $orout->get(-name) ,
		target => 'source',
	      },$orout->get(-name)) ;
	    } else {
	      if ($main::opt_nocolor) {
		print $handle "$pname" ;
	      } else {
		print $handle "<FONT COLOR=RED>$pname</FONT>" ;
	      }
	    }
	    $ok++ ;
	  }
	  print $handle "\n" ;
	}
      }
    } else {
      print $handle "<BR>" . $stuff . "\n" ;
    }
  }
  print $handle $cgi->end_ol() ."\n" ;
}
###########################################################################
sub print_index {
  my $outdir = shift ;
  my $reldir = shift ;
  my $type = shift ;
  print STDOUT (caller(0))[3] .": Printing $type\n" if $debug ;
  my(@objects) = (sort byname @_) ;
  my(%hash) = () ;
  if (open(OUT,">$outdir/${type}index_c.html")) {
    my $handle = \*OUT ;
    my $self = new ihtml(-name => "$type", -handle => $handle, -file => "$type", -reldir => $reldir) ;
    my $cgi = $self->{"cgi"} ;
    my $title = "Index for " . ucfirst($type) ;
    $self->set(-title => $title) ;
    $self->start_header() ;
    print $handle $cgi->h1("$title") . "\n" ;
    $self->print_table() ;
    print $handle $cgi->start_ol() ."\n" ;
    my $pfile = undef ;
    foreach my $object (@objects) {
      my $name = $object->get(-name) ;
      my $file = $object->get(-file) ;
      print $handle $cgi->li() ;
      my $target ;
      if ($file && -f "$file") {
	$target = &targetname(${file}) ;
      } else {
	$target =  &targetinter($name);
      }
      if ($file) {
	unless ($hash{$file} && $pfile && $pfile eq $file) {
	  print STDOUT "Setting hash for $file\n" if $debug ;
	  $hash{$file} = 1 ;
	  print $handle "<A Name=\"R_" . $target . "\"></A>\n" ;
	  $pfile = $file ;
	}
	print $handle $cgi->a({
	  href => $target . "_c.html#R_" . $name ,
	  target => 'source',
	}, $name) . "\n";
      } else {
	print $handle $cgi->a({
	  href => $target . "_i.html" ,
	  target => 'index',
	}, $name) . "\n";
      }
    }
    print $handle $cgi->end_ol() ."\n" ;
    $self->end_html() ;
    close(OUT) ;
  }
  print STDOUT (caller(0))[3] .": Printed $type\n" if $debug ;
}
###########################################################################
sub printf_index {
  my $outdir = shift ;
  my $reldir = shift ;
  my $type = shift ;
  my $top = shift ;
  print STDOUT (caller(0))[3] .": Printing $type\n" if $debug ;
  my(%ofiles) = (@_) ;
  my(@objects) = (sort byname values %ofiles) ;
  my(%hash) = () ;
  if (open(OUT,">$outdir/${type}index_c.html")) {
    my $handle = \*OUT ;
    my $self = new ihtml(-name => "$type", -handle => $handle, -file => "$type", -reldir => $reldir) ;
    my $cgi = $self->{"cgi"} ;
    my $title = "Index for " . ucfirst($type) ;
    $self->set(-title => $title) ;
    $self->start_header() ;
    print $handle $cgi->h1("$title") . "\n" ;
    $self->print_table() ;
    # Make the all dependencies list:
    my (@deps) = $top->getCompiles(%ofiles) ;
    if (@deps) {
      print STDOUT "\tAll levels compilation dependencies is @deps\n" if $main::opt_v ;
      $self->print_deps(@deps) ;
      if ($main::opt_u) {
	print $handle "<HR>\n" ;
      } else {
	print $handle $cgi->img({src => "color.gif", width => '100%'}) . "\n" ;
      }
    }
    print $handle $cgi->start_ol() ."\n" ;
    my $pfile = undef ;
    foreach my $object (@objects) {
      my $name = $object->get(-name) ;
      my $file = $object->get(-file) ;
      print $handle $cgi->li() ;
      my $target ;
      if ($file && -f "$file") {
	$target = &targetname(${file}) ;
      } else {
	$target =  &targetinter($name);
      }
      if ($file) {
	unless ($hash{$file} && $pfile && $pfile eq $file) {
	  print STDOUT "Setting hash for $file\n" if $debug ;
	  $hash{$file} = 1 ;
	  print $handle "<A Name=\"R_" . $target . "\"></A>\n" ;
	  $pfile = $file ;
	}
	print $handle $cgi->a({
	  href => $target . "_c.html#R_" . $name ,
	  target => 'source',
	}, $name) . "\n";
      } else {
	print $handle $cgi->a({
	  href => $target . "_i.html" ,
	  target => 'index',
	}, $name) . "\n";
      }
    }
    print $handle $cgi->end_ol() ."\n" ;
    $self->end_html() ;
    close(OUT) ;
  }
  print STDOUT (caller(0))[3] .": Printed $type\n" if $debug ;
}
###########################################################################
# Special print for routines and functions:
sub printr_index {
  my $outdir = shift ;
  my $reldir = shift ;
  my $type = shift ;
  print STDOUT (caller(0))[3] .": Printing $type\n" if $debug ;
  my(@objects) = (sort byname @_) ;
  my(%hash) = () ;
  if (open(OUT,">$outdir/${type}index_c.html")) {
    my $handle = \*OUT ;
    my $self = new ihtml(-name => "$type", -handle => $handle, -file => "$type", -reldir => $reldir) ;
    my $cgi = $self->{"cgi"} ;
    my $title = "Index for " . ucfirst($type) ;
    $self->set(-title => $title) ;
    $self->start_header() ;
    print $handle $cgi->h1("$title") . "\n" ;
    $self->print_table() ;
    my $pfile = undef ;
    # Write the globals:
    my $nbglo = &fthrout::getCounter(-type => $type, -mode => "glo");
    my $nbmod = &fthrout::getCounter(-type => $type, -mode => "mod");
    my $nbpri = &fthrout::getCounter(-type => $type, -mode => "pri");
    print STDOUT "\t\t\t" . (caller(0))[3] .": Type is $type, nbglo=$nbglo, nbmod=$nbmod, nbpri=$nbpri\n" if $main::opt_v  ;
    if ($nbglo && $nbmod || $nbglo && $nbpri || $nbmod && $nbpri) {
      print $handle $cgi->start_ol() ."\n" ;
      print $handle "<LI><A Href=\#globals>Globals</A>\n" if $nbglo ;
      print $handle "<LI><A Href=\#modules>Modules</A>\n" if $nbmod ;
      print $handle "<LI><A Href=\#privates>Privates</A>\n" if $nbpri ;
      print $handle $cgi->end_ol() ."\n" ;
    }
    if ($nbglo) {
      print $handle "<BR><A Name=globals>" . $cgi->h2("Globals") . "</A>\n" ;
      print $handle $cgi->start_ol() ."\n" ;
      $pfile = undef ;
      foreach my $object (@objects) {
	next unless $object->get(-global) ;
	my $name = $object->get(-name) ;
	my $file = $object->get(-file) ;
	print $handle $cgi->li() ;
	my $target ;
	if ($file && -f "$file") {
	  $target = &targetname(${file}) ;
	} else {
	  $target =  &targetinter($name);
	}
	if ($file) {
	  unless ($hash{$file} && $pfile && $pfile eq $file) {
	    print STDOUT "Setting hash for $file\n" if $debug ;
	    $hash{$file} = 1 ;
	    print $handle "<A Name=\"R_" . $target . "\"></A>\n" ;
	    $pfile = $file ;
	  }
	  print $handle $cgi->a({
	    href => $target . "_c.html#R_" . $name ,
	    target => 'source',
	  }, $name) . "\n";
	} else {
	  print $handle $cgi->a({
	    href => $target . "_i.html" ,
	    target => 'index',
	  }, $name) . "\n";
	}
      }
      print $handle $cgi->end_ol() ."\n" ;
    }
    if ($nbmod) {
      print $handle "<A Name=modules>" . $cgi->h2("Modules") . "</A>\n" ;
      print $handle $cgi->start_ol() ."\n" ;
      $pfile = undef ;
      # Write the modules or private ones:
      foreach my $object (@objects) {
	next if $object->get(-global) ;
	my $mname = $object->get(-module) ;
	my $module = undef ;
	if ($mname) {
	  $module = &fthmod::getObject($mname) ;
	}
	next unless $module ;
	my $name = $object->get(-name) ;
	my $file = $object->get(-file) ;
	print $handle $cgi->li() ;
	my $target ;
	if ($file && -f "$file") {
	  $target = &targetname(${file}) ;
	} else {
	  $target =  &targetinter($name);
	}
	if ($file) {
	  unless ($hash{$file} && $pfile && $pfile eq $file) {
	    print STDOUT "Setting hash for $file\n" if $debug ;
	    $hash{$file} = 1 ;
	    print $handle "<A Name=\"R_" . $target . "\"></A>\n" ;
	    $pfile = $file ;
	  }
	  print $handle $cgi->a({
	    href => $target . "_c.html#R_" . $name ,
	    target => 'source',
	  }, $name);
	  print $handle " from module " . $cgi->a({
	    href => $target . "_c.html#R_" . $module->get(-name) ,
	    target => 'source',
	  }, $module->get(-name)) . "\n";
	} else {
	  print $handle $cgi->a({
	    href => $target . "_i.html" ,
	    target => 'index',
	  }, $name) . "\n";
	}
      }
      print $handle $cgi->end_ol() ."\n" ;
    }
    if ($nbpri) {
      print $handle "<A Name=privates>" . $cgi->h2("Privates") . "</A>\n" ;
      print $handle $cgi->start_ol() ."\n" ;
      $pfile = undef ;
      # Write the modules or private ones:
      foreach my $object (@objects) {
	next if $object->get(-global) ;
	my $prout = $object->get(-private) ;
	next unless $prout ;
	my $name = $object->get(-name) ;
	my $file = $object->get(-file) ;
	print $handle $cgi->li() ;
	my $target ;
	if ($file && -f "$file") {
	  $target = &targetname(${file}) ;
	} else {
	  $target =  &targetinter($name);
	}
	if ($file) {
	  unless ($hash{$file} && $pfile && $pfile eq $file) {
	    print STDOUT "Setting hash for $file\n" if $debug ;
	    $hash{$file} = 1 ;
	    print $handle "<A Name=\"R_" . $target . "\"></A>\n" ;
	    $pfile = $file ;
	  }
	  print $handle $cgi->a({
	    href => $target . "_c.html#R_" . $name ,
	    target => 'source',
	  }, $name);
	  print $handle " from private " . $cgi->a({
	    href => $target . "_c.html#R_" . $prout->get(-name) ,
	    target => 'source',
	  }, $prout->get(-name)) . "\n";
	} else {
	  print $handle $cgi->a({
	    href => $target . "_i.html" ,
	    target => 'index',
	  }, $name) . "\n";
	}
      }
      print $handle $cgi->end_ol() ."\n" ;
    }
    if ($nbglo && $nbmod || $nbglo && $nbpri || $nbmod && $nbpri) {
      print $handle $cgi->start_ol() ."\n" ;
      print $handle "<LI><A Href=\#globals>Globals</A>\n" if $nbglo ;
      print $handle "<LI><A Href=\#modules>Modules</A>\n" if $nbmod ;
      print $handle "<LI><A Href=\#privates>Privates</A>\n" if $nbpri ;
      print $handle $cgi->end_ol() ."\n" ;
    }
    $self->end_html() ;
    close(OUT) ;
  }
  print STDOUT (caller(0))[3] .": Printed $type\n" if $debug ;
}
###########################################################################
# A basic function that ought to be a copy of ftagshtml::Build_Index_Chapter:
sub print_latex {
  my $handle = shift ;
  my $type = shift ;
  my(@objects) = (sort byname @_) ;
  print STDOUT (caller(0))[3] .": Printing $type\n" if $debug ;
  print $handle "\\section{" . ucfirst($type). "}\\label{$type}\n" ;
  print $handle "\\begin{enumerate}\n" ;
  foreach my $object (@objects) {
    my $name = $object->get(-name) ;
    my $file = $object->get(-file) ;
    my $pname = &fthinp::LaTeXProtect($name) ;
    print $handle "\\item " . &fthinp::LaTeXProtect(&ihtml::targetname(${file})) . ": $pname\\label{$pname}\n";
  }
  print $handle "\\end{enumerate}\n" ;
  print STDOUT (caller(0))[3] .": Printed $type\n" if $debug ;
}
###########################################################################
sub print_bcorext {
  my $outdir = shift ;
  my $reldir = shift ;
  my $type = shift ;
  my $counter = 0 ;
  print STDOUT (caller(0))[3] .": Printing $type\n" if $debug ;
  my(@objects) = (sort byname @_) ;
  my(%hash) = () ;
  if (open(OUT,">$outdir/${type}index_c.html")) {
    my $handle = \*OUT ;
    my $self = new ihtml(-name => "$type", -handle => $handle, -file => "$type", -reldir => $reldir) ;
    my $cgi = $self->{"cgi"} ;
    my $title = "Index for " ;
    if ($type eq "badcall") {
      $title .= "either incorrect or simply unrecognized calls by ftagshtml limitations" ;
    } else {
      $title .= ucfirst($type) ;
    }
    $self->set(-title => $title) ;
    $self->start_header() ;
    print $handle $cgi->h1("$title") . "\n" ;
    $self->print_table() ;
    print $handle $cgi->start_ol() ;
    foreach my $object (@objects) {
      my $name = $object->get(-name) ;
      my (@stuffs) = $object->getStuffs() ;
      $counter += ( $#stuffs + 1) ;
      print $handle $cgi->li() ;
      if (&fthobj::getObject($name)) {
	my $fthobj = &fthobj::getObject($name) ;
	$name = $fthobj->get(-name) ;
	print $handle "\n<A Name=R_" . lc($name) .">" . $cgi->a({
	  href => &targetinter($name) . "_i.html" ,
	  target => 'index',
	}, $name ) . "</A>(". ( $#stuffs + 1) . "): " ;
      } else {
	print $handle "\n<A Name=R_" . lc($name) .">$name</A>(". ( $#stuffs + 1) . "): " ;
      }
      print $handle $cgi->start_ol() ."\n" ;
      my $ofather = undef ;
      my $ok ;
      foreach my $stuff (sort byhashcall @stuffs) {
	my $number = $stuff->get(-number) ;
	my $file = $stuff->get(-file) ;
	my $father = $stuff->get(-father) ;
	my $rame = $father->get(-name) ;
	unless ($ofather && $ofather eq $father) {
	  print $handle $cgi->li() ;
	  $ok = 0 ;
	  print $handle $cgi->a({
	    href => &targetname($father->get(-file)) . "_c.html#R_" . $rame ,
	    target => 'source',
	  }, $rame) . " (" . $stuff->get(-file) . "):";
	  $ofather = $father ;
	}
	print $handle ", " if $ok ;
	print $handle $cgi->a({
	  href =>  &targetname(${file})  . "_c.html#C_" . $number ,
	  target => 'source',
	}, $number) ;
	$ok++ ;
      }
      print $handle $cgi->end_ol() ."\n" ;
    }
    print $handle $cgi->end_ol() ."\n" ;
    $self->end_html() ;
    close(OUT) ;
  }
  print STDOUT (caller(0))[3] .": Printed $type\n" if $debug ;
  return($counter) ;
}
###########################################################################
sub print_java {
  my $self = shift ;
  my $outdir = shift ;
  my $reldir = shift ;
  my $type = shift ;
  print STDOUT (caller(0))[3] .": Printing $type\n" if $debug ;
  my(@objects) = (sort byname @_) ;
  my $handle = $self->{"handle"} ;
  my $cgi = $self->{"cgi"} ;
  print $handle "<TR>\n";
  print $handle "<TH><INPUT TYPE=button NAME=\"N$type\" VALUE=\"". ucfirst($type). "\" ONCLICK=lance(document.listform.my$type" . "[my$type.options.selectedIndex].value)></TH>\n";
  print $handle "<TH><SELECT NAME=\"my$type\">\n";
  foreach my $object (@objects) {
    my $name = $object->get(-name) ;
    my $file = $object->get(-file) ;
    print $handle "<OPTION VALUE=\"" . &ihtml::targetname(${file}) . "_c.html#R_" . $name . "\">$name\n";
  }
  print $handle "</SELECT></TH>\n";
  print $handle "</TR>\n";
}
###########################################################################
sub open_java {
  my $outdir = shift ;
  my $reldir = shift ;
  my $self = undef ;
  if (open(JVS,">$outdir/javanavig_c.html")) {
    my $handle = \*JVS ;
    $self = new ihtml(-name => "javanavig", -handle => $handle, -file => "javanavig", -reldir => $reldir) ;
    my $cgi = $self->{"cgi"} ;
    $self->set(-title => "JavaScript Form for $main::RepCodeSource") ;
    $self->start_java() ;
    my $title = "Java Script Index" ;
    print $handle $cgi->h1("$title") . "\n" ;
    $self->print_table() ;
    if ($main::opt_u) {
      # Defines the separtor:
      print $handle "<HR>\n" ;
    } else {
      print $handle "<IMG SRC=\"color.gif\" WIDTH=\"100%\"><BR>\n" ;
    }
    print $handle "<H2>Select an entry below and click on its type button:</H2>\n" ;
    print $handle "<FORM NAME=\"listform\">\n";
    print $handle "<TABLE BORDER=2 CELLPADDING=3 BGCOLOR=\"#AFFFC0\">\n";
  }
  return($self) ;
}
###########################################################################
sub close_java {
  my $self = shift ;
  my $outdir = shift ;
  my $reldir = shift ;
  my(@objects) = (sort byname @_) ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  print $handle "<TR>\n";
  print $handle "<TH><INPUT TYPE=button NAME=\"Nfile\" VALUE=\"Files Index\" ONCLICK=fileindex(document.listform.fileselect[fileselect.options.selectedIndex].value)></TH>\n";
  print $handle "<TH><SELECT NAME=\"fileselect\">\n";
  foreach my $object (@objects) {
    my $name = $object->get(-name) ;
    my $file = $object->get(-file) ;
    # print $handle "<OPTION VALUE=\"" . &ihtml::targetname(${file}) . "_c.html#R_" . $name . "\">$file\n";
    print $handle "<OPTION VALUE=\"R_" . &ihtml::targetname(${file}) . "\">$file\n";
  }
  print $handle "</SELECT></TH>\n";
  print $handle "</TR>\n";
  print $handle "</TABLE>\n";
  print $handle "</FORM>\n";
  $self->end_html() ;
  close($handle) ;
}
###########################################################################
sub start_interface {
  my $self = shift ;
  my $rel = shift ;
  $rel = "" unless $rel ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $title = $self->{"title"} ;
  my $file = $self->{"file"} ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  my $script = "// Javascript functions for $file.
function LienDynamique(document2){
   top.index.location=document2 ;
}
function JavaDynamique(document2){
   top.source.location=document2 ;
}
" ;
  my $base ;
  if ($main::opt_www) {
    $base = $main::opt_www ;
  } else {
    $base = '.' ;
  }
  print STDOUT (caller(0))[3] . ": Called by " . (caller(1))[3] . " line " . (caller(0))[2] . ": Starting on $self\n" if $debug ;
  print $handle $cgi->start_html
    (
     -BGCOLOR => "WHITE",
     -script => "$script",
     -title  => "$title",
     -author => "$main::USER",
     -base   => 'true',
     -xbase   => $base,
     -target => "source",
     -style  => { -src => "ftagshtml.css"},
     -TEXT   => 'BLACK',
     -LINK   => 'BLUE',
     -VLINK  => '#9910DD',
     -ALINK  => 'GREEN',
    )."\n" ;
  print $handle "<CENTER>\n" ;
  print $handle $cgi->start_table
    (
     {-border => '2',
     -CELLPADDING => '7',
     -width => '100%',
     -BGCOLOR => "#AAFFC0",}
    )."\n" ;
  my $level = 0 ;
  print $handle "<TR>\n" ;
  if ($level >= 8) {
    print $handle "</tr><tr>\n" ;
    $level = 0 ;
  }
  print $handle $cgi->th
    (
     {-align => 'right'},
     $cgi->a({
	      href=> $rel . "fileindex_c.html#R_". &targetname(${file}),
	      target => 'source',
	      onMouseOver => "window.status='Files Index: $file.'; return true"
	     },"F")) ."\n" ;
  if ($main::listpro) {
    $level++ ;
    if ($level >= 8) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "programindex_c.html#R_". &targetname(${file}),
	   target => 'source',
	   onMouseOver => "window.status='Program Index.'; return true"
	     },"p")) ."\n" ;
  }
  if ($main::listsub) {
    $level++ ;
    if ($level >= 8) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "routineindex_c.html#R_". &targetname(${file}),
	   target => 'source',
	   onMouseOver => "window.status='Routine Index.'; return true"
	     },"r")) ."\n" ;
  }
  if ($main::listfun) {
    $level++ ;
    if ($level >= 8) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "functionindex_c.html#R_". &targetname(${file}),
	   target => 'source',
	   onMouseOver => "window.status='Functions Index.'; return true"
	     },"f")) ."\n" ;
  }
  if (&fthmod::any()) {
    $level++ ;
    if ($level >= 8) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "moduleindex_c.html#R_". &targetname(${file}),
	   target => 'source',
	   onMouseOver => "window.status='Modules Index.'; return true"
	     },"M")) ."\n" ;
  }
  if (&fthtype::any()) {
    $level++ ;
    if ($level >= 8) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "typeindex_c.html#R_". &targetname(${file}),
	   target => 'source',
	   onMouseOver => "window.status='Types Index.'; return true"
	     },"T")) ."\n" ;
  }
  if (&fthinc::any()) {
    $level++ ;
    if ($level >= 8) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "includeindex_c.html",
	   target => 'source',
	   onMouseOver => "window.status='Includes Index.'; return true"
	     },"I")) ."\n" ;
  }
  if (&fthinp::any()) {
    $level++ ;
    if ($level >= 8) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "inputindex_c.html",
	   target => 'source',
	   onMouseOver => "window.status='Inputs Index.'; return true"
	     },"i")) ."\n" ;
  }
  if (&fthobj::any()) {
    $level++ ;
    if ($level >= 8) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "objectindex_c.html",
	   target => 'source',
	     },"o")) ."\n" ;
  }
  if (&fthext::any()) {
    $level++ ;
    if ($level >= 8) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "externalindex_c.html",
	   target => 'source',
	     },"E")) ."\n" ;
  }
  if (&fthbc::any()) {
    $level++ ;
    if ($level >= 8) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "badcallindex_c.html",
	   target => 'source',
	     },"?")) ."\n" ;
  }
  $level++ ;
  if ($level >= 8) {
    print $handle "</tr><tr>\n" ;
    $level = 0 ;
  }
  print $handle $cgi->th
      (
       {-align => 'right'},
       $cgi->a({
	 href=> $rel . "javanavig_c.html",
	 target => 'source',
	 onMouseOver => "window.status='Search Index.'; return true"
	   },"J")) ."\n" ;
  print $handle "</TR>\n" ;
  print $handle $cgi->end_table()."\n" ;
  print $handle "</CENTER>\n" ;
  # Reference to the dynamic Java Tree:
  unless ($file =~ /html/i) {
    # Hum, interface for object has no c file...
    print $handle "<H1>" . $cgi->a({
      href => &targetname($file) . "_LF.html",
      target => 'java',
      onClick => "JavaDynamique('" . &targetname(${file}). "_c.html')"
	},"$file")."</H1>\n" ;
  }
  if ($main::opt_u) {
    print $handle "<HR>\n" ;
  } else {
    print $handle $cgi->img({src => "color.gif", width => '100%'}) . "\n" ;
  }
}
###########################################################################
# NAJ: not very clean function, borrowed from the old ftagshtml
sub last_modified {
  my $self = shift ;
  my $file = $self->{"file"} ;
  my(@mx) = stat $file ;
  my($lastmodified) = $mx[9] ;
  my(@lt) = localtime($lastmodified);
  my($d,$m,$y) = @lt[3,4,5];
  $y += 1900 ; $m++ ;
  if (time()-$lastmodified <= 10*24*3600) { #24*3600) {
    return("$d/$m/$y".'@'."$lt[2]h$lt[1]m$lt[0]s");
  } else {
    return("$d/$m/$y");
  }
}
###########################################################################
sub print_top {
  my $outdir = shift ;
  my $object = shift ;
  if (open(TOP,">$outdir/index.html")) {
    my $handle = \*TOP ;
    my $title = "Base for $main::RepCodeSource user $main::USER" ;
    my $target = &targetname($object->get(-file)) ;
    print $handle "<html> <head> <title>$title</title></head>\n" ;
    print $handle "<frameset COLS=\"30%,*\" BORDER=\"1\" MARGINHEIGHT=\"0\" MARGINWIDTH=\"0\">\n" ;
    if ($main::opt_nojava) {
      print $handle "  <frame SRC=\"${target}_i.html\" NAME=\"index\">\n" ;
    } else {
      print $handle "<FRAMESET ROWS=\"50%,*\">\n" ;
      print $handle "  <frame SRC=\"${target}_LF.html\" NAME=\"java\">\n" ;
      print $handle "  <frame SRC=\"${target}_i.html\" NAME=\"index\">\n" ;
      print $handle "</FRAMESET> \n" ;
    }
    if (1) {
      print $handle "  <frame SRC=\"${target}_c.html\" NAME=\"source\">\n" ;
    } else {
      print $handle "  <frame SRC=\"fileindex_c.html\" NAME=\"source\">\n" ;
    }
    print $handle "</FRAMESET>\n" ;
    print $handle "</html>\n" ;
    close($handle) ;
  }
}
###########################################################################
# Prints the master cascading style sheet if does not exist.
# Do not erase it otherwise.
sub print_css {
  my $outdir = shift ;
  return if (-f "$outdir/ftagshtml.css") ;
  if (open(CSS,">$outdir/ftagshtml.css")) {
    my $handle = \*CSS ;
    # Enforced style sheet: make more down there:
    if ($main::opt_nocolor) {
      print $handle 'CODE {font-family: Arial, serif;}'."\n" ;
      print $handle 'COMMENT {text-style:bold; font-family: Gothic, roman; font-weight: bold;}'."\n" ;
      print $handle 'STRING {text-style:italic; font-family: Gothic, roman; font-weight: italic;}'."\n" ;
      print $handle 'ARGLIST {font-weight: bold;}'."\n" ;
      print $handle 'DECLARE {text-style:normal;}'."\n" ;
      print $handle 'VARIABLE {text-style:underline;}'."\n" ;
      print $handle 'BRANCHEMENT {font-style: italic;}'."\n" ;
      print $handle 'SUBR {font-family: Arial, serif;}'."\n" ;
      print $handle 'HEADER {text-style:normal;}'."\n" ;
      print $handle 'VERBATIM {text-style:normal;}'."\n" ;
      print $handle 'LIENSINDEX {text-style:normal;}'."\n" ;
      print $handle 'TEST {text-style:normal;}'."\n" ;
      print $handle 'LABEL {text-style:normal;}'."\n" ;
      print $handle 'READ {text-style:underline;}'."\n" ;
      print $handle 'LOOP {text-style:underline;}'."\n" ;
      print $handle 'INFO {text-style:normal;}'."\n" ;
      print $handle 'GETIN {text-style:normal;}'."\n" ;
    } else {
      print $handle 'CODE {color:#660000; font-family: Arial, serif; background-color:#ddffff;}'."\n" ;
      print $handle 'COMMENT {color:BLACK; text-style:bold; background-color:#ffffdd; font-family: Gothic, roman; font-weight: bold;}'."\n" ;
      print $handle 'ARGLIST {color:MAGENTA; background-color:#ccffdd; font-weight: bold}'."\n" ;
      #print $handle 'STRING {font-size:14pt; text-decoration:normal; font-style:small-caps; text-style:italic; font-family: Arial, serif;}'."\n" ;
      #print $handle 'STRING {color:BLACK; font-size:14pt; text-decoration:normal; font-style:small-caps; text-style:italic; font-family: Arial, serif; background-color:#ffffdd}'."\n" ;
      print $handle 'STRING {color:BROWN; text-style:italic; background-color:#ffffdd; font-family: Gothic, roman; font-weight: italic;}'."\n" ;
      print $handle 'DECLARE {color:#FF9900}'."\n" ;
      print $handle 'VARIABLE {color:GREEN; text-style:underline}'."\n" ;
      print $handle 'BRANCHEMENT {color:#6600EE; background-color:#aaffaa; font-style: italic}'."\n" ;
      print $handle 'SUBR {color:#660000; font-family: Arial, serif; background-color:#ddffff;}'."\n" ;
      print $handle 'HEADER {color:MAGENTA; background-color:#ccffdd}'."\n" ;
      print $handle 'VERBATIM {color:GREEN; background-color:#ffaadd}'."\n" ;
      print $handle 'LIENSINDEX {color:GREEN; background-color:#ccbbee}'."\n" ;
      print $handle 'TEST {color:#33DDDF}'."\n" ;
      print $handle 'LABEL {color:GREEN}'."\n" ;
      print $handle 'READ {color:#FF7799; text-style:underline}'."\n" ;
      print $handle 'LOOP {color:#33DD99; text-style:underline}'."\n" ;
      print $handle 'INFO {color:RED; text-style:normal; background-color:WHITE}'."\n" ;
      print $handle 'GETIN {color:RED; text-style:normal; background-color:GREEN}'."\n" ;
    }
    close(CSS) ;
  }
  if (open(CSS,">$outdir/ftie4style.css")) {
    my $handle = \*CSS ;
    print $handle "BODY {background-color: white}\n";
    close(CSS) ;
  }
}
###########################################################################
sub start_header {
  my $self = shift ;
  my $rel = shift ;
  $rel = "" unless $rel ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $title = $self->{"title"} ;
  my $file = $self->{"file"} ;
  my $target = &targetname(${file}) ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  my $script = "// Javascript functions for $file.
function LienDynamique(document2){
   top.index.location=document2 ;
}
function JavaDynamique(document2){
   top.source.location=document2 ;
}
" ;
  my $base ;
  if ($main::opt_www) {
    $base = $main::opt_www ;
  } else {
    $base = '.' ;
  }
  print $handle $cgi->start_html
    (
     -BGCOLOR => "WHITE",
     -script => "$script",
     -title  => "$title",
     -author => "$main::USER",
     -base   => 'true',
     -xbase   => $base,
     -target => "source",
     -style  => { -src => "ftagshtml.css"},
     -TEXT   => 'BLACK',
     -LINK   => 'BLUE',
     -VLINK  => '#9910DD',
     -ALINK  => 'GREEN',
    )."\n" ;
}
###########################################################################
sub start_java {
  my $self = shift ;
  my $rel = shift ;
  $rel = "" unless $rel ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $title = $self->{"title"} ;
  my $file = $self->{"file"} ;
  my $target = &targetname(${file}) ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  my $script = "// Javascript functions for $file.
function LienDynamique(document2){
   top.index.location=document2 ;
}
var fic=\"\";
function lance(commande){
   top.source.location=commande;
}
function monlance(commande){
   top.source.location=commande;
}
function fileindex(fic){
   top.source.location='fileindex_c.html#'+fic;
}
function programindex(fic){
   top.source.location='programindex_c.html#'+fic;
}
function routineindex(fic){
   top.source.location='routineindex_c.html#'+fic;
}
function moduleindex(fic){
   top.source.location='moduleindex_c.html#'+fic;
}
function functionindex(fic){
   top.source.location='functionindex_c.html#'+fic;
}
function includeindex(fic){
   top.source.location='includeindex_c.html#'+fic;
}
function inputindex(fic){
   top.source.location='inputindex_c.html#'+fic+'.tex';
}
" ;
  my $base ;
  if ($main::opt_www) {
    $base = $main::opt_www ;
  } else {
    $base = '.' ;
  }
  print $handle $cgi->start_html
    (
     -BGCOLOR => "WHITE",
     -script => "$script",
     -title  => "$title",
     -author => "$main::USER",
     -base   => 'true',
     -xbase   => $base,
     -target => "source",
     -style  => { -src => "ftagshtml.css"},
     -TEXT   => 'BLACK',
     -LINK   => 'BLUE',
     -VLINK  => '#9910DD',
     -ALINK  => 'GREEN',
    )."\n" ;
}
###########################################################################
sub print_table {
  my $self = shift ;
  my $rel = shift ;
  $rel = "" unless $rel ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $title = $self->{"title"} ;
  my $file = $self->{"file"} ;
  my $target = &targetname(${file}) ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  print $handle "<CENTER>\n" ;
  print $handle $cgi->start_table
      (
       {-border => '2',
	-CELLPADDING => '7',
	-width => '100%',
	-BGCOLOR => "#AAFFC0",}
       )."\n" ;
  print $handle "<TR>\n" ;
  my $level = 0 ;
  if ($level >= 5) {
    print $handle "</tr><tr>\n" ;
    $level = 0 ;
  }
  print $handle $cgi->th
      (
       {-align => 'right'},
       $cgi->a({
	 href=> $rel . "fileindex_c.html#R_". &targetname(${file}),
	 target => 'source',
	 onMouseOver => "window.status='Files Index: $file.'; return true"
	   },"Files")) ."\n" ;
  if ($main::listpro) {
    $level++ ;
    if ($level >= 5) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "programindex_c.html#R_". &targetname(${file}),
	   target => 'source',
	   onMouseOver => "window.status='Program Index.'; return true"
	     },"Programs")) ."\n" ;
  }
  if ($main::listsub) {
    $level++ ;
    if ($level >= 5) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "routineindex_c.html#R_". &targetname(${file}),
	   target => 'source',
	   onMouseOver => "window.status='Routine Index.'; return true"
	     },"Routines")) ."\n" ;
  }
  if ($main::listfun) {
    $level++ ;
    if ($level >= 5) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "functionindex_c.html#R_". &targetname(${file}),
	   target => 'source',
	   onMouseOver => "window.status='Functions Index.'; return true"
	     },"Functions")) ."\n" ;
  }
  if (&fthmod::any()) {
    $level++ ;
    if ($level >= 5) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "moduleindex_c.html#R_". &targetname(${file}),
	   target => 'source',
	   onMouseOver => "window.status='Modules Index.'; return true"
	     },"Modules")) ."\n" ;
  }
  if (&fthtype::any()) {
    $level++ ;
    if ($level >= 5) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "typeindex_c.html#R_". &targetname(${file}),
	   target => 'source',
	   onMouseOver => "window.status='Types Index.'; return true"
	     },"Types")) ."\n" ;
  }
  if (&fthinc::any()) {
    $level++ ;
    if ($level >= 5) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "includeindex_c.html",
	   target => 'source',
	   onMouseOver => "window.status='Includes Index.'; return true"
	     },"Includes")) ."\n" ;
  }
  if (&fthinp::any()) {
    $level++ ;
    if ($level >= 5) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "inputindex_c.html",
	   target => 'source',
	   onMouseOver => "window.status='Inputs Index.'; return true"
	     },"Inputs")) ."\n" ;
  }
  if (&fthobj::any()) {
    $level++ ;
    if ($level >= 5) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "objectindex_c.html",
	   target => 'source',
	     },"Objects")) ."\n" ;
  }
  if (&fthext::any()) {
    $level++ ;
    if ($level >= 5) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "externalindex_c.html",
	   target => 'source',
	     },"Externals")) ."\n" ;
  }
  if (&fthbc::any()) {
    $level++ ;
    if ($level >= 5) {
      print $handle "</tr><tr>\n" ;
      $level = 0 ;
    }
    print $handle $cgi->th
	(
	 {-align => 'right'},
	 $cgi->a({
	   href=> $rel . "badcallindex_c.html",
	   target => 'source',
	     },"Bad Calls")) ."\n" ;
  }
  $level++ ;
  if ($level >= 5) {
    print $handle "</tr><tr>\n" ;
    $level = 0 ;
  }
  print $handle $cgi->th
      (
       {-align => 'right'},
       $cgi->a({
	 href=> $rel . "javanavig_c.html",
	 target => 'source',
	 onMouseOver => "window.status='Search Index.'; return true"
	   },"JavaForm")) ."\n" ;
  print $handle "</TR>\n" ;
  print $handle $cgi->end_table()."\n" ;
  print $handle "</CENTER>\n" ;
}
#|#######################################################
sub CGI_Print {
  my $self = shift ;
  my $file = $self->{"file"} ;
  my $handle = $self->{"handle"} ;
  if ($main::opt_conf) {
    print $handle "<BR>
<FORM METHOD=\"POST\" ACTION=\"$main::opt_cgi/getin.pl\" ENCTYPE=\"application/x-www-form-urlencoded\" TARGET=\"source\">
<INPUT TYPE=\"hidden\" NAME=\"conf\" VALUE=\"$main::opt_conf\">
<INPUT TYPE=\"hidden\" NAME=\"fichier\" VALUE=\"$file\">
<INPUT TYPE=\"text\" NAME=\"ARGV\"  SIZE=20 MAXLENGTH=500>
<INPUT TYPE=\"submit\" NAME=\"Search\" VALUE=\"Search\">
</FORM>
<BR><A HREF=\"$main::opt_cgi/getin.pl?conf=$main::opt_conf\" TARGET=\"source\">Extended getin search</A>\n"
      ;
  } else {
    print $handle "<BR>
<FORM METHOD=\"POST\" ACTION=\"$main::opt_cgi/getin.pl\" ENCTYPE=\"application/x-www-form-urlencoded\" TARGET=\"source\">
<INPUT TYPE=\"hidden\" NAME=\"dir\" VALUE=\"$main::DIRW\">
<INPUT TYPE=\"hidden\" NAME=\"html\" VALUE=\"$main::DIRE\">
<INPUT TYPE=\"hidden\" NAME=\"www\" VALUE=\"$main::opt_www\">
<INPUT TYPE=\"hidden\" NAME=\"I\" VALUE=\"$main::opt_I\">
<INPUT TYPE=\"hidden\" NAME=\"getin_format\" VALUE=\"$main::opt_getin_format\">
<INPUT TYPE=\"hidden\" NAME=\"fichier\" VALUE=\"$file\">
<INPUT TYPE=\"text\" NAME=\"ARGV\"  SIZE=20 MAXLENGTH=500>
<INPUT TYPE=\"submit\" NAME=\"Search\" VALUE=\"Search\">
</FORM>
<BR><A HREF=\"$main::opt_cgi/getin.pl?dir=$main::DIRW&html=$main::DIRE\" TARGET=\"source\">Extended getin search</A>\n" ;
  }
}
###########################################################################
sub start_implementation {
  my $self = shift ;
  my $rel = shift ;
  $rel = "" unless $rel ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $title = $self->{"title"} ;
  my $file = $self->{"file"} ;
  my $target = &targetname(${file}) ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  $self->start_header($rel) ;
  print $handle "<A NAME=R_$target></A>" ;
  print $handle "<H1 align=center>" ;
  # Where on earth is the js flag being set ?
  if (!$main::opt_nojava && $self->{"js"}) {
    print $handle $cgi->a({
      href => $target . "_LF.html",
      target => 'java',
      onClick => "LienDynamique('${target}_i.html#R_$target')"
	},"${file}") ;
  } else {
    print $handle $cgi->a({
      href => $target . "_i.html",
      target => 'index',
    },"${file}") ;
  }
  print $handle "</H1>\n" ;
  if (exists $self->{"file"} && -f $self->{"file"}) {
    print $handle "<H2 align=center><H3 align=center>Last modified " ;
    if ($main::opt_cvsweb) {
      print $handle $cgi->a({
	href => "$main::opt_cvsweb/" . $self->{"file"},
	target => 'source',
	  }, $self->last_modified() )."\n" ;
    } else {
      print $handle $self->last_modified() ;
    }
    print $handle ".</H3></H2>\n" ;
  }

  $self->print_table($rel) ;
  my $reldir = $self->{"reldir"} ;
  print $handle $cgi->a({
	      href=>"$reldir$file",
	      target => 'source',
	      onMouseOver => "window.status='Source file $file'; return true"
	     },"View source file")."\n" ;
  if ($main::opt_cgi) {
    $self->CGI_Print() ;
  }
  if ($main::opt_u) {
    print $handle "<HR>\n" ;
  } else {
    print $handle $cgi->img({src => "color.gif", width => '100%'})."\n" ;
  }
}
###########################################################################
# Sorts called stuffs by name:
sub numeric {
  $a <=> $b ;
}
###########################################################################
sub alphabetic {
  lc($a) cmp lc($b) ||  $a cmp $b ;
}
###########################################################################
###########################################################################
# Sorts called stuffs by name:
sub byname {
  lc($a->get(-name)) cmp lc($b->get(-name)) ||  $a->get(-name) cmp $b->get(-name) ;
}
###########################################################################
# Sorts called stuffs by first appearance number in call:
sub bynumber {
  $a->get(-number) <=> $b->get(-number) ;
}
###########################################################################
# Sorts called stuffs by first appearance number in call:
sub byhashnumber {
  my ($a1,$a2,$a3) = split(/\:/, $a) ;
  my ($b1,$b2,$b3) = split(/\:/, $b) ;
  unless (exists $hash{$a} && exists $hash{$b} ) {
    #print STDERR (caller(1))[3] . " : hash $a:$hash{$a} or $b:$hash{$b} is wrong\n" if $debug ;
    return($a1 cmp $b1) ;
  }
  my ($a4, @a5) = (@{$hash{$a}}) ;
  my ($b4, @b5) = (@{$hash{$b}}) ;
  $a5[0]->get(-number) <=> $b5[0]->get(-number) || $a1 cmp $b1 ;
}
###########################################################################
# Sorts by father's name, file and then by called line number in self
sub byhashname {
  # my ($name,$fic,$number) = split(/\:/, $cle) ;
  my ($a1,$a2,$a3) = split(/\:/, $a) ;
  my ($b1,$b2,$b3) = split(/\:/, $b) ;
  $a1 cmp $b1 || $a2 cmp $b2 || $a3 <=> $b3 ;
}
###########################################################################
# Sorts by father's name, file and then by called line number in self
sub byhashcall {
  my $a1 = $a->get(-name) ;
  my $x2 = $a->get(-father) ;
  my $a3 = $a->get(-number) ;
  my $b1 = $b->get(-name) ;
  my $y2 = $b->get(-father) ;
  my $b3 = $b->get(-number) ;
  my $a2 = $x2 ;
  my $b2 = $y2 ;
  if ($x2 && ref($x2)) {
    $a2 = $x2->get(-file);
    $a1 = $x2->get(-name) ;
  }
  if ($y2 && ref($y2)) {
    $b2 = $y2->get(-file);
    $b1 = $y2->get(-name) ;
  }
  $a1 cmp $b1 || $a2 cmp $b2 || $a3 <=> $b3 ;
}
###########################################################################
sub imods_up_tree {
  my $self = shift ;
  my $cmod = shift ;
  my $file = $cmod->get(-file) ;
  my $target=  &targetname(${file}) ;
  my $name = $cmod->get(-name) ;
  my $ligne ;
  if ($opt_oldmm) {
    $ligne = "\taux2 = insFld(aux1,gFld(\"$name\",\"${target}_c.html#R_$name\"))\n" ;
  } else {
    $ligne = "\t[\"$name\", \"${target}_c.html#R_$name\", \n";
  }
  my(%hash) = $cmod->listCalled() ;
  if (%hash) {
    foreach my $mname (sort alphabetic keys %hash) {
      my $omod = &fthmod::getObject($mname) ;
      if ($omod) {
	my $hame = $omod->get(-name) ;
	my $href = &targetname($omod->get(-file)) . "_c.html#R_" . $hame ;
	if ($opt_oldmm) {
	  $ligne .= "\t\taux3 = insFld(aux2,gFld(\"$hame\",\"$href\"))\n" ;
	} else {
	  $ligne .= "\t\t[\"$hame\", \"$href\", \n";
	}
      } else {
	if ($opt_oldmm) {
	  $ligne .= "\t\taux3 = insFld(aux2,gFld(\"Global procedures:\",\"empty_reference.html\"))\n" ;
	} else {
	  $ligne .= "\t\t[\"Global procedures:\", \"empty_reference.html\", \n";
	}
      }
      my $orout ;
      foreach my $rname (sort alphabetic keys %{$hash{$mname}}) {
	foreach my $number (sort numeric keys %{$hash{$mname}->{$rname}}) {
	  my $ocall = $hash{$mname}->{$rname}->{$number} ;
	  #$orout = $ocall->getLink() ;
	  $orout = $ocall->get(-father) ;
	  my $hame = $orout->get(-name);
	  my $href = &targetname($orout->get(-file)) . "_c.html#R_" . $orout->get(-name);
	  if ($opt_oldmm) {
	    $ligne .= "\t\t\taux4 = insFld(aux3,gFld(\"$hame\",\"$href\"))\n" ;
	  } else {
	    $ligne .= "\t\t\t[\"$hame\", \"$href\", ],\n";
	  }
	  last ;
	}
      }
      unless ($opt_oldmm) {
	$ligne .="\t\t],\n" ;
      }
    }
  }
  unless ($opt_oldmm) {
    $ligne .="\t],\n" ;
  }
  return($ligne);
}
###########################################################################
sub imods_down_tree {
  my $self = shift ;
  my $cmod = shift ;
  my $file = $cmod->get(-file) ;
  my $target=  &targetname(${file}) ;
  my $name = $cmod->get(-name) ;
  my $ligne ;
  if ($opt_oldmm) {
    $ligne = "\taux2 = insFld(aux1,gFld(\"$name\",\"${target}_c.html#R_$name\"))\n" ;
  } else {
    $ligne = "\t[\"$name\", \"${target}_c.html#R_$name\", \n";
  }
  local(%hash) = $cmod->getCalls() ;
  if (%hash) {
    foreach my $mname (sort alphabetic keys %hash) {
      my $omod = &fthmod::getObject($mname) ;
      if ($omod) {
	my $hame = $omod->get(-name) ;
	my $href = &targetname($omod->get(-file)) . "_c.html#R_" . $hame ;
	if ($opt_oldmm) {
	  $ligne .= "\t\taux3 = insFld(aux2,gFld(\"$hame\",\"$href\"))\n" ;
	} else {
	  $ligne .= "\t\t[\"$hame\", \"$href\", \n";
	}
      } else {
	if ($opt_oldmm) {
	  $ligne .= "\t\taux3 = insFld(aux2,gFld(\"Global procedures:\",\"empty_reference.html\"))\n" ;
	} else {
	  $ligne .= "\t\t[\"Global procedures:\", \"empty_reference.html\", \n";
	}
      }
      my $ok ;
      my $title ;
      foreach my $rname (sort alphabetic keys %{$hash{$mname}}) {
	$ok = 0 ;
	foreach my $number (sort numeric keys %{$hash{$mname}->{$rname}}) {
	  my $ocall = $hash{$mname}->{$rname}->{$number} ;
	  my $orout = $ocall->getLink() ;
	  unless ($ok) {
	    if ($omod) {
	      $title = $omod->get(-name) . "->" . $orout->get(-name) ;
	    } else {
	      $title = $orout->get(-name) ;
	    }
	    my $hame = $orout->get(-name);
	    my $href = &targetname($orout->get(-file)) . "_c.html#R_" . $orout->get(-name);
	    if ($opt_oldmm) {
	      $ligne .= "\t\t\taux4 = insFld(aux3,gFld(\"$hame\",\"$href\"))\n" ;
	    } else {
	      $ligne .= "\t\t\t[\"$hame\", \"$href\", ],\n";
	    }
	  }
	  $ok++ ;
	  last ;
	}
      }
      unless ($opt_oldmm) {
	$ligne .="\t\t],\n" ;
      }
    }
  }
  unless ($opt_oldmm) {
    $ligne .="\t],\n" ;
  }
  return($ligne) ;
}
###########################################################################
sub printi_recurse {
  my $self = shift ;
  my $orout = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $file = $self->{"file"} ;
  local(%hash) = $orout->tellRouts() ;
  if (%hash) {
    print $handle "</BR>" . $orout->get(-name) . " defines private procedures:\n" ;
    print $handle $cgi->start_ol() ."\n";
    # A sort bynumber would result in usual by first appearance number, sorting by name may be easier
    foreach my $srout (sort byname values %hash) {
      print $handle $cgi->start_li() ;
      print $handle "<A Name=R_" .  $srout->get(-name) . "></A>" . $cgi->a({
	href => &targetname($srout->get(-file)) . "_c.html#R_" . $srout->get(-name),
	target => 'source',
      }, $srout->get(-name) ) ;
      # Now prints the is called by for the routine:
      my(%iscalledby) = $srout->listCalled() ;
      if (%iscalledby) {
	print $handle " called by:\n" ;
	my $olname = "" ;
	my $ok ;
	print $handle $cgi->start_ol() ."\n";
	foreach my $cle (sort byhashname keys %iscalledby) {
	  my ($lname,$fic,$number) = split(/\:/, $cle) ;
	  my $called = $iscalledby{$cle} ;
	  if ($olname) {
	    if ($olname eq "$lname") {
	      $ok = 0 ;
	      print $handle ",\n" ;
	    } else {
	      print $handle ";\n" ;
	      $ok = 1 ;
	      $olname = $lname ;
	    }
	  } else {
	    $ok = 1 ;
	    $olname = $lname ;
	  }
	  if ($ok) {
	    print $handle $cgi->li() . "\n" ;
	    my $urout = $called->get(-father) ;
	    if ($urout) {
	      print $handle $cgi->a({
		href => &targetname($urout->get(-file)) . "_c.html#R_" . $urout->get(-name),
		target => 'source',
	      }, $urout->get(-name) ) . ": " ;
	    } else {
	      # This ought not to happen
	      if ($main::opt_nocolor) {
		print $handle "$lname" ;
	      } else {
		print $handle "<FONT COLOR=RED>$lname</FONT>\n" ;
	      }
	    }
	  }
	  print $handle $cgi->a({
	    href => &targetname($called->get(-file)) . "_c.html#C_" . $called->get(-number),
	    target => 'source',
	  }, "$number" )  ;
	}
	print $handle ".\n" ;
	print $handle $cgi->end_ol() ."\n";
      }
      print $handle "\n";
      my(@poscalledby) = $srout->getPossibles() ;
      if (@poscalledby) {
	#NAJ518
	print $handle "<BR>May be called:\n" ;
	print $handle $cgi->start_ol() ."\n";
	my $oname = "" ;
	my $ok ;
	foreach my $ocall (sort byhashcall @poscalledby) {
	  my $urout = $ocall->get(-father) ;
	  my $name = $ocall->get(-name) ;
	  if ($urout) {
	    $name = $urout->get(-file) .':' . $urout->get(-name) ;
	  }
	  my $number = $ocall->get(-number) ;
	  if ($oname) {
	    if ($oname eq "$name") {
	      $ok = 0 ;
	      print $handle ",\n" ;
	    } else {
	      print $handle ".\n" ;
	      $ok = 1 ;
	      $oname = $name ;
	    }
	  } else {
	    $ok = 1 ;
	    $oname = $name ;
	  }
	  if ($ok) {
	    print $handle $cgi->li() . "\n" ;
	    if ($urout) {
	      print $handle $cgi->a({
		href => &targetname($urout->get(-file)) . "_c.html#R_" . $urout->get(-name),
		target => 'source',
	      }, $urout->get(-name) ) . ": " ;
	    } else {
	      # This ought not to happen
	      if ($main::opt_nocolor) {
		print $handle "$name" ;
	      } else {
		print $handle "<FONT COLOR=RED>$name</FONT>\n" ;
	      }
	    }
	  }
	  print $handle $cgi->a({
	    href => &targetname($ocall->get(-file)) . "_c.html#C_" . $ocall->get(-number),
	    target => 'source',
	  }, "$number" )  ;
	}
	print $handle $cgi->end_ol() ."\n";
      }
      $self->printi_recurse($srout) ;
    }
    print $handle $cgi->end_ol() . "\n";
  }
}
###########################################################################
sub printi_mods {
  my $self = shift ;
  my $module = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $file = $self->{"file"} ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  my $name = $module->get(-name) ;
  print $handle "<A Name=\"R_$name\"></A>\n" ;
  if ($main::opt_u) {
    print $handle "<HR>\n" ;
  } else {
    print $handle $cgi->img({src => "color.gif", width => '100%'}) . "\n" ;
  }
  print $handle $cgi->br . "Module " . $cgi->a({
	      href => &targetname(${file}) . "_LF.html",
	      onClick => "JavaDynamique('" . &targetname(${file}) . "_c.html#R_$name')",
	      target => 'java',
	     },"$name")."\n" ;
  # Lists all the implemented stuffs:
  local(%hash) = $module->getRouts() ;
  if (%hash) {
    print $handle "<BR>$name defines routines and procedures:\n" ;
    print $handle $cgi->start_ol() ."\n";
    # A sort bynumber would result in usual by first appearance number, sorting by name may be easier
    foreach my $orout (sort byname values %hash) {
      print $handle $cgi->start_li() ;
      print $handle "<A Name=R_" .  $orout->get(-name) . "></A>" . $cgi->a({
	href => &targetname($orout->get(-file)) . "_c.html#R_" . $orout->get(-name),
	target => 'source',
      }, $orout->get(-name) ) ;
      # Now prints the is called by for the routine:
      my(%iscalledby) = $orout->listCalled() ;
      if (%iscalledby) {
	print $handle " called by:\n" ;
	my $oname = "" ;
	my $ok ;
	print $handle $cgi->start_ol() ."\n";
	foreach my $cle (sort byhashname keys %iscalledby) {
	  my ($name,$fic,$number) = split(/\:/, $cle) ;
	  my $called = $iscalledby{$cle} ;
	  if ($oname) {
	    if ($oname eq "$name") {
	      $ok = 0 ;
	      print $handle ",\n" ;
	    } else {
	      print $handle ";\n" ;
	      $ok = 1 ;
	      $oname = $name ;
	    }
	  } else {
	    $ok = 1 ;
	    $oname = $name ;
	  }
	  if ($ok) {
	    print $handle $cgi->li() . "\n" ;
	    my $urout = $called->get(-father) ;
	    if ($urout) {
	      print $handle $cgi->a({
		href => &targetname($urout->get(-file)) . "_c.html#R_" . $urout->get(-name),
		target => 'source',
	      }, $urout->get(-name) ) . ": " ;
	    } else {
	      # This ought not to happen
	      if ($main::opt_nocolor) {
		print $handle "$name" ;
	      } else {
		print $handle "<FONT COLOR=RED>$name</FONT>\n" ;
	      }
	    }
	  }
	  print $handle $cgi->a({
	    href => &targetname($called->get(-file)) . "_c.html#C_" . $called->get(-number),
	    target => 'source',
	  }, "$number" )  ;
	}
	print $handle ".\n" ;
	print $handle $cgi->end_ol() . "\n";
      }
      print $handle "\n";
      my(@poscalledby) = $orout->getPossibles() ;
      if (@poscalledby) {
	#NAJ518
	print $handle "<BR>May be called:\n" ;
	print $handle $cgi->start_ol() ."\n";
	my $oname = "" ;
	my $ok ;
	foreach my $ocall (sort byhashcall @poscalledby) {
	  my $urout = $ocall->get(-father) ;
	  my $name = $ocall->get(-name) ;
	  if ($urout) {
	    $name = $urout->get(-file) .':' . $urout->get(-name) ;
	  }
	  my $number = $ocall->get(-number) ;
	  if ($oname) {
	    if ($oname eq "$name") {
	      $ok = 0 ;
	      print $handle ",\n" ;
	    } else {
	      print $handle ".\n" ;
	      $ok = 1 ;
	      $oname = $name ;
	    }
	  } else {
	    $ok = 1 ;
	    $oname = $name ;
	  }
	  if ($ok) {
	    print $handle $cgi->li() . "\n" ;
	    if ($urout) {
	      print $handle $cgi->a({
		href => &targetname($urout->get(-file)) . "_c.html#R_" . $urout->get(-name),
		target => 'source',
	      }, $urout->get(-name) ) . ": " ;
	    } else {
	      # This ought not to happen
	      if ($main::opt_nocolor) {
		print $handle "$name" ;
	      } else {
		print $handle "<FONT COLOR=RED>$name</FONT>\n" ;
	      }
	    }
	  }
	  print $handle $cgi->a({
	    href => &targetname($ocall->get(-file)) . "_c.html#C_" . $ocall->get(-number),
	    target => 'source',
	  }, "$number" )  ;
	}
	print $handle $cgi->end_ol() ."\n";
      }
      # Lists the routines that are contained by the available subroutines:
      $self->printi_recurse($orout) ;
    }
    print $handle $cgi->end_ol() . "\n";
  }
  local(%hash) = $module->getTypes() ;
  if (%hash) {
    print $handle "<BR>$name defines types:\n" ;
    print $handle $cgi->start_ol() ."\n";
    foreach my $type (sort bynumber values %hash) {
      print $handle "\t" . $cgi->start_li();
      print $handle "<A Name=R_" .  $type->get(-name) . "></A>" . $cgi->a({
	href => &targetname($type->get(-file)) . "_c.html#R_" . $type->get(-name),
	target => 'source',
      }, $type->get(-name) )."\n" ;
    }
    print $handle $cgi->end_ol() ."\n";
  }
  local(%hash) = $module->getDepends() ;
  if (%hash) {
    print $handle "<BR>$name recursively uses:\n" ;
    print $handle $cgi->start_ol() ."\n";
    foreach my $omod (sort byname values %hash) {
      print $handle "\t" . $cgi->start_li();
      print $handle $cgi->a({
	href => &targetname($omod->get(-file)) . "_c.html#R_" . $omod->get(-name),
	target => 'source',
      }, $omod->get(-name) )."\n" ;
    }
    print $handle $cgi->end_ol() ."\n";
  }
  local(%hash) = $module->getCalls() ;
  if (%hash) {
    print $handle "<BR>$name calls:\n" ;
    print $handle $cgi->start_ol() ."\n";
    foreach my $mname (sort alphabetic keys %hash) {
      my $omod = &fthmod::getObject($mname) ;
      print $handle "\t" . $cgi->start_li();
      if ($omod) {
	print $handle $cgi->a({
	  href => &targetname($omod->get(-file)) . "_c.html#R_" . $omod->get(-name),
	  target => 'source',
	}, $omod->get(-name) ) . "\n" ;
      } else {
	print $handle "Global procedures:" . "\n" ;
      }
      my $ok ;
      my $title ;
      foreach my $rname (sort alphabetic keys %{$hash{$mname}}) {
	$ok = 0 ;
	foreach my $number (sort numeric keys %{$hash{$mname}->{$rname}}) {
	  my $ocall = $hash{$mname}->{$rname}->{$number} ;
	  my $orout = $ocall->getLink() ;
	  unless ($ok) {
	    print $handle $cgi->start_ul() ."\n";
	    if ($omod) {
	      $title = $omod->get(-name) . "->" . $orout->get(-name) ;
	    } else {
	      $title = $orout->get(-name) ;
	    }
	    print $handle "\t\t" . $cgi->start_li(). $cgi->a({
	      href => &targetname($orout->get(-file)) . "_c.html#R_" . $orout->get(-name),
	      target => 'source',
	    }, "$title: " ) ;
	  }
	  if ($ok) {
	    print $handle ", " ;
	  }
	  print $handle $cgi->a({
	    href => &targetname($ocall->get(-file)) . "_c.html#C_" . $ocall->get(-number),
	    target => 'source',
	  }, $ocall->get(-number) ) ;
	  $ok++ ;
	}
	print $handle ".\n" . $cgi->end_ul() . "\n";
      }
    }
    print $handle $cgi->end_ol() ."\n";
  }
  my (%mhash) ;
  (%mhash) = $module->listCalled() ;
  if (%mhash) {
    print $handle "<BR>$name is called by:\n" ;
    print $handle $cgi->start_ol() ."\n" ;
    # Sorts by calling module name
    foreach my $mname (sort alphabetic keys %mhash) {
      my $omod = &fthmod::getObject($mname) ;
      print $handle "\t" . $cgi->li();
      if ($omod) {
	print $handle $cgi->a({
	  href => &targetname($omod->get(-file)) . "_c.html#R_" . $omod->get(-name),
	  target => 'source',
	}, $omod->get(-name) ) . ":\n" ;
      } else {
	print $handle "Global procedures:" . "\n" ;
      }
      my $ok ;
      my $title ;
      my $orout ;
      my $father ;
      # Sorting by name of calling routine inside the module:
      foreach my $rname (sort alphabetic keys %{$mhash{$mname}}) {
	$ok = 0 ;
	foreach my $number (sort numeric keys %{$mhash{$mname}->{$rname}}) {
	  my $ocall = $mhash{$mname}->{$rname}->{$number} ;
	  # Provide the name of the function that shall be called:
	  $orout = $ocall->getLink() ;
	  $father = $ocall->get(-father) ;
	  unless ($ok) {
	    print $handle $cgi->start_ol() ."\n";
	    if ($omod) {
	      $title = $omod->get(-name) . "::" . $orout->get(-name) ;
	      # Shorter title is more readable indeed:
	      $title = $orout->get(-name) ;
	    } else {
	      $title = $orout->get(-name) ;
	    }
	    print $handle $cgi->start_li(). $cgi->a({
	      href => &targetname($father->get(-file)) . "_c.html#R_" . $father->get(-name),
	      target => 'source',
	    }, $father->get(-name) ) . '->' ;
	    print $handle $cgi->a({
	      href => &targetname($orout->get(-file)) . "_c.html#R_" . $orout->get(-name),
	      target => 'source',
	    }, "$title: ") ;
	  }
	  if ($ok) {
	    print $handle ", " ;
	  }
	  print $handle $cgi->a({
	    href => &targetname($ocall->get(-file)) . "_c.html#C_" . $ocall->get(-number),
	    target => 'source',
	  }, $ocall->get(-number) ) ;
	  $ok++ ;
	}
	if ($ok) {
	  print $handle ".\n" . $cgi->end_ol() . "\n";
	}
      }
      print $handle $cgi->end_ul() . "\n";
    }
    print $handle $cgi->end_ol() . "\n";
  }
  (%mhash) = $module->getInputs() ;
  if (%mhash) {
    print $handle "<BR>$name inputs:\n" ;
    print $handle $cgi->start_ol() ."\n";
    foreach my $cle (sort alphabetic keys %mhash) {
      my $inpo = $mhash{$cle} ;
      print $handle $cgi->start_li() ."\n";
      $self->printi_incp($inpo) ;
    }
    print $handle $cgi->end_ol() ."\n";
  }
  (%mhash) = $module->getIncludes() ;
  if (%mhash) {
    print $handle "<BR>$name includes:\n" ;
    print $handle $cgi->start_ol() ."\n";
    foreach my $cle (sort alphabetic keys %mhash) {
      my $incl = $mhash{$cle} ;
      print $handle $cgi->start_li() ."\n";
      $self->printi_incl($incl) ;
    }
    print $handle $cgi->end_ol() ."\n";
  }
}
###########################################################################
sub addFolder {
  my $self = shift ;
  my $titre = shift ;
  my $file = shift ;
  my $target=  &targetname(${file}) ;
  if ($opt_oldmm) {
    return("aux1 = insFld(foldersTree,gFld(\"<i>$titre</i>\",\"${target}_c.html\"))") ;
  } else {
    return("  [\"<i>$titre</i>\", \"${target}_c.html\", ");
  }
}
###########################################################################
# Builds a two levels descending tree:
sub irouts_down_tree {
  my $self = shift ;
  my $orout = shift ;
  my $file = $orout->get(-file) ;
  my $target=  &targetname(${file}) ;
  my $name = $orout->get(-name) ;
  my $ligne ;
  if ($opt_oldmm) {
    $ligne = "\taux2 = insFld(aux1,gFld(\"$name\",\"${target}_c.html#R_$name\"))\n" ;
  } else {
    $ligne = "\t[\"$name\", \"${target}_c.html#R_$name\", \n";
  }
  local(%hash) = $orout->getCalled() ;
  if (%hash) {
    my $prout ;
    foreach my $cle (sort byhashnumber keys %hash) {
      my($crout, @called) = (@{$hash{$cle}}) ;
      if ($crout && ref($crout) eq 'fthrout') {
	my $hame = $crout->get(-name) ;
	my $href = &targetname($crout->get(-file)) . "_c.html#R_" . $hame ;
	if ($opt_oldmm) {
	  $ligne .= "\t\taux3 = insFld(aux2,gFld(\"$hame\",\"$href\"))\n" ;
	} else {
	  $ligne .= "\t\t[\"$hame\", \"$href\", \n";
	}
	my (%sash) = $crout->getCalled() ;
	if (%sash) {
	  foreach my $sle (sort byhashname keys %sash) {
	    my($brout, @nalled) = (@{$sash{$sle}}) ;
	    if ($brout && ref($brout) eq 'fthrout') {
	      my $same = $brout->get(-name) ;
	      my $sref = &targetname($brout->get(-file)) . "_c.html#R_" . $same ;
	      if ($opt_oldmm) {
		$ligne .= "\t\taux4 = insFld(aux3,gFld(\"$same\",\"$sref\"))\n" ;
	      } else {
		$ligne .= "\t\t\t[\"$same\", \"$sref\", ],\n";
	      }
	    } else {
	      my $ocall = $nalled[0] ;
	      my $same = $ocall->get(-name) ;
	      my $number = $ocall->get(-number) ;
	      my $sref = &targetname($ocall->get(-file)) . "_c.html#C_" . $number ;
	      if ($opt_oldmm) {
		$ligne .= "\t\taux4 = insFld(aux3,gFld(\"<FONT COLOR=RED>$same</FONT>\",\"$sref\"))\n" ;
	      } else {
		$ligne .= "\t\t\t[\"<FONT COLOR=RED>$same</FONT>\", \"$sref\", ],\n";
	      }
	    }
	  }
	}
	unless ($opt_oldmm) {
	  $ligne .="\t\t],\n" ;
	}
      } elsif (!$prout || lc($prout) ne lc($crout)) {
	$prout = $crout ;
	my $ocall = $called[0] ;
	my $same = $ocall->get(-name) ;
	my $number = $ocall->get(-number) ;
	my $sref = &targetname($ocall->get(-file)) . "_c.html#C_" . $number ;
	if ($opt_oldmm) {
	  $ligne .= "\t\taux3 = insFld(aux2,gFld(\"<FONT COLOR=RED>$same</FONT>\",\"$sref\"))\n" ;
	  # $ligne .= "\t\taux3 = insFld(aux2,gFld(\"<FONT COLOR=RED>$crout</FONT>\",\"empty_reference.html\"))\n" ;
	} else {
	  $ligne .= "\t\t[\"<FONT COLOR=RED>$same</FONT>\", \"$sref\", ],\n";
	}
      }
    }
  }
  unless ($opt_oldmm) {
    $ligne .= "\t],\n" ;
  }
  return($ligne) ;
}
###########################################################################
sub irouts_up_tree {
  my $self = shift ;
  my $orout = shift ;
  my $file = $orout->get(-file) ;
  my $target=  &targetname(${file}) ;
  my $name = $orout->get(-name) ;
  my $ligne ;
  if ($opt_oldmm) {
    $ligne = "\taux2 = insFld(aux1,gFld(\"$name\",\"${target}_c.html#R_$name\"))\n" ;
  } else {
    $ligne = "\t[\"$name\", \"${target}_c.html#R_$name\", \n";
  }
  local(%hash) = $orout->listCalled() ;
  if (%hash) {
    my $oname = "" ;
    my $pname = 0 ;
    foreach my $cle (sort byhashname keys %hash) {
      my ($name,$fic,$number) = split(/\:/, $cle) ;
      if ($oname) {
	if ($oname eq "$name") {
	  $pname = 0 ;
	} else {
	  $pname = 1 ;
	  $oname = $name ;
	}
      } else {
	$pname = 1 ;
	$oname = $name ;
      }
      if ($pname) {
	my $called = $hash{$cle} ;
	my $father = $called->get(-father) ;
	my $hame = $father->get(-name) ;
	my $href = &targetname($father->get(-file)) . "_c.html#R_" . $hame ;
	if ($opt_oldmm) {
	  $ligne .= "\t\taux3 = insFld(aux2,gFld(\"$hame\",\"$href\"))\n" ;
	} else {
	  $ligne .= "\t\t[\"$hame\", \"$href\", \n";
	}
	my(%sash) = $father->listCalled() ;
	if (%sash) {
	  my $o2name = "" ;
	  my $p2name = 0 ;
	  foreach my $sle (sort byhashname keys %sash) {
	    my ($n2ame,$fic2,$number2) = split(/\:/, $sle) ;
	    if ($o2name) {
	      if ($o2name eq "$n2ame") {
		$p2name = 0 ;
	      } else {
		$p2name = 1 ;
		$o2name = $n2ame ;
	      }
	    } else {
	      $p2name = 1 ;
	      $o2name = $n2ame ;
	    }
	    if ($p2name) {
	      my $called2 = $sash{$sle} ;
	      my $f2ather = $called2->get(-father) ;
	      my $same = $f2ather->get(-name) ;
	      my $sref = &targetname($f2ather->get(-file)) . "_c.html#R_" . $same ;
	      if ($opt_oldmm) {
		$ligne .= "\t\taux4 = insFld(aux3,gFld(\"$same\",\"$sref\"))\n" ;
	      } else {
		$ligne .= "\t\t\t[\"$same\", \"$sref\", ],\n";
	      }
	    }
	  }
	}
	unless ($opt_oldmm) {
	  $ligne .= "\t\t],\n" ;
	}
      }
    }
  }
  unless ($opt_oldmm) {
    $ligne .= "\t],\n" ;
  }
  return($ligne) ;
}
###########################################################################
sub print_deps {
  my $self = shift ;
  my (@deps) = (@_) ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $file = $self->{"file"} ;
  print $handle "<H2>The ". ($#deps + 1) . " compilation dependencies for $file:</H2>\n" ;
  foreach my $fic (@deps) {
    print $handle $cgi->a({
      href => &ihtml::targetname($fic) . "_c.html",
      target => 'source',
    }, $fic ). " " ;
  }
  print $handle "\n" ;
  
}
###########################################################################
sub printi_routs {
  my $self = shift ;
  my $orout = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  my $file = $self->{"file"} ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  my $name = $orout->get(-name) ;
  if ($main::opt_u) {
    print $handle "<HR>\n" ;
  } else {
    print $handle $cgi->img({src => "color.gif", width => '100%'}) . "\n" ;
  }
  print $handle "<A Name=\"R_$name\"></A>\n" ;
  my $proto = $orout->get(-theproto) ;
  if ($proto && $proto ne "1" && $proto ne "0" && $proto ne "2" && $proto ne "$name") {
    print $handle $orout->get(-theproto) . " " ;
  }
  # Reference to the dynamic Java Tree:
  print $handle $cgi->a({
	      href => &targetname(${file}) . "_LF.html",
	      target => 'java',
	      onClick => "JavaDynamique('" . &targetname(${file}). "_c.html#R_$name')"
	     },"$name")."\n" ;
  # NAJ: what do we need to do out of that ?
  # Now unroll all the calls:
  local(%hash) = $orout->getDepends() ;
  if (%hash) {
    print $handle "<BR>$name recursively uses:\n" ;
    print $handle $cgi->start_ol() ."\n";
    foreach my $omod (sort byname values %hash) {
      print $handle "\t" . $cgi->start_li();
      print $handle $cgi->a({
	href => &targetname($omod->get(-file)) . "_c.html#R_" . $omod->get(-name),
	target => 'source',
      }, $omod->get(-name) )."\n" ;
    }
    print $handle $cgi->end_ol() ."\n";
  }
  local(%hash) = $orout->tellRouts() ;
  if (%hash) {
    print $handle "<BR>$name defines private procedures:\n" ;
    print $handle $cgi->start_ol() ."\n";
    # A sort bynumber would result in usual by first appearance number, sorting by name may be easier
    foreach my $srout (sort byname values %hash) {
      print $handle $cgi->start_li() ;
      print $handle "<A Name=R_" .  $srout->get(-name) . "></A>" . $cgi->a({
	href => &targetname($srout->get(-file)) . "_c.html#R_" . $srout->get(-name),
	target => 'source',
      }, $srout->get(-name) ) ;
      # Now prints the is called by for the routine:
      my(%iscalledby) = $srout->listCalled() ;
      if (%iscalledby) {
	print $handle " called by:\n" ;
	my $oname = "" ;
	my $ok ;
	print $handle $cgi->start_ol() ."\n";
	foreach my $cle (sort byhashname keys %iscalledby) {
	  my ($name,$fic,$number) = split(/\:/, $cle) ;
	  my $called = $iscalledby{$cle} ;
	  if ($oname) {
	    if ($oname eq "$name") {
	      $ok = 0 ;
	      print $handle ",\n" ;
	    } else {
	      print $handle ".\n" ;
	      $ok = 1 ;
	      $oname = $name ;
	    }
	  } else {
	    $ok = 1 ;
	    $oname = $name ;
	  }
	  if ($ok) {
	    print $handle $cgi->li() . "\n" ;
	    my $urout = $called->get(-father) ;
	    if ($urout) {
	      print $handle $cgi->a({
		href => &targetname($urout->get(-file)) . "_c.html#R_" . $urout->get(-name),
		target => 'source',
	      }, $urout->get(-name) ) . ": " ;
	    } else {
	      # This ought not to happen
	      if ($main::opt_nocolor) {
		print $handle "$name" ;
	      } else {
		print $handle "<FONT COLOR=RED>$name</FONT>\n" ;
	      }
	    }
	  }
	  print $handle $cgi->a({
	    href => &targetname($called->get(-file)) . "_c.html#C_" . $called->get(-number),
	    target => 'source',
	  }, "$number" )  ;
	}
	print $handle $cgi->end_ol() ."\n";
      }
      print $handle "\n";
      my(@poscalledby) = $srout->getPossibles() ;
      if (@poscalledby) {
	#NAJ518
	print $handle "<BR>May be called:\n" ;
	print $handle $cgi->start_ol() ."\n";
	my $oname = "" ;
	my $ok ;
	foreach my $ocall (sort byhashcall @poscalledby) {
	  my $urout = $ocall->get(-father) ;
	  my $name = $ocall->get(-name) ;
	  if ($urout) {
	    $name = $urout->get(-file) .':' . $urout->get(-name) ;
	  }
	  my $number = $ocall->get(-number) ;
	  if ($oname) {
	    if ($oname eq "$name") {
	      $ok = 0 ;
	      print $handle ",\n" ;
	    } else {
	      print $handle ".\n" ;
	      $ok = 1 ;
	      $oname = $name ;
	    }
	  } else {
	    $ok = 1 ;
	    $oname = $name ;
	  }
	  if ($ok) {
	    print $handle $cgi->li() . "\n" ;
	    if ($urout) {
	      print $handle $cgi->a({
		href => &targetname($urout->get(-file)) . "_c.html#R_" . $urout->get(-name),
		target => 'source',
	      }, $urout->get(-name) ) . ": " ;
	    } else {
	      # This ought not to happen
	      if ($main::opt_nocolor) {
		print $handle "$name" ;
	      } else {
		print $handle "<FONT COLOR=RED>$name</FONT>\n" ;
	      }
	    }
	  }
	  print $handle $cgi->a({
	    href => &targetname($ocall->get(-file)) . "_c.html#C_" . $ocall->get(-number),
	    target => 'source',
	  }, "$number" )  ;
	}
	print $handle $cgi->end_ol() ."\n";
      }
    }
    print $handle $cgi->end_ol() . "\n";
  }
  local(%hash) = $orout->getCalled() ;
  if (%hash) {
    print $handle "<BR>$name calls\n" ;
    print $handle $cgi->start_ol() ."\n";
    foreach my $cle (sort byhashnumber keys %hash) {
      my($crout, @called) = (@{$hash{$cle}}) ;
      print $handle $cgi->start_li() ."\n";
      if ($crout && ref($crout) eq 'fthrout') {
	print $handle $cgi->a({
	  href => &targetname($crout->get(-file)) . "_c.html#R_" . $crout->get(-name),
	  target => 'source',
	}, $crout->get(-name) )."\n" ;
      } else {
	if ($main::opt_nocolor) {
	  print $handle "$crout" ;
	} else {
	  print $handle "<FONT COLOR=RED>$crout</FONT>\n" ;
	}
      }
      print $handle "\t(\n" ;
      my $count = 0 ;
      foreach my $ocall (@called) {
	print $handle " - " if $count ;
	print $handle $cgi->a({
	  href => &targetname($file) . "_c.html#C_" . $ocall->get(-number),
	  target => 'source',
	}, $ocall->get(-number) )."\n" ;
	$count++ ;
      }
      print $handle "\t)\n" ;
    }
    print $handle $cgi->end_ol() ."\n";
  }
  local(%hash) = $orout->listCalled() ;
  if (%hash) {
    print $handle "<BR>$name is called by\n" ;
    my $oname = "" ;
    my $ok ;
    print $handle $cgi->start_ol() ."\n";
    foreach my $cle (sort byhashname keys %hash) {
      my ($name,$fic,$number) = split(/\:/, $cle) ;
      my $called = $hash{$cle} ;
      if ($oname) {
	if ($oname eq "$name") {
	  $ok = 0 ;
	  print $handle ",\n" ;
	} else {
	  print $handle ".\n" ;
	  $ok = 1 ;
	  $oname = $name ;
	}
      } else {
	$ok = 1 ;
	$oname = $name ;
      }
      if ($ok) {
	print $handle $cgi->li() . "\n" ;
	my $urout = $called->get(-father) ;
	if ($urout) {
	  print $handle $cgi->a({
	    href => &targetname($urout->get(-file)) . "_c.html#R_" . $urout->get(-name),
	    target => 'source',
	  }, $urout->get(-name) ) . ": " ;
	} else {
	  # This ought not to happen
	  if ($main::opt_nocolor) {
	    print $handle "$name" ;
	  } else {
	    print $handle "<FONT COLOR=RED>$name</FONT>\n" ;
	  }
	}
      }
      print $handle $cgi->a({
	href => &targetname($called->get(-file)) . "_c.html#C_" . $called->get(-number),
	target => 'source',
      }, "$number" )  ;
    }
    print $handle $cgi->end_ol() ."\n";
  }
  local(%hash) = $orout->getInputs() ;
  if (%hash) {
    print $handle "<BR>$name inputs:\n" ;
    print $handle $cgi->start_ol() ."\n";
    foreach my $cle (sort alphabetic keys %hash) {
      my $inpo = $hash{$cle} ;
      print $handle $cgi->start_li() ."\n";
      $self->printi_incp($inpo) ;
    }
    print $handle $cgi->end_ol() ."\n";
  }
  local(%hash) = $orout->getIncludes() ;
  if (%hash) {
    print $handle "<BR>$name includes:\n" ;
    print $handle $cgi->start_ol() ."\n";
    foreach my $cle (sort alphabetic keys %hash) {
      my $incl = $hash{$cle} ;
      print $handle $cgi->start_li() ."\n";
      $self->printi_incl($incl) ;
    }
    print $handle $cgi->end_ol() ."\n";
  }
}
###########################################################################
sub targetinter {
  my $name = shift ;
  $name = "inter/$name" ;
  return($name) ;
}
###########################################################################
sub targetname {
  my $file = shift ;
  # print STDOUT (caller(0))[3] . ": On file $file, called by " . (caller(2))[3] ."". (caller(1))[3] . (caller(1))[2] . "\n" ;
  unless ($file) {
    print STDOUT "" . (caller(0))[3] . (caller(0))[2] .":" . (caller(1))[3] . (caller(1))[2] . ": Invalid API\n" ;
    $debug = 1 ;
    &ihtml::debug("No file argument to targetname") ;
    croak("Invalid API(ihtml::targetname)") ;
  }
  $file = basename($file) ;
  if ($file =~ /(.*)\.(.*)/) {
    $file = $1 . '_' . $2 ;
  }
  return($file) ;
}
###########################################################################
sub printi_type {
  my $self = shift ;
  my $object = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  print $handle $cgi->a({
    href => &targetname($object->get(-file)) . "_c.html",
    target => 'source',
  },  $object->get(-name) )."\n" ;
}
###########################################################################
sub printi_incp {
  my $self = shift ;
  my $inpo = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  print $handle $cgi->a({
    href => &targetname($inpo->get(-file)) . "_c.html",
    target => 'source',
  },  $inpo->get(-name) )."\n" ;
}
###########################################################################
sub printi_incl {
  my $self = shift ;
  my $incl = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  unless ($handle) {
    $handle = \*STDOUT ;
  }
  print $handle $cgi->a({
    href => &targetname($incl->get(-file)) . "_c.html",
    target => 'source',
  },  $incl->get(-name) )."\n" ;
}
###########################################################################
sub printstuff {
  my $self = shift ;
  my $stuff = shift ;
  my $cgi = $self->{"cgi"} ;
  my $handle = $self->{"handle"} ;
  if ($stuff && ref($stuff)) {
    my $masfic = $stuff->get(-masfic) ;
    $masfic = $stuff->get(-file) unless $masfic ;
    # print STDOUT "Master fic $stuff = ".$stuff->get(-masfic)." file=".$stuff->get(-file)."\n" ;
    my $name = $stuff->get(-name) ;
    if (ref($stuff) eq 'oref') {
      my $type = $stuff->get(-type) ;
      my $oname = $stuff->get(-name) ;
      my $object ;
      print STDOUT (caller(1))[3] .": Refering to a void $name($type) object $stuff\n" if $debug ;
      if ($type eq 'fthtype') {
	$object = &fthtype::getObject($oname) ;
      } elsif ($type eq 'fthrout') {
	$object = &fthrout::getObject($oname) ;
	my $mname = $oname ;
	if ($oname =~ /(.*)\:\:(.*)/) {
	  # name of the module
	  $mname = $1 ;
	  $oname = $2 ;
	}
	unless ($object) {
	  # Reference to a module interface ?
	  my $omod =  &fthmod::getObject($mname) ;
	  $object = $omod->getInter($oname) if $omod ;
	}
	unless ($object) {
	  # Another attempt:
	  my $fobj = &fthobj::getObject($oname) ;
	  $object = $fobj->getProcedure($oname) if $fobj ;
	}
      } elsif ($type eq 'uses') {
	# Try direct link to a module procedure:
	$object = &fthrout::getObject($oname) ;
	my $mname = $oname ;
	if ($oname =~ /(.*)\:\:(.*)/) {
	  # name of the module
	  $mname = $1 ;
	  $oname = $2 ;
	}
	unless ($object) {
	  # Reference to a module interface ?
	  my $omod =  &fthmod::getObject($mname) ;
	  $object = $omod->getInter($oname) if $omod ;
	}
	unless ($object) {
	  # It might be a structure ?
	  $object =  &fthtype::getObject($oname) ;
	}
      } elsif ($type eq 'fthinc') {
	$object = &fthinc::getObject($oname) ;
      } elsif ($type eq 'fthinp') {
	$object = &fthinp::getObject($oname) ;
      } elsif ($type eq 'fthmod') {
	$object = &fthmod::getObject($oname) ;
      } elsif ($type eq 'irout') {
	# This is a prototype for a routine, can be referenced, also try to find the link:
	my $orout = $stuff->get(-fthrout) ;
	$object = $orout->getLink() if $orout ;
	if ($object) {
	  # Links to implementation first then to the interface index
	  print $handle "<!new fthrout " . $stuff->get(-name) . ">" ;
	  print $handle "<A Name='R_$name'>" ;
	  # $object->datadump("\t") ;
	  print $handle $cgi->a({
	    href => &targetname($object->get(-file)) . "_c.html#R_" . $object->get(-name),
	    target => 'source',
	    onClick => "LienDynamique('" . &targetname($orout->get(-masfic)) . "_i.html#R_" . $orout->get(-name) . "')",
	  },"$oname") ;
	  print $handle "</A>" ;
	} elsif ($orout) {
	  print $handle "<!new fthrout " . $orout->get(-name) . ">" ;
	  print $handle "<A Name='R_$name'>" ;
	  print $handle $cgi->a({
	    href => &targetname($orout->get(-masfic)) . "_i.html#R_" . $orout->get(-name),
	    target => 'index',
	  },"$oname") ;
	  print $handle "</A>" ;
	} else {
	  if ($main::opt_nocolor) {
	    print $handle "<A Name='R_$oname'>$oname</A>" ;
	  } else {
	    print $handle "<A Name='R_$oname'><FONT COLOR=RED>$oname</FONT></A>" ;
	  }
	}
	return ;
      } elsif ($type eq 'fthlabel') {
	$object = &fthlabel::getObject($oname) ;
	if ($object) {
	  print $handle $cgi->a({
	    href => &targetname($object->get(-file)) . "_c.html#" . $object->get(-name),
	  },$oname) ;
	} else {
	  # The reference may be an external one !
	  if (exists $main::external_labels{$oname}) {
	    print $handle $cgi->a({
	      href => $main::external_labels{$oname} . "#" . $oname, 
				  },$oname) ;
	    print STDOUT "External reference to $oname points to $main::external_labels{$oname}\n" if $debug;
	  } elsif ($main::opt_nocolor) {
	    print $handle "$oname" ;
	  } else {
	    print $handle "<FONT COLOR=RED>$oname</FONT>" ;
	  }
	}
	return ;
      }
      if ($object) {
	print $handle $cgi->a({
	  href => &targetname($object->get(-file)) . "_c.html#R_" . $object->get(-name),
	},$oname) ;
      } else {
	if ($main::opt_nocolor) {
	  print $handle "$oname" ;
	} else {
	  print $handle "<FONT COLOR=RED>$oname</FONT>" ;
	}
      }
    } elsif (ref($stuff) eq 'fthtype') {
      print $handle "<!new fthtype " . $stuff->get(-name) . ">" ;
      print $handle "<A Name='R_$name'>" ;
      print $handle $cgi->a({
	href => &targetname($stuff->get(-masfic)) . "_i.html#R_" . $stuff->get(-name),
	target => 'index',
      },"$name") ;
      print $handle "</A>" ;
    } elsif (ref($stuff) eq 'fthrout') {
      # Add a reference in the tree stuff:
      print $handle "<!new fthrout " . $stuff->get(-name) . ">" ;
      print $handle "<A Name='R_$name'>" ;
      print $handle $cgi->a({
	href => &targetname($stuff->get(-masfic)) . "_i.html#R_" . $stuff->get(-name),
	target => 'index',
      },"$name") ;
      print $handle "</A>" ;
      if ($main::opt_show) {
	my (@cproto) = $stuff->getProto() ;
	print $handle "<U>[@cproto]</U>" ;
      }
    } elsif (ref($stuff) eq 'fthinc') {
      print $handle $cgi->a({
	href => &targetname($stuff->get(-file)) . "_c.html" ,
      },$stuff->get(-name)) ;
    } elsif (ref($stuff) eq 'fthinp') {
      print $handle $cgi->a({
	href => &targetname($stuff->get(-file)) . "_c.html" ,
      },$stuff->get(-name)) ;
    } elsif (ref($stuff) eq 'fthinter') {
      if ($stuff->get(-name)) {
	print $handle "<A Name='R_" . $stuff->get(-name) . "'></A>" ;
	print $handle $cgi->a({
	  href => &targetinter($stuff->get(-name)) . "_i.html" ,
	  target => 'index',
	},$stuff->get(-name)) ;
      } else {
	# Void interface must present the interface to the routines ?
	print $handle "<!No named interface, pointing to>:" ;
	print $handle "(" ;
	my(@procs) = $stuff->getProcedures() ;
	my $ok = 0 ;
	foreach my $pname (sort @procs) {
	  my $orout = $stuff->getProcedure($pname) ;
	  if ($orout) {
	    print $handle ", " if $ok ;
	    print $handle $cgi->a({
	      href => &targetname($orout->get(-file)) . "_c.html#R_" . $orout->get(-name),
	    }, $pname) ;
	    $ok++ ;
	  }
	}
	print $handle ")" ;
      }
    } elsif (ref($stuff) eq 'fthcall') {
      print $handle "<!new fthcall " . $stuff->get(-name) . ">" ;
      my (@args) = $stuff->getArguments() ;
      my $number = $stuff->get(-number) ;
      my $orout = $stuff->getLink() ;
      if ($orout) {
	my $stat = $stuff->okLink() ;
	if ($stat) {
	  my $fobj = &fthobj::getObject($stuff->get(-name)) ;
	  print $handle "<A Name='C_$number'>" ;
	  if ($fobj) {
	    # Probably, an interface exists for the call, let the user decide upon which one it is:
	    # NAJ518: probably time is left to decide which is the proper interface:
	    print $handle $cgi->a({
	      href => &ihtml::targetinter($stuff->get(-name)) . "_i.html" ,
	      target => 'index',
	    }, $stuff->get(-name)) ;
	  } else {
	    # Object not found:
	    print $handle "<I>" ;
	  }
	  print $handle "?" . $cgi->a({
	    href => &targetname($orout->get(-file)) . "_c.html#R_" . $orout->get(-name) ,
	  }, $orout->get(-name)) ;
	  unless ($fobj) {
	    print $handle "</I>" ;
	  }
	  print $handle "</A>" ;
	} else {
	  print $handle "<A Name='C_$number'>" . $cgi->a({
	    href => &targetname($orout->get(-file)) . "_c.html#R_" . $orout->get(-name) ,
	  }, $stuff->get(-name)) . "</A>" ;
	}
      } else {
	my $fobj = &fthobj::getObject($stuff->get(-name)) ;
	my(@fpossibles) = $stuff->getPossibles() ;
	if ($stuff->get(-fake)) {
	  # NAJ: If this is a fake call, we favour that:
	  if ($main::opt_show) {
	    print $handle "<A Name='C_$number'><FONT COLOR=GREEN>" . $stuff->get(-name) . "</FONT></A>" ;
	  } else {
	    # Normal print in noshow mode: eliminate the artefact.
	    print $handle $stuff->get(-name) ;
	  }
	  $delbr = 2 ;
	  if ($main::opt_show) {
	    print $handle "<I><B>{" . $stuff->get(-fake) . "}</B></I>" ;
	  }
	  return ;
	} elsif (@fpossibles) {
	  my $ok = 1 ;
	  my $name = $fpossibles[$#fpossibles]->get(-name) ;
	  my $finter = $stuff->getInterface($name) ;
	  for (my $i=0; $i <= $#fpossibles; $i++) {
	    if ($finter ne $stuff->getInterface($fpossibles[$i]->get(-name))) {
	      $ok = 0 ;
	    }
	  }
	  if ($ok) {
	    # Same interface:
	    print $handle "<A Name='C_$number'>" . $cgi->a({
	      href => &ihtml::targetname($finter->get(-file)) . "_c.html#R_" . $finter->get(-name) ,
	      target => 'source',
							   }, $stuff->get(-name)) . "</A>" ;
	  } else {
	    print $handle "<A Name='C_$number'>" . $cgi->a({
	      href => &ihtml::targetinter($stuff->get(-name)) . "_i.html" ,
	      target => 'index',
							   }, $stuff->get(-name)) . "</A>" ;
	  }
	  $ok = 0 ;
	  for (my $i=0; $i <= $#fpossibles; $i++) {
	    my $lorout = $fpossibles[$i] ;
	    if ($lorout) {
	      print $handle ", " if $ok ;
	      print $handle "[" unless $ok ;
	      print $handle "?" . $cgi->a({
		href => &ihtml::targetname($lorout->get(-file)) . "_c.html#R_" . $lorout->get(-name) ,
		target => 'source',
					  },$lorout->get(-name)) ;
	      $ok++ ;
	    }
	  }
	  print $handle "]" if $ok ;
	} elsif ($fobj) {
	  # First look for possible interfaces to the call:
	  # Probably, an interface exists for the call, let the user decide upon which one it is:
	  # NAJ518: probably time is left to decide which is the proper interface:
	  my (@linters) = $fobj->getStuffs() ;
	  # NAJ518: 0==1 for testing...
	  if (@linters && $#linters == 0) {
	    my $finter = $linters[0] ;
	    my $ok = 0 ;
	    if (ref($finter) eq 'fthinter') {
	      my (@procs) = $finter->getProcedures() ;
	      print $handle "<A Name='C_$number'>" . $cgi->a({
		href => &ihtml::targetname($finter->get(-file)) . "_c.html#R_" . $finter->get(-name) ,
		target => 'source',
							     }, $stuff->get(-name)) . "</A>" ;
	      # Now writes the direct links:
	      foreach my $lpname (@procs) {
		my $lorout = &fthrout::getObject($lpname) ;
		unless ($lorout) {
		  # Another attempt:
		  my $lfobj = &fthobj::getObject($lpname) ;
		  $lorout = $lfobj->getProcedure($lpname) if $lfobj ;
		}
		if ($lorout) {
		  print $handle ", " if $ok ;
		  print $handle "[" unless $ok ;
		  print $handle "?" . $cgi->a({
		    href => &ihtml::targetname($lorout->get(-file)) . "_c.html#R_" . $lorout->get(-name) ,
		    target => 'source',
					      },$lorout->get(-name)) ;
		  $ok++ ;
		}
	      }
	    } elsif(ref($finter) eq 'fthrout') {
	      print $handle ", " if $ok ;
	      print $handle "[" unless $ok ;
	      print $handle "<A Name='C_$number'>?" . $cgi->a({
		href => &ihtml::targetname($finter->get(-file)) . "_c.html#R_" . $finter->get(-name) ,
		target => 'source',
							      }, $stuff->get(-name)) . "</A>" ;
	      $ok++ ;
	    } else {
	      print $handle "<A Name='C_$number'>" . $cgi->a({
		href => &ihtml::targetinter($stuff->get(-name)) . "_i.html" ,
		target => 'index',
							     }, $stuff->get(-name)) . "</A>" ;
	    }
	    print $handle "]" if $ok ;
	  } else {
	    print $handle "<A Name='C_$number'>" . $cgi->a({
	      href => &ihtml::targetinter($stuff->get(-name)) . "_i.html" ,
	      target => 'index',
							   }, $stuff->get(-name)) . "</A>" ;
	    # Now try to find an equivalent proto:
	    my (@cproto) = $stuff->getProto() ;
	    # This call proto should have been transformed into a nicer one !
	    # Loops over the interfaces, get their common proto and compare, add a ? if matched !
	    print STDOUT "cproto='@cproto'\n" if $debug ;
	    if (@linters) {
	      my $ok = 0 ;
	      foreach my $finter (@linters) {
		if (ref($finter) eq 'fthinter') {
		  my (@procs) = $finter->getProcedures() ;
		  foreach my $lpname (@procs) {
		    my $lorout = &fthrout::getObject($lpname) ;
		    unless ($lorout) {
		      # Another attempt:
		      my $lfobj = &fthobj::getObject($lpname) ;
		      $lorout = $lfobj->getProcedure($lpname) if $lfobj ;
		    }
		    if ($lorout) {
		      # Compare the protos ?
		      my(@mproto) = $lorout->getProto() ;
		      my $ko = &fthrout::relProto(\@mproto, \@cproto) ;
		      unless ($ko) {
			# Add a possible link:
			if ($lorout) {
			  print $handle ", " if $ok ;
			  print $handle "[" unless $ok ;
			  print $handle "?" . $cgi->a({
			    href => &ihtml::targetname($lorout->get(-file)) . "_c.html#R_" . $lorout->get(-name) ,
			    target => 'source',
						      },$lorout->get(-name)) ;
			  $ok++ ;
			}
		      }
		    }
		  }
		} elsif (ref($finter) eq 'fthrout') {
		  print $handle ", " if $ok ;
		  print $handle "[" unless $ok ;
		  print $handle "<A Name='C_$number'>?" . $cgi->a({
		    href => &ihtml::targetname($finter->get(-file)) . "_c.html#R_" . $finter->get(-name) ,
		    target => 'source',
								  }, $stuff->get(-name)) . "</A>" ;
		  $ok++ ;
		}
	      }
	      print $handle "]" if $ok ;
	    }
	  }
	} else {
	  my $rame = $stuff->get(-name) ;
	  if (&fthbc::getObject($rame)) {
	    print $handle "<A Name='C_$number'><FONT COLOR=RED>" . $cgi->a({
	      href => "badcallindex_c.html#R_" . lc($rame) ,
	      target => 'source',
	    }, $stuff->get(-name)) . "</FONT></A>" ;
	  } elsif (&fthext::getObject($rame)) {
	    print $handle "<A Name='C_$number'><FONT COLOR=GREEN>" . $cgi->a({
	      href => "externalindex_c.html#R_" . lc($rame) ,
	      target => 'source',
	    }, $stuff->get(-name)) . "</FONT></A>" ;
	  } else {
	    print $handle "<A Name='C_$number'><FONT COLOR=RED>" . $stuff->get(-name) . "</FONT></A>" ;
	  }
	}
      }
      if ($main::opt_show) {
	my (@cproto) = $stuff->getProto() ;
	print $handle "<U>[@cproto]</U>" ;
	# Gets the reason for the link from top
	print $handle "<B>" . $stuff->debugLink() ."</B>" if ($main::opt_v >= 2) ;
      }
    } elsif (ref($stuff) eq 'fthmod') {
      print $handle "<!New module " . $stuff->get(-name) . ">" ;
      print $handle "<A Name='R_$name'>" ;
      print $handle $cgi->a({
	href => &targetname($stuff->get(-masfic)) . "_i.html#R_" . $stuff->get(-name),
	target => 'index',
      },"$name") ;
      print $handle "</A>" ;
    }
  } elsif ($stuff) {
    if ($delbr) {
      if ($delbr == 2 && $stuff eq '<BRANCHEMENT>') {
	$delbr = 1 ;
      } elsif ($delbr == 1 && $stuff eq '</BRANCHEMENT>') {
	$delbr = 0 ;
      } else {
	print $handle $stuff ;
      }
    } else {
      print $handle $stuff ;
    }
  } else {
    print $handle $stuff ;
  }
}
###########################################################################
1;

