package oref ;
###########################################################################
# Licence: 
# Copyright (C) 2000 Olivier Cessenat.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
###########################################################################
# A stuff container
###########################################################################
use strict;
use Carp ;
###########################################################################
use fthrout ;
use fthcall ;
use fthtype ;
use fthmod ;
use fthinter ;
###########################################################################
use vars qw($debug) ;
$debug = 0 ;
###########################################################################
sub new {
  my $self = {} ;
  shift ;
  my(%args) = (@_) ;
  bless $self ;
  if (%args) {
    $self->init(%args) ;
    $debug = $self->get(-debug) unless $debug;
  }
  return $self ;
}
###########################################################################
sub init {
  my $self = shift ;
  my(%args) = (@_) ;
  my($cle,$val) = () ;
  while(($cle,$val) = each %args) {
    if ($cle =~ /-(\w+)/) {
      $cle = $1 ;
      $self->{"$cle"} = $val ;
    }
  }
}
###########################################################################
sub get {
  my $self = shift ;
  my $param = shift ;
  if ($param =~ /-(\w+)/) {
    return($self->{"$1"}) ;
  } else {
    croak("Attempt to fetch a badly formed attribute") ;
    return(undef) ;
  }
}
###########################################################################
sub set {
  my $self = shift ;
  my $param = shift ;
  my $value = shift ;
  if ($param =~ /-(\w+)/) {
    $self->{"$1"} = $value ;
  } else {
    croak("Attempt to set a badly formed attribute") ;
    return(0) ;
  }
}
###########################################################################
# Debug function.
sub debug {
  return unless $debug ;
  my $self = shift ;
  my $message = shift ;
  unless (ref($self)) {
    $message = $self ;
  } 
  my ($a,$file,$b,$line0) = caller(0) ;
  my ($c,$fileb,$line1,$subr1) = caller(1) ;
  my ($d,$e,$line2,$subr2) = caller(2) ; 
  print STDOUT "$message: file $file routine $subr1 called at line $line1 by $subr2 said\n" ;
}
###########################################################################
# Frees the data structure
sub free {
  my $self = shift ;
  foreach my $cle (keys %{$self}) {
    delete $self->{"$cle"} ;
  }
}
###########################################################################
1;

